// This file defines the format of task type group files.

#ifndef _TASK_TYPE_GROUPS_H
#define _TASK_TYPE_GROUPS_H

#include "Schedule.h"

///////////////////////////////////////////////////////////////////////////
// MACROS
///////////////////////////////////////////////////////////////////////////

// This is the version of task type group file currently generated:
#define TTG_INFO_VERSION                      0x000007

// This is the signature written to the beginning of a task type
// group file to identify it as such:
#define TTG_INFO_SIGNATURE                    'TTGF'

// This macro converts a TTG_INFO pointer and an index into a pointer
// to the TT_INFO structure corresponding to the task type of the
// given index:
#define GET_TT_INFO(ttgInfo, index) \
  ((PTT_INFO) ((ttgInfo)->ttInfoOffsets[index] + (PCHAR) (ttgInfo)))

// This macro converts a TT_INFO pointer into a pointer to the
// algorithm data within that TT_INFO structure, and converts it
// to the given type:
#define GET_ALGORITHM_DATA(ttInfo, type) \
  ((type) ((ttInfo)->algorithmDataOffset + (PCHAR) (ttInfo)))

// This macro gives the size of the task type group info preamble,
// i.e., the part of the task type group file before the array of
// offsets for task type information structures:
#define TTG_INFO_PREAMBLE_SIZE (sizeof(TTG_INFO) - sizeof(unsigned long))

///////////////////////////////////////////////////////////////////////////
// STRUCTURES
///////////////////////////////////////////////////////////////////////////

// This structure describes the information for a single task type:

typedef struct _task_type_information {

  // This is the deadline in hundreds of nanoseconds:
  unsigned long deadlineHns;

  // This is the algorithm to use for scheduling tasks of this type:
  long algorithmType;

  // This is the offset in bytes from the beginning of this structure
  // to the beginning of the algorithm data structure (not included
  // in this definition because it comes after the variable-length
  // schedule at the end):
  unsigned long algorithmDataOffset;

  // This boolean indicates whether this algorithm includes gathering
  // a sample of recent task work requirements:
  BOOLEAN useSample;

  // This boolean indicates whether this algorithm includes in its
  // sample the work requirements of tasks that performed I/O:
  BOOLEAN sampleIfIo;

  // This is the schedule to use for this task type. Note that this
  // structure is of variable length, depending on how many phases
  // there are room for:
  SPEED_SCHEDULE schedule;

} TT_INFO, *PTT_INFO, *PPTT_INFO;

// This structure describes the information for a task type group.
// Thus, it's both a description of a task type group file as well as
// an in-memory copy of such a file. It doesn't really describe the
// whole file; it only describes the beginning of the file. In other
// words, there is data past the end of this structure that's also
// part of the task type group file.

typedef struct _ttg_info {

  // The file starts with a signature to identify it as a task type
  // group file:
  unsigned long signature;

  // Next comes the version identifier that describes the format
  // of this file:
  unsigned long version;

  // This is the size of this file:
  unsigned long size;

  // This is the number of task types in this group:
  unsigned long numTaskTypes;

  // This is the time this file's contents were last successfully
  // checked:
  LARGE_INTEGER timeLastChecked;

  // This is the scheduling interval in hundreds of nanoseconds
  // that was in force on the machine when this file was last
  // successfully checked:
  unsigned long schedulingIntervalHnsWhenLastChecked;

  // This array, although it appears to only have 1 entry, actually
  // is of variable length and has one entry for each task type.
  // Each entry gives the offset of the corresponding task type
  // information, in bytes from the beginning of the file:
  unsigned long ttInfoOffsets[1];

} TTG_INFO, *PTTG_INFO;

#endif
