// This file contains routines for deducing CPU information and adding
// it to the registry.

#include "stdafx.h"
#include <list>
#include "../Common/Parameters.h"
#include "Cpu.h"
#include "Install.h"
#include "InstallDlg.h"
#include "InstallRoutines.h"
#include "Speeds.h"
#include "SpeedCurves.h"
using namespace std;

/////////////////////////////////////////////////////////////////////////////
// MACROS
/////////////////////////////////////////////////////////////////////////////

// This is the number of milliseconds to sleep during the test for
// how quickly the timestamp counter advances
#define TSC_DELAY_MS                  30000

/////////////////////////////////////////////////////////////////////////////
// FORWARD DEFINITIONS
/////////////////////////////////////////////////////////////////////////////

BOOL WriteSpeedsInfoToRegistry (
  list<AvailableSpeedInfo>& availableSpeeds,
  list<WorthwhileSettingInfo>& worthwhileSettings,
  HKEY hKey
  );

/////////////////////////////////////////////////////////////////////////////
// CPU INFO
/////////////////////////////////////////////////////////////////////////////

// This function deduces the CPU information for this CPU. The passed
// parameter 'hWnd' is a handle to the main window, for use in
// displaying messages. This function returns a boolean indicating if
// it was successful.

BOOL AddCpuInformationToRegistry (HWND hWnd, CString& mainDirectory)
{
  // Open the registry key for RightSpeed.

  LONG err;
  HKEY hKey;
  err = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                     "Software\\Jay Lorch\\RightSpeed\\CurrentVersion",
                     0, // reserved
                     KEY_SET_VALUE,
                     &hKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not open RightSpeed registry key to write CPU info");
    return FALSE;
  }

  // Determine the system time and write it to the registry as the
  // last time we updated the CPU information.

  FILETIME fileTime;
  GetSystemTimeAsFileTime(&fileTime);
  if (!MyRegSetValue(hKey,
                     "CpuInfoModificationTime",
                     (BYTE *) &fileTime,
                     REG_BINARY,
                     sizeof(FILETIME),
                     "Last CPU information modification time")) {
    RegCloseKey(hKey);
    return FALSE;
  }

  // Determine what speeds are available on this CPU.

  PostMessage(hWnd, WM_CHANGE_STATUS, (WPARAM)
              "Determining what speeds are available on this CPU",
              0);
  
  list<AvailableSpeedInfo> availableSpeeds;
  if (!GetAvailableSpeeds(availableSpeeds)) {
    RegCloseKey(hKey);
    return FALSE;
  }

  // Make sure there are at least two speeds available, or there's
  // no sense installing RightSpeed.

  if (availableSpeeds.size() < 2) {
    MessageBox(hWnd, "Found fewer than two available speeds", "Error",
               MB_OK | MB_ICONEXCLAMATION);
    RegCloseKey(hKey);
    return FALSE;
  }

  // Determine the wortwhile settings among these available speeds.

  list<WorthwhileSettingInfo> worthwhileSettings;
  if (!GetWorthwhileSettings(worthwhileSettings, availableSpeeds)) {
    RegCloseKey(hKey);
    return FALSE;
  }

  // Write the speeds information to the registry.

  if (!WriteSpeedsInfoToRegistry(availableSpeeds, worthwhileSettings, hKey)) {
    RegCloseKey(hKey);
    return FALSE;
  }

  // We're done with the registry key now, so close it.

  RegCloseKey(hKey);

  // Create the speed curves file.

  if (!CreateSpeedCurvesFile(worthwhileSettings, mainDirectory, hWnd))
    return FALSE;

  // We're done, so return success.

  return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// TIMESTAMP COUNTER
/////////////////////////////////////////////////////////////////////////////

// This function reads the current timestamp counter and puts it
// in 'tscPtr'.

void GetTSC (LONGLONG *tscPtr)
{
  __asm {
    _emit 0xF
    _emit 0x31
    mov ebx, tscPtr
    mov [ebx], eax
    mov [ebx+4], edx
  }
}

// This function writes information about the available speeds and
// worthwhile settings on this CPU to the registry. The parameter
// 'hKey' gives an open registry key to write to.

BOOL WriteSpeedsInfoToRegistry (
  list<AvailableSpeedInfo>& availableSpeeds,
  list<WorthwhileSettingInfo>& worthwhileSettings,
  HKEY hKey
  )
{
  DWORD numAvailableSpeeds = availableSpeeds.size();
  DWORD numWorthwhileSettings = worthwhileSettings.size();
  DWORD *availableSpeedsArray = new DWORD[numAvailableSpeeds];
  short *worthwhileSettingsArray = new short[numWorthwhileSettings];
  DWORD *codesArray = new DWORD[numAvailableSpeeds];

  // For each speed, fill in the corresponding element of the arrays
  // availableSpeedsArray and codesArray.
  
  list<AvailableSpeedInfo>::iterator it;
  int i;
  for (it = availableSpeeds.begin(), i = 0; it != availableSpeeds.end();
       ++it, ++i) {
    codesArray[i] = it->fidVidCode;
    availableSpeedsArray[i] = it->khz;
  }

  list<WorthwhileSettingInfo>::iterator it2;
  for (it2 = worthwhileSettings.begin(), i = 0;
       it2 != worthwhileSettings.end();
       ++it2, ++i)
    worthwhileSettingsArray[i] = it2->speedIndex;

  // Write the arrays to the registry.

  if (!MyRegSetValue(hKey,
                     "AvailableSpeeds",
                     (BYTE *) availableSpeedsArray,
                     REG_BINARY,
                     sizeof(DWORD) * numAvailableSpeeds,
                     "Available speeds") ||
      !MyRegSetValue(hKey,
                     "WorthwhileSettings",
                     (BYTE *) worthwhileSettingsArray,
                     REG_BINARY,
                     sizeof(short) * numWorthwhileSettings,
                     "Worthwhile speed settings") ||
      !MyRegSetValue(hKey,
                     "AmdFidVidCodes",
                     (BYTE *) codesArray,
                     REG_BINARY,
                     sizeof(DWORD) * numAvailableSpeeds,
                     "Transmeta speed codes")) {
    delete availableSpeedsArray;
    delete worthwhileSettingsArray;
    delete codesArray;
    return FALSE;
  }

  // Free the arrays we allocated and return success.

  delete availableSpeedsArray;
  delete worthwhileSettingsArray;
  delete codesArray;
  return TRUE;
}
