// This file implements the RSInitHandle class, which represents an
// open handle to the RSInit device.

#include "stdafx.h"
#include <windows.h>
#include <winioctl.h>
#include <iostream>
#include <winsvc.h>
#include "InstallRoutines.h"
#include "RSInitHandle.h"
#include "../Common/MSR.h"

//////////////////////////////////////////////////////////////////////////////
// INITIALIZATION / TERMINATION
//////////////////////////////////////////////////////////////////////////////

// This function initializes the object to represent an unopened
// handle.

RSInitHandle::RSInitHandle() :
  h(INVALID_HANDLE_VALUE),
  scManagerHandle(NULL),
  serviceHandle(NULL),
  mustStopRSInit(false)
{
}

// This function initializes the handle by opening a handle to the
// RSInit device.

bool RSInitHandle::Initialize ()
{
  // Get a handle to the service control manager.

  scManagerHandle = OpenSCManager(NULL, // which machine: default
                                  NULL, // which database: default
                                  SC_MANAGER_ALL_ACCESS);
  if (scManagerHandle == NULL) {
    DisplayError("Could not open the service control manager");
    return false;
  }

  // Get a handle to the RSInit service.

  serviceHandle = OpenService(scManagerHandle,
                              "RSInit",
                              SERVICE_ALL_ACCESS);
  if (serviceHandle == NULL) {
    DisplayError("Could not open a handle to the RSInit service");
    return false;
  }

  // Start the RSInit service if it isn't started already.

  if (StartService(serviceHandle,
                   NULL, // service arguments
                   NULL)) { // argument vectors
    mustStopRSInit = true;
  }
  else if (GetLastError() != ERROR_SERVICE_ALREADY_RUNNING) {
    DisplayError("Could not start RSInit service");
    return false;
  }

  // Open the handle by opening the file \\.\RSInit.

  h = CreateFile("\\\\.\\RSInit",
                 GENERIC_READ | GENERIC_WRITE,
                 0,     // share mode
                 NULL,  // security attributes
                 OPEN_EXISTING,
                 FILE_ATTRIBUTE_NORMAL,
                 NULL); // template
  if (h == INVALID_HANDLE_VALUE)
    return false;

  // Return success.

  return true;
}

// This function closes the handle, if it is open. It also closes any
// other resources that we have open.

void RSInitHandle::Close ()
{
  // Close the handle if it's open.
  
  if (h != INVALID_HANDLE_VALUE) {
    CloseHandle(h);
    h = INVALID_HANDLE_VALUE;
  }

  // If we have to stop RSInit, do so.

  if (mustStopRSInit && serviceHandle != NULL) {
    SERVICE_STATUS serviceStatus;
    ControlService(serviceHandle,
                   SERVICE_CONTROL_STOP,
                   &serviceStatus);
    mustStopRSInit = false;
  }

  // If we have an open handle to RSInit, close it.

  if (serviceHandle != NULL) {
    CloseServiceHandle(serviceHandle);
    serviceHandle = NULL;
  }

  // If we have an open handle to the service control manager, close it.

  if (scManagerHandle != NULL) {
    CloseServiceHandle(scManagerHandle);
    scManagerHandle = NULL;
  }
}

// This function sets the speed according to the FID/VID code given.

bool RSInitHandle::WriteSpeedMsr (DWORD fidVidCode)
{
  DWORD returnSize;
  struct {
    DWORD value1;
    DWORD value2;
  } sendBuf;

  sendBuf.value1 = fidVidCode;
  sendBuf.value2 = AMD_SPEED_VOLTAGE_SET_DELAY;
  if (!DeviceIoControl(h,
                       RS_INIT_IOCTL_WRITE_SPEED_MSR,
                       &sendBuf,
                       sizeof(sendBuf),
                       0,
                       0,
                       &returnSize,
                       NULL))
    return false;
  else
    return true;
}

// This function obtains the FID/VID status register contents, and
// puts them in the 8-byte buffer pointed to by 'fidVidStatusPtr'.
// It returns a boolean indicating if it was successful.

bool RSInitHandle::GetFidVidStatus (char *fidVidStatusPtr)
{
  DWORD returnSize;

  if (!DeviceIoControl(h,
                       RS_INIT_IOCTL_GET_FID_VID_STATUS,
                       0,
                       0,
                       fidVidStatusPtr,
                       8,
                       &returnSize,
                       NULL) ||
      returnSize != 8)
    return false;
  else
    return true;
}

// This function reads up to 'size' bytes from the PowerNow table into
// the buffer 'table' of size 'size'.  It returns the actual number
// of bytes copied to there.

DWORD RSInitHandle::ReadPowerNowTable (char *table, DWORD size)
{
  DWORD returnSize;

  if (!DeviceIoControl(h,
                       RS_INIT_IOCTL_READ_POWERNOW_TABLE,
                       0,
                       0,
                       table,
                       size,
                       &returnSize,
                       NULL))
    return 0;
  else
    return returnSize;
}
