/////////////////////////////////////////////////////////////////////////////
// SPEED CURVES
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <queue>
#include "../Common/Parameters.h"
#include "InstallRoutines.h"
#include "SpeedCurves.h"

namespace {

  // A "K value" represents a value of k at which one interval will
  // make a transition from one speed to another. The speed index i
  // indicates that the transition would be from i to i + 1. The
  // phase index j indicates that the interval is the one preceding
  // quantile number j.

  struct KValue {
    KValue(double kVal, short settingIndexVal, short phaseIndexVal) :
      k(kVal), settingIndex(settingIndexVal), phaseIndex(phaseIndexVal) { }

    // The operator < is defined backwards so that a priority queue,
    // which always provides the largest value, actually provides the
    // smallest one.

    bool operator< (const KValue& other) const { return k > other.k; }

    double k;
    short settingIndex;
    short phaseIndex;
  };

  double avgCDFTail[IDEAL_SCHEDULE_NUM_PHASES] =
  { 1,             0.9085996679, 0.7441598046, 0.6094805384,  0.4991757475,
    0.4088340992,  0.3348426311, 0.2742422606, 0.224609445,   0.183959258,
    0.1506660087,  0.1233982265, 0.1010654124, 0.08277442768, 0.06779377553,
    0.05552434646, 0.0425,       0.0275,       0.0125,        0.0025 };

  double HarmonicMean (double a, double b)
  {
    return (2 * a * b) / (a + b);
  }

}

BOOL CreateSpeedCurvesFile (
  list<WorthwhileSettingInfo>& worthwhileSettings,
  CString& mainDirectory,
  HWND hWnd
  )
{
  priority_queue<KValue> kValues;
  list<WorthwhileSettingInfo>::iterator currentSetting, nextSetting;
  short numWorthwhileSettings = worthwhileSettings.size();
  DWORD currentKHz, nextKHz, currentEnergy, nextEnergy, bytesWritten;
  double averageKHz, energyFunctionDerivative, multiplier;
  short settingIndex, phaseIndex;
  HANDLE hFile;
  char *currentTransitionSet;

  // Create the file to contain the speed curves information.

  hFile = CreateFile(mainDirectory + "\\SpeedCurves.txt",
                     GENERIC_WRITE,
                     0, // share mode,
                     NULL, // security attributes
                     CREATE_ALWAYS,
                     FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
                     NULL); // template
  if (hFile == INVALID_HANDLE_VALUE) {
    DisplayError("Could not create speed curves file");
    return FALSE;
  }

  // Initialize the first transition set to be such that all phases run
  // at the minimum speed.

  currentTransitionSet = new char[numWorthwhileSettings - 1];
  if (!currentTransitionSet) {
    CloseHandle(hFile);
    MessageBox(hWnd, "Could not allocate space for transition set", "Error",
               MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }
  for (settingIndex = 0; settingIndex < numWorthwhileSettings - 1;
       ++settingIndex)
    currentTransitionSet[settingIndex] = (char) IDEAL_SCHEDULE_NUM_PHASES;

  // Write the first transition set to the file.

  if (!WriteFile(hFile,
                 currentTransitionSet,
                 numWorthwhileSettings - 1,
                 &bytesWritten,
                 NULL) ||
      bytesWritten != numWorthwhileSettings - 1) { // overlapped
    DisplayError("Could not write initial transition set to speed curves "
                 "file");
    CloseHandle(hFile);
    delete currentTransitionSet;
    return FALSE;
  }

  // Loop through the speeds.  For each speed, compute E'(s) x s^2.
  // Then, for each phase, compute the K value that would cause a
  // transition from this speed to the next speed.

  for (currentSetting = nextSetting = worthwhileSettings.begin(),
         ++nextSetting, settingIndex = 0;
       nextSetting != worthwhileSettings.end();
       currentSetting = nextSetting, ++nextSetting, ++settingIndex) {

    currentKHz = currentSetting->khz;
    nextKHz = nextSetting->khz;

    currentEnergy = currentSetting->mv * currentSetting->mv;
    nextEnergy = nextSetting->mv * nextSetting->mv;

    averageKHz = HarmonicMean(currentKHz, nextKHz);
    energyFunctionDerivative = (nextEnergy - currentEnergy) * 1.0 /
                               (nextKHz - currentKHz);
    multiplier = energyFunctionDerivative * averageKHz * averageKHz;

    for (phaseIndex = 0; phaseIndex < IDEAL_SCHEDULE_NUM_PHASES; ++phaseIndex)
      kValues.push(KValue(multiplier * avgCDFTail[phaseIndex],
                          settingIndex, phaseIndex));
  }

  // Loop through all the K values in increasing order.  (Note that <
  // is defined backwards for K values so that using a priority queue
  // to always obtain the largest one actually always retrieves the
  // smallest one.)  For each such K value, update the current
  // transition set and write the consequent transition set to the
  // file.

  while (kValues.size()) {
    currentTransitionSet[kValues.top().settingIndex] =
      (char) kValues.top().phaseIndex;
    kValues.pop();

    // Write this next transition set to the file.

    if (!WriteFile(hFile,
                   currentTransitionSet,
                   numWorthwhileSettings - 1,
                   &bytesWritten,
                   NULL) ||
        bytesWritten != numWorthwhileSettings - 1) { // overlapped
      DisplayError("Could not write transition set to speed curves file");
      CloseHandle(hFile);
      delete currentTransitionSet;
      return FALSE;
    }
  }

  // Free the resources we obtained in this function.

  CloseHandle(hFile);
  delete currentTransitionSet;
  return TRUE;  
}
