// This file contains functions for dealing with CPU speeds
// available on this CPU.

#include "stdafx.h"
#include <vector>
#include "../Common/Parameters.h"
#include "Cpu.h"
#include "FidVid.h"
#include "InstallRoutines.h"
#include "RSInitHandle.h"
#include "Speeds.h"

  *** ELIDED DUE TO CONFIDENTIALITY AGREEMENTS ***

/////////////////////////////////////////////////////////////////////////////
// SPEEDS INFO
/////////////////////////////////////////////////////////////////////////////

DWORD GetCurrentCpuKHz (void)
{
  LARGE_INTEGER performanceFrequency, endPC, startPC;
  LONGLONG startTSC, endTSC, cycles;
  double rate, msec;

  QueryPerformanceFrequency(&performanceFrequency);
  rate = 1.0 / performanceFrequency.QuadPart;

  GetTSC(&startTSC);
  QueryPerformanceCounter(&startPC);
  Sleep(2000);
  GetTSC(&endTSC);
  QueryPerformanceCounter(&endPC);
  cycles = endTSC - startTSC;
  msec = (endPC.QuadPart - startPC.QuadPart) * rate * 1000.0;

  return DWORD(cycles / msec);
}

// This function determines what speeds are available on this CPU and
// fills in the list 'availableSpeeds' with them. It returns a boolean
// indicating if it was successful.

BOOL GetAvailableSpeeds (list<AvailableSpeedInfo>& availableSpeeds)
{
  *** ELIDED DUE TO CONFIDENTIALITY AGREEMENTS ***

  // Make sure we found some available speeds for this CPU.

  if (availableSpeeds.size() == 0) {
    MessageBox(NULL, "Could not find performance state table entry for "
               "this CPU.", "Error",
               MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }

  // Make sure there aren't too many available speeds.

  if (availableSpeeds.size() > MAX_NUM_AVAILABLE_SPEEDS) {
    MessageBox(NULL, "There are too many available speeds.", "Error",
               MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }

  // Determine the true speed at each available speed.

  *** ELIDED DUE TO CONFIDENTIALITY AGREEMENTS ***

  // Return success.

  return TRUE;
}

// This function computes the worthwhile settings given a list of
// available speeds, and puts those worthwhile settings in
// 'worthwhileSettings'.

BOOL GetWorthwhileSettings (list<WorthwhileSettingInfo>& worthwhileSettings,
                            list<AvailableSpeedInfo>& availableSpeeds)
{
  // First, convert the available speeds into arrays so that we can
  // easily access entries by index.

  short numAvailableSpeeds = availableSpeeds.size();
  DWORD *avail_khz = new DWORD[numAvailableSpeeds];
  DWORD *avail_mv = new DWORD[numAvailableSpeeds];
  list<AvailableSpeedInfo>::iterator it;
  short i;
  for (i = 0, it = availableSpeeds.begin(); it != availableSpeeds.end();
       ++i, ++it) {
    avail_khz[i] = it->khz;
    avail_mv[i] = it->mv;
  }

  // Compute the first worthwhile setting as the one with the lowest
  // voltage and thus the lowest energy consumption.  Among those
  // tied, prefer ones with higher speeds.

  short best = 0;
  for (i = 1; i < numAvailableSpeeds; ++i)
    if (avail_mv[i] <= avail_mv[best])
      best = i;

  worthwhileSettings.push_back(WorthwhileSettingInfo(best,
                                                     avail_khz[best],
                                                     avail_mv[best]));

  // Loop through the rest of the speeds, each time setting 'last' to
  // the last worthwhile setting and computing the next 'best' worthwhile
  // setting.  This is the one with a higher voltage than the last one
  // which minimizes the slope of the power-vs-speed line passing between
  // it and the last one.  In other words, it's the one that minimizes
  // (P_i - P_last) / (s_i - s_last), preferring ones with higher speeds
  // in case of ties.

  short last;
  double ratio, bestRatio;
  while (1) {
    last = best;

    // Our first guess for the best one is the smallest successor speed
    // that has higher energy consumption.

    for (best = last + 1; best < numAvailableSpeeds; ++best)
      if (avail_mv[best] > avail_mv[last])
        break;

    // If there is no such successor, we're done, so break out of the
    // loop.

    if (best >= numAvailableSpeeds)
      break;

    // Compute the ratio (P_i - P_last) / (s_i - s_last) for this initial
    // guess at the best i.

    bestRatio = (1.0 * avail_mv[best] * avail_mv[best] * avail_khz[best] -
                 1.0 * avail_mv[last] * avail_mv[last] * avail_khz[last]) /
                 (avail_khz[best] - avail_khz[last]);

    // Loop through the remaining speeds looking for one with a better
    // ratio (or an equal ratio, considering that we prefer higher speeds
    // when ratios are tied).

    for (i = best + 1; i < numAvailableSpeeds; ++i) {
      if (avail_mv[i] > avail_mv[last]) {
        ratio = (1.0 * avail_mv[i] * avail_mv[i] * avail_khz[i] -
                 1.0 * avail_mv[last] * avail_mv[last] * avail_khz[last]) /
                 (avail_khz[i] - avail_khz[last]);
        if (ratio <= bestRatio) {
          best = i;
          bestRatio = ratio;
        }
      }
    }

    // Add the new worthwhile setting to the set of worthwhile settings.

    worthwhileSettings.push_back(WorthwhileSettingInfo(best,
                                                       avail_khz[best],
                                                       avail_mv[best]));
  }

  // Free the arrays we allocated and return success.

  delete avail_khz;
  delete avail_mv;
  return TRUE;
}
