// This file contains the main routine for Mkttg, a program that creates
// a task type group file.

#pragma warning ( disable : 4786 )

#include "stdafx.h"
#include <windows.h>
#include <iostream>
#include <list>
#include <string>
#include "../Common/Parameters.h"
#include "../Common/RSLib.h"

using namespace std;

int main(int argc, char* argv[])
{
  // If there aren't enough command-line arguments, print a usage
  // message.

  if (argc < 2) {
    cerr << "Usage: " << argv[0] << " TTG-file-name < descriptions" << endl;
    cerr << "The first line of the descriptions should contain the number "
         << endl
         << "of task types. Each subsequent line should describe a task "
         << endl
         << "type. If fewer lines are provided than the number of task types,"
         << endl
         << "the last line is used for all subsequent task types. If no"
         << endl
         << "descriptions at all are provided, default parameters are"
         << endl
         << "assumed for all task types."
         << endl;
    return -1;
  }

  // Read the first line of input. It must exist since it should contain
  // the number of task types.

  string s;
  if (!getline(cin, s)) {
    cerr << "Error: Empty description set; there should at least be the "
         << "number of task types on the first line." << endl;
    return -1;
  }

  // Convert the first line into a number of task types, and make sure
  // this number is reasonable.

  int numTaskTypes = atoi(s.c_str());
  if (numTaskTypes < 1 || numTaskTypes > MAX_NUM_TASK_TYPES) {
    cerr << "ERROR - Number of task types on first line is " << numTaskTypes
         << ", but it must be between 1 and " << MAX_NUM_TASK_TYPES
         << endl;
    return -1;
  }

  // Read the rest of the input, one line at a time, into a list of
  // strings. As we go along, store pointers to those lines in the
  // 'descriptions' array. So, 'descriptions[i]' will point to a
  // string containing line number i. That string is kept allocated
  // because it is in the list 'strings'.

  LPCSTR *descriptions = new LPCSTR[numTaskTypes];
  list<string> strings;

  for (int i = 0; i < numTaskTypes; ++i) {
    if (getline(cin, s)) {
      strings.push_back(s);
      descriptions[i] = strings.back().c_str();
    }
    else {
      // If we've run out of input lines before we've finished finding
      // descriptions for all tasks, assume that the last description
      // applies to all remaining descriptions. If there never was a
      // last description, i.e., if this is the first task type, just
      // use 0 as the descriptor.

      descriptions[i] = (i > 0) ? descriptions[i-1] : 0;
    }
  }

  // Create the task type group file.

  HANDLE h = RightSpeedCreateFile(argv[1],
                                  CREATE_ALWAYS,
                                  numTaskTypes,
                                  descriptions);
  if (h == INVALID_HANDLE_VALUE) {
    // If we failed, generate an error message.
    char buf[512];
    RightSpeedFormatMessage(GetLastError(), buf, 512);
    buf[0] = tolower(buf[0]);
    cerr << "ERROR - Could not create TTG file because " << buf;
    delete descriptions;
    return -1;
  }

  // We're done, so close the handle of the file we created, delete
  // the descriptions array, and finish.

  CloseHandle(h);
  delete descriptions;
  return 0;
}
