// This file contains routines for performing I/O control requests
// for RSInit.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/MSR.h"
#include "Globals.h"
#include "Log.h"

////////////////////////////////////////////////////////////////////////////
// I/O CONTROL ROUTINES
////////////////////////////////////////////////////////////////////////////

// This function returns the pointer to the global shared state. This
// I/O control call is made from kernel level, so we just copy the
// pointer directly to the output buffer.

NTSTATUS IoctlGetSharedState (
  PCHAR outBuf,
  ULONG outBufLength,
  PULONG returnedBytesPtr
  )
{
  DebugPrint(("RSInit: Responding to get-shared-state request.\n"));
  if (outBufLength < sizeof(SHARED_STATE *))
    return STATUS_INVALID_PARAMETER;

  * (SHARED_STATE **) outBuf = g.sharedState;
  *returnedBytesPtr = sizeof(SHARED_STATE *);
  return STATUS_SUCCESS;
}

// This function maps the kernel-user shared state into the address space
// of the calling process and returns a pointer to it. It records in
// the associated file object enough information to free this mapping
// when the file object is closed.

NTSTATUS IoctlGetKernelUserSharedState (
  PFILE_OBJECT fileObject,
  PCHAR outBuf,
  ULONG outBufLength,
  PULONG returnedBytesPtr
  )
{
  PVOID address;
  PMDL mdl;

  // Make sure there's enough room allocated for the pointer to be
  // returned.

  if (outBufLength < sizeof(PVOID)) {
    DebugPrint(("RSInit: ERROR - Request for user-shared state had "
                "too short an output buffer.\n"));
    return STATUS_INVALID_PARAMETER;
  }

  // Make sure there's an associated file object where we can store
  // the information necessary to undo this mapping when it closes.

  if (!fileObject) {
    DebugPrint(("RSInit: ERROR - Can't share state with user level "
                "because file object is null.\n"));
    if (LOGGING_WARNINGS)
      LogError(RS_INIT_WARNING_CANT_SHARE_STATE_WITH_NULL_FILE_OBJECT);
    return STATUS_UNSUCCESSFUL;
  }

  // If we've already created a mapping for a region of memory using this
  // file object, fail this call.

  if (fileObject->FsContext || fileObject->FsContext2) {
    DebugPrint(("RSInit: ERROR - Can't share state with user level because "
                "we're already mapping memory with this file object.\n"));
    if (LOGGING_WARNINGS)
      LogError(RS_INIT_WARNING_CANT_SHARE_STATE_BECAUSE_ACTIVE_MAPPING);
    return STATUS_UNSUCCESSFUL;
  }

  // Allocate an MDL big enough to map the kernel-user shared state.

  mdl = IoAllocateMdl((PVOID) &g.sharedState->kernelUserSharedState,
                      sizeof(KERNEL_USER_SHARED_STATE),
                      FALSE,  // secondary buffer
                      FALSE,  // charge quota
                      NULL);  // associated IRP
  if (!mdl) {
    DebugPrint(("RSInit: ERROR - Can't get user shared state because "
                "MDL initialization failed.\n"));
    BothLogShort(RS_INIT_ERROR_CANT_ALLOCATE_MDL_FOR_STATE,
                 STATUS_INSUFFICIENT_RESOURCES);
    return STATUS_INSUFFICIENT_RESOURCES;
  }

  // Build the MDL for the kernel-user shared state.

  MmBuildMdlForNonPagedPool(mdl);

  // Map the locked pages into user mode and get the address.
  
  address = MmMapLockedPages(mdl, UserMode);
  if (!address) {
    DebugPrint(("RSInit: ERROR - Can't get user shared state "
                "because MmMapLockedPages failed.\n"));
    BothLogShort(RS_INIT_ERROR_CANT_MAP_SHARED_STATE_TO_USER,
                 STATUS_UNSUCCESSFUL);
    IoFreeMdl(mdl);
    return STATUS_UNSUCCESSFUL;
  }

  // Record the MDL and address in the file object so that the address
  // can be unmapped when the file object is closed.

  fileObject->FsContext = address;
  fileObject->FsContext2 = mdl;

  // If this is NT 4.0, it might be Service Pack 3 or earlier, in which
  // case we have been given the base address rather than the actual
  // address of the kernel-user shared state. So, for NT 4.0, we convert
  // to an address in the proper location on the page.

#if TARGET_NT_VERSION == 4
  address = (PVOID) ((ULONG)PAGE_ALIGN(address) + MmGetMdlByteOffset(mdl));
#endif

  DebugPrint(("RSInit: Mapped user shared state to %x\n", address));

  // Return the address in the output buffer.

  * (PVOID *) outBuf = address;
  *returnedBytesPtr = sizeof(address);

  return STATUS_SUCCESS;
}

// This function logs the event passed in the input buffer. The first
// four bytes are interpreted as the event type.

NTSTATUS IoctlLogEvent (
  PCHAR inBuf,
  ULONG inBufLength
  )
{
  if (inBufLength < 4) {
    DebugPrint(("RSInit: Received invalid log-event request.\n"));
    return STATUS_INVALID_PARAMETER;
  }

  LogAnyEventTypeAndBytes(* (short *) inBuf, &inBuf[4], inBufLength - 4);
  return STATUS_SUCCESS;
}

// This function writes the speed MSR with the 8 bytes passed in the
// input buffer.

NTSTATUS IoctlWriteSpeedMsr (
  PCHAR inBuf,
  ULONG inBufLength
  )
{
  ULONG value1, value2;

  if (inBufLength < 8) {
    DebugPrint(("RSInit: Received invalid write-speed-MSR request.\n"));
    return STATUS_INVALID_PARAMETER;
  }

  value1 = * (ULONG *) inBuf;
  value2 = * (ULONG *) &inBuf[4];

  __asm {
    mov ecx, AMD_MSR_SET_SPEED
    mov eax, value1
    mov edx, value2
    wrmsr
  }

  return STATUS_SUCCESS;
}

// This function satisfies an I/O request to get the FID/VID status.

NTSTATUS IoctlGetFidVidStatus (
  PCHAR outBuf,
  ULONG outBufLength,
  PULONG returnedBytesPtr
  )
{
  // Make sure there's enough room to return the FID/VID status.

  if (outBufLength < 8)
    return STATUS_BUFFER_TOO_SMALL;

  __asm {
    mov ecx, AMD_MSR_GET_FID_VID_STATUS
    rdmsr
    mov ebx, outBuf
    mov [ebx], eax
    mov [ebx+4], edx
  }

  // Indicate the number of bytes returned and return success.

  *returnedBytesPtr = 8;
  return STATUS_SUCCESS;
}

// This function satisfies an I/O request to read from the PowerNow!
// table.

NTSTATUS IoctlReadPowerNowTable (
  PCHAR outBuf,
  ULONG outBufLength,
  PULONG returnedBytesPtr
  )
{
  PCHAR addr, p;
  PHYSICAL_ADDRESS physicalAddress;
  ULONG i;

  // Map the 0x40000 bytes between 0xC0000 and 0xFFFFF in physical
  // memory to 'addr' so that we can scan for the PowerNow state table.

  physicalAddress.HighPart = 0;
  physicalAddress.LowPart = 0xC0000;
  addr = MmMapIoSpace(physicalAddress, 0x40000, MmCached);
  if (addr == NULL) {
    DebugPrint(("RSInit: ERROR - Can't map I/O space at 0xC0000 to find "
                "PowerNow state table.\n"));
    return STATUS_UNSUCCESSFUL;
  }

  // Search through the memory on every paragraph (16-byte) boundary
  // for the beginning signature of the table.

  for (i = 0; i < 0x40000; i += 0x10) {
    if (RtlCompareMemory(addr + i, "AMDK7PNOW!", 10) == 10) {

      // We've found it, so copy the number of bytes requested, but
      // not more than will go past 0xFFFFF.

      if (outBufLength > 0x40000 - i)
        outBufLength = 0x40000 - i;
      RtlCopyMemory(outBuf, addr + i, outBufLength);
      *returnedBytesPtr = outBufLength;

      // Unmap memory and return success.

      MmUnmapIoSpace(addr, 0x40000);
      return STATUS_SUCCESS;
    }
  }

  // The search failed, so unmap memory and return failure.

  DebugPrint(("RSInit: ERROR - Can't find PowerNow state table between "
              "addresses 0xC0000 and 0xFFFFF.\n"));
  MmUnmapIoSpace(addr, 0x40000);
  return STATUS_UNSUCCESSFUL;
}
