// This file contains the main functionality for RSInit, the
// initializer driver for RightSpeed.

#include <ntddk.h>
#include "../Common/Version.h"
#include "../Common/Drivers.h"
#include "../Common/DrvShare.h"
#include "Activate.h"
#include "Globals.h"
#include "Ioctls.h"
#include "Log.h"

////////////////////////////////////////////////////////////////////////////
// GLOBALS
////////////////////////////////////////////////////////////////////////////

RS_INIT_GLOBALS g;

////////////////////////////////////////////////////////////////////////////
// FORWARD DEFINITIONS
////////////////////////////////////////////////////////////////////////////

NTSTATUS DriverEntry (IN PDRIVER_OBJECT  DriverObject,
                      IN PUNICODE_STRING RegistryPath);
NTSTATUS DispatchDeviceControl (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS DispatchCreate (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS DispatchClose (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS DispatchGeneral (IN PDEVICE_OBJECT devObj, IN PIRP irp);
void Unload (PDRIVER_OBJECT driverObject);

void ReleaseResources (void);

char *DummyLogEvent (short type, ULONG length);
void DummyNotifyIosComplete (void);

////////////////////////////////////////////////////////////////////////////
// SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#endif

////////////////////////////////////////////////////////////////////////////
// DRIVER ENTRY
////////////////////////////////////////////////////////////////////////////

// DriverEntry is the installable driver initialization.  Here we set
// up the driver, create our one logger device, and do all sorts of
// miscellaneous initialization.

NTSTATUS DriverEntry (IN PDRIVER_OBJECT  DriverObject,
                      IN PUNICODE_STRING RegistryPath)
{
  NTSTATUS status;
  UNICODE_STRING deviceNameUnicodeString, deviceLinkUnicodeString;
  int i;

  DebugPrint(("RSInit: Entering DriverEntry.\n"));

  // Determine if we should activate RightSpeed. If not, return failure.

  if (!ShouldActivateRightSpeed()) {
    DebugPrint(("RSInit: Not initializing.\n"));
    return STATUS_UNSUCCESSFUL;
  }

  // Initialize "resources obtained" to 0 and save the driver object
  // so that we can deal with errors.

  g.resourcesObtained = 0;
  g.ourDriver = DriverObject;

  // Allocate space for shared state. If we can't, we fail.

  g.sharedState = ExAllocatePool(NonPagedPool, sizeof(SHARED_STATE));
  if (!g.sharedState) {
    DebugPrint(("RSInit: ERROR - Could not allocate shared state.\n"));
    SysLogShort(RS_INIT_ERROR_CANT_ALLOCATE_SHARED_STATE,
                STATUS_INSUFFICIENT_RESOURCES);
    return STATUS_INSUFFICIENT_RESOURCES;
  }
  g.resourcesObtained |= SHARED_STATE_RESOURCE_FLAG;

  DebugPrint(("RSInit: Allocated shared state.\n"));

  // Initialize driver-shared state fields

  KeInitializeSpinLock(&g.sharedState->spinLock);
  g.sharedState->logEventFunc = &DummyLogEvent;
  g.sharedState->numOngoingIos = 0;
  g.sharedState->notifyIosCompleteFunc = &DummyNotifyIosComplete;
  g.sharedState->kernelUserSharedState.loggedEventCategorySet = 0;
  g.sharedState->numOngoingTasks = 0;

  // Set up the device driver entry points.

  DebugPrint(("RSInit: Setting up driver entry points.\n"));

  for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++)
    DriverObject->MajorFunction[i] = DispatchGeneral;
  DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = DispatchDeviceControl;
  DriverObject->MajorFunction[IRP_MJ_CREATE] = DispatchCreate;
  DriverObject->MajorFunction[IRP_MJ_CLOSE] = DispatchClose;

  DriverObject->DriverUnload = Unload;

  // Create the main interface device for RSInit.

  RtlInitUnicodeString(&deviceNameUnicodeString, L"\\Device\\RSInit");
  status = IoCreateDevice(DriverObject,
                          0,  // size of device extension
                          &deviceNameUnicodeString,
                          FILE_DEVICE_RS_INIT,
                          0,  // device characteristics
                          FALSE,  // exclusive access
                          &g.interfaceDeviceObject);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSInit: ERROR - Could not create interface device (%x).\n",
                status));
    SysLogShort(RS_INIT_ERROR_CANT_CREATE_INTERFACE_DEVICE,
                status);
    ReleaseResources();
    return status;
  }
  g.resourcesObtained |= INTERFACE_DEVICE_RESOURCE_FLAG;

  DebugPrint(("RSInit: Created interface device.\n"));

  // Create a link to this device so it can be accessed from user level

  RtlInitUnicodeString(&deviceLinkUnicodeString, L"\\DosDevices\\RSInit");
  status = IoCreateSymbolicLink(&deviceLinkUnicodeString,
                                &deviceNameUnicodeString);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSInit: ERROR - Could not create symbolic link to "
                "interface device (%x).\n", status));
    SysLogShort(RS_INIT_ERROR_CANT_CREATE_INTERFACE_SYMBOLIC_LINK,
                status);
    ReleaseResources();
    return status;
  }
  g.resourcesObtained |= SYMBOLIC_LINK_RESOURCE_FLAG;

  // Return success.

  DebugPrint(("RSInit: Initialized successfully.\n"));
  return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////
// DISPATCH ROUTINES
////////////////////////////////////////////////////////////////////////////

// DispatchDeviceControl is the dispatch routine called to handle any
// device control IRP sent to our RSInit device. So, we check what
// kind of device control request it is and handle it accordingly.

NTSTATUS DispatchDeviceControl (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation(irp);
  ULONG ioControlCode;
  NTSTATUS status;

  // Set the request up as returning no information, by default.

  irp->IoStatus.Information = 0;

  // Get the control code (what kind of request this is)

  ioControlCode = irpStack->Parameters.DeviceIoControl.IoControlCode;

  // Do the appropriate thing, depending on the control code.

  switch (ioControlCode) {
  case RS_INIT_IOCTL_GET_SHARED_STATE:
    status = IoctlGetSharedState(
               irp->UserBuffer,
               irpStack->Parameters.DeviceIoControl.OutputBufferLength,
               &irp->IoStatus.Information
               );
    break;

  case RS_INIT_IOCTL_GET_KERNEL_USER_SHARED_STATE :
    status = IoctlGetKernelUserSharedState(
               irpStack->FileObject,
               irp->AssociatedIrp.SystemBuffer,
               irpStack->Parameters.DeviceIoControl.OutputBufferLength,
               &irp->IoStatus.Information
               );
    break;

  case RS_INIT_IOCTL_LOG_EVENT :
    status = IoctlLogEvent(
               irp->AssociatedIrp.SystemBuffer,
               irpStack->Parameters.DeviceIoControl.InputBufferLength
               );
    break;

  case RS_INIT_IOCTL_WRITE_SPEED_MSR :
    status = IoctlWriteSpeedMsr(
               irp->AssociatedIrp.SystemBuffer,
               irpStack->Parameters.DeviceIoControl.InputBufferLength
               );
    break;

  case RS_INIT_IOCTL_READ_POWERNOW_TABLE :
    status = IoctlReadPowerNowTable(
               irp->AssociatedIrp.SystemBuffer,
               irpStack->Parameters.DeviceIoControl.OutputBufferLength,
               &irp->IoStatus.Information
               );
    break;

  case RS_INIT_IOCTL_GET_FID_VID_STATUS :
    status = IoctlGetFidVidStatus(
               irp->AssociatedIrp.SystemBuffer,
               irpStack->Parameters.DeviceIoControl.OutputBufferLength,
               &irp->IoStatus.Information
               );
    break;

  default :
    status = STATUS_INVALID_DEVICE_REQUEST;
    break;
  }

  // Complete the IRP.

  irp->IoStatus.Status = status;
  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return status;
}

// DispatchCreate is the dispatch routine called to handle any create
// IRP sent to our RSInit device. The main thing to do here is initialize
// the FsContext and FsContext2 fields to NULL so that we can tell at close
// if we've used them.

NTSTATUS DispatchCreate (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation(irp);
  PFILE_OBJECT fileObject = irpStack->FileObject;

  // Set the context values of the file to NULL so that we can recognize
  // if we've set them.

  if (fileObject != NULL) {
    fileObject->FsContext = NULL;
    fileObject->FsContext2 = NULL;  
  }

  // Set the request up as successful.

  irp->IoStatus.Status = STATUS_SUCCESS;
  irp->IoStatus.Information = 0;
    
  // Complete the IRP.

  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

// DispatchClose is the dispatch routine called to handle any close
// IRP sent to our RSInit device. Here we examine the FsContext and
// FsContext2 fields of the file object to determine if we have used this
// file object to map the kernel-user shared state into user space. If we
// have, we now free any resources associated with that mapping.

NTSTATUS DispatchClose (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation(irp);
  PFILE_OBJECT fileObject = irpStack->FileObject;
  PMDL mdl;
  PVOID address;

  // If we created a mapping for the kernel-user shared state using
  // this file object, we free it now.

  if (fileObject != NULL) {
    address = fileObject->FsContext;
    mdl = fileObject->FsContext2;
    if (address != NULL && mdl != NULL) {
      DebugPrint(("RSInit: Unmapping address %x\n", address));
      MmUnmapLockedPages(address, mdl);
      IoFreeMdl(mdl);
    }
  }

  // Set the request up as successful.

  irp->IoStatus.Status = STATUS_SUCCESS;
  irp->IoStatus.Information = 0;
    
  // Complete the IRP.

  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

// DispatchGeneral is the dispatch routine called to handle any
// miscellaneous IRP sent to our RSInit device. We just complete it.

NTSTATUS DispatchGeneral (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  // Set the request up as successful.

  irp->IoStatus.Status = STATUS_SUCCESS;
  irp->IoStatus.Information = 0;
    
  // Complete the IRP.

  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////
// FREEING RESOURCES
////////////////////////////////////////////////////////////////////////////

// This function releases all resources that we've obtained so that the
// driver can be unloaded.

void ReleaseResources (void)
{
  UNICODE_STRING deviceLinkUnicodeString;

  DebugPrint(("RSInit: Releasing resources.\n"));

  // If we have created a symbolic link to our main device object,
  // delete the symbolic link.

  if (g.resourcesObtained & SYMBOLIC_LINK_RESOURCE_FLAG) {
    RtlInitUnicodeString(&deviceLinkUnicodeString, L"\\DosDevices\\RSInit");
    IoDeleteSymbolicLink(&deviceLinkUnicodeString);
    g.resourcesObtained &= ~SYMBOLIC_LINK_RESOURCE_FLAG;
  }

  // If we've created an interface device, delete it.

  if (g.resourcesObtained & INTERFACE_DEVICE_RESOURCE_FLAG) {
    IoDeleteDevice(g.interfaceDeviceObject);
    g.resourcesObtained &= ~INTERFACE_DEVICE_RESOURCE_FLAG;
  }

  // If we allocated shared state, free it.
  
  if (g.resourcesObtained & SHARED_STATE_RESOURCE_FLAG) {
    ExFreePool(g.sharedState);
    g.resourcesObtained &= ~SHARED_STATE_RESOURCE_FLAG;
  }
}

// This is the unload routine for our driver.

void Unload (PDRIVER_OBJECT driverObject)
{
  DebugPrint(("RSInit: Unloading\n"));

  ReleaseResources();

  DebugPrint(("RSInit: Completed unload.\n"));
}

////////////////////////////////////////////////////////////////////////////
// DUMMY SHARED FUNCTIONS
////////////////////////////////////////////////////////////////////////////

// This is a dummy routine for logging events, used when RSLog is not
// running. All it does is return NULL, indicating the event should
// not be traced.

char *DummyLogEvent (short type, ULONG length)
{
  return NULL;
}

// This is a dummy routine for notifying when all I/O's have completed.
// It doesn't do anything. It's just used when RSTask isn't loaded.

void DummyNotifyIosComplete (void)
{
}
