// This file contains fast I/O routines for our driver as well as the
// driver's fast I/O dispatch table. All these routines really do
// is pass on the request. An exception is the fast detach-device
// routine.
//
// (Uses some code by Mark Russinovich and Bryce Cogswell)

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "Detach.h"
#include "Globals.h"
#include "FastIo.h"
#include "Log.h"

///////////////////////////////////////////////////////////////////////////
// FAST I/O DISPATCH ROUTINES
///////////////////////////////////////////////////////////////////////////

// RSIoCntFastIoCheckIfPossible returns true if the real fast I/O routine
// exists and returns true

BOOLEAN
RSIoCntFastIoCheckIfPossible (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN BOOLEAN Wait,
    IN ULONG LockKey,
    IN BOOLEAN CheckForReadOperation,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 8 ||
      origFastIoTable->FastIoCheckIfPossible == NULL)
    return FALSE;

  // Call the original driver's fast I/O dispatch routine.

  return origFastIoTable->FastIoCheckIfPossible(FileObject,
                                                FileOffset,
                                                Length,
                                                Wait,
                                                LockKey,
                                                CheckForReadOperation,
                                                IoStatus,
                                                filteredDevice);
}

// RSIoCntFastIoRead

BOOLEAN
RSIoCntFastIoRead (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN BOOLEAN Wait,
    IN ULONG LockKey,
    OUT PVOID Buffer,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 12 ||
      origFastIoTable->FastIoRead == NULL)
    return FALSE;

  // Call the real fast I/O routine.

  return origFastIoTable->FastIoRead(FileObject,
                                     FileOffset,
                                     Length,
                                     Wait,
                                     LockKey,
                                     Buffer,
                                     IoStatus,
                                     filteredDevice);
}

// RSIoCntFastIoWrite

BOOLEAN
RSIoCntFastIoWrite (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN BOOLEAN Wait,
    IN ULONG LockKey,
    IN PVOID Buffer,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 16 ||
      origFastIoTable->FastIoWrite == NULL)
    return FALSE;

  // Call the real fast I/O routine.

  return origFastIoTable->FastIoWrite(FileObject,
                                      FileOffset,
                                      Length,
                                      Wait,
                                      LockKey,
                                      Buffer,
                                      IoStatus,
                                      filteredDevice);
}

// RSIoCntFastIoQueryBasicinfo

BOOLEAN
RSIoCntFastIoQueryBasicInfo (
    IN struct _FILE_OBJECT *FileObject,
    IN BOOLEAN Wait,
    OUT PFILE_BASIC_INFORMATION Buffer,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 20 ||
      origFastIoTable->FastIoQueryBasicInfo == NULL)
    return FALSE;

  // Call the real fast I/O routine.

  return origFastIoTable->FastIoQueryBasicInfo(FileObject,
                                               Wait,
                                               Buffer,
                                               IoStatus,
                                               filteredDevice);
}

// RSIoCntFastIoQueryStandardInfo

BOOLEAN
RSIoCntFastIoQueryStandardInfo (
    IN struct _FILE_OBJECT *FileObject,
    IN BOOLEAN Wait,
    OUT PFILE_STANDARD_INFORMATION Buffer,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 24 ||
      origFastIoTable->FastIoQueryStandardInfo == NULL)
    return FALSE;

  // Call the real fast I/O routine.

  return origFastIoTable->FastIoQueryStandardInfo(FileObject,
                                                  Wait,
                                                  Buffer,
                                                  IoStatus,
                                                  filteredDevice);
}

// RSIoCntFastIoLock

BOOLEAN
RSIoCntFastIoLock (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN PLARGE_INTEGER Length,
    PEPROCESS ProcessId,
    ULONG Key,
    BOOLEAN FailImmediately,
    BOOLEAN ExclusiveLock,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 28 ||
      origFastIoTable->FastIoLock == NULL)
    return FALSE;

  // Call the original driver's fast I/O routine

  return origFastIoTable->FastIoLock(FileObject,
                                     FileOffset,
                                     Length,
                                     ProcessId,
                                     Key,
                                     FailImmediately,
                                     ExclusiveLock,
                                     IoStatus,
                                     filteredDevice);
}

// RSIoCntFastIoUnlockSingle

BOOLEAN
RSIoCntFastIoUnlockSingle (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN PLARGE_INTEGER Length,
    PEPROCESS ProcessId,
    ULONG Key,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 32 ||
      origFastIoTable->FastIoUnlockSingle == NULL)
    return FALSE;

  // Call the original driver's fast I/O routine

  return origFastIoTable->FastIoUnlockSingle(FileObject,
                                             FileOffset,
                                             Length,
                                             ProcessId,
                                             Key,
                                             IoStatus,
                                             filteredDevice);
}

// RSIoCntFastIoUnlockAll

BOOLEAN
RSIoCntFastIoUnlockAll (
    IN struct _FILE_OBJECT *FileObject,
    PEPROCESS ProcessId,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 36 ||
      origFastIoTable->FastIoUnlockAll == NULL)
    return FALSE;

  // Call the original driver's fast I/O routine.

  return origFastIoTable->FastIoUnlockAll(FileObject,
                                          ProcessId,
                                          IoStatus,
                                          filteredDevice);
}

// RSIoCntFastIoUnlockAllByKey

BOOLEAN
RSIoCntFastIoUnlockAllByKey (
    IN struct _FILE_OBJECT *FileObject,
    PVOID ProcessId,
    ULONG Key,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 40 ||
      origFastIoTable->FastIoUnlockAllByKey == NULL)
    return FALSE;

  // Call the original driver's fast I/O routine

  return origFastIoTable->FastIoUnlockAllByKey(FileObject,
                                               ProcessId,
                                               Key,
                                               IoStatus,
                                               filteredDevice);
}

// RSIoCntFastIoDeviceControl

BOOLEAN
RSIoCntFastIoDeviceControl (
    IN struct _FILE_OBJECT *FileObject,
    IN BOOLEAN Wait,
    IN PVOID InputBuffer OPTIONAL,
    IN ULONG InputBufferLength,
    OUT PVOID OutputBuffer OPTIONAL,
    IN ULONG OutputBufferLength,
    IN ULONG IoControlCode,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 44 ||
      origFastIoTable->FastIoDeviceControl == NULL)
    return FALSE;

  // Call the original driver's fast I/O routine.

  return origFastIoTable->FastIoDeviceControl(FileObject,
                                              Wait,
                                              InputBuffer,
                                              InputBufferLength,
                                              OutputBuffer,
                                              OutputBufferLength,
                                              IoControlCode,
                                              IoStatus,
                                              filteredDevice);
}

// RSIoCntAcquireFileForNtCreateSection

VOID
RSIoCntAcquireFileForNtCreateSection (
    IN struct _FILE_OBJECT *FileObject
    )
{
  PDEVICE_OBJECT deviceObject = IoGetRelatedDeviceObject(FileObject);
  POUR_DEVICE_EXTENSION ourDeviceExtension = deviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 48 ||
      origFastIoTable->AcquireFileForNtCreateSection == NULL)
    return;

  // Call the original driver's fast I/O routine.

  origFastIoTable->AcquireFileForNtCreateSection(FileObject);
}

// RSIoCntReleaseFileForNtCreateSection

VOID
RSIoCntReleaseFileForNtCreateSection (
    IN struct _FILE_OBJECT *FileObject
    )
{
  PDEVICE_OBJECT deviceObject = IoGetRelatedDeviceObject(FileObject);
  POUR_DEVICE_EXTENSION ourDeviceExtension = deviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return;

  // If this routine doesn't exist in the original driver's fast I/O
  // dispatch table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 52 ||
      origFastIoTable->ReleaseFileForNtCreateSection == NULL)
    return;

  // Call the original driver's fast I/O routine.

  origFastIoTable->ReleaseFileForNtCreateSection(FileObject);
}

// RSIoCntFastIoDetachDevice is called when a device that we have hooked
// is being deleted. This happens when, for example, a floppy is formatted.
// We have to detach from it and delete our device.

VOID
RSIoCntFastIoDetachDevice (
    IN struct _DEVICE_OBJECT *SourceDevice,
    IN struct _DEVICE_OBJECT *TargetDevice
    )
{
  // Make sure that the source device really is from our driver.

  if (SourceDevice->DriverObject != g.ourDriver) {
    DebugPrint(("RSIoCnt: ERROR - Received fast I/O detach request for a "
                "device that isn't ours (%x).\n", SourceDevice));
    BothLogShort(RS_IOCNT_ERROR_RECEIVED_FAST_DETACH_FOR_UNRELATED_DEVICE,
                 STATUS_UNSUCCESSFUL);
    return;
  }

  // Detach from the device we're filtering.

  SafeDetachDevice(SourceDevice);

  // Delete our device now as part of the normal fast I/O detach protocol.

  IoDeleteDevice(SourceDevice);

  // There's no need to call the original detach device routine, since
  // it will be called automatically by the I/O manager.
}

BOOLEAN
RSIoCntFastIoQueryNetworkOpenInfo (
    IN struct _FILE_OBJECT *FileObject,
    IN BOOLEAN Wait,
    OUT struct _FILE_NETWORK_OPEN_INFORMATION *Buffer,
    OUT struct _IO_STATUS_BLOCK *IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 60 ||
      origFastIoTable->FastIoQueryNetworkOpenInfo == NULL)
    return FALSE;

  // Call the real fast I/O routine.

  return origFastIoTable->FastIoQueryNetworkOpenInfo(FileObject,
                                                     Wait,
                                                     Buffer,
                                                     IoStatus,
                                                     filteredDevice);
}

// RSIoCntMdlRead

BOOLEAN
RSIoCntMdlRead (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN ULONG LockKey,
    OUT PMDL *MdlChain,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 68 ||
      origFastIoTable->MdlRead == NULL)
    return FALSE;

  // Call the real fast I/O routine.

  return origFastIoTable->MdlRead(FileObject,
                                  FileOffset,
                                  Length,
                                  LockKey,
                                  MdlChain,
                                  IoStatus,
                                  filteredDevice);
}

// RSIoCntMdlReadComplete

BOOLEAN
RSIoCntMdlReadComplete (
    IN struct _FILE_OBJECT *FileObject,
    IN PMDL MdlChain,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 72 ||
      origFastIoTable->MdlReadComplete == NULL)
    return FALSE;

  // Call the original driver's fast I/O routine.

  return origFastIoTable->MdlReadComplete(FileObject,
                                          MdlChain,
                                          filteredDevice);
}

// RSIoCntFastIoPrepareMdlWrite

BOOLEAN
RSIoCntPrepareMdlWrite (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN ULONG LockKey,
    OUT PMDL *MdlChain,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // For some reason, NtFilemon had the following code, and we figure
  // it can't hurt.

  IoStatus->Status = STATUS_NOT_SUPPORTED;
  IoStatus->Information = 0;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 76 ||
      origFastIoTable->PrepareMdlWrite == NULL)
    return FALSE;

  // Call the original driver's fast I/O routine

  return origFastIoTable->PrepareMdlWrite(FileObject,
                                          FileOffset,
                                          Length,
                                          LockKey,
                                          MdlChain,
                                          IoStatus,
                                          filteredDevice);
}

// RSIoCntMdlWriteComplete

BOOLEAN
RSIoCntMdlWriteComplete (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN PMDL MdlChain,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 80 ||
      origFastIoTable->MdlWriteComplete == NULL)
    return FALSE;

  // Call the original driver's fast I/O routine.

  return origFastIoTable->MdlWriteComplete(FileObject,
                                           FileOffset,
                                           MdlChain,
                                           filteredDevice);
}

// RSIoCntAcquireForModWrite

NTSTATUS
RSIoCntAcquireForModWrite (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER EndingOffset,
    OUT struct _ERESOURCE **ResourceToRelease,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return STATUS_NOT_SUPPORTED;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return STATUS_NOT_SUPPORTED.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 64 ||
      origFastIoTable->AcquireForModWrite == NULL)
    return STATUS_NOT_SUPPORTED;

  // Call the original driver's fast I/O routine

  return origFastIoTable->AcquireForModWrite(FileObject,
                                             EndingOffset,
                                             ResourceToRelease,
                                             filteredDevice);
}

// RSIoCntReleaseForModWrite

NTSTATUS
RSIoCntReleaseForModWrite (
    IN struct _FILE_OBJECT *FileObject,
    IN struct _ERESOURCE *ResourceToRelease,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return STATUS_NOT_SUPPORTED;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return STATUS_NOT_SUPPORTED.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 104 ||
      origFastIoTable->ReleaseForModWrite == NULL)
    return STATUS_NOT_SUPPORTED;

  // Call the original driver's fast I/O routine.

  return origFastIoTable->ReleaseForModWrite(FileObject,
                                             ResourceToRelease,
                                             filteredDevice);
}

// RSIoCntAcquireForCcFlush

NTSTATUS
RSIoCntAcquireForCcFlush (
    IN struct _FILE_OBJECT *FileObject,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return STATUS_NOT_SUPPORTED;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return STATUS_NOT_SUPPORTED.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 108 ||
      origFastIoTable->AcquireForCcFlush == NULL)
    return STATUS_NOT_SUPPORTED;

  // Call the original driver's fast I/O routine.

  return origFastIoTable->AcquireForCcFlush(FileObject,
                                            filteredDevice);
}

// RSIoCntFastIoReleaseForCcFlush

NTSTATUS
RSIoCntReleaseForCcFlush (
    IN struct _FILE_OBJECT *FileObject,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return STATUS_NOT_SUPPORTED;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return STATUS_NOT_SUPPORTED.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 112 ||
      origFastIoTable->ReleaseForCcFlush == NULL)
    return STATUS_NOT_SUPPORTED;

  // Call the original driver's fast I/O routine.

  return origFastIoTable->ReleaseForCcFlush(FileObject,
                                            filteredDevice);
}

// RSIoCntFastIoReadCompressed

BOOLEAN
RSIoCntFastIoReadCompressed (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN ULONG LockKey,
    OUT PVOID Buffer,
    OUT PMDL *MdlChain,
    OUT PIO_STATUS_BLOCK IoStatus,
    OUT struct _COMPRESSED_DATA_INFO *CompressedDataInfo,
    IN ULONG CompressedDataInfoLength,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O
  // dispatch table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 84 ||
      origFastIoTable->FastIoReadCompressed == NULL)
    return FALSE;

  // Call the real fast I/O routine.

  return origFastIoTable->FastIoReadCompressed(FileObject,
                                               FileOffset,
                                               Length,
                                               LockKey,
                                               Buffer,
                                               MdlChain,
                                               IoStatus,
                                               CompressedDataInfo,
                                               CompressedDataInfoLength,
                                               filteredDevice);
}

// RSIoCntFastIoWriteCompressed

BOOLEAN
RSIoCntFastIoWriteCompressed (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN ULONG LockKey,
    IN PVOID Buffer,
    OUT PMDL *MdlChain,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _COMPRESSED_DATA_INFO *CompressedDataInfo,
    IN ULONG CompressedDataInfoLength,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 88 ||
      origFastIoTable->FastIoWriteCompressed == NULL)
    return FALSE;

  // Call the real fast I/O routine.

  return origFastIoTable->FastIoWriteCompressed(FileObject,
                                                FileOffset,
                                                Length,
                                                LockKey,
                                                Buffer,
                                                MdlChain,
                                                IoStatus,
                                                CompressedDataInfo,
                                                CompressedDataInfoLength,
                                                filteredDevice);
}

// RSIoCntMdlReadCompleteCompressed

BOOLEAN
RSIoCntMdlReadCompleteCompressed (
    IN struct _FILE_OBJECT *FileObject,
    IN PMDL MdlChain,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 92 ||
      origFastIoTable->MdlReadCompleteCompressed == NULL)
    return FALSE;

  // Call the original driver's fast I/O routine

  return origFastIoTable->MdlReadCompleteCompressed(FileObject,
                                                    MdlChain,
                                                    filteredDevice);
}

// RSIoCntMdlWriteCompleteCompressed

BOOLEAN
RSIoCntMdlWriteCompleteCompressed (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN PMDL MdlChain,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 96 ||
      origFastIoTable->MdlWriteCompleteCompressed == NULL)
    return FALSE;

  // Call the original driver's fast I/O routine

  return origFastIoTable->MdlWriteCompleteCompressed(FileObject,
                                                     FileOffset,
                                                     MdlChain,
                                                     filteredDevice);
}

// RSIoCntFastIoQueryOpen (This call actually passes an IRP!)

BOOLEAN
RSIoCntFastIoQueryOpen (
    IN struct _IRP *Irp,
    OUT PFILE_NETWORK_OPEN_INFORMATION NetworkInformation,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;
  PFAST_IO_DISPATCH origFastIoTable;
  PIO_STACK_LOCATION irpSp;
  BOOLEAN retVal;

  if (!filteredDevice)
    return FALSE;

  // If this routine doesn't exist in the original driver's fast I/O
  // dispatch table, return false.

  origFastIoTable = filteredDevice->DriverObject->FastIoDispatch;
  if (!origFastIoTable ||
      origFastIoTable->SizeOfFastIoDispatch < 100 ||
      origFastIoTable->FastIoQueryOpen == NULL)
    return FALSE;

  // Replace the device in the IRP current stack location with the
  // filtered device.

  irpSp = IoGetCurrentIrpStackLocation(Irp);
  irpSp->DeviceObject = filteredDevice;

  // Call the real fast I/O routine.

  retVal = origFastIoTable->FastIoQueryOpen(Irp,
                                            NetworkInformation,
                                            filteredDevice);

  // If the call was not successful, put the device object of
  // the IRP stack location back the way it was.

  if (!retVal)
    irpSp->DeviceObject = DeviceObject;

  // Return what the real fast I/O routine returned.

  return retVal;
}

///////////////////////////////////////////////////////////////////////////
// DISPATCH TABLE
///////////////////////////////////////////////////////////////////////////

// This is RSIoCnt's fast I/O dispatch table. Note that NT assumes that
// file system drivers support some fast I/O calls, so this table must
// be present for a file system filter driver.

FAST_IO_DISPATCH FastIoDispatchTable = {
  sizeof(FAST_IO_DISPATCH),
  RSIoCntFastIoCheckIfPossible,
  RSIoCntFastIoRead,
  RSIoCntFastIoWrite,
  RSIoCntFastIoQueryBasicInfo,
  RSIoCntFastIoQueryStandardInfo,
  RSIoCntFastIoLock,
  RSIoCntFastIoUnlockSingle,
  RSIoCntFastIoUnlockAll,
  RSIoCntFastIoUnlockAllByKey,
  RSIoCntFastIoDeviceControl,
  RSIoCntAcquireFileForNtCreateSection,
  RSIoCntReleaseFileForNtCreateSection,
  RSIoCntFastIoDetachDevice,
  RSIoCntFastIoQueryNetworkOpenInfo,      // From here on is new for NT 4.0
  RSIoCntAcquireForModWrite,
  RSIoCntMdlRead,
  RSIoCntMdlReadComplete,
  RSIoCntPrepareMdlWrite,
  RSIoCntMdlWriteComplete,
  RSIoCntFastIoReadCompressed,
  RSIoCntFastIoWriteCompressed,
  RSIoCntMdlReadCompleteCompressed,
  RSIoCntMdlWriteCompleteCompressed,
  RSIoCntFastIoQueryOpen,
  RSIoCntReleaseForModWrite,
  RSIoCntAcquireForCcFlush,
  RSIoCntReleaseForCcFlush
  };
