// This file contains routines for filtering file system and file
// system recognizer devices.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/ErrCodes.h"
#include "Globals.h"
#include "Detach.h"
#include "Log.h"

////////////////////////////////////////////////////////////////////////////
// FILTERING
////////////////////////////////////////////////////////////////////////////

// This function attaches the given filter device to the given device
// to be filtered.

NTSTATUS SafeAttachDevice (PDEVICE_OBJECT filterDevice,
                           PDEVICE_OBJECT filteredDevice)
{
  PDEVICE_OBJECT layeredDevice;
  POUR_DEVICE_EXTENSION ourDeviceExtension = filterDevice->DeviceExtension;
  POUR_DEVICE_EXTENSION layeredDeviceExtension;

  // Acquire the lock to perform this attachment.

  ExAcquireResourceExclusiveLite(&g.deviceFilteringLock, TRUE);

  // First, make sure we're not one of the devices already layered on
  // top of the given device to filter.

  for (layeredDevice = filteredDevice;
       layeredDevice != NULL;
       layeredDevice = layeredDevice->AttachedDevice) {
    if (layeredDevice->DriverObject == g.ourDriver) {
      ExReleaseResourceLite(&g.deviceFilteringLock);
      DebugPrint(("RSIoCnt: WARNING - Would have attached to our own device "
                  "%x already attached to %x if it weren't for this check.\n",
                  layeredDevice, filteredDevice));
      if (LOGGING_WARNINGS) {
        layeredDeviceExtension = layeredDevice->DeviceExtension;
        LogErrorTwoParameters(RS_IOCNT_WARNING_ATTEMPT_TO_ATTACH_TWICE,
                              filteredDevice->DeviceType,
                              layeredDeviceExtension->ourDeviceType);
      }
      return RIGHTSPEED_ERROR_FS_ALREADY_FILTERED;
    }
  }

  ourDeviceExtension->filteredDevice =
    IoAttachDeviceToDeviceStack(filterDevice, filteredDevice);

  // Release the lock now that we've attached.

  ExReleaseResourceLite(&g.deviceFilteringLock);

  // Make sure we actually succeeded in the attach.

  if (ourDeviceExtension->filteredDevice == NULL) {
    DebugPrint(("RSIoCnt: Failed to attach device %x to device %x.\n",
                filterDevice, filteredDevice));
    BothLogOneParameter(RS_IOCNT_ERROR_CANT_ATTACH_TO_DEVICE,
                        STATUS_UNSUCCESSFUL,
                        filteredDevice->DeviceType);
    return STATUS_UNSUCCESSFUL;
  }

  // Return success.

  return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////
// NOTIFICATION CALLBACK FUNCTIONS
////////////////////////////////////////////////////////////////////////////

// This function gets called whenever a device is registering or
// deregistering as a file system recognizer. The boolean parameter
// indicates if it's registering (as opposed to deregistering).

void NotifyFsRegistrationChange (PDEVICE_OBJECT fsRecognizerDevice,
                                 BOOLEAN registering)
{
  NTSTATUS status;
  PDEVICE_OBJECT recognizerFilterDevice, layeredDevice;
  POUR_DEVICE_EXTENSION recognizerFilterDeviceExtension;

  DebugPrint(("RSIoCnt: Notified that the file system recognizer %x of type "
              "%lu is %s.\n",
              fsRecognizerDevice,
              fsRecognizerDevice->DeviceType,
              registering ? "registering" : "unregistering"));

  if (registering) {
    // If this device is registering as a file system, we create a new
    // RSIoCnt device to attach to it. Note that we must *not* use the
    // same device type as the filtered device, or else things like SIS
    // (the single-instance store service in Windows 2000) may assume we
    // are actually one of those file systems and treat us as such.

    status = IoCreateDevice(g.ourDriver,
                            sizeof(OUR_DEVICE_EXTENSION),
                            NULL,    // device name
                            FILE_DEVICE_FILE_SYSTEM,
                            0,       // device characteristics
                            FALSE,   // exclusive?
                            &recognizerFilterDevice);
    if (!NT_SUCCESS(status)) {
      DebugPrint(("RSIoCnt: ERROR - Could not create a device to filter the "
                  "file system recognizer device %x (%x).\n",
                  fsRecognizerDevice, status));
      BothLogShort(RS_IOCNT_ERROR_CANT_CREATE_RECOGNIZER_FILTER_DEVICE,
                  status);
      return;
    }

    // Initialize the device extension.

    recognizerFilterDeviceExtension = recognizerFilterDevice->DeviceExtension;
    recognizerFilterDeviceExtension->ourDeviceType =
                                         OUR_DEVICE_TYPE_FS_RECOGNIZER_FILTER;
    recognizerFilterDeviceExtension->filteredDevice = NULL;

    // Attach to the file system recognizer.

    DebugPrint(("RSIoCnt: About to filter recognizer %x with %x.\n",
                fsRecognizerDevice, recognizerFilterDevice));

    status = SafeAttachDevice(recognizerFilterDevice, fsRecognizerDevice);
    if (!NT_SUCCESS(status)) {
      DebugPrint(("RSIoCnt: Failed to attach %x to %x (%x).\n",
                  recognizerFilterDevice, fsRecognizerDevice, status));
      BothLogShort(RS_IOCNT_ERROR_CANT_ATTACH_TO_RECOGNIZER,
                   status);
      IoDeleteDevice(recognizerFilterDevice);
    }
    else {
      recognizerFilterDevice->Flags &= ~DO_DEVICE_INITIALIZING;
      DebugPrint(("RSIoCnt: Successfully started filtering %x.\n",
                  recognizerFilterDeviceExtension->filteredDevice));
    }
  }
  else {
    // If this device is deregistering as a file system, we should detach
    // ourselves from it. First, we need to find which device belongs to us
    // and is attached.

    for (recognizerFilterDevice = fsRecognizerDevice->AttachedDevice;
         recognizerFilterDevice != NULL;
         recognizerFilterDevice = recognizerFilterDevice->AttachedDevice) {

      if (recognizerFilterDevice->DriverObject == g.ourDriver) {
        // Once we find the attached device that's from our driver...

        // Make sure that this device really is a file system recognizer
        // filter.

        recognizerFilterDeviceExtension =
          recognizerFilterDevice->DeviceExtension;
        if (recognizerFilterDeviceExtension->ourDeviceType !=
            OUR_DEVICE_TYPE_FS_RECOGNIZER_FILTER) {
          DebugPrint(("RSIoCnt: ERROR - The device attached to the "
                      "deregistering file system is of the wrong type "
                      "(%lu).\n",
                      recognizerFilterDeviceExtension->ourDeviceType));
          BothLogShort(RS_IOCNT_ERROR_WRONG_TYPE_ATTACHED_TO_RECOGNIZER,
                      STATUS_UNSUCCESSFUL);
          return;
        }
        else {
          DebugPrint(("RSIoCnt: Determined that the recognizer that's "
                      "deregistering is attached to by our recognizer filter "
                      "device %x.\n", recognizerFilterDevice));
        }

        // Detach from the file system recognizer we're attached to, and,
        // if this is successful, delete our now-useless device.

        if (SafeDetachDevice(recognizerFilterDevice))
          SafeDeleteDevice(recognizerFilterDevice);

        // Now that we've found a match, return. We don't want to
        // continue the loop anyway because the next thing it would do
        // is to access the AttachedDevice field of the device we may
        // have just deleted.

        return;
      }
    }
  }
}

////////////////////////////////////////////////////////////////////////////
// FILTERING FILE SYSTEMS
////////////////////////////////////////////////////////////////////////////

// FilterFileSystem creates and attaches a new RSIoCnt device to the
// given file system device.

NTSTATUS FilterFileSystem (PDEVICE_OBJECT fsDevice)
{
  PDEVICE_OBJECT filterDevice;
  POUR_DEVICE_EXTENSION ourDeviceExtension;
  NTSTATUS status;

  DebugPrint(("RSIoCnt: Filtering file system device %x of type %d.\n",
              fsDevice, fsDevice->DeviceType));

  // Make sure the device exists.

  if (fsDevice == NULL) {
    DebugPrint(("RSIoCnt: ERROR - Attempt to filter a nonexistent file "
                "system.\n"));
    BothLogShort(RS_IOCNT_ERROR_CANT_ATTACH_TO_NULL_FILE_SYSTEM,
                STATUS_INVALID_PARAMETER);
    return STATUS_INVALID_PARAMETER;
  }

  // Create a new RSIoCnt device to attach to the file system.

  status = IoCreateDevice(g.ourDriver,
                          sizeof(OUR_DEVICE_EXTENSION),
                          NULL,   // device name
                          fsDevice->DeviceType,
                          0,      // device characteristics
                          FALSE,  // exclusive?
                          &filterDevice);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSIoCnt: ERROR - Could not create device to filter "
                "file system (%x).\n", status));
    BothLogShort(RS_IOCNT_ERROR_CANT_CREATE_FILE_SYSTEM_FILTER_DEVICE,
                status);
    return status;
  }

  // Initialize our device extension.
  
  ourDeviceExtension = filterDevice->DeviceExtension;
  ourDeviceExtension->ourDeviceType = OUR_DEVICE_TYPE_FS_FILTER;
  ourDeviceExtension->filteredDevice = NULL;

  // Attach to the given file system device.

  status = SafeAttachDevice(filterDevice, fsDevice);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSIoCnt: ERROR - Failed to attach device %x to device %x.\n",
                filterDevice, fsDevice));
    BothLogOneParameter(RS_IOCNT_ERROR_CANT_ATTACH_TO_FILE_SYSTEM,
                        status,
                        fsDevice->DeviceType);
    IoDeleteDevice(filterDevice);
    return status;
  }

  // Determine the highest-level device we actually attached to.

  fsDevice = ourDeviceExtension->filteredDevice;
  DebugPrint(("RSIoCnt: Successfully attached filter device %x to "
              "device %x.\n", filterDevice, fsDevice));

  // Configure our device to match the device we attached to.

  if (fsDevice->Flags & DO_BUFFERED_IO)
    filterDevice->Flags |= DO_BUFFERED_IO;
  if (fsDevice->Flags & DO_DIRECT_IO)
    filterDevice->Flags |= DO_DIRECT_IO;

  // Indicate we're no longer initializing our filter device.
  
  filterDevice->Flags &= ~DO_DEVICE_INITIALIZING;

  // Return success.

  return STATUS_SUCCESS;
}
