// This file contains functions for obtaining the shared state pointer
// from RSInit.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/DrvShare.h"
#include "GetState.h"
#include "Globals.h"
#include "Log.h"

////////////////////////////////////////////////////////////////////////////
// FORWARD DEFINITIONS
////////////////////////////////////////////////////////////////////////////

NTSTATUS GetSharedStateCompletionRoutine (
  PDEVICE_OBJECT DeviceObject,
  PIRP Irp,
  PVOID Context
  );

////////////////////////////////////////////////////////////////////////////
// SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, GetSharedState)
#pragma alloc_text (INIT, GetSharedStateCompletionRoutine)
#endif

////////////////////////////////////////////////////////////////////////////
// CONNECTING TO RSINIT
////////////////////////////////////////////////////////////////////////////

// GetSharedState gets the driver-shared state from the RSInit driver by
// sending it a device control IRP.

NTSTATUS GetSharedState (void)
{
  PIRP                    irp;
  PIO_STACK_LOCATION      irpStack;
  IO_STATUS_BLOCK         ioStatusBlock;
  KEVENT                  event;
  UNICODE_STRING          deviceNameUnicodeString;
  PDEVICE_OBJECT          deviceObject;
  NTSTATUS                status;

  // Get a device object pointer to the RSInit device. Save a file
  // object pointer to it in g.RSInitFileObject to ensure that RSInit
  // sticks around until we dereference it.

  RtlInitUnicodeString(&deviceNameUnicodeString, L"\\Device\\RSInit");
  status = IoGetDeviceObjectPointer(&deviceNameUnicodeString,
                                    FILE_ALL_ACCESS,
                                    &g.RSInitFileObject,
                                    &deviceObject);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSIoCnt: Could not get device pointer for RSInit (%x).\n",
                status));
    SysLogShort(RS_IOCNT_ERROR_CANT_GET_RSINIT_DEVICE_POINTER,
                status);
    return status;
  }
  g.resourcesObtained |= RSINIT_FILE_OBJECT_RESOURCE_FLAG;

  // Allocate an IRP for the device request.  We only need a stack
  // size of 1, since RSInit isn't going to pass the IRP on.

  irp = IoAllocateIrp(1, FALSE);
  if (!irp) {
    DebugPrint(("RSIoCnt: Could not allocate IRP to get shared state.\n"));
    SysLogShort(RS_IOCNT_ERROR_CANT_ALLOCATE_IRP_FOR_SHARED_STATE,
                STATUS_INSUFFICIENT_RESOURCES);
    return STATUS_INSUFFICIENT_RESOURCES;
  }

  ///////////////////////////////////////
  // Initialize the fields in the IRP. //
  ///////////////////////////////////////

  // Set the system buffer parameter to the address of g.sharedState, since
  // that is where we want the driver-shared state put.

  irp->UserBuffer = &g.sharedState;

  // Set the user IO status block to the address of ioStatusBlock so that
  // we can have access to the return value.

  irp->UserIosb = &ioStatusBlock;

  // Initialize an event, and make it the user event of the IRP, so that
  // we can know when the IRP is completed.

  KeInitializeEvent(&event, SynchronizationEvent, FALSE);
  irp->UserEvent = &event;

  // Set other miscellaneous IRP info.

  irp->Tail.Overlay.Thread = PsGetCurrentThread();
  irp->RequestorMode = KernelMode;

  // Set information in the first stack location indicating what type
  // of request this is, and other info.

  irpStack = IoGetNextIrpStackLocation(irp);
  irpStack->MajorFunction = IRP_MJ_DEVICE_CONTROL;
  irpStack->MinorFunction = 0;
  irpStack->Parameters.DeviceIoControl.Type3InputBuffer = NULL;
  irpStack->Parameters.DeviceIoControl.InputBufferLength = 0;
  irpStack->Parameters.DeviceIoControl.OutputBufferLength =
                                                       sizeof(SHARED_STATE *);
  irpStack->Parameters.DeviceIoControl.IoControlCode =
                                               RS_INIT_IOCTL_GET_SHARED_STATE;
  irpStack->DeviceObject = deviceObject;

  // Set our completion routine so that we're called when the IRP completes.

  IoSetCompletionRoutine(irp, GetSharedStateCompletionRoutine, 0, TRUE,
                         TRUE, TRUE);

  // Pass on the IRP.

  IoCallDriver(deviceObject, irp);

  // Wait for the IRP to complete.

  while (KeWaitForSingleObject(&event, Executive, KernelMode, TRUE, 0) !=
         STATUS_SUCCESS)
    ;

  // Status is returned in ioStatusBlock, so check it to see how the
  // request went.

  if (!NT_SUCCESS(ioStatusBlock.Status)) {
    DebugPrint(("RSIoCnt: Couldn't read shared state (%x).\n",
                ioStatusBlock.Status));
    SysLogShort(RS_IOCNT_ERROR_CANT_GET_SHARED_STATE,
                ioStatusBlock.Status);
  }
  return ioStatusBlock.Status;
}

// GetSharedStateCompletionRoutine is the completion routine called
// when the GetSharedState IRP completes. (This IRP gets the shared
// state pointer from RSInit.) The main job of this completion routine
// is to free the IRP we allocated.

NTSTATUS GetSharedStateCompletionRoutine (
  PDEVICE_OBJECT DeviceObject,
  PIRP Irp,
  PVOID Context
  )
{
  // Copy status information to user status buffer.

  *Irp->UserIosb = Irp->IoStatus;

  // Set event so that the routine waiting on this IRP can continue.

  KeSetEvent(Irp->UserEvent, 0, FALSE);

  // Free the IRP.

  IoFreeIrp(Irp);

  // Return STATUS_MORE_PROCESSING_REQUIRED to indicate that this driver
  // is taking control of the IRP.  This is necessary because the IRP has
  // just been freed, so we don't want anyone else trying to work on it.
  // Indeed, we have just finished processing the IRP by freeing it.

  return STATUS_MORE_PROCESSING_REQUIRED;
}
