// This file contains routines pertaining to the log buffer.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/DrvShare.h"
#include "Globals.h"
#include "Log.h"
#include "Buffer.h"
#include "StartStop.h"
#include "TSC.h"

/////////////////////////////////////////////////////////////////////////////
// INITIALIZATION AND FREEING
/////////////////////////////////////////////////////////////////////////////

// This function initializes the log buffer.

NTSTATUS InitializeLogBuffer (PLOG_BUFFER_INFO info)
{
  info->bufferSize = g.registryInfo.logBufferSize;
  info->allocatedSpace = ExAllocatePool(NonPagedPool, info->bufferSize);
  if (!info->allocatedSpace) {
    DebugPrint(("RSLog: Could not allocate log buffer.\n"));
    BothLogShort(RS_LOG_ERROR_CANT_ALLOCATE_LOG_BUFFER,
                 STATUS_INSUFFICIENT_RESOURCES);
    return STATUS_INSUFFICIENT_RESOURCES;
  }

  info->endOfAllocatedSpace = info->allocatedSpace + info->bufferSize;
  info->startPos = info->allocatedSpace;
  info->endPos = info->allocatedSpace;
  info->unflushedBytes = 0;
  info->wrapped = FALSE;
  info->emptySpaceAtEnd = 0;
  info->dangerousLength = info->bufferSize * 3 / 4;

  return STATUS_SUCCESS;
}

// This function frees the log buffer.

void FreeLogBuffer (PLOG_BUFFER_INFO info)
{
  ExFreePool(info->allocatedSpace);
}

/////////////////////////////////////////////////////////////////////////////
// LOGGING
/////////////////////////////////////////////////////////////////////////////

// This function writes everything but the data part of a new log
// entry into the log, and returns a pointer to the place to write
// the data part. It returns NULL if it fails. The passed parameters
// indicate the type of event and the length of data to reserve.
//
// It is assumed that the spin lock g.sharedState->spinLock is held
// when this routine is called.

PCHAR LogEvent (short eventType, ULONG dataLength)
{
  ULONG entryLength;
  PLOG_BUFFER_INFO info = &g.logBufferInfo;
  PCHAR dataPtr;
  LARGE_INTEGER currentTime;

  // If logging is disabled, or if the amount of data is too long,
  // return NULL immediately.

  if (g.whyLoggingDisabled || dataLength > MAX_LOG_ENTRY_DATA_LENGTH)
    return NULL;

  // Determine the time of the event.  We must put this value in
  // aligned memory, rather than directly into the log.

  KeQuerySystemTime(&currentTime);

  // Compute the total entry length as the data length plus 11. This
  // extra amount is 8 bytes for the timestamp count, 2 bytes for the
  // event type, and one byte for the data length.

  entryLength = dataLength + 11;

  // If we need to wrap past the end of the buffer, do this now. Only
  // do it if we haven't already wrapped around. (If we have already
  // wrapped around, we'll soon find out that the buffer overflowed,
  // so there's no need to do anything now.)

  if (info->endPos + entryLength > info->endOfAllocatedSpace &&
      !info->wrapped) {

    info->wrapped = TRUE;
    info->emptySpaceAtEnd = info->endOfAllocatedSpace - info->endPos;
    info->unflushedBytes += info->emptySpaceAtEnd;
    info->endPos = info->allocatedSpace;

  }

  // Determine if we're in danger of running out of room, and if
  // we have actually run out of room. In either case, signal the
  // log file thread to wake up and flush data from the log buffer.
  // Note that if we are indicating the buffer overflowed, we must
  // do this *before* signalling the thread to wake up, so that
  // when it wakes up it will notice the buffer overflowed and
  // eventually remove that reason for not logging.

  if (info->unflushedBytes > info->dangerousLength) {
    if (info->unflushedBytes + entryLength > info->bufferSize) {
      AddReasonsForNotLogging(BECAUSE_BUFFER_OVERFLOWED);
      KeSetEvent(&g.logFileInfo.waitEvent,
                 IO_NO_INCREMENT,
                 FALSE);  // wait
      return NULL;
    }
    else
      KeSetEvent(&g.logFileInfo.waitEvent,
                 IO_NO_INCREMENT,
                 FALSE);  // wait
  }

  // Fill in the new entry.

  RtlCopyMemory(info->endPos, &currentTime, 8);
  * (short *) &info->endPos[8] = eventType;
  info->endPos[10] = (unsigned char) dataLength;
  dataPtr = &info->endPos[11];

  // Reserve space for the entry.

  info->unflushedBytes += entryLength;
  info->endPos += entryLength;

  // Return the pointer to the data part of the entry.

  return dataPtr;
}

/////////////////////////////////////////////////////////////////////////////
// FLUSHING
/////////////////////////////////////////////////////////////////////////////

// This function simply writes the given number of bytes at the given
// location to the given file handle. It returns a status value.

NTSTATUS SimpleWriteFile (HANDLE fileHandle, PCHAR buf, ULONG bufLength)
{
  IO_STATUS_BLOCK ioStatusBlock;
  NTSTATUS status;

  // Perform the write.

  status = ZwWriteFile(fileHandle,
                       NULL, // event
                       NULL, // APC routine
                       NULL, // APC context
                       &ioStatusBlock,
                       buf,
                       bufLength,
                       NULL,  // byte offset: default to current position
                       NULL); // key
  if (!NT_SUCCESS(status))
    BothLogShort(RS_LOG_ERROR_FLUSH_FAILED,
                 status);

  return status;
}

// This function flushes the given number of log bytes from the buffer
// to the given log file handle. If 'wrapped' is true, then this number
// of log bytes wraps around the end of the buffer, and we need not
// flush the empty space at the end of the buffer.

NTSTATUS FlushLogBytes (
  HANDLE fileHandle,
  ULONG bytesToFlush,
  BOOLEAN wrapped
  )
{
  ULONG bytesToWrite;
  KIRQL oldirql;
  NTSTATUS status;
  PLOG_BUFFER_INFO info = &g.logBufferInfo;

  // Do different things depending on whether the bytes to flush are
  // wrapped around the end of the buffer.

  if (wrapped) {

    //////////////////////////////////
    //////////// WRAPPED /////////////
    //////////////////////////////////

    // First, write the bytes from the start position to the end of
    // the buffer, excluding the empty space left at the end.

    bytesToWrite = info->endOfAllocatedSpace - info->startPos -
                   info->emptySpaceAtEnd;
    status = SimpleWriteFile(fileHandle, info->startPos, bytesToWrite);
    if (!NT_SUCCESS(status))
      return status;

    // Next, write the remaining bytes to flush, starting from the
    // beginning of the buffer.

    bytesToWrite = bytesToFlush - bytesToWrite - info->emptySpaceAtEnd;
    status = SimpleWriteFile(fileHandle, info->allocatedSpace, bytesToWrite);
    if (!NT_SUCCESS(status))
      return status;

    // Finally, update the buffer information to reflect the new start
    // position of the unflushed bytes as well as the lesser number of
    // unflushed bytes and the fact that these bytes are no longer wrapped
    // past the end of the buffer.

    KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
    info->startPos = info->allocatedSpace + bytesToWrite;
    info->unflushedBytes -= bytesToFlush;
    info->wrapped = FALSE;
    info->emptySpaceAtEnd = 0;
    KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);
  }
  else {

    /////////////////////////////////////
    ///////////// UNWRAPPED /////////////
    /////////////////////////////////////

    // Write the number of bytes requested.
    
    status = SimpleWriteFile(fileHandle, info->startPos, bytesToFlush);
    if (!NT_SUCCESS(status))
      return status;

    // Update the log buffer information to reflect the new place that
    // the unflushed bytes start as well as the reduced number of
    // unflushed bytes.

    KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
    info->startPos += bytesToFlush;
    info->unflushedBytes -= bytesToFlush;
    KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);
  }

  return STATUS_SUCCESS;
}
