// This file contains routines implementing I/O control requests.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/DrvShare.h"
#include "Globals.h"
#include "Ioctls.h"
#include "Log.h"
#include "LogFile.h"
#include "StartStop.h"
#include "RegInfo.h"

// This function logs the event passed. The first four bytes of the buffer
// are considered the event type, and the rest are considered the event
// data.

NTSTATUS IoctlLogEvent (
  PCHAR inBuf,
  ULONG inBufLength
  )
{
  // Make sure there are at least four bytes for the event type.

  if (inBufLength < 4)
    return STATUS_INVALID_PARAMETER;

  LogAnyEventTypeAndBytes(* (short *) inBuf, &inBuf[4], inBufLength - 4);

  return STATUS_SUCCESS;
}

// This function stops logging for the reason "because the user
// requested it".

NTSTATUS IoctlStopLogging (void)
{
  KIRQL oldirql;

  KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
  AddReasonsForNotLogging(BECAUSE_USER_REQUESTED);
  KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);

  return STATUS_SUCCESS;
}

// This function removes the reason "because the user requested it"
// from the reasons to not log.

NTSTATUS IoctlStartLogging (void)
{
  KIRQL oldirql;

  KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
  RemoveReasonsForNotLogging(BECAUSE_USER_REQUESTED);
  KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);

  return STATUS_SUCCESS;
}

// This function starts logging in a new log file.

NTSTATUS IoctlStartNewLogFile (void)
{
  // If there is a log file currently active, terminate it first.

  TerminateCurrentLogFile();

  // Start the next log file.

  return StartNextLogFile();
}

// This function terminates the current log file.

NTSTATUS IoctlTerminateCurrentLogFile (void)
{
  return TerminateCurrentLogFile();
}

// This function signals the log file thread to flush the log buffer
// to the log file.

NTSTATUS IoctlFlush (void)
{
  KeSetEvent(&g.logFileInfo.waitEvent,
             IO_NO_INCREMENT,
             FALSE);  // wait

  return STATUS_SUCCESS;
}

// This function reloads the registry information.

NTSTATUS IoctlReloadRegistry (void)
{
  return ReloadRegistryInfo(&g.registryInfo);
}
