// This file contains routines for the log-flushing thread of RSLog.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/ErrCodes.h"
#include "../Common/Version.h"
#include "Globals.h"
#include "Buffer.h"
#include "Log.h"
#include "LogFile.h"
#include "StartStop.h"
#include "TSC.h"

////////////////////////////////////////////////////////////////////////////
// MACROS
////////////////////////////////////////////////////////////////////////////

// The following maximum log file name length is the main directory's
// maximum length plus enough room for a leading \??\, a trailing
// \Logs\R4000000000.log, and an extra couple of bytes just for safety:

#define MAX_LOG_FILE_NAME_LENGTH     (MAX_RIGHTSPEED_DIRECTORY_LENGTH + 26)

////////////////////////////////////////////////////////////////////////////
// FORWARD DEFINITIONS
////////////////////////////////////////////////////////////////////////////

void LogFileThreadRoutine (PVOID StartContext);

NTSTATUS InitializeLogFile (HANDLE *fileHandlePtr);
NTSTATUS GetNextLogFileNumber (ULONG *nextLogFileNumberPtr);
NTSTATUS FillInLogFileName (
  PUNICODE_STRING logFileName,
  ULONG logFileNumber
  );
NTSTATUS OpenFileForWriting (
  PUNICODE_STRING fileName,
  HANDLE *fileHandlePtr
  );
NTSTATUS WritePreamble (HANDLE fileHandle);

NTSTATUS StartNextLogFileUnchecked (PLOG_FILE_INFO info);
NTSTATUS TerminateCurrentLogFileUnchecked (PLOG_FILE_INFO info);

////////////////////////////////////////////////////////////////////////////
// THREAD MANAGEMENT
////////////////////////////////////////////////////////////////////////////

// This function initializes handling of log files.

void InitializeLogFileInfo (PLOG_FILE_INFO info)
{
  // Initialize the event the thread will wait for.

  KeInitializeSpinLock(&info->spinLock);
  KeInitializeEvent(&info->waitEvent,
                    SynchronizationEvent,
                    FALSE);
  info->threadObject = NULL;
  info->threadStartingOrStopping = FALSE;
}

// This function closes resources obtained pertaining to log files.

void CloseLogFileInfo (PLOG_FILE_INFO info)
{
  TerminateCurrentLogFile();
}

// This function begins a new log file.

NTSTATUS StartNextLogFile (void)
{
  PLOG_FILE_INFO info = &g.logFileInfo;
  NTSTATUS status = STATUS_SUCCESS;
  KIRQL oldirql;

  // Make sure that there is no thread ongoing or being started or
  // stopped. If this check turns out fine, indicate that a thread is
  // being started or stopped.

  KeAcquireSpinLock(&info->spinLock, &oldirql);
  if (info->threadObject != NULL)
    status = RIGHTSPEED_ERROR_LOG_FILE_ACTIVE;
  else if (info->threadStartingOrStopping)
    status = RIGHTSPEED_ERROR_LOG_FILE_STARTING_OR_STOPPING;
  else
    info->threadStartingOrStopping = TRUE;
  KeReleaseSpinLock(&info->spinLock, oldirql);

  // If one of the checks failed, return the status code.

  if (!NT_SUCCESS(status))
    return status;

  // Actually start the next log file.

  status = StartNextLogFileUnchecked(info);

  // Indicate that a thread is no longer being started or stopped.

  KeAcquireSpinLock(&info->spinLock, &oldirql);
  info->threadStartingOrStopping = FALSE;
  KeReleaseSpinLock(&info->spinLock, oldirql);

  // Return the status value from the actual call to start the next
  // log file.

  return status;
}

// This function terminates the current log file.

NTSTATUS TerminateCurrentLogFile (void)
{
  PLOG_FILE_INFO info = &g.logFileInfo;
  NTSTATUS status = STATUS_SUCCESS;
  BOOLEAN threadExists = TRUE;
  KIRQL oldirql;

  // Make sure that there is a thread ongoing but not being started or
  // stopped. If this is the case, indicate that a thread is being
  // started or stopped.

  KeAcquireSpinLock(&info->spinLock, &oldirql);
  if (info->threadObject == NULL)
    threadExists = FALSE;
  else if (info->threadStartingOrStopping)
    status = RIGHTSPEED_ERROR_LOG_FILE_STARTING_OR_STOPPING;
  else
    info->threadStartingOrStopping = TRUE;
  KeReleaseSpinLock(&info->spinLock, oldirql);

  // If one of the checks failed or if there is no thread, return the
  // status code.

  if (!NT_SUCCESS(status) || !threadExists)
    return status;

  // Actually terminate the current log file.

  status = TerminateCurrentLogFileUnchecked(info);

  // Indicate that a thread is no longer being started or stopped.

  KeAcquireSpinLock(&info->spinLock, &oldirql);
  info->threadStartingOrStopping = FALSE;
  KeReleaseSpinLock(&info->spinLock, oldirql);

  // Return the status value from the actual call to terminate the
  // current log file.

  return status;
}

// This function starts the thread that will write the next log
// file. It returns a status value indicating whether it succeeded.
// It is "unchecked" in that it makes no checks to ensure that the
// thread info is not being used by someone else simultaneously.

NTSTATUS StartNextLogFileUnchecked (PLOG_FILE_INFO info)
{
  OBJECT_ATTRIBUTES objectAttributes;
  CLIENT_ID threadId;
  NTSTATUS status;
  HANDLE threadHandle;

  DebugPrint(("RSLog: Creating thread from process %d.\n",
              PsGetCurrentProcessId()));

  // Create a new system thread to run the log file thread routine.

  InitializeObjectAttributes(&objectAttributes,
                             NULL,  // object name
                             OBJ_KERNEL_HANDLE, // attributes
                             NULL,  // root directory
                             NULL); // security descriptor
  status = PsCreateSystemThread(&threadHandle,
                                THREAD_ALL_ACCESS,
                                &objectAttributes,
                                NULL,
                                &threadId,
                                LogFileThreadRoutine,
                                NULL);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: ERROR - Could not create log file thread "
                "(%x).\n", status));
    BothLogShort(RS_LOG_ERROR_CANT_CREATE_LOG_FILE_THREAD,
                 status);
    return status;
  }

  // Obtain a pointer to the thread object to reference it.

  status = ObReferenceObjectByHandle(threadHandle,
                                     SYNCHRONIZE,
                                     NULL,
                                     KernelMode,
                                     &info->threadObject,
                                     NULL);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: ERROR - Could not get thread object pointer (%x).\n",
                status));
    BothLogShort(RS_LOG_ERROR_CANT_GET_THREAD_OBJECT_POINTER,
                 status);
    return status;
  }

  DebugPrint(("RSLog: Created log file thread with handle %x and "
              "pointer %x.\n", threadHandle, info->threadObject));

  // Close the handle to the thread since we don't need it any more.

  status = ZwClose(threadHandle);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: ERROR - Could not close thread handle (%x).\n",
                status));
    BothLogShort(RS_LOG_ERROR_CANT_CLOSE_THREAD_HANDLE,
                 status);
  }

  // Indicate success.

  return STATUS_SUCCESS;
}

// This function tells the log file thread to flush and terminate the
// current log file, then to terminate itself. It is "unchecked" in
// that it makes no checks to ensure that such a thread is ongoing,
// and that no other simultaneous access is being made to the thread
// information structures it will be manipulating.

NTSTATUS TerminateCurrentLogFileUnchecked (PLOG_FILE_INFO info)
{
  NTSTATUS status;
  KIRQL oldirql;

  // Indicate that the end of the log has been reached.

  KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
  AddReasonsForNotLogging(BECAUSE_END_REACHED);
  KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);

  // Set the event the thread is waiting for so that it can wake up and
  // notice it should flush and then end.

  KeSetEvent(&info->waitEvent, IO_NO_INCREMENT, FALSE);

  // Wait for the thread to terminate.

  DebugPrint(("RSLog: Waiting for thread to terminate.\n"));

  do {
    status = KeWaitForSingleObject(info->threadObject,
                                   Executive,
                                   KernelMode,
                                   TRUE,  // alertable
                                   NULL); // no timeout
  } while (status == STATUS_ALERTED || status == STATUS_USER_APC);

  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: Wait for thread to terminate failed (%x).\n",
                status));
    BothLogShort(RS_LOG_ERROR_WAIT_FOR_THREAD_FAILED,
                 status);
  }

  DebugPrint(("RSLog: Finished waiting for thread to terminate.\n"));

  // Dereference the thread object since we don't need it any more.
  // Set the thread object to NULL to indicate there is no thread any
  // more.

  ObDereferenceObject(info->threadObject);
  info->threadObject = NULL;

  // Remove "the end of the log was reached" as a reason to not log.

  KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
  RemoveReasonsForNotLogging(BECAUSE_END_REACHED);
  KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);

  // Return success.

  return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////
// THREAD ROUTINE
////////////////////////////////////////////////////////////////////////////

// This is the main routine for the log file thread. It creates the
// log file, writes the preamble information to it, then loops waiting
// to be signalled to know when it should flush the log buffer to the
// file. It ends when the log is finished.

void LogFileThreadRoutine (PVOID StartContext)
{
  NTSTATUS status;
  BOOLEAN done = FALSE, wrapped;
  HANDLE fileHandle;
  ULONG whyLoggingDisabled, bytesToFlush;
  KIRQL oldirql;

  DebugPrint(("RSLog: Thread created in process %d.\n",
              PsGetCurrentProcessId()));

  // Initialize the log file. If we fail, terminate the thread. If we
  // succeed, remove "no log file" from the set of reasons to not log.

  status = InitializeLogFile(&fileHandle);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: ERROR - Can't initialize log file, so terminating "
                "log file thread (%x).\n", status));
    PsTerminateSystemThread(status);
    return;
  }
  else {
    KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
    RemoveReasonsForNotLogging(BECAUSE_NO_LOG_FILE);
    KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);
  }

  // Loop continuously until we are done.

  while (!done) {

    // Wait to be signalled.

    status = KeWaitForSingleObject(&g.logFileInfo.waitEvent,
                                   Executive,
                                   KernelMode,
                                   TRUE,  // alertable
                                   NULL); // no timeout
    if (status == STATUS_WAIT_0) {

      // If we were signalled, determine why we were signalled and how
      // many bytes are unflushed and whether those bytes wrap past the
      // end of the log buffer.

      KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
      whyLoggingDisabled = g.whyLoggingDisabled;
      bytesToFlush = g.logBufferInfo.unflushedBytes;
      wrapped = g.logBufferInfo.wrapped;
      KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);

      // Try to flush that many bytes.

      if (!NT_SUCCESS(FlushLogBytes(fileHandle, bytesToFlush, wrapped))) {

        // If we can't, indicate that flushing failed.

        KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
        AddReasonsForNotLogging(BECAUSE_FLUSH_FAILED);
        KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);
      }
      else if (whyLoggingDisabled & BECAUSE_BUFFER_OVERFLOWED) {

        // If we succeeded in flushing, and we were signalled because
        // the buffer overflowed, indicate the buffer is no longer
        // overflowed.
        
        KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
        RemoveReasonsForNotLogging(BECAUSE_BUFFER_OVERFLOWED);
        LogAnyEventTypeAndBytes(EVENT_TYPE_BUFFER_OVERFLOWED, &oldirql, 0);
        KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);
      }

      // If we were signalled because the log was complete, we're done.

      if (whyLoggingDisabled & BECAUSE_END_REACHED)
        done = TRUE;

    }
    else if (!NT_SUCCESS(status)) {
      DebugPrint(("RSLog: ERROR - Thread wait failed (%x).\n", status));
      BothLogShort(RS_LOG_ERROR_LOG_FILE_THREAD_WAIT_FAILED,
                   status);
      done = TRUE;
    }
  }

  // Close the log file.

  ZwClose(fileHandle);

  // Indicate that there is currently no log file.

  KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
  AddReasonsForNotLogging(BECAUSE_NO_LOG_FILE);
  KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);

  // Exit the thread.

  DebugPrint(("RSLog: Exiting log file thread successfully.\n"));
  PsTerminateSystemThread(STATUS_SUCCESS);
}

////////////////////////////////////////////////////////////////////////////
// LOG FILE INITIALIZATION
////////////////////////////////////////////////////////////////////////////

// This function initializes the next log file and puts a handle to it
// in the given pointer. It returns a status value.

NTSTATUS InitializeLogFile (HANDLE *fileHandlePtr)
{
  NTSTATUS status;
  ULONG logFileNumber;
  UNICODE_STRING logFileName;
  WCHAR logFileNameBuffer[MAX_LOG_FILE_NAME_LENGTH] = L"";;

  // Obtain the number for the next log file.

  status = GetNextLogFileNumber(&logFileNumber);
  if (!NT_SUCCESS(status))
    return status;

  // Fill in the log file name using this number.

  logFileName.Length = 0;
  logFileName.MaximumLength = MAX_LOG_FILE_NAME_LENGTH * 2; // sizeof(WCHAR)==2
  logFileName.Buffer = logFileNameBuffer;
  status = FillInLogFileName(&logFileName, logFileNumber);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: ERROR - Could not make log file name (%x).\n",
                status));
    BothLogShort(RS_LOG_ERROR_CANT_FILL_IN_LOG_FILE_NAME,
                 status);
    return status;
  }

  // Open the log file of this name for writing.

  status = OpenFileForWriting(&logFileName, fileHandlePtr);
  if (!NT_SUCCESS(status))
    return status;

  // Write the log file preamble to the file.

  status = WritePreamble(*fileHandlePtr);
  if (!NT_SUCCESS(status)) {
    ZwClose(*fileHandlePtr);
    return status;
  }

  // Return success.

  return STATUS_SUCCESS;
}

// This function obtains and reserves the next log file number using the
// registry. It puts the next log file number in the pointer location
// passed. It returns a status value.

NTSTATUS GetNextLogFileNumber (ULONG *nextLogFileNumberPtr)
{
  struct {
    ULONG   TitleIndex;
    ULONG   Type;
    ULONG   DataLength;
    ULONG   DataDword;
  } keyValueInfo;
  HANDLE keyHandle;
  ULONG keyValueInfoLength, lastLogFileNumber;
  UNICODE_STRING miscUnicodeString;
  OBJECT_ATTRIBUTES objectAttributes;
  NTSTATUS status;

  // Open the registry key for RightSpeed.

  RtlInitUnicodeString(&miscUnicodeString,
    L"\\Registry\\Machine\\Software\\Jay Lorch\\RightSpeed\\CurrentVersion");
  InitializeObjectAttributes(&objectAttributes,
                             &miscUnicodeString,
                             OBJ_CASE_INSENSITIVE,
                             NULL,  // root directory
                             NULL); // security descriptor
  status = ZwOpenKey(&keyHandle,
                     KEY_QUERY_VALUE | KEY_SET_VALUE,
                     &objectAttributes);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: ERROR - Could not open main registry key (%x).\n",
                status));
    BothLogShortRepeat(RS_LOG_ERROR_CANT_OPEN_MAIN_REGISTRY_KEY,
                       status,
                       2);
    return status;
  }

  // Read the last log file number.

  RtlInitUnicodeString(&miscUnicodeString, L"LastLogNumber");
  status = ZwQueryValueKey(keyHandle,
                           &miscUnicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (status == STATUS_SUCCESS && keyValueInfo.Type == REG_DWORD)
    lastLogFileNumber = keyValueInfo.DataDword;
  else
    lastLogFileNumber = 0;

  // Compute the next log file number and set it in the registry as the
  // last file number.

  *nextLogFileNumberPtr = lastLogFileNumber + 1;
  status = ZwSetValueKey(keyHandle,
                         &miscUnicodeString,
                         0, // title index: must be 0
                         REG_DWORD,
                         nextLogFileNumberPtr,
                         4);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: ERROR - Could not write next log file number to "
                "registry (%x).\n", status));
    BothLogShort(RS_LOG_ERROR_CANT_WRITE_NEXT_LOG_FILE_NUMBER,
                 status);
    ZwClose(keyHandle);
    return status;
  }

  // Close the registry key now that we're done with it.

  ZwClose(keyHandle);

  // Return success.
  
  return STATUS_SUCCESS;
}

// This function fills in the log file name given the log file number.
// It assumes it is passed a Unicode string that starts out empty, and
// it fills it up. It returns a status value.

NTSTATUS FillInLogFileName (
  PUNICODE_STRING logFileName,
  ULONG logFileNumber
  )
{
  NTSTATUS status;
  WCHAR numberBuf[12] = L"";
  UNICODE_STRING numberUnicodeString;

  // Convert the log file number to a Unicode string.

  numberUnicodeString.Length = 0;
  numberUnicodeString.MaximumLength = 24; // 12 * sizeof(WCHAR)
  numberUnicodeString.Buffer = numberBuf;
  status = RtlIntegerToUnicodeString(logFileNumber,
                                     10,
                                     &numberUnicodeString);
  if (!NT_SUCCESS(status))
    return status;

  // Start the log file name with \??\.

  status = RtlAppendUnicodeToString(logFileName, L"\\??\\");
  if (!NT_SUCCESS(status))
    return status;

  // Next, append the main directory string.
  
  status = RtlAppendUnicodeStringToString(logFileName,
                                          &g.registryInfo.mainDirectory);
  if (!NT_SUCCESS(status))
    return status;

  // Next, append \Logs\R.
  
  status = RtlAppendUnicodeToString(logFileName, L"\\Logs\\R");
  if (!NT_SUCCESS(status))
    return status;

  // Next, append the log file number.

  status = RtlAppendUnicodeStringToString(logFileName,
                                          &numberUnicodeString);
  if (!NT_SUCCESS(status))
    return status;

  // Finally, append .log.
  
  status = RtlAppendUnicodeToString(logFileName, L".log");

  // Return the status.
  
  return status;
}

// This function opens the file of the given name for writing and puts
// a handle to it in the given parameter location. It returns a status
// value.

NTSTATUS OpenFileForWriting (
  PUNICODE_STRING fileName,
  HANDLE *fileHandlePtr
  )
{
  OBJECT_ATTRIBUTES objectAttributes;
  IO_STATUS_BLOCK ioStatusBlock;
  NTSTATUS status;

  // Create the file.

  InitializeObjectAttributes(&objectAttributes,
                             fileName,
                             OBJ_CASE_INSENSITIVE,
                             NULL, // root directory
                             NULL); // security descriptor
  status = ZwCreateFile(fileHandlePtr,
                        FILE_WRITE_DATA | SYNCHRONIZE,
                        &objectAttributes,
                        &ioStatusBlock,
                        0, // allocation size
                        FILE_ATTRIBUTE_NORMAL,
                        FILE_SHARE_READ,
                        FILE_CREATE,
                        FILE_WRITE_THROUGH | FILE_SEQUENTIAL_ONLY |
                        FILE_SYNCHRONOUS_IO_ALERT,
                        NULL, // extended attributes buffer
                        0); // extended attributes length
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: ERROR - Could not create log file (%x).\n", status));
    BothLogShort(RS_LOG_ERROR_CANT_CREATE_LOG_FILE,
                 status);
    return status;
  }

  // Return success.

  return STATUS_SUCCESS;
}

// This function writes the log file preamble to the log file whose
// handle is passed.

NTSTATUS WritePreamble (HANDLE fileHandle)
{
  LOG_FILE_PREAMBLE preamble;
  IO_STATUS_BLOCK ioStatusBlock;
  NTSTATUS status;

  // Fill in the preamble data structure.

  preamble.version = g.registryInfo.version;
  preamble.cpuType = RIGHTSPEED_CPU_TYPE;
  preamble.maxCycPerMs = g.registryInfo.maxCycPerMs;
  GetTSC(&preamble.startTSC);
  KeQuerySystemTime(&preamble.startGMTTime);
  ExSystemTimeToLocalTime(&preamble.startGMTTime, &preamble.startLocalTime);

  // Write the preamble to the file.

  status = ZwWriteFile(fileHandle,
                       NULL,  // event
                       NULL,  // APC routine,
                       NULL,  // APC context
                       &ioStatusBlock,
                       &preamble,
                       sizeof(preamble),
                       NULL,  // byte offset: default to current position
                       NULL); // key
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: Could not write preamble of log file (%x).\n",
                status));
    BothLogShort(RS_LOG_ERROR_CANT_WRITE_LOG_FILE_PREAMBLE,
                 status);
    return status;
  }

  // Return success.

  return STATUS_SUCCESS;
}
