// This file contains the main functionality of the RSLog driver.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/DrvShare.h"
#include "Globals.h"
#include "Buffer.h"
#include "GetState.h"
#include "Ioctls.h"
#include "Log.h"
#include "LogFile.h"
#include "RegInfo.h"
#include "StartStop.h"

////////////////////////////////////////////////////////////////////////////
//                               GLOBALS
////////////////////////////////////////////////////////////////////////////

// All the global variables are in one neat structure:

RS_LOG_GLOBALS g;

////////////////////////////////////////////////////////////////////////////
//                           FORWARD DEFINES
////////////////////////////////////////////////////////////////////////////

NTSTATUS DriverEntry (IN PDRIVER_OBJECT  DriverObject,
                      IN PUNICODE_STRING RegistryPath);
NTSTATUS DispatchDeviceControl (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS DispatchGeneral (IN PDEVICE_OBJECT devObj, IN PIRP irp);
void Unload (PDRIVER_OBJECT driverObject);

void ReleaseResources (void);

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#endif

////////////////////////////////////////////////////////////////////////////
//                             DRIVER ENTRY
////////////////////////////////////////////////////////////////////////////

// DriverEntry is the installable driver initialization.  Here we set
// up the driver, create our one logger device, and do all sorts of
// miscellaneous initialization.

NTSTATUS DriverEntry (IN PDRIVER_OBJECT  DriverObject,
                      IN PUNICODE_STRING RegistryPath)
{
  NTSTATUS status;
  UNICODE_STRING deviceNameUnicodeString, miscUnicodeString;
  KIRQL oldirql;
  int i;

  DebugPrint(("RSLog: Entering DriverEntry.\n"));

  // Initialize resources-obtained field so that we can know what resources
  // have ben obtained if we need to free them in case of error. Also
  // initialize the driver object, which is needed to log errors.

  g.resourcesObtained = 0;
  g.ourDriver = DriverObject;

  // Get a pointer to the shared state from RSInit. If we can't get
  // it, RightSpeed is deactivated, and we should exit. (Or something
  // even more wrong is going on, in which case we should also exit.)

  status = GetSharedState();
  if (!NT_SUCCESS(status)) {
    ReleaseResources();
    return status;
  }

  // Initialize logging to "disabled because RSLog has not yet loaded".

  g.whyLoggingDisabled = BECAUSE_NO_LOG_FILE;

  // Read registry information.

  status = ReadRegistryInfo(&g.registryInfo);
  if (!NT_SUCCESS(status)) {
    ReleaseResources();
    return status;
  }
  g.resourcesObtained |= REGISTRY_INFO_RESOURCE_FLAG;

  // Initialize the log buffer.

  status = InitializeLogBuffer(&g.logBufferInfo);
  if (!NT_SUCCESS(status)) {
    ReleaseResources();
    return status;
  }
  g.resourcesObtained |= LOG_BUFFER_RESOURCE_FLAG;

  // Initialize shared-state hooks.

  KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
  g.replacedLogEventFunc = g.sharedState->logEventFunc;
  g.sharedState->logEventFunc = &LogEvent;
  KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);
  g.resourcesObtained |= SHARED_STATE_HOOKS_RESOURCE_FLAG;

  // Initialize the log file info.

  InitializeLogFileInfo(&g.logFileInfo);
  g.resourcesObtained |= LOG_FILE_RESOURCE_FLAG;

  // Start the next log file.

  status = StartNextLogFile();
  if (!NT_SUCCESS(status)) {
    ReleaseResources();
    return status;
  }

  // Create dispatch points for all the IRPs we handle.

  for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; ++i)
    DriverObject->MajorFunction[i] = DispatchGeneral;
  DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = DispatchDeviceControl;

  DriverObject->DriverUnload = Unload;

  // Create the one and only RSLog interface device.

  RtlInitUnicodeString(&deviceNameUnicodeString, L"\\Device\\RSLog");
  status = IoCreateDevice(DriverObject,
                          0, // device extension size
                          &deviceNameUnicodeString,
                          FILE_DEVICE_RS_LOG,
                          0, // device characteristics
                          FALSE, // exclusive
                          &g.interfaceDeviceObject);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: ERROR - Could not create interface device (%x).\n",
                status));
    BothLogShort(RS_LOG_ERROR_CANT_CREATE_INTERFACE_DEVICE,
                 status);
    ReleaseResources();
    return status;
  }
  g.resourcesObtained |= INTERFACE_DEVICE_RESOURCE_FLAG;

  // Create a symbolic link for the RSLog device so that it can be accessed
  // from user level.

  RtlInitUnicodeString(&miscUnicodeString, L"\\DosDevices\\RSLog");
  status = IoCreateSymbolicLink(&miscUnicodeString,
                                &deviceNameUnicodeString);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: ERROR - Could not create symbolic link to "
                "interface device (%x).\n", status));
    BothLogShort(RS_LOG_ERROR_CANT_CREATE_INTERFACE_SYMBOLIC_LINK,
                 status);
    ReleaseResources();
    return status;
  }
  g.resourcesObtained |= SYMBOLIC_LINK_RESOURCE_FLAG;

  DebugPrint(("RSLog: Successfully initialized.\n"));

  return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////
//                           RESOURCE MANAGEMENT
////////////////////////////////////////////////////////////////////////////

// ReleaseResources releases any resources we have allocated
// for RSLog so far. The flags in g.resourcesObtained lists those
// resources that have been allocated.

void ReleaseResources (void)
{
  UNICODE_STRING deviceLinkUnicodeString;
  KIRQL oldirql;

  if (g.resourcesObtained & SYMBOLIC_LINK_RESOURCE_FLAG) {
    RtlInitUnicodeString(&deviceLinkUnicodeString, L"\\DosDevices\\RSLog");
    IoDeleteSymbolicLink(&deviceLinkUnicodeString);
    g.resourcesObtained &= ~SYMBOLIC_LINK_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & INTERFACE_DEVICE_RESOURCE_FLAG) {
    IoDeleteDevice(g.interfaceDeviceObject);
    g.resourcesObtained &= ~INTERFACE_DEVICE_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & LOG_FILE_RESOURCE_FLAG) {
    CloseLogFileInfo(&g.logFileInfo);
    g.resourcesObtained &= ~LOG_FILE_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & SHARED_STATE_HOOKS_RESOURCE_FLAG) {
    KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
    g.sharedState->logEventFunc = g.replacedLogEventFunc;
    g.sharedState->kernelUserSharedState.loggedEventCategorySet = 0;
    KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);
    g.resourcesObtained &= ~SHARED_STATE_HOOKS_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & REGISTRY_INFO_RESOURCE_FLAG) {
    FreeRegistryInfo(&g.registryInfo);
    g.resourcesObtained &= ~REGISTRY_INFO_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & LOG_BUFFER_RESOURCE_FLAG) {
    FreeLogBuffer(&g.logBufferInfo);
    g.resourcesObtained &= ~LOG_BUFFER_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & RSINIT_FILE_OBJECT_RESOURCE_FLAG) {
    ObDereferenceObject(g.RSInitFileObject);
    g.resourcesObtained &= ~RSINIT_FILE_OBJECT_RESOURCE_FLAG;
  }
}

////////////////////////////////////////////////////////////////////////////
//                          DISPATCH ROUTINES
////////////////////////////////////////////////////////////////////////////

// DispatchDeviceControl is the dispatch routine called to handle any
// device control IRP sent to our RSLog device. The main thing to do
// here is check what kind of device control request it is and handle
// it accordingly.

NTSTATUS DispatchDeviceControl (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  PIO_STACK_LOCATION  irpStack = IoGetCurrentIrpStackLocation(irp);
  ULONG               ioControlCode;
  NTSTATUS            status;

  // Set the request up as returning no information, by default.

  irp->IoStatus.Information = 0;

  // Get the control code (what kind of request this is)

  ioControlCode = irpStack->Parameters.DeviceIoControl.IoControlCode;

  // Do the appropriate thing, depending on the control code

  switch (ioControlCode) {
  case RS_LOG_IOCTL_LOG_EVENT :
    status = IoctlLogEvent(
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.InputBufferLength
      );
    break;

  case RS_LOG_IOCTL_STOP_LOGGING :
    status = IoctlStopLogging();
    break;

  case RS_LOG_IOCTL_START_LOGGING :
    status = IoctlStartLogging();
    break;

  case RS_LOG_IOCTL_START_NEW_LOG_FILE :
    status = IoctlStartNewLogFile();
    break;

  case RS_LOG_IOCTL_TERMINATE_CURRENT_LOG_FILE :
    status = IoctlTerminateCurrentLogFile();
    break;

  case RS_LOG_IOCTL_FLUSH :
    status = IoctlFlush();
    break;

  case RS_LOG_IOCTL_RELOAD_REGISTRY :
    status = IoctlReloadRegistry();
    break;

  default :
    status = STATUS_INVALID_DEVICE_REQUEST;
    break;
  }

  // Complete the IRP.

  irp->IoStatus.Status = status;
  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return status;
}

// DispatchGeneral is the dispatch routine called to handle most IRP
// types sent to our RSLog device. The main thing to do here is to
// simply complete the IRP.

NTSTATUS DispatchGeneral (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  // Set the request up as successful.

  irp->IoStatus.Status = STATUS_SUCCESS;
  irp->IoStatus.Information = 0;
    
  // Complete the IRP.

  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

// This is the unload routine for our driver. Here we release any
// resources obtained by the driver before we unload.

void Unload (PDRIVER_OBJECT driverObject)
{
  DebugPrint(("RSLog: Unloading.\n"));

  ReleaseResources();

  DebugPrint(("RSLog: Completed unload.\n"));
}
