#include "stdafx.h"
#include <ctime>
#include <string>
#include "filetime.h"
#include "nsconv.h"
#include "typeconv.h"

// This file contains definitions for the FileTime class member functions.

using Windows::FileTime;

/////////////////////////////////////////////////////////////////////////////
// The local namespace for this file
/////////////////////////////////////////////////////////////////////////////

namespace {

  // The FileTimeFlags namespace contains bit flags useful when writing
  // a FileTime to a file or reading it back from such a file.

  namespace FileTimeFlags {
    const int KNOWN         = 1;
    const int ABSOLUT       = 2;
    const int INFINIT       = 4;
  };

  // The following conversion constants come in handy:

  const int SEC_PER_HOUR = 3600;
  const unsigned long NS_PER_SEC = 1000000000;
  const unsigned long NS_PER_MS = 1000000;

  // The number of hundred-nanosecond units between January 1, 1601
  // (the base of NT time) and January 1, 1970 (the base of Unix time,
  // and also the base of absolute Nanosec's) is 0x19db1ded53e8000.

  inline ULONGLONG HnsBetween1601and1970 ()
  {
    return jrl::MakeUlonglong(0x019db1de, 0xd53e8000);
  }

}

/////////////////////////////////////////////////////////////////////////////
// FileTime class member functions
/////////////////////////////////////////////////////////////////////////////

// This function initializes a FileTime using a local NT time. That NT
// time is passed as a count of the number of nanoseconds since 1601.
// To perform the conversion to UTC, we require additional information
// about the local time zone this NT time was read in. First, we need
// the offset in hours between local time and UTC. Second, we need to
// know whether the local time zone uses daylight savings time.

FileTime::FileTime(LONGLONG hnsSince1601, int localUTCOffset,
                   bool daylightUsed)
  : known(true), absolute(true), infinite(false)
{
  // The number of hundreds of nanoseconds that passed between January 1,
  // 1601 and January 1, 1970 (the base of UNIX times) is 0x19db1ded53e8000.

  time = (hnsSince1601 - HnsBetween1601and1970()) * 100;

  // Subtract the UTC offset, converted from hours to ns.

  time -= jrl::Nanosec(localUTCOffset * SEC_PER_HOUR) * NS_PER_SEC;

  // If daylight savings time is used in the time zone, convert the
  // time to a UNIX local time to determine if daylight savings time
  // is in effect at this time. If it is, subtract an additional hour
  // (converted to ns, of course) to "fall back" to UTC.

  if (daylightUsed) {
    time_t unixTime = (time_t) (time / NS_PER_SEC);
    if (localtime(&unixTime)->tm_isdst)
      time -= jrl::Nanosec(SEC_PER_HOUR) * NS_PER_SEC;
  }
}

// This function initializes a FileTime from a relative number of
// milliseconds. We take into account the fact that NT uses
// 0xFFFFFFFF to indicate an infinite amount of time.

FileTime::FileTime(unsigned long milliseconds)
  : known(true), absolute(false)
{
  if (milliseconds == 0xFFFFFFFFUL) {
    infinite = true;
    time = 0;
  }
  else {
    infinite = false;
    time = jrl::Nanosec(milliseconds) * NS_PER_MS;
  }
}

// This function initializes a FileTime from an 8-byte NT time read
// from a VTrace trace. We take into account the fact that VTrace
// sometimes uses "????????" or "Infinite" to denote an infinite time,
// and the fact that NT uses 0x7FFF...FFF to denote an infinite
// time. Also, we take account of the fact that NT uses nonpositive
// times to denote relative times and positive times to denote
// absolute times.

FileTime::FileTime(char ntTime[8])
  : known(true), absolute(false), infinite(false)
{
  // VTrace sometimes uses ???????? or Infinite to denote a nonexistent
  // (and thus infinite) time; NT sometimes uses 0x7FFF...FFF.

  if (!strncmp(ntTime, "????????", 8) ||
      !strncmp(ntTime, "Infinite", 8) ||
      !strncmp(ntTime, "\377\377\377\377\377\377\377\177", 8)) {
    infinite = true;
    time = 0;
    return;
  }

  // If the time as a LONGLONG is non-positive, it's a relative value,
  // given in units of 100 ns. If it's positive, it's an absolute value,
  // which we can translate into a real time.

  LONGLONG ntTimeAsLonglong = jrl::ConvertToLonglong(ntTime);
  if (ntTimeAsLonglong <= 0)
    time = -ntTimeAsLonglong * 100;
  else {
    absolute = true;
    time = (ntTimeAsLonglong - HnsBetween1601and1970()) * 100;
  }
}

// This function returns the time as a number of nanoseconds. It
// returns 0 if the time is unknown, and 0x7FF...FFF if it is infinite.

jrl::Nanosec FileTime::GetTime ()
{
  if (!known)
    return 0;
  else if (infinite)
    return jrl::MaxNanosec();
  else
    return time;
}

// This function prints the FileTime to the given output stream.

void FileTime::Print (ostream& os) const
{
  if (!known)
    os << "unknown";
  else if (infinite)
    os << "infinite";
  else if (absolute)
    os << jrl::PrintableNs(time);
  else
    os << "for " << jrl::PrintableNs(time);
}

// This function outputs the FileTime to the given output destination.

void FileTime::Write (jrl::RawOutput& od) const
{
  int flags = 0;
  if (known) flags |= FileTimeFlags::KNOWN;
  if (absolute) flags |= FileTimeFlags::ABSOLUT;
  if (infinite) flags |= FileTimeFlags::INFINIT;
  od.WriteChar(flags);
  if (known && !infinite)
    od << time;
}

// This function inputs the FileTime from the given input source.

void FileTime::Read (jrl::RawInput& is)
{
  int flags = is.ReadChar();
  known = (flags & FileTimeFlags::KNOWN);
  absolute = (flags & FileTimeFlags::ABSOLUT) != 0;
  infinite = (flags & FileTimeFlags::INFINIT) != 0;
  if (known && !infinite)
    is >> time;
  else
    time = 0;
}
