#ifndef _JRL_GENIO_H
#define _JRL_GENIO_H

// This file contains definitions of general I/O classes. These
// classes represent input sources and output destinations. These
// classes use a raw, computer-readable (as opposed to human-readable)
// format. For instance, a 4-byte long would be read and written in
// 4-byte little endian format instead of as a sequence of digit
// characters.

#pragma warning ( disable : 4250 )

#include <windows.h>
#include <string>
#include <cstdio>
#include "typeconv.h"

using namespace std;

#define NO_COPYING(Type) private: Type(const Type&); \
                                  Type& operator= (const Type&)

namespace jrl {

  // This namespace contains exceptions that can be thrown by these
  // general I/O routines. All of them derive from the general
  // exception class "Any".

  namespace GenIoExceptions {

    class Any { };
    class Eof : public Any { };
    class Read : public Any { };
    class Write : public Any { };
    class Open : public Any { };
    class StringTooLong : public Any { };

  }

  // This namespace contains constants used by general I/O classes.

  namespace GenIoConstants {

    const int INTERNAL_READ_BUFFER_SIZE = 1024;
    const int INTERNAL_WRITE_BUFFER_SIZE = 4096;

  }

  ////////////////////////////////////////////////////////////////////////////
  // I/O objects with primitive operations
  ////////////////////////////////////////////////////////////////////////////

  // This abstract class represents any thing that can be read, such
  // as a file descriptor or a FILE *. It is a virtual class, in that
  // it is always inherited using the keyword 'virtual'.  This is
  // because it can be inherited via multiple paths, and all paths
  // should refer to the same object (set of functions).

  class ReadableObject {
    NO_COPYING(ReadableObject);
  public:
    ReadableObject() { }
    virtual ~ReadableObject() { }

  protected:
    // The semantics of the following function are as follows. It
    // reads bytes into the buffer 'buf', but no more than 'length'
    // bytes. The value returned is the number of bytes read. It's
    // always either positive or equal to length. If such a value
    // cannot be legitimately returned, it will throw either Eof() or
    // Read(). The former means the end of input was reached; the
    // latter means a read operation failed.
    
    virtual int UnbufferedRead (char *buf, int length) = 0;
  };

  // This abstract class represents any thing that can be written,
  // such as a file descriptor or a FILE *. It is a virtual class, in
  // that it is always inherited using the keyword 'virtual'.  This is
  // because it can be inherited via multiple paths, and all paths
  // should refer to the same object (set of functions).

  class WritableObject {
    NO_COPYING(WritableObject);
  public:
    WritableObject() { }
    virtual ~WritableObject() { }

  protected:
    // The semantics of the following function are as follows. It
    // writes bytes from the buffer 'buf', but no more than 'length'
    // bytes. The value returned is the number of bytes written. It's
    // always either positive or equal to length. If such a value
    // cannot be legitimately returned, it will throw Write().
    
    virtual int UnbufferedWrite (const char *buf, int length) = 0;
  };

  // This class represents an open file descriptor.

  class WindowsHandle : public virtual ReadableObject,
                        public virtual WritableObject {
    NO_COPYING(WindowsHandle);
  public:
    WindowsHandle() : fileHandle(INVALID_HANDLE_VALUE) { }
    WindowsHandle(HANDLE fileHandleVal) : fileHandle(fileHandleVal) { }
    virtual ~WindowsHandle() { Close(); }

    virtual void Close ();

  protected:
    HANDLE fileHandle;     // the file handle (INVALID_HANDLE_VALUE
                           //  if invalid)

    int UnbufferedRead (char *buf, int length);
    int UnbufferedWrite (const char *buf, int length);
  };

  // This class represents a particular type of file descriptor,
  // namely one that is opened for reading of a file.

  class InputFile : public WindowsHandle {
    NO_COPYING(InputFile);
  public:
    InputFile() { }

    void Open (string fileName);
  };

  // This class represents a particular type of file descriptor,
  // namely one that is opened for reading a compressed file. It
  // launches a gunzip process so that the data read is the
  // uncompressed data.

  class InputCompressedFile : public WindowsHandle {
    NO_COPYING(InputCompressedFile);
  public:
    InputCompressedFile() { }
    ~InputCompressedFile() { Close(); }

    void Open (string fileName);
    void Close ();

  protected:
    HANDLE childHandle;
  };

  // This class represents a particular type of file descriptor,
  // namely one that is opened for reading of a general input
  // source. That input source could be standard input, a simple file,
  // or the output of gunzip decompressing a file.

  class GeneralInput : public WindowsHandle {
    NO_COPYING(GeneralInput);
  public:
    GeneralInput() { }
    ~GeneralInput() { Close(); }

    void Open (string fileName);
    void Close ();

  protected:
    HANDLE childHandle;
  };

  // This class represents a particular type of file descriptor,
  // namely one that is opened for writing to a file.

  class OutputFile : public WindowsHandle {
    NO_COPYING(OutputFile);
  public:
    OutputFile() { }

    void Open (string fileName, bool deleteIfExists = false,
               bool append = false);
  };

  // This class represents a particular type of file descriptor,
  // namely one that is opened for writing to a compressed file. It
  // launches a gzip process to do the actual compression of what
  // we write.

  class OutputCompressedFile : public WindowsHandle {
    NO_COPYING(OutputCompressedFile);
  public:
    OutputCompressedFile() { }
    ~OutputCompressedFile() { Close(); }

    void Open (string fileName, bool deleteIfExists = false);
    void Close ();

  protected:
    HANDLE childHandle;
  };

  // This class represents a particular type of file descriptor,
  // namely one representing a general type of output. This output
  // could be the standard output, a simple file, or a compressed
  // file.

  class GeneralOutput : public WindowsHandle {
    NO_COPYING(GeneralOutput);
  public:
    GeneralOutput() { }
    ~GeneralOutput() { Close(); }

    void Open (string fileName, bool deleteIfExists = false,
               bool append = false);
    void Close ();

  protected:
    HANDLE childHandle;
  };

  // This class represents a writable object that doesn't actually
  // write its output anywhere.

  class OutputNowhere : public virtual WritableObject {
    NO_COPYING(OutputNowhere);
  public:
    OutputNowhere() { }

    int UnbufferedWrite (const char *buf, int length) { return length; }
  };

  ////////////////////////////////////////////////////////////////////////////
  // Abstract objects supporting raw data I/O
  ////////////////////////////////////////////////////////////////////////////

  // This abstract class represents an object that raw I/O can be
  // performed on.

  class RawInput : public virtual ReadableObject {
    NO_COPYING(RawInput);
  public:
    RawInput() { }
    virtual ~RawInput() { }

    // The following two functions can be overridden by buffered
    // subclasses.

    virtual int ReadNoMoreThan (char *buf, int length)
    { return UnbufferedRead(buf, length); }
    virtual void SkipBytes (int length);

    // This function isn't to be overridden, but it uses the above
    // functions, which can be.

    void ReadBytes (char *buf, int length);

    // The following can be overridden. If they aren't overridden,
    // they use ReadBytes() for general applicability.

    virtual bool ReadBool ();
    virtual char ReadChar ();
    virtual unsigned char ReadUchar ();
    virtual short ReadShort ();
    virtual unsigned short ReadUshort ();
    virtual long ReadLong ();
    virtual unsigned long ReadUlong ();
    virtual LONGLONG ReadLonglong ();
    virtual ULONGLONG ReadUlonglong ();
    virtual ULONGLONG Read5ByteUlonglong ();
    virtual ULONGLONG Read6ByteUlonglong ();

    // The following functions aren't to be overridden. They use
    // ReadBytes() for general applicability.

    string ReadString ();
    void ReadCString (char *buf, int maxLength = 0xFFFF);
    string ReadBytes (int length);
  };

  // RawInputs can also be read using >>.

  inline RawInput& operator>> (RawInput& is, bool& b)
  { b = (is.ReadChar() != '\0'); return is; }

  inline RawInput& operator>> (RawInput& is, char& c)
  { c = is.ReadChar(); return is; }

  inline RawInput& operator>> (RawInput& is, unsigned char& c)
  { c = is.ReadUchar(); return is; }

  inline RawInput& operator>> (RawInput& is, short& i)
  { i = is.ReadShort(); return is; }

  inline RawInput& operator>> (RawInput& is, unsigned short& i)
  { i = is.ReadUshort(); return is; }

  inline RawInput& operator>> (RawInput& is, long& i)
  { i = is.ReadLong(); return is; }

  inline RawInput& operator>> (RawInput& is, unsigned long& i)
  { i = is.ReadUlong(); return is; }

  inline RawInput& operator>> (RawInput& is, LONGLONG& i)
  { i = is.ReadLonglong(); return is; }

  inline RawInput& operator>> (RawInput& is, ULONGLONG& i)
  { i = is.ReadUlonglong(); return is; }

  inline RawInput& operator>> (RawInput& is, string& s)
  { s = is.ReadString(); return is; }

  inline RawInput& operator>> (RawInput& is, char *s)
  { is.ReadCString(s); return is; }

  // This abstract class represents an object that can do raw input.
  // It also implements a buffering scheme for that raw input, so that
  // unbuffered reads are made large and infrequent.

  class RawBufferedInput : public RawInput {
    NO_COPYING(RawBufferedInput);
  public:
    RawBufferedInput() :
        internalBufSize(GenIoConstants::INTERNAL_READ_BUFFER_SIZE),
      validBytesInInternalBuf(0), bytesReadFromInternalBuf(0) { }

    int ReadNoMoreThan (char *buf, int length);
    void SkipBytes (int length);

    bool ReadBool ();
    char ReadChar ();
    unsigned char ReadUchar ();
    short ReadShort ();
    unsigned short ReadUshort ();
    long ReadLong ();
    unsigned long ReadUlong ();
    LONGLONG ReadLonglong ();
    ULONGLONG ReadUlonglong ();
    ULONGLONG Read5ByteUlonglong ();
    ULONGLONG Read6ByteUlonglong ();

  protected:
    const int internalBufSize;       // size of internal buffer
    char internalBuf[GenIoConstants::INTERNAL_READ_BUFFER_SIZE];
    int validBytesInInternalBuf;     // # of valid bytes in internal buffer
    int bytesReadFromInternalBuf;    // how many of the first valid bytes
                                     //  have already been used
  private:
    void FillBuffer ();
  };

  // This abstract class represents an object that can do raw writes.

  class RawOutput : public virtual WritableObject {
    NO_COPYING(RawOutput);
  public:
    RawOutput() { }
    virtual ~RawOutput() { }

    // This function can be overridden. If it is, the functions below
    // it work differently since they use this as a subroutine.

    virtual int WriteNoMoreThan (const char *buf, int length)
    { return UnbufferedWrite(buf, length); }

    void WriteBytes (const char *buf, int length);

    void WriteBool (bool b);
    void WriteChar (char c);
    void WriteUchar (unsigned char c);
    void WriteShort (short i);
    void WriteUshort (unsigned short i);
    void WriteLong (long i);
    void WriteUlong (unsigned long i);
    void WriteLonglong (LONGLONG i);
    void WriteUlonglong (ULONGLONG i);
    void WriteString (const string& s);
    void WriteCString (const char *buf);
    void WriteBytes (const string& s);
  };

  // RawOutput objects can also be written to using >>.

  inline RawOutput& operator<< (RawOutput& od, bool b)
  { od.WriteBool(b); return od; }

  inline RawOutput& operator<< (RawOutput& od, char c)
  { od.WriteChar(c); return od; }

  inline RawOutput& operator<< (RawOutput& od, unsigned char c)
  { od.WriteUchar(c); return od; }

  inline RawOutput& operator<< (RawOutput& od, short i)
  { od.WriteShort(i); return od; }

  inline RawOutput& operator<< (RawOutput& od, unsigned short i)
  { od.WriteUshort(i); return od; }

  inline RawOutput& operator<< (RawOutput& od, long i)
  { od.WriteLong(i); return od; }

  inline RawOutput& operator<< (RawOutput& od, unsigned long i)
  { od.WriteUlong(i); return od; }

  inline RawOutput& operator<< (RawOutput& od, LONGLONG i)
  { od.WriteLonglong(i); return od; }

  inline RawOutput& operator<< (RawOutput& od, ULONGLONG i)
  { od.WriteUlonglong(i); return od; }

  inline RawOutput& operator<< (RawOutput& od, const string& s)
  { od.WriteString(s); return od; }

  inline RawOutput& operator<< (RawOutput& od, const char *s)
  { od.WriteCString(s); return od; }

  // This abstract class represents an object that can perform raw
  // writes. It also implements a buffering scheme so that unbuffered
  // writes can be large and infrequent.

  class RawBufferedOutput : public RawOutput {
    NO_COPYING(RawBufferedOutput);
  public:
    RawBufferedOutput() :
      internalBufSize(GenIoConstants::INTERNAL_WRITE_BUFFER_SIZE),
      validBytesInInternalBuf(0) { }

    // This function is an overriding version, that does buffering.

    int WriteNoMoreThan (const char *buf, int length);

    void Flush ();

  protected:
    const int internalBufSize;       // size of internal buffer
    char internalBuf[GenIoConstants::INTERNAL_WRITE_BUFFER_SIZE];
    int validBytesInInternalBuf;     // bytes not yet flushed from
                                     //  internal buffer
  };

  ////////////////////////////////////////////////////////////////////////////
  // Objects supporting raw data I/O
  ////////////////////////////////////////////////////////////////////////////

  // This class represents a file to be read with raw operations.

  class RawInputFile : public InputFile, public RawInput {
    NO_COPYING(RawInputFile);
  public:
    RawInputFile() { }
  };

  // This class represents a file to be read, buffered, with raw
  // operations.

  class RawBufferedInputFile : public InputFile, public RawBufferedInput {
    NO_COPYING(RawBufferedInputFile);
  public:
    RawBufferedInputFile() { }
  };

  // This class represents a compressed file to be read with raw
  // operations.

  class RawInputCompressedFile : public InputCompressedFile, public RawInput {
    NO_COPYING(RawInputCompressedFile);
  public:
    RawInputCompressedFile() { }
  };

  // This class represents a compressed file to be read, buffered,
  // with raw operations.

  class RawBufferedInputCompressedFile : public RawBufferedInput,
                                         public InputCompressedFile {
    NO_COPYING(RawBufferedInputCompressedFile);
  public:
    RawBufferedInputCompressedFile() { }
  };

  // This class represents a general input source to be read with raw
  // operations.

  class RawGeneralInput : public RawInput, public GeneralInput {
    NO_COPYING(RawGeneralInput);
  public:
    RawGeneralInput() { }
  };

  // This class represents a general input source to be read,
  // buffered, with raw operations.

  class RawBufferedGeneralInput : public RawBufferedInput,
                                  public GeneralInput {
    NO_COPYING(RawBufferedGeneralInput);
  public:
    RawBufferedGeneralInput() { }

    void StupidCompilerWorkaround (string fileName) { Open(fileName); }
  };

  // This class represents a file to be written with raw operations.

  class RawOutputFile : public RawOutput, public OutputFile {
    NO_COPYING(RawOutputFile);
  public:
    RawOutputFile() { }
  };

  // This class represents a file to be written, buffered, with raw
  // operations.

  class RawBufferedOutputFile : public RawBufferedOutput, public OutputFile {
    NO_COPYING(RawBufferedOutputFile);
  public:
    RawBufferedOutputFile() { }
    ~RawBufferedOutputFile() { Close(); }

    void Close ();
  };

  // This class represents a compressed file to be written with raw
  // operations.

  class RawOutputCompressedFile : public RawOutput,
                                  public OutputCompressedFile {
    NO_COPYING(RawOutputCompressedFile);
  public:
    RawOutputCompressedFile() { }
  };

  // This class represents a compressed file to be written, buffered,
  // with raw operations.

  class RawBufferedOutputCompressedFile : public RawBufferedOutput,
                                          public OutputCompressedFile {
    NO_COPYING(RawBufferedOutputCompressedFile);
  public:
    RawBufferedOutputCompressedFile() { }
    ~RawBufferedOutputCompressedFile() { Close(); }

    void Close ();
  };

  // This class represents a general output destination to be written
  // with raw operations.

  class RawGeneralOutput : public RawOutput, public GeneralOutput {
    NO_COPYING(RawGeneralOutput);
  public:
    RawGeneralOutput() { }
  };

  // This class represents a general output destination to be written,
  // buffered, with raw operations.

  class RawBufferedGeneralOutput : public RawBufferedOutput,
                                   public GeneralOutput {
    NO_COPYING(RawBufferedGeneralOutput);
  public:
    RawBufferedGeneralOutput() { }
    ~RawBufferedGeneralOutput() { Close(); }

    void Close ();
  };

  // This class represents an output destination that doesn't actually
  // put its data anywhere, to be written with raw operations.

  class RawOutputNowhere : public RawOutput, public OutputNowhere {
    NO_COPYING(RawOutputNowhere);
  public:
    RawOutputNowhere() { }
  };
  
}

#endif
