// This file contains fast I/O routines for our driver. Mostly they
// do nothing, but they do nothing fast! The exception is the device
// I/O control routine, which actually performs I/O control requests.
//
// (Uses some code by Mark Russinovich and Bryce Cogswell)

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/KernelUser.h"
#include "Ioctls.h"
#include "TTGFile.h"

///////////////////////////////////////////////////////////////////////////
// FAST I/O DISPATCH ROUTINES
///////////////////////////////////////////////////////////////////////////

// RSTaskFastIoCheckIfPossible returns true if the real fast I/O routine
// exists and returns true

BOOLEAN
RSTaskFastIoCheckIfPossible (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN BOOLEAN Wait,
    IN ULONG LockKey,
    IN BOOLEAN CheckForReadOperation,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  return TRUE;
}

// RSTaskFastIoRead

BOOLEAN
RSTaskFastIoRead (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN BOOLEAN Wait,
    IN ULONG LockKey,
    OUT PVOID Buffer,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  IoStatus->Status = STATUS_SUCCESS;
  IoStatus->Information = 0;
  return TRUE;
}

// RSTaskFastIoWrite

BOOLEAN
RSTaskFastIoWrite (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN BOOLEAN Wait,
    IN ULONG LockKey,
    IN PVOID Buffer,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  IoStatus->Status = STATUS_SUCCESS;
  IoStatus->Information = 0;
  return TRUE;
}

// RSTaskFastIoQueryBasicinfo

BOOLEAN
RSTaskFastIoQueryBasicInfo (
    IN struct _FILE_OBJECT *FileObject,
    IN BOOLEAN Wait,
    OUT PFILE_BASIC_INFORMATION Buffer,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  IoStatus->Status = STATUS_SUCCESS;
  IoStatus->Information = 0;
  return TRUE;
}

// RSTaskFastIoQueryStandardInfo

BOOLEAN
RSTaskFastIoQueryStandardInfo (
    IN struct _FILE_OBJECT *FileObject,
    IN BOOLEAN Wait,
    OUT PFILE_STANDARD_INFORMATION Buffer,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  IoStatus->Status = STATUS_SUCCESS;
  IoStatus->Information = 0;
  return TRUE;
}

// RSTaskFastIoLock

BOOLEAN
RSTaskFastIoLock (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN PLARGE_INTEGER Length,
    PEPROCESS ProcessId,
    ULONG Key,
    BOOLEAN FailImmediately,
    BOOLEAN ExclusiveLock,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  IoStatus->Status = STATUS_SUCCESS;
  IoStatus->Information = 0;
  return TRUE;
}

// RSTaskFastIoUnlockSingle

BOOLEAN
RSTaskFastIoUnlockSingle (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN PLARGE_INTEGER Length,
    PEPROCESS ProcessId,
    ULONG Key,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  IoStatus->Status = STATUS_SUCCESS;
  IoStatus->Information = 0;
  return TRUE;
}

// RSTaskFastIoUnlockAll

BOOLEAN
RSTaskFastIoUnlockAll (
    IN struct _FILE_OBJECT *FileObject,
    PEPROCESS ProcessId,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  IoStatus->Status = STATUS_SUCCESS;
  IoStatus->Information = 0;
  return TRUE;
}

// RSTaskFastIoUnlockAllByKey

BOOLEAN
RSTaskFastIoUnlockAllByKey (
    IN struct _FILE_OBJECT *FileObject,
    PVOID ProcessId,
    ULONG Key,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  IoStatus->Status = STATUS_SUCCESS;
  IoStatus->Information = 0;
  return TRUE;
}

// RSTaskFastIoDeviceControl actually performs device control requests.

BOOLEAN
RSTaskFastIoDeviceControl (
    IN struct _FILE_OBJECT *FileObject,
    IN BOOLEAN Wait,
    IN PVOID InputBuffer OPTIONAL,
    IN ULONG InputBufferLength,
    OUT PVOID OutputBuffer OPTIONAL,
    IN ULONG OutputBufferLength,
    IN ULONG IoControlCode,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  // Initialize the status block with 'no information returned'.

  IoStatus->Information = 0;

  // Do different things depending on the I/O control code. Each of
  // these possibilities sets IoStatus->Status to a status value
  // depending on how the operation went.

  switch (IoControlCode) {
  case RS_TASK_IOCTL_END_TASK_AND_BEGIN_ANOTHER :
    IoStatus->Status = IoctlEndTaskAndBeginAnother(
      (PTTG_FILE) FileObject->FsContext,
      InputBuffer,
      InputBufferLength,
      OutputBuffer,
      OutputBufferLength,
      &IoStatus->Information
      );
    break;

  case RS_TASK_IOCTL_BEGIN_TASK :
    IoStatus->Status = IoctlBeginTask(
      (PTTG_FILE) FileObject->FsContext,
      InputBuffer,
      InputBufferLength,
      OutputBuffer,
      OutputBufferLength,
      &IoStatus->Information
      );
    break;

  case RS_TASK_IOCTL_END_TASK :
    IoStatus->Status = IoctlEndTask(
      InputBuffer,
      InputBufferLength
      );
    break;

  case RS_TASK_IOCTL_SET_GLOBAL_POLICY :
    IoStatus->Status = IoctlSetGlobalPolicy(
      InputBuffer,
      InputBufferLength
      );
    break;

  case RS_TASK_IOCTL_QUERY_DEADLINE :
    IoStatus->Status = IoctlQueryDeadline(
      (PTTG_FILE) FileObject->FsContext,
      InputBuffer,
      InputBufferLength,
      OutputBuffer,
      OutputBufferLength,
      &IoStatus->Information
      );
    break;

  case RS_TASK_IOCTL_CHANGE_DEADLINE :
    IoStatus->Status = IoctlChangeDeadline(
      (PTTG_FILE) FileObject->FsContext,
      InputBuffer,
      InputBufferLength
      );
    break;

  case RS_TASK_IOCTL_RESET_SAMPLE :
    IoStatus->Status = IoctlResetSample(
      (PTTG_FILE) FileObject->FsContext,
      InputBuffer,
      InputBufferLength
      );
    break;

  case RS_TASK_IOCTL_RESET_ALL_SAMPLES :
    IoStatus->Status = IoctlResetAllSamples(
      (PTTG_FILE) FileObject->FsContext
      );
    break;

  case RS_TASK_IOCTL_ALLOW_FILE_MODIFICATION :
    IoStatus->Status = IoctlAllowFileModification(
      (PTTG_FILE) FileObject->FsContext
      );
    break;

  case RS_TASK_IOCTL_FLUSH :
    IoStatus->Status = IoctlFlush(
      (PTTG_FILE) FileObject->FsContext
      );
    break;

  case RS_TASK_IOCTL_FLUSH_ALL :
    IoStatus->Status = IoctlFlushAllTTGFiles();
    break;

  case RS_TASK_IOCTL_QUERY_TTG_FILE_PREAMBLE :
    IoStatus->Status = IoctlQueryTTGFilePreamble(
      (PTTG_FILE) FileObject->FsContext,
      OutputBuffer,
      OutputBufferLength,
      &IoStatus->Information
      );
    break;

  case RS_TASK_IOCTL_QUERY_NUM_SPEEDS :
    IoStatus->Status = IoctlQueryNumSpeeds(
      OutputBuffer,
      OutputBufferLength,
      &IoStatus->Information
      );
    break;

  case RS_TASK_IOCTL_QUERY_TTG_DIRECTORY :
    IoStatus->Status = IoctlQueryTTGDirectory(
      OutputBuffer,
      OutputBufferLength,
      &IoStatus->Information
      );
    break;

  case RS_TASK_IOCTL_PREPARE_FOR_UNINSTALL :
    IoStatus->Status = IoctlPrepareForUninstall();
    break;

  case RS_TASK_IOCTL_GET_FUNCTION_POINTER :
    IoStatus->Status = IoctlGetFunctionPointer(
      InputBuffer,
      InputBufferLength,
      OutputBuffer,
      OutputBufferLength,
      &IoStatus->Information
      );
    break;

  case RS_TASK_IOCTL_QUERY_AVAILABLE_SPEEDS_KHZ :
    IoStatus->Status = IoctlQueryAvailableSpeedsKHz(
      OutputBuffer,
      OutputBufferLength,
      &IoStatus->Information
      );
    break;

  case RS_TASK_IOCTL_QUERY_IDLE_THREAD_TIME :
    IoStatus->Status = IoctlQueryIdleThreadTime(
      OutputBuffer,
      OutputBufferLength,
      &IoStatus->Information
      );
    break;

  default :
    DebugPrint(("RSTask: WARNING - Unrecognized I/O control code %x.\n",
                IoControlCode));
    IoStatus->Status = STATUS_NOT_IMPLEMENTED;
    break;
  }

  // Return TRUE indicating that the request was handled completely
  // (though not necessarily successfully) and thus no IRP is
  // necessary.
  
  return TRUE;
}

// RSTaskAcquireFileForNtCreateSection

VOID
RSTaskAcquireFileForNtCreateSection (
    IN struct _FILE_OBJECT *FileObject
    )
{
}

// RSTaskReleaseFileForNtCreateSection

VOID
RSTaskReleaseFileForNtCreateSection (
    IN struct _FILE_OBJECT *FileObject
    )
{
}

// RSTaskFastIoDetachDevice is called when a device that we have hooked
// is being deleted. This happens when, for example, a floppy is formatted.
// We have to detach from it and delete our device.

VOID
RSTaskFastIoDetachDevice (
    IN struct _DEVICE_OBJECT *SourceDevice,
    IN struct _DEVICE_OBJECT *TargetDevice
    )
{
}

BOOLEAN
RSTaskFastIoQueryNetworkOpenInfo (
    IN struct _FILE_OBJECT *FileObject,
    IN BOOLEAN Wait,
    OUT struct _FILE_NETWORK_OPEN_INFORMATION *Buffer,
    OUT struct _IO_STATUS_BLOCK *IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  IoStatus->Status = STATUS_SUCCESS;
  IoStatus->Information = 0;
  return TRUE;
}

// RSTaskMdlRead

BOOLEAN
RSTaskMdlRead (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN ULONG LockKey,
    OUT PMDL *MdlChain,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  IoStatus->Status = STATUS_SUCCESS;
  IoStatus->Information = 0;
  return TRUE;
}

// RSTaskMdlReadComplete

BOOLEAN
RSTaskMdlReadComplete (
    IN struct _FILE_OBJECT *FileObject,
    IN PMDL MdlChain,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  return TRUE;
}

// RSTaskFastIoPrepareMdlWrite

BOOLEAN
RSTaskPrepareMdlWrite (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN ULONG LockKey,
    OUT PMDL *MdlChain,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  IoStatus->Status = STATUS_SUCCESS;
  IoStatus->Information = 0;
  return TRUE;
}

// RSTaskMdlWriteComplete

BOOLEAN
RSTaskMdlWriteComplete (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN PMDL MdlChain,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  return TRUE;
}

// RSTaskAcquireForModWrite

NTSTATUS
RSTaskAcquireForModWrite (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER EndingOffset,
    OUT struct _ERESOURCE **ResourceToRelease,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  return STATUS_NOT_SUPPORTED;
}

// RSTaskReleaseForModWrite

NTSTATUS
RSTaskReleaseForModWrite (
    IN struct _FILE_OBJECT *FileObject,
    IN struct _ERESOURCE *ResourceToRelease,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  return STATUS_NOT_SUPPORTED;
}

// RSTaskAcquireForCcFlush

NTSTATUS
RSTaskAcquireForCcFlush (
    IN struct _FILE_OBJECT *FileObject,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  return STATUS_NOT_SUPPORTED;
}

// RSTaskFastIoReleaseForCcFlush

NTSTATUS
RSTaskReleaseForCcFlush (
    IN struct _FILE_OBJECT *FileObject,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  return STATUS_NOT_SUPPORTED;
}

// RSTaskFastIoReadCompressed

BOOLEAN
RSTaskFastIoReadCompressed (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN ULONG LockKey,
    OUT PVOID Buffer,
    OUT PMDL *MdlChain,
    OUT PIO_STATUS_BLOCK IoStatus,
    OUT struct _COMPRESSED_DATA_INFO *CompressedDataInfo,
    IN ULONG CompressedDataInfoLength,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  IoStatus->Status = STATUS_SUCCESS;
  IoStatus->Information = 0;
  return TRUE;
}

// RSTaskFastIoWriteCompressed

BOOLEAN
RSTaskFastIoWriteCompressed (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN ULONG LockKey,
    IN PVOID Buffer,
    OUT PMDL *MdlChain,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN struct _COMPRESSED_DATA_INFO *CompressedDataInfo,
    IN ULONG CompressedDataInfoLength,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  IoStatus->Status = STATUS_SUCCESS;
  IoStatus->Information = 0;
  return TRUE;
}

// RSTaskMdlReadCompleteCompressed

BOOLEAN
RSTaskMdlReadCompleteCompressed (
    IN struct _FILE_OBJECT *FileObject,
    IN PMDL MdlChain,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  return TRUE;
}

// RSTaskMdlWriteCompleteCompressed

BOOLEAN
RSTaskMdlWriteCompleteCompressed (
    IN struct _FILE_OBJECT *FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN PMDL MdlChain,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  return TRUE;
}

// RSTaskFastIoQueryOpen (This call actually passes an IRP!)

BOOLEAN
RSTaskFastIoQueryOpen (
    IN struct _IRP *Irp,
    OUT PFILE_NETWORK_OPEN_INFORMATION NetworkInformation,
    IN struct _DEVICE_OBJECT *DeviceObject
    )
{
  DebugPrint(("RSTask: WARNING - Received fast I/O query open.\n"));
  return FALSE;
}

///////////////////////////////////////////////////////////////////////////
// DISPATCH TABLE
///////////////////////////////////////////////////////////////////////////

// This is RSTask's fast I/O dispatch table. Note that NT assumes that
// file system drivers support some fast I/O calls, so this table must
// be present for a file system filter driver.

FAST_IO_DISPATCH FastIoDispatchTable = {
  sizeof(FAST_IO_DISPATCH),
  RSTaskFastIoCheckIfPossible,
  RSTaskFastIoRead,
  RSTaskFastIoWrite,
  RSTaskFastIoQueryBasicInfo,
  RSTaskFastIoQueryStandardInfo,
  RSTaskFastIoLock,
  RSTaskFastIoUnlockSingle,
  RSTaskFastIoUnlockAll,
  RSTaskFastIoUnlockAllByKey,
  RSTaskFastIoDeviceControl,
  RSTaskAcquireFileForNtCreateSection,
  RSTaskReleaseFileForNtCreateSection,
  RSTaskFastIoDetachDevice,
  RSTaskFastIoQueryNetworkOpenInfo,      // From here on is new for NT 4.0
  RSTaskAcquireForModWrite,
  RSTaskMdlRead,
  RSTaskMdlReadComplete,
  RSTaskPrepareMdlWrite,
  RSTaskMdlWriteComplete,
  RSTaskFastIoReadCompressed,
  RSTaskFastIoWriteCompressed,
  RSTaskMdlReadCompleteCompressed,
  RSTaskMdlWriteCompleteCompressed,
  RSTaskFastIoQueryOpen,
  RSTaskReleaseForModWrite,
  RSTaskAcquireForCcFlush,
  RSTaskReleaseForCcFlush
  };
