// This file contains functions related to the idle thread.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "IdleThread.h"
#include "Globals.h"
#include "Log.h"
#include "SampleQ.h"

/////////////////////////////////////////////////////////////////////////////
// MACROS
/////////////////////////////////////////////////////////////////////////////

// This is the priority the idle thread usually runs at, which is just
// above the normal "Idle" level:
#define DESIRED_THREAD_PRIORITY                      5

// This is the priority threads usually run at:
#define NORMAL_THREAD_PRIORITY                       8

/////////////////////////////////////////////////////////////////////////////
// INFO FROM GARY NEBBETT'S BOOK
/////////////////////////////////////////////////////////////////////////////

#define THREAD_QUERY_INFORMATION       0x0040

NTSYSAPI
NTSTATUS
NTAPI
ZwOpenThread(
    OUT PHANDLE ThreadHandle,
    IN ACCESS_MASK AccessMask,
    IN POBJECT_ATTRIBUTES ObjectAttributes,
    IN PCLIENT_ID ClientId
    );

_declspec(naked) NTSTATUS
NTAPI
ZwQueryInformationThread(
    IN HANDLE FileHandle,
    IN THREADINFOCLASS ThreadInformationClass,
    OUT PVOID ThreadInformation,
    IN ULONG ThreadInformationLength,
    OUT PULONG ReturnLength OPTIONAL
    )
{
  _asm {
    mov eax, 87h
    lea edx, [esp+4]
    int 2eh
    ret 14h
  }
}

/////////////////////////////////////////////////////////////////////////////
// THREAD ROUTINE
/////////////////////////////////////////////////////////////////////////////

// This is the routine that the idle thread runs.

VOID IdleThreadRoutine (PVOID StartContext)
{
  NTSTATUS status;
  PKTHREAD thisThread = KeGetCurrentThread();
  BOOLEAN done = FALSE;
  KFLOATING_SAVE floatingPointState;

  // This thread performs some floating point operations (when it
  // processes sample values in the queue), so we save the
  // floating-point state before it runs and will restore it before it
  // terminates. This may seem unnecessary, since it's a completely
  // separate thread, but I've found that if you don't save floating-
  // point state, floating-point operations don't work right.

  KeSaveFloatingPointState(&floatingPointState);

  // Keep looping until something stops us, either an error condition
  // or the setting of the boolean that tells us to terminate.

  while (!done) {

    // Wait to be signalled by someone indicating we have something to do.

    status = KeWaitForSingleObject(&g.idleThreadInfo.threadWaitEvent,
                                   Executive,
                                   KernelMode,
                                   TRUE,  // alertable
                                   NULL); // no timeout
    if (status == STATUS_WAIT_0) {

      // If we were signalled by someone, make sure our priority is set
      // so that we only proceed when there's nothing else important
      // going on on the system. (Our priority may have been boosted
      // due to starvation.) Note an exception: if we should terminate,
      // there's no need to ensure our priority is low.

      while ((KeQueryPriorityThread(thisThread) > DESIRED_THREAD_PRIORITY) &&
             !g.idleThreadInfo.shouldTerminate)
        KeSetPriorityThread(thisThread, DESIRED_THREAD_PRIORITY);

      // If we should terminate, we're done, but only after we finish
      // one last round of ending all tasks and processing sample
      // values. We check for possible termination here rather than
      // later because we don't want to finish processing the queue,
      // then get context-switched out while some other thread adds
      // a sample value to the queue and then some other thread
      // tells us to terminate.

      if (g.idleThreadInfo.shouldTerminate)
        done = TRUE;

      // At this point, we should end all tasks since the system is
      // idle. However, if there are I/O's ongoing, we don't end
      // all tasks, but rather pause the current schedule (if there
      // is one).

      if (g.sharedState->numOngoingIos) {
        if (g.ongoingTasksInfo.ttInfoApplies) {
          // If there is I/O ongoing during a single ongoing task,
          // pause its schedule and also record that I/O occurred
          // during it.

          g.ongoingTasksInfo.ioOccurredDuringTask = TRUE;
          PauseSchedule();

          DebugPrint(("RSTask: Pausing schedule because I/O's are "
                      "ongoing.\n"));
        }
      }
      else {
        DebugPrint(("RSTask: Ending all tasks because no I/O's are "
                    "ongoing.\n"));
        EndAllTasks();
      }

      // The last thing we do (after ending all tasks, since that can
      // add sample values to the queue) is process all the sample values
      // in the queue.

      ProcessSamplesInQueue();

    }
    else if (!NT_SUCCESS(status)) {
      // If there was an error in the wait, end this thread.
      
      DebugPrint(("RSTask: ERROR - Thread wait failed (%x).\n", status));
      BothLogShort(RS_TASK_ERROR_IDLE_THREAD_WAIT_FAILED,
                   status);
      done = TRUE;
    }
  }

  // We're done with the main loop of the thread, so exit. Restore the
  // floating-point state before we exit, just out of paranoia.

  DebugPrint(("RSTask: Exiting thread.\n"));
  KeRestoreFloatingPointState(&floatingPointState);
  PsTerminateSystemThread(STATUS_SUCCESS);
}

/////////////////////////////////////////////////////////////////////////////
// STARTING AND STOPPING THREAD
/////////////////////////////////////////////////////////////////////////////

// This function initializes the idle thread structure and also starts
// the idle thread.

NTSTATUS InitializeIdleThread (PIDLE_THREAD_INFO info)
{
  NTSTATUS status;
  OBJECT_ATTRIBUTES objectAttributes;
  HANDLE handle;
  long data;

  // Initialize the fields of the idle thread structure.

  KeInitializeEvent(&info->threadWaitEvent,
                    SynchronizationEvent,
                    FALSE);
  info->shouldTerminate = FALSE;

  // Create a system thread to run the idle thread routine.

  InitializeObjectAttributes(&objectAttributes,
                             NULL,  // object name
                             0,     // attributes
                             NULL,  // root directory
                             NULL); // security descriptor
  status = PsCreateSystemThread(&handle,
                                THREAD_ALL_ACCESS,
                                &objectAttributes,
                                NULL, // process handle: default to system
                                &info->clientId,
                                IdleThreadRoutine,
                                NULL); // StartContext parameter to routine
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not create idle thread "
                "(%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_CREATE_IDLE_THREAD,
                 status);
    return status;
  }
  DebugPrint(("RSTask: Started thread with handle %lx\n", handle));

  // Get a pointer to the thread to save for later.

  status = ObReferenceObjectByHandle(handle,
                                     SYNCHRONIZE,
                                     NULL, // object type: unused
                                     KernelMode,
                                     &info->threadObject,
                                     NULL); // place to put handle information
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not get thread object pointer "
                "(%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_GET_THREAD_OBJECT_POINTER,
                 status);
    ZwClose(handle);
    return status;
  }

  // We no longer need the handle as we're referencing the thread
  // object, so close the handle.

  status = ZwClose(handle);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not close thread handle (%x).\n",
                status));
    BothLogShort(RS_TASK_ERROR_CANT_CLOSE_THREAD_HANDLE,
                 status);
  }

  // There's no need to set the idle thread to the desired priority, since
  // the first time it wakes up it'll set itself to that priority.

  // Return success.

  return STATUS_SUCCESS;
}

// This routine terminates the idle thread, assumed to have been
// started.

void CloseIdleThread (PIDLE_THREAD_INFO info)
{
  NTSTATUS status;
  long data;

  // Indicate the thread should terminate.

  info->shouldTerminate = TRUE;

  // Set the event the thread is waiting for so that it can wake up and
  // notice it should terminate. Boost its priority by the difference
  // between its priority and the "normal" thread priority so that it
  // wakes up soon.

  KeSetEvent(&info->threadWaitEvent,
             NORMAL_THREAD_PRIORITY - DESIRED_THREAD_PRIORITY,
             FALSE); // wait immediately after

  // Wait for the thread object, which will be signalled when the
  // thread actually terminates.

  DebugPrint(("RSTask: Waiting for thread to terminate.\n"));
  do {
    status = KeWaitForSingleObject(info->threadObject,
                                   Executive,
                                   KernelMode,
                                   TRUE,  // alertable
                                   NULL); // timeout
  } while (status == STATUS_ALERTED || status == STATUS_USER_APC);
  DebugPrint(("RSTask: Finished waiting for thread to terminate.\n"));

  // We no longer need to reference the thread, so dereference the
  // thread object.
  
  ObDereferenceObject(info->threadObject);
}

// This function determines how long the thread has been running.  It
// puts the answer, in hundreds of nanoseconds, in hnsPtr.  It returns a
// status value indicating if it succeeded.

NTSTATUS GetIdleThreadTime (PLARGE_INTEGER hnsPtr)
{
  HANDLE handle;
  KERNEL_USER_TIMES timesInfo;
  NTSTATUS status;
  OBJECT_ATTRIBUTES objectAttributes;

  // Open a handle to the idle thread using its client ID.

  InitializeObjectAttributes(&objectAttributes,
                             NULL,
                             0,
                             NULL,
                             NULL);
  status = ZwOpenThread(&handle,
                        THREAD_QUERY_INFORMATION,
                        &objectAttributes,
                        &g.idleThreadInfo.clientId);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Can't open idle thread to query info (%x).\n",
                status));
    return status;
  }

  // Query the times information for that thread.

  status = ZwQueryInformationThread(handle,
                                    ThreadTimes,
                                    &timesInfo,
                                    sizeof(timesInfo),
                                    NULL);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Can't query times information of idle "
                "thread (%x).\n", status));
    ZwClose(handle);
    return status;
  }

  // Put the kernel time in hnsPtr, since this thread only runs in
  // kernel mode.  Then, return success.

  *hnsPtr = timesInfo.KernelTime;
  ZwClose(handle);
  return STATUS_SUCCESS;
}
