// This file contains routines for logging, both to RSLog and to the
// system logger.

#include <ntddk.h>
#include "Log.h"
#include "../Common/Drivers.h"
#include "../Common/Logging.h"
#include "../Common/Schedule.h"
#include "Globals.h"

/////////////////////////////////////////////////////////////////////////
// MACROS
/////////////////////////////////////////////////////////////////////////

// LOG_EVENT is a macro that acts as a function call to log an event.

#define LOG_EVENT           (*(g.sharedState->logEventFunc))

////////////////////////////////////////////////////////////////////////////
// RSLOG ROUTINES
////////////////////////////////////////////////////////////////////////////

// This function writes an RSLog entry consisting of the given log
// type and no data.

void LogAnyEventType (short eventType)
{
  KIRQL oldirql;

  KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
  LOG_EVENT(eventType, 0);
  KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);
}

// This function writes an RSLog entry consisting of the given log
// type, including the given data of the given length.

void LogAnyEventTypeAndBytes (short eventType, PVOID data, ULONG dataLength)
{
  KIRQL oldirql;
  PCHAR dataPtr;

  KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
  dataPtr = LOG_EVENT(eventType, dataLength);
  if (dataPtr) {
    RtlCopyMemory(dataPtr, data, dataLength);
  }
  KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);
}

// This function writes an RSLog entry consisting of the given log
// type, including the given long number as its data.

void LogAnyEventTypeAndLong (short eventType, long n)
{
  LogAnyEventTypeAndBytes(eventType, &n, sizeof(long));
}

// This function writes an RSLog entry consisting of the given log
// type, including the given two long numbers as its data.

void LogAnyEventTypeAndTwoLongs (short eventType, long n1, long n2)
{
  long data[2];
  data[0] = n1;
  data[1] = n2;
  LogAnyEventTypeAndBytes(eventType, data, sizeof(long) * 2);
}

// This function writes an RSLog EVENT_TYPE_LONG_AND_DOUBLE entry with
// the given values as data.

void LogLongAndDouble (long n, double x)
{
  KIRQL oldirql;
  PCHAR dataPtr;

  KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
  dataPtr = LOG_EVENT(EVENT_TYPE_LONG_AND_DOUBLE, 12);
  if (dataPtr) {
    RtlCopyMemory(dataPtr, &n, 4);
    RtlCopyMemory(dataPtr + 4, &x, 8);
  }
  KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);
}

// This function writes an RSLog EVENT_TYPE_STRING entry with the given
// string as data.

void LogString (char *str)
{
  char *p = str;
  long length = 0;

  // Determine the length of the string.

  while (*p++)
    ++length;

  // Log the entry.

  LogAnyEventTypeAndBytes(EVENT_TYPE_STRING, str, length);
}

// This function writes an RSLog EVENT_TYPE_SET_SCHEDULE entry with the
// given schedule as data.

void LogSchedule (PSPEED_SCHEDULE schedule)
{
  // Compute the length as the number of phases times the size of a phase.

  LogAnyEventTypeAndBytes(
    EVENT_TYPE_SET_SCHEDULE,
    schedule->phase,
    schedule->numPhases * sizeof(SPEED_SCHEDULE_PHASE)
    );
}

// This function writes an RSLog EVENT_TYPE_ERROR entry with the given
// error code as data.

void LogError (
  NTSTATUS errorCode
  )
{
  LogAnyEventTypeAndBytes(EVENT_TYPE_ERROR, &errorCode, sizeof(NTSTATUS));
}

// This function writes an RSLog EVENT_TYPE_ERROR entry with the given
// error code and the given single parameter as data.

void LogErrorOneParameter (
  NTSTATUS errorCode,
  ULONG param1
  )
{
  PCHAR dataPtr;
  KIRQL oldirql;

  KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
  dataPtr = LOG_EVENT(EVENT_TYPE_ERROR, 8);
  if (dataPtr) {
    RtlCopyMemory(dataPtr, &errorCode, 4);
    RtlCopyMemory(&dataPtr[4], &param1, 4);
  }
  KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);
}

// This function writes an RSLog EVENT_TYPE_ERROR entry with the given
// error code and the given two parameters as data.

void LogErrorTwoParameters (
  NTSTATUS errorCode,
  ULONG param1,
  ULONG param2)
{
  PCHAR dataPtr;
  KIRQL oldirql;

  KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
  dataPtr = LOG_EVENT(EVENT_TYPE_ERROR, 12);
  if (dataPtr) {
    RtlCopyMemory(dataPtr, &errorCode, 4);
    RtlCopyMemory(&dataPtr[4], &param1, 4);
    RtlCopyMemory(&dataPtr[8], &param2, 4);
  }
  KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);
}

// This function writes an RSLog EVENT_TYPE_ERROR entry with the given
// error code and the given three parameters as data.

void LogErrorThreeParameters (
  NTSTATUS errorCode,
  ULONG param1,
  ULONG param2,
  ULONG param3)
{
  PCHAR dataPtr;
  KIRQL oldirql;

  KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
  dataPtr = LOG_EVENT(EVENT_TYPE_ERROR, 16);
  if (dataPtr) {
    RtlCopyMemory(dataPtr, &errorCode, 4);
    RtlCopyMemory(&dataPtr[4], &param1, 4);
    RtlCopyMemory(&dataPtr[8], &param2, 4);
    RtlCopyMemory(&dataPtr[12], &param3, 4);
  }
  KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);
}

////////////////////////////////////////////////////////////////////////////
// SYSTEM LOGGING
////////////////////////////////////////////////////////////////////////////

// This routine writes an entry to the system logger describing an
// "error" condition (which could actually be just a warning, or
// information, or even success, depending on the error code).

void SysLog (
  PVOID DeviceOrDriverObject,
  NTSTATUS ErrorCode,
  NTSTATUS FinalStatus,
  ULONG UniqueErrorValue,
  UCHAR MajorFunctionCode,
  ULONG IoControlCode,
  USHORT NumDumpDataEntries,
  ULONG *DumpData
  )
{
  IO_ERROR_LOG_PACKET *errorPacket;
  USHORT i;

  errorPacket =
    IoAllocateErrorLogEntry(DeviceOrDriverObject,
                            (UCHAR)(sizeof(IO_ERROR_LOG_PACKET) +
                                    NumDumpDataEntries * sizeof(ULONG)));
  if (errorPacket) {
    errorPacket->MajorFunctionCode = MajorFunctionCode;
    errorPacket->RetryCount = 0;
    errorPacket->DumpDataSize = NumDumpDataEntries * sizeof(ULONG);
    errorPacket->ErrorCode = ErrorCode;
    errorPacket->UniqueErrorValue = UniqueErrorValue;
    errorPacket->FinalStatus = FinalStatus;
    errorPacket->SequenceNumber = 0;
    errorPacket->IoControlCode = 0;
    for (i = 0; i < NumDumpDataEntries; ++i)
      errorPacket->DumpData[i] = DumpData[i];
    IoWriteErrorLogEntry(errorPacket);
  }
  else {
    DebugPrint(("RightSpeed: ERROR - Could not allocate error packet.\n"));
  }
}

// This routine writes an entry to the system logger describing an
// "error" condition (which could actually be just a warning, or
// information, or even success, depending on the error code). It uses
// a truncated set of parameters.

void SysLogShort (
  NTSTATUS ErrorCode,
  NTSTATUS FinalStatus
  )
{
  SysLog(g.ourDriver,
         ErrorCode,
         FinalStatus,
         0,  // unique error value
         0,  // major function code
         0,  // I/O control code
         0,  // dump data size
         0); // dump data
}

// This routine writes an entry to the system logger describing an
// "error" condition (which could actually be just a warning, or
// information, or even success, depending on the error code). It uses
// a truncated set of parameters that reflects just the error code, the
// status, and the unique error value that encodes this particular place
// in the code where this error occurs.

void SysLogShortRepeat (
  NTSTATUS ErrorCode,
  NTSTATUS FinalStatus,
  ULONG UniqueErrorValue
  )
{
  SysLog(g.ourDriver,
           ErrorCode,
           FinalStatus,
           UniqueErrorValue,
           0,  // major function code
           0,  // I/O control code
           0,  // dump data size
           0); // dump data
}

// This routine writes an entry to the system logger describing an
// "error" condition (which could actually be just a warning, or
// information, or even success, depending on the error code). It uses
// a truncated set of parameters that reflects just the error code,
// the status, and the one parameter that describes the context of the
// error.

void SysLogOneParameter (
  NTSTATUS ErrorCode,
  NTSTATUS FinalStatus,
  ULONG DumpParameter
  )
{
  SysLog(g.ourDriver,
           ErrorCode,
           FinalStatus,
           0,  // unique error value
           0,  // major function code
           0,  // I/O control code
           1,  // dump data size
           &DumpParameter);
}

// This routine writes an entry to the system logger describing an
// "error" condition (which could actually be just a warning, or
// information, or even success, depending on the error code). It uses
// a truncated set of parameters that reflects just the error code,
// the status, and the two parameters that describe the context of the
// error.

void SysLogTwoParameters (
  NTSTATUS ErrorCode,
  NTSTATUS FinalStatus,
  ULONG DumpParameter0,
  ULONG DumpParameter1
  )
{
  ULONG DumpParameter[2];

  DumpParameter[0] = DumpParameter0;
  DumpParameter[1] = DumpParameter1;
  
  SysLog(g.ourDriver,
           ErrorCode,
           FinalStatus,
           0,  // unique error value
           0,  // major function code
           0,  // I/O control code
           2,  // dump data size
           DumpParameter);
}

////////////////////////////////////////////////////////////////////////////
// BOTH RSLOG AND SYSTEM LOGGING
////////////////////////////////////////////////////////////////////////////

// This routine writes a log entry indicating an error to both the system
// log and, if RSLog is logging errors, the RSLog log.

void BothLogShort (
  NTSTATUS ErrorCode,
  NTSTATUS FinalStatus
  )
{
  SysLog(g.ourDriver,
         ErrorCode,
         FinalStatus,
         0,  // unique error value
         0,  // major function code
         0,  // I/O control code
         0,  // dump data size
         0); // dump data

  if (LOGGING_ERRORS)
    LogErrorOneParameter(ErrorCode, FinalStatus);
}

// This routine writes a log entry indicating an error to both the
// system log and, if RSLog is logging errors, the RSLog log.

void BothLogShortRepeat (
  NTSTATUS ErrorCode,
  NTSTATUS FinalStatus,
  ULONG UniqueErrorValue
  )
{
  SysLog(g.ourDriver,
         ErrorCode,
         FinalStatus,
         UniqueErrorValue,
         0,  // major function code
         0,  // I/O control code
         0,  // dump data size
         0); // dump data

  if (LOGGING_ERRORS)
    LogErrorTwoParameters(ErrorCode, FinalStatus, UniqueErrorValue);
}

// This routine writes a log entry indicating an error to both the
// system log and, if RSLog is logging errors, the RSLog log.

void BothLogOneParameter (
  NTSTATUS ErrorCode,
  NTSTATUS FinalStatus,
  ULONG DumpParameter
  )
{
  SysLog(g.ourDriver,
         ErrorCode,
         FinalStatus,
         0,  // unique error value
         0,  // major function code
         0,  // I/O control code
         1,  // dump data size
         &DumpParameter);

  if (LOGGING_ERRORS)
    LogErrorTwoParameters(ErrorCode, FinalStatus, DumpParameter);
}

// This routine writes a log entry indicating an error to both the
// system log and, if RSLog is logging errors, the RSLog log.

void BothLogTwoParameters (
  NTSTATUS ErrorCode,
  NTSTATUS FinalStatus,
  ULONG DumpParameter0,
  ULONG DumpParameter1
  )
{
  ULONG DumpParameter[2];

  DumpParameter[0] = DumpParameter0;
  DumpParameter[1] = DumpParameter1;
  
  SysLog(g.ourDriver,
         ErrorCode,
         FinalStatus,
         0,  // unique error value
         0,  // major function code
         0,  // I/O control code
         2,  // dump data size
         DumpParameter);

  if (LOGGING_ERRORS)
    LogErrorThreeParameters(ErrorCode, FinalStatus, DumpParameter0,
                            DumpParameter1);
}
