// This file contains routines for reading configuration information
// from the registry.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/DrvShare.h"
#include "../Common/ErrCodes.h"
#include "../Common/Parameters.h"
#include "Gamma.h"
#include "Globals.h"
#include "Log.h"
#include "MyFloat.h"
#include "RegInfo.h"
#include "Unicode.h"
#include "Convert.h"

/////////////////////////////////////////////////////////////////////////////
// MACROS
/////////////////////////////////////////////////////////////////////////////

// This macro computes the maximum of two numbers:
#define MAX(a,b)                    ((a)>(b)?(a):(b))

// This macro gives the maximum number of bytes a registry entry can
// have. It's the maximum of the main directory length (times 2 because
// each character is 2 bytes) and the number of available speeds (times 4
// because each speed stores 4 bytes of information):
#define MAX_REGISTRY_DATA_LENGTH    MAX(MAX_RIGHTSPEED_DIRECTORY_LENGTH * 2, \
                                        MAX_NUM_AVAILABLE_SPEEDS * 4)

// This is the number of microcycles per cycle:
#define UCYC_PER_CYC                  1000000

/////////////////////////////////////////////////////////////////////////////
// FORWARD DEFINITIONS
/////////////////////////////////////////////////////////////////////////////

NTSTATUS ReadSpeedCurvesInformation (PWCHAR mainDirectory);

/////////////////////////////////////////////////////////////////////////////
// NUMERICAL CONVERSION
/////////////////////////////////////////////////////////////////////////////

// This function converts a fraction (given as a numerator and
// denominator) to a new fraction with a new denominator. It returns
// the new numerator.

ULONG GetFraction (ULONG numerator, ULONG denominator, ULONG newDenominator)
{
  ULONGLONG n;

  n = numerator;
  n *= newDenominator;
  n += (denominator >> 1);
  n /= denominator;

  return (long)(n);
}

// This function computes the harmonic mean of two numbers.

double HarmonicMean (double a, double b)
{
  return (2 * a * b) / (a + b);
}

/////////////////////////////////////////////////////////////////////////////
// STRING MANIPULATION
/////////////////////////////////////////////////////////////////////////////

// This function fills in a TTG directory string given the main directory
// string. It assumes the TTG directory string starts out empty.

NTSTATUS FillInTTGDirectoryString (PUNICODE_STRING ttgDirectory,
                                   PWCHAR mainDirectory)
{
  NTSTATUS status;

  // First, append \??\.

  status = RtlAppendUnicodeToString(ttgDirectory, L"\\??\\");
  if (!NT_SUCCESS(status))
    return status;

  // Next, append the main directory string.

  status = RtlAppendUnicodeToString(ttgDirectory, mainDirectory);
  if (!NT_SUCCESS(status))
    return status;

  // Finally, append \TTG.

  return RtlAppendUnicodeToString(ttgDirectory, L"\\TTG");
}

// This function fills in a file name Unicode string given the main
// directory string. It assumes the file name Unicode string starts
// out empty.

NTSTATUS FillInSpeedCurvesFileName (PUNICODE_STRING fileName,
                                    PWCHAR mainDirectory)
{
  NTSTATUS status;

  // First, append \??\.

  status = RtlAppendUnicodeToString(fileName, L"\\??\\");
  if (!NT_SUCCESS(status))
    return status;

  // Next, append the main directory string.

  status = RtlAppendUnicodeToString(fileName, mainDirectory);
  if (!NT_SUCCESS(status))
    return status;

  // Finally, append \SpeedCurves.txt.

  return RtlAppendUnicodeToString(fileName, L"\\SpeedCurves.txt");
}

/////////////////////////////////////////////////////////////////////////////
// READING REGISTRY
/////////////////////////////////////////////////////////////////////////////

// This function reads the configuration information for RightSpeed
// from the registry and puts it in the structure passed.

NTSTATUS ReadRegistryInfo (REGISTRY_INFO *info)
{
  NTSTATUS status;
  OBJECT_ATTRIBUTES objectAttributes;
  struct {
    ULONG   TitleIndex;
    ULONG   Type;
    ULONG   DataLength;
    char    Data[MAX_REGISTRY_DATA_LENGTH];
  } keyValueInfo;
  HANDLE keyHandle;
  ULONG keyValueInfoLength;
  short i;
  UNICODE_STRING miscUnicodeString;
  ULONG *speedsKHz, *speedCodes;
  short *settings;

  // Open the registry key for RightSpeed.

  RtlInitUnicodeString(&miscUnicodeString,
    L"\\Registry\\Machine\\Software\\Jay Lorch\\RightSpeed\\CurrentVersion");
  InitializeObjectAttributes(&objectAttributes,
                             &miscUnicodeString,
                             OBJ_CASE_INSENSITIVE,
                             NULL,  // root directory
                             NULL); // security descriptor
  status = ZwOpenKey(&keyHandle,
                     KEY_QUERY_VALUE | KEY_SET_VALUE,
                     &objectAttributes);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not open main registry key (%x).\n",
                status));
    BothLogShort(RS_TASK_ERROR_CANT_OPEN_MAIN_REGISTRY_KEY,
                 status);
    return status;
  }

  // Read the TTG flush period in seconds.

  DebugPrint(("RSTask: Reading TTG flush period.\n"));
  RtlInitUnicodeString(&miscUnicodeString, L"TTGFlushPeriodSec");
  status = ZwQueryValueKey(keyHandle,
                           &miscUnicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (NT_SUCCESS(status) && keyValueInfo.Type == REG_DWORD) {
    info->ttgFlushPeriodSec = * (ULONG *) keyValueInfo.Data;
    if (info->ttgFlushPeriodSec < MIN_TTG_FLUSH_PERIOD_SEC ||
        info->ttgFlushPeriodSec > MAX_TTG_FLUSH_PERIOD_SEC)
      info->ttgFlushPeriodSec = DEFAULT_TTG_FLUSH_PERIOD_SEC;
  }
  else
    info->ttgFlushPeriodSec = DEFAULT_TTG_FLUSH_PERIOD_SEC;

  // Read the sample queue size.

  DebugPrint(("RSTask: Reading sample queue size.\n"));
  RtlInitUnicodeString(&miscUnicodeString, L"SampleQueueSize");
  status = ZwQueryValueKey(keyHandle,
                           &miscUnicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (NT_SUCCESS(status) && keyValueInfo.Type == REG_DWORD) {
    info->sampleQueueSize = * (ULONG *) keyValueInfo.Data;
    if (info->sampleQueueSize < MIN_SAMPLE_QUEUE_SIZE ||
        info->sampleQueueSize > MAX_SAMPLE_QUEUE_SIZE)
      info->sampleQueueSize = DEFAULT_SAMPLE_QUEUE_SIZE;
  }
  else
    info->sampleQueueSize = DEFAULT_SAMPLE_QUEUE_SIZE;

  // Read the last time registry information about this CPU was modified.

  DebugPrint(("RSTask: Reading last CPU info modification time.\n"));
  RtlInitUnicodeString(&miscUnicodeString, L"CpuInfoModificationTime");
  status = ZwQueryValueKey(keyHandle,
                           &miscUnicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Can't read CPU info modification time "
                "from registry (%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_READ_CPU_MODIFICATION_TIME,
                 status);
    ZwClose(keyHandle);
    return status;
  }
  if (keyValueInfo.Type != REG_BINARY ||
      keyValueInfo.DataLength != sizeof(LARGE_INTEGER)) {
    DebugPrint(("RSTask: ERROR - CPU modification time in registry is of "
                "an invalid type or length.\n"));
    BothLogShort(RS_TASK_ERROR_CPU_MODIFICATION_TIME_INVALID,
                 RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
    ZwClose(keyHandle);
    return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
  }
  RtlCopyMemory(&info->cpuInfoModificationTime,
                keyValueInfo.Data,
                sizeof(LARGE_INTEGER));

  // Read the available speeds.

  RtlInitUnicodeString(&miscUnicodeString, L"AvailableSpeeds");
  status = ZwQueryValueKey(keyHandle,
                           &miscUnicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not read available speeds (%x).\n",
                status));
    BothLogShort(RS_TASK_ERROR_CANT_READ_AVAILABLE_SPEEDS_REGISTRY_KEY,
                 status);
    ZwClose(keyHandle);
    return status;
  }
  if (keyValueInfo.Type != REG_BINARY) {
    DebugPrint(("RSTask: ERROR - Available speeds registry key has type %d, "
                "not REG_BINARY.\n", keyValueInfo.Type));
    BothLogShort(RS_TASK_ERROR_AVAILABLE_SPEEDS_REGISTRY_KEY_NOT_BINARY,
                 RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
    ZwClose(keyHandle);
    return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
  }
  if (keyValueInfo.DataLength < 8) {
    DebugPrint(("RSTask: ERROR - Fewer than 2 available speeds.\n"));
    BothLogShort(RS_TASK_ERROR_AVAILABLE_SPEEDS_REGISTRY_KEY_TOO_SMALL,
                 RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
    ZwClose(keyHandle);
    return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
  }
  if (keyValueInfo.DataLength % 4) {
    DebugPrint(("RSTask: ERROR - Available speeds length not a multiple "
                "of 4.\n"));
    BothLogShort(RS_TASK_ERROR_AVAILABLE_SPEEDS_REGISTRY_KEY_BAD_SIZE,
                 RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
    ZwClose(keyHandle);
    return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
  }
  info->numAvailableSpeeds = (short) (keyValueInfo.DataLength / 4);

  // Make sure the speeds are positive and in strictly increasing order.
  
  speedsKHz = (ULONG *) keyValueInfo.Data;
  if (speedsKHz[0] <= 0) {
    DebugPrint(("RSTask: ERROR - Minimum speed is not positive.\n"));
    BothLogShort(RS_TASK_ERROR_MINIMUM_SPEED_NOT_POSITIVE,
                 RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
    ZwClose(keyHandle);
    return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
  }
  for (i = 0; i < info->numAvailableSpeeds - 1; ++i) {
    if (speedsKHz[i] >= speedsKHz[i+1]) {
      DebugPrint(("RSTask: ERROR - Speeds are not in strictly increasing "
                  "order (%lu comes before %lu).\n",
                  speedsKHz[i], speedsKHz[i+1]));
      BothLogShort(RS_TASK_ERROR_AVAILABLE_SPEEDS_OUT_OF_ORDER,
                   RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
      ZwClose(keyHandle);
      return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
    }
  }

  // Allocate space for the available speeds info. There should be room
  // for all the available speeds, plus one extra entry to represent
  // using a variable-speed setting.

  info->availableSpeedsInfo = ExAllocatePool(NonPagedPool,
                                             sizeof(AVAILABLE_SPEED_INFO) *
                                             (info->numAvailableSpeeds + 1));
  if (!info->availableSpeedsInfo) {
    DebugPrint(("RSTask: ERROR - Could not allocate space for available "
                "speed info.\n"));
    BothLogShort(RS_TASK_ERROR_CANT_ALLOCATE_AVAILABLE_SPEED_INFO,
                 STATUS_INSUFFICIENT_RESOURCES);
    ZwClose(keyHandle);
    return STATUS_INSUFFICIENT_RESOURCES;
  }
  g.resourcesObtained |= AVAILABLE_SPEEDS_INFO_RESOURCE_FLAG;

  // Fill in as much as possible of the information about the available
  // speeds.
  
  for (i = 0; i < info->numAvailableSpeeds; ++i) {
    info->availableSpeedsInfo[i].cycPerMs = speedsKHz[i];
    info->availableSpeedsInfo[i].cpi = CycPerMsToCpi(speedsKHz[i]);
  }

  // Fill in information for the extra "speed", representing the
  // variable-speed setting. Treat it as if it were the same as the
  // lowest available speed.

  RtlCopyMemory(&info->availableSpeedsInfo[info->numAvailableSpeeds],
                &info->availableSpeedsInfo[0],
                sizeof(AVAILABLE_SPEED_INFO));

  // Set the shortcut fields related to available speeds:

  info->maxAvailableSpeedIndex = info->numAvailableSpeeds - 1;
  info->minAvailableCpi =
    info->availableSpeedsInfo[0].cpi;
  info->maxAvailableCpi =
    info->availableSpeedsInfo[info->maxAvailableSpeedIndex].cpi;

  // Read the speed codes for the available speeds.

  DebugPrint(("RSTask: Reading speed codes.\n"));
  RtlInitUnicodeString(&miscUnicodeString, L"AmdFidVidCodes");
  status = ZwQueryValueKey(keyHandle,
                           &miscUnicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not read speed codes (%x).\n",
                status));
    BothLogShort(RS_TASK_ERROR_CANT_READ_SPEED_CODES,
                 status);
    ZwClose(keyHandle);
    return status;
  }
  if (keyValueInfo.Type != REG_BINARY) {
    DebugPrint(("RSTask: ERROR - Speed code list has type %d, "
                "not REG_BINARY.\n", keyValueInfo.Type));
    BothLogShort(RS_TASK_ERROR_SPEED_CODES_NOT_BINARY,
                 RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
    ZwClose(keyHandle);
    return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
  }
  if (keyValueInfo.DataLength != (ULONG) (info->numAvailableSpeeds * 4)) {
    DebugPrint(("RSTask: ERROR - Speed code list has wrong length "
                "(%d instead of %d).\n", keyValueInfo.DataLength,
                info->numAvailableSpeeds * 4));
    BothLogShort(RS_TASK_ERROR_SPEED_CODES_WRONG_SIZE,
                 RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
    ZwClose(keyHandle);
    return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
  }

  // Copy the information about the speed codes into the available
  // speed info.

  speedCodes = (ULONG *) keyValueInfo.Data;
  for (i = 0; i < info->numAvailableSpeeds; ++i)
    info->availableSpeedsInfo[i].speedCode = speedCodes[i];
  info->minSpeedCode = speedCodes[0];
  info->maxSpeedCode = speedCodes[info->maxAvailableSpeedIndex];

  // Read the worthwhile settings information.

  DebugPrint(("RSTask: Reading worthwhile settings.\n"));
  RtlInitUnicodeString(&miscUnicodeString, L"WorthwhileSettings");
  status = ZwQueryValueKey(keyHandle,
                           &miscUnicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not read worthwhile settings (%x).\n",
                status));
    BothLogShort(RS_TASK_ERROR_CANT_READ_WORTHWHILE_SETTINGS_REGISTRY_KEY,
                 status);
    ZwClose(keyHandle);
    return status;
  }
  if (keyValueInfo.Type != REG_BINARY) {
    DebugPrint(("RSTask: ERROR - Worthwhile settings registry key has "
                "type %d, not REG_BINARY.\n", keyValueInfo.Type));
    BothLogShort(RS_TASK_ERROR_WORTHWHILE_SETTINGS_REGISTRY_KEY_NOT_BINARY,
                 RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
    ZwClose(keyHandle);
    return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
  }
  if (keyValueInfo.DataLength < 4) {
    DebugPrint(("RSTask: ERROR - Fewer than 2 worthwhile settings.\n"));
    BothLogShort(RS_TASK_ERROR_WORTHWHILE_SETTINGS_REGISTRY_KEY_TOO_SMALL,
                 RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
    ZwClose(keyHandle);
    return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
  }
  if (keyValueInfo.DataLength % 2) {
    DebugPrint(("RSTask: ERROR - Worthwhile settings length not a multiple "
                "of 2.\n"));
    BothLogShort(RS_TASK_ERROR_WORTHWHILE_SETTINGS_REGISTRY_KEY_BAD_SIZE,
                 RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
    ZwClose(keyHandle);
    return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
  }
  info->numWorthwhileSettings = (short) (keyValueInfo.DataLength / 2);
  info->maxWorthwhileSetting = info->numWorthwhileSettings - 1;

  // Make sure the worthwhile settings are valid, i.e., that they are
  // in strictly increasing order and use only available speeds.

  settings = (short *) keyValueInfo.Data;
  for (i = 0; i < info->maxWorthwhileSetting; ++i) {
    if (settings[i] >= settings[i+1]) {
      DebugPrint(("RSTask: ERROR - Worthwhile settings are not in strictly "
                  "increasing order (%d comes before %d).\n",
                  settings[i], settings[i+1]));
      BothLogShort(RS_TASK_ERROR_WORTHWHILE_SETTINGS_OUT_OF_ORDER,
                   RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
      ZwClose(keyHandle);
      return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
    }
  }
  if (settings[0] < 0 ||
      settings[info->maxWorthwhileSetting] >= info->numAvailableSpeeds) {
    DebugPrint(("RSTask: ERROR - Worthwhile settings go beyond available "
                "speeds.\n"));
    BothLogShort(RS_TASK_ERROR_WORTHWHILE_SETTINGS_GO_BEYOND_AVAILABLE,
                 RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
    ZwClose(keyHandle);
    return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
  }

  // Allocate space for the worthwhile setting information.

  info->worthwhileSettingsInfo =
    ExAllocatePool(NonPagedPool, sizeof(WORTHWHILE_SETTING_INFO) *
                   info->numWorthwhileSettings);
  if (!info->worthwhileSettingsInfo) {
    DebugPrint(("RSTask: ERROR - Could not allocate space for worthwhile "
                "settings info.\n"));
    BothLogShort(RS_TASK_ERROR_CANT_ALLOCATE_WORTHWHILE_SETTINGS_INFO,
                 STATUS_INSUFFICIENT_RESOURCES);
    return STATUS_INSUFFICIENT_RESOURCES;
  }
  g.resourcesObtained |= WORTHWHILE_SETTINGS_INFO_RESOURCE_FLAG;

  // Fill in the worthwhile settings information.

  for (i = 0; i < info->numWorthwhileSettings; ++i) {
    info->worthwhileSettingsInfo[i].cpi =
      info->availableSpeedsInfo[settings[i]].cpi;
    info->worthwhileSettingsInfo[i].speedIndex = settings[i];
  }

  // Set the shortcut fields related to worthwhile speeds:

  info->minWorthwhileSpeedIndex =
    info->worthwhileSettingsInfo[0].speedIndex;
  info->maxWorthwhileSpeedIndex =
    info->worthwhileSettingsInfo[info->maxWorthwhileSetting].speedIndex;
  info->minWorthwhileCpi =
    info->worthwhileSettingsInfo[0].cpi;
  info->maxWorthwhileCpi =
    info->worthwhileSettingsInfo[info->maxWorthwhileSetting].cpi;
  info->minWorthwhileCycPerMs =
    CYC_PER_MS_OF_INDEX(info->minWorthwhileSpeedIndex);
  info->maxWorthwhileCycPerMs =
    CYC_PER_MS_OF_INDEX(info->maxWorthwhileSpeedIndex);

  // Read the main directory of RightSpeed.

  DebugPrint(("RSTask: Reading main directory.\n"));
  RtlInitUnicodeString(&miscUnicodeString, L"MainDirectory");
  status = ZwQueryValueKey(keyHandle,
                           &miscUnicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not read main directory registry key "
                " (%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_READ_MAIN_DIRECTORY,
                 status);
    ZwClose(keyHandle);
    return status;
  }
  if (keyValueInfo.Type != REG_SZ) {
    DebugPrint(("RSTask: ERROR - Main directory registry entry is %lu, "
                "not REG_SZ\n", keyValueInfo.Type));
    BothLogShort(RS_TASK_ERROR_MAIN_DIRECTORY_IS_NOT_STRING,
                 RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
    ZwClose(keyHandle);
    return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
  }
  if (keyValueInfo.DataLength > MAX_RIGHTSPEED_DIRECTORY_LENGTH * 2) {
    DebugPrint(("RSTask: ERROR - Main directory is too big (%lu bytes).\n",
                keyValueInfo.DataLength));
    BothLogShort(RS_TASK_ERROR_MAIN_DIRECTORY_TOO_BIG,
                 RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
    ZwClose(keyHandle);
    return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
  }

  // Allocate the TTG directory Unicode string. Leave enough room for
  // the main directory plus \??\ and \TTG (plus a little extra),
  // remembering that we need 2 bytes per character due to Unicode.
  // So, allocate a total of 20 extra bytes.

  DebugPrint(("RSTask: Allocating and filling in TTG directory string.\n"));
  status = MyAllocateUnicodeString(&info->ttgDirectory,
                                   (USHORT)(keyValueInfo.DataLength + 20));
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not initialize TTG directory string "
                "(%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_INITIALIZE_TTG_DIRECTORY_STRING,
                 status);
    ZwClose(keyHandle);
    return status;
  }
  g.resourcesObtained |= TTG_DIRECTORY_RESOURCE_FLAG;

  // Fill in the TTG directory Unicode string using the main directory
  // string.

  status = FillInTTGDirectoryString(&info->ttgDirectory,
                                    (PWCHAR) keyValueInfo.Data);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not fill in TTG directory string "
                "using main directory string (%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_FILL_IN_TTG_DIRECTORY_STRING,
                 status);
    ZwClose(keyHandle);
    return status;
  }

  // We no longer need the registry key, so close the handle.

  ZwClose(keyHandle);

  // Read the speed curves information using the main directory information,
  // still in keyValueInfo.Data.

  return ReadSpeedCurvesInformation((PWCHAR) keyValueInfo.Data);
}

// This function reloads the registry information for configuring
// RSTask. If the TTG flush period has changed, it restarts the
// periodic flush.

NTSTATUS ReloadRegistryInfo (REGISTRY_INFO *info)
{
  NTSTATUS status;
  OBJECT_ATTRIBUTES objectAttributes;
  struct {
    ULONG   TitleIndex;
    ULONG   Type;
    ULONG   DataLength;
    ULONG   DataDword;
  } keyValueInfo;
  HANDLE keyHandle;
  ULONG keyValueInfoLength, newTTGFlushPeriodSec;
  UNICODE_STRING miscUnicodeString;

  // Open the registry key for RightSpeed.

  RtlInitUnicodeString(&miscUnicodeString,
    L"\\Registry\\Machine\\Software\\Jay Lorch\\RightSpeed\\CurrentVersion");
  InitializeObjectAttributes(&objectAttributes,
                             &miscUnicodeString,
                             OBJ_CASE_INSENSITIVE,
                             NULL,  // root directory
                             NULL); // security descriptor
  status = ZwOpenKey(&keyHandle,
                     KEY_QUERY_VALUE | KEY_SET_VALUE,
                     &objectAttributes);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not open main registry key (%x).\n",
                status));
    BothLogShortRepeat(RS_TASK_ERROR_CANT_OPEN_MAIN_REGISTRY_KEY,
                       status,
                       1);
    return status;
  }

  // Read the TTG flush period in seconds.

  RtlInitUnicodeString(&miscUnicodeString, L"TTGFlushPeriodSec");
  status = ZwQueryValueKey(keyHandle,
                           &miscUnicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (NT_SUCCESS(status) && keyValueInfo.Type == REG_DWORD) {
    newTTGFlushPeriodSec = keyValueInfo.DataDword;
    if (newTTGFlushPeriodSec < MIN_TTG_FLUSH_PERIOD_SEC ||
        newTTGFlushPeriodSec > MAX_TTG_FLUSH_PERIOD_SEC)
      newTTGFlushPeriodSec = DEFAULT_TTG_FLUSH_PERIOD_SEC;
  }
  else
    newTTGFlushPeriodSec = DEFAULT_TTG_FLUSH_PERIOD_SEC;

  // If the TTG flush period has changed, restart the periodic flush.

  if (newTTGFlushPeriodSec != info->ttgFlushPeriodSec) {
    info->ttgFlushPeriodSec = newTTGFlushPeriodSec;
    StartPeriodicFlush(&g.periodicFlushInfo);
  }

  // We no longer need the registry key, so close the handle.

  ZwClose(keyHandle);

  // Return success.

  return STATUS_SUCCESS;
}

// This function reads the speed curves information from the file on disk.

NTSTATUS ReadSpeedCurvesInformation (PWCHAR mainDirectory)
{
  WCHAR fileNameStr[MAX_RIGHTSPEED_DIRECTORY_LENGTH + 24];
  UNICODE_STRING fileNameUnicodeString;
  NTSTATUS status;
  IO_STATUS_BLOCK ioStatusBlock;
  OBJECT_ATTRIBUTES objectAttributes;
  HANDLE fileHandle;
  FILE_STANDARD_INFORMATION fileInformation;
  ULONG startingOffset, currentOffset, totalSpeedCurvesLength;
  ULONG whichValue, whichCurve, prevValue, currentValue;
  short entryLength;

  // Determine how big the speed curves data should be and allocate
  // space for it.

  g.registryInfo.speedCurveLength = g.registryInfo.numWorthwhileSettings - 1;
  g.registryInfo.numSpeedCurves =
    IDEAL_SCHEDULE_NUM_PHASES * g.registryInfo.speedCurveLength + 1;
  totalSpeedCurvesLength = g.registryInfo.numSpeedCurves *
                           g.registryInfo.speedCurveLength;
  g.registryInfo.speedCurves =
    (unsigned char *) ExAllocatePool(NonPagedPool, totalSpeedCurvesLength);
  if (!g.registryInfo.speedCurves) {
    DebugPrint(("RSTask: ERROR - Couldn't allocate space for speed curves "
                "data.\n"));
    BothLogShort(RS_TASK_ERROR_CANT_ALLOCATE_SPEED_CURVES,
                 STATUS_INSUFFICIENT_RESOURCES);
    return STATUS_INSUFFICIENT_RESOURCES;
  }
  g.resourcesObtained |= SPEED_CURVES_RESOURCE_FLAG;

  // Create the file name Unicode string from the main directory.

  fileNameUnicodeString.Buffer = fileNameStr;
  fileNameUnicodeString.Length = 0;
  fileNameUnicodeString.MaximumLength =
    (MAX_RIGHTSPEED_DIRECTORY_LENGTH + 20) * sizeof(WCHAR);
  status = FillInSpeedCurvesFileName(&fileNameUnicodeString,
                                     mainDirectory);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not create speed curves file name "
                "using main directory string (%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_MAKE_SPEED_CURVES_FILE_NAME,
                 status);
    return status;
  }

  // Open the speed curves file.

  InitializeObjectAttributes(&objectAttributes,
                             &fileNameUnicodeString,
                             OBJ_CASE_INSENSITIVE,
                             NULL, // root directory
                             NULL); // security descriptor
  status = ZwOpenFile(&fileHandle,
                      FILE_READ_DATA,
                      &objectAttributes,
                      &ioStatusBlock,
                      FILE_SHARE_READ,
                      FILE_SEQUENTIAL_ONLY | FILE_SYNCHRONOUS_IO_ALERT);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not open speed curves file (%x).\n",
                status));
    BothLogShort(RS_TASK_ERROR_CANT_OPEN_SPEED_CURVES_FILE,
                 status);
    return status;
  }

  // Make sure the file has the right length.

  status = ZwQueryInformationFile(fileHandle,
                                  &ioStatusBlock,
                                  &fileInformation,
                                  sizeof(fileInformation),
                                  FileStandardInformation);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not query size of speed curves file "
                "(%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_QUERY_SIZE_SPEED_CURVES_FILE,
                 status);
    ZwClose(fileHandle);
    return status;
  }
  if (fileInformation.EndOfFile.LowPart != totalSpeedCurvesLength) {
    DebugPrint(("RSTask: ERROR - Speed curves file has invalid size "
                "(%lu instead of %lu).\n",
                fileInformation.EndOfFile.LowPart,
                totalSpeedCurvesLength));
    BothLogShort(RS_TASK_ERROR_INVALID_SPEED_CURVES_FILE_SIZE,
                 RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
    ZwClose(fileHandle);
    return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
  }

  // Read the data from the file into the allocated buffer.

  status = ZwReadFile(fileHandle,
                      NULL, // event
                      NULL, // APC routine
                      NULL, // APC context
                      &ioStatusBlock,
                      g.registryInfo.speedCurves,
                      totalSpeedCurvesLength,
                      NULL, // byte offset: default 0
                      NULL); // key
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not read speed curves data (%x).\n",
                status));
    BothLogShort(RS_TASK_ERROR_CANT_READ_SPEED_CURVES,
                 status);
    ZwClose(fileHandle);
    return status;
  }

  // Close the file handle as we no longer need it open.

  ZwClose(fileHandle);

  // Check that the speed curve data is valid.  In particular, make sure
  // each curve is in nondecreasing order, and make sure that each element
  // of each curve is a value not greater that IDEAL_SCHEDULE_NUM_PHASES.

  for (whichCurve = 0; whichCurve < g.registryInfo.numSpeedCurves;
       ++whichCurve) {
    for (prevValue = 0, whichValue = 0;
         whichValue < g.registryInfo.speedCurveLength;
         prevValue = currentValue, ++whichValue) {
      currentValue = SPEED_CURVE_VALUE(whichCurve, whichValue);
      if (currentValue < prevValue) {
        DebugPrint(("RSTask: ERROR - Invalid speed curve; value #%lu in curve "
                    "#%lu is %lu, which is less than previous value %lu.\n",
                    whichValue,
                    whichCurve,
                    currentValue,
                    prevValue));
        BothLogTwoParameters(RS_TASK_ERROR_BAD_ORDER_IN_SPEED_CURVES,
                             RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA,
                             whichCurve,
                             whichValue);
        return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
      }
      if (currentValue > IDEAL_SCHEDULE_NUM_PHASES) {
        DebugPrint(("RSTask: ERROR - Invalid speed curve; value #%lu in curve "
                    "#%lu is %lu, which is too big.\n",
                    whichValue,
                    whichCurve,
                    currentValue));
        BothLogTwoParameters(RS_TASK_ERROR_BAD_ENTRY_IN_SPEED_CURVES,
                             RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA,
                             whichCurve,
                             whichValue);
        return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
      }
    }
  }

  // Everything is OK, so return success.

  return STATUS_SUCCESS;
}
