// This file contains routines dealing with the queue of sample values.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "Globals.h"
#include "Log.h"
#include "SampleQ.h"
#include "Update.h"

/////////////////////////////////////////////////////////////////////////////
// INITIALIZATION / FREEING
/////////////////////////////////////////////////////////////////////////////

// This function initializes the sample queue. It allocates space for
// the sample queue entries.

NTSTATUS InitializeSampleQueue (PSAMPLE_QUEUE_INFO info)
{
  ULONG i;

  // Initialize the fields of the info structure.

  KeInitializeSpinLock(&info->spinLock);
  info->maxSize = g.registryInfo.sampleQueueSize;
  info->next = 0;
  info->size = 0;

  // Allocate space for the sample queue entries.

  info->entries = ExAllocatePool(NonPagedPool,
                                 sizeof(SAMPLE_QUEUE_ENTRY) *
                                 g.registryInfo.sampleQueueSize);
  if (!info->entries) {
    DebugPrint(("RSTask: ERROR - Can't allocate space for sample queue.\n"));
    BothLogShort(RS_TASK_ERROR_CANT_ALLOCATE_SAMPLE_QUEUE,
                 STATUS_INSUFFICIENT_RESOURCES);
    return STATUS_INSUFFICIENT_RESOURCES;
  }

  // Return success.

  return STATUS_SUCCESS;
}

// This function frees any resources consumed by the sample queue.

void CloseSampleQueue (PSAMPLE_QUEUE_INFO info)
{
  ExFreePool(info->entries);
}

/////////////////////////////////////////////////////////////////////////////
// SAMPLE QUEUE ROUTINES
/////////////////////////////////////////////////////////////////////////////

// This function adds a sample value to the queue to be processed
// later. It assumes that the caller is currently referencing the task
// type group file passed, and is passing on the responsibility for
// that referencing to the sample queue structure with this call.

void AddSampleValueToQueue (
  LONGLONG cycles,    // the number of cycles this run took
  PTT_INFO ttInfo,    // the task type of this task
  PTTG_FILE ttgFile   // the task type group this task type belongs to
  )
{
  KIRQL oldirql;
  PSAMPLE_QUEUE_ENTRY entry;

  // Acquire the spin lock controlling access to the sample queue.

  KeAcquireSpinLock(&g.sampleQueueInfo.spinLock, &oldirql);

  // If there isn't room in the sample queue for this entry, drop it.
  // This requires dereferencing the task type group file immediately.

  if (g.sampleQueueInfo.size >= g.sampleQueueInfo.maxSize) {
    KeReleaseSpinLock(&g.sampleQueueInfo.spinLock, oldirql);
    DereferenceTTGFile(ttgFile);

    DebugPrint(("RSTask: WARNING - Sample queue full, so dropping sample "
                "value.\n"));
    if (LOGGING_WARNINGS)
      LogError(RS_TASK_WARNING_SAMPLE_QUEUE_FULL);
    return;
  }

  // Fill in the next entry with the data about this sample value.

  entry = &g.sampleQueueInfo.entries[g.sampleQueueInfo.next];
  entry->ttgFile = ttgFile;
  entry->ttInfo = ttInfo;
  entry->cycles = cycles;

  // Add the entry to the queue by incrementing the size and the
  // 'next' index.

  InterlockedIncrement(&g.sampleQueueInfo.size);
  ++g.sampleQueueInfo.next;
  if (g.sampleQueueInfo.next >= g.sampleQueueInfo.maxSize)
    g.sampleQueueInfo.next = 0;

  // We're done with the sample queue info structure, so release the
  // spin lock.

  KeReleaseSpinLock(&g.sampleQueueInfo.spinLock, oldirql);
}

// This function processes and removes all sample values in the queue.
// It assumes that floating-point operations can be done without
// saving and restoring floating-point state. It also assumes that
// only one thread, the one calling this routine, ever removes entries
// from the queue.

void ProcessSamplesInQueue (void)
{
  LONG numToProcess, currentPos, nextPos;
  PSAMPLE_QUEUE_ENTRY currentEntry, nextEntry;
  KIRQL oldirql;

  DebugPrint(("RSTask: Processing all samples in queue.\n"));

  // Acquire the spin lock just long enough to determine how many
  // entries we have to process and where those entries start.

  KeAcquireSpinLock(&g.sampleQueueInfo.spinLock, &oldirql);
  numToProcess = g.sampleQueueInfo.size;
  currentPos = g.sampleQueueInfo.next - g.sampleQueueInfo.size;
  KeReleaseSpinLock(&g.sampleQueueInfo.spinLock, oldirql);

  // If there are no entries to process, we're done.

  if (numToProcess < 1)
    return;

  // Correct the 'current' position if it's less than 0, and compute
  // the location of the current entry.

  if (currentPos < 0)
    currentPos += g.sampleQueueInfo.maxSize;
  currentEntry = &g.sampleQueueInfo.entries[currentPos];

  // Loop through all the entries that we have to process except the
  // last one.

  while (numToProcess > 1) {
    // Find the position and location of the next entry.

    nextPos = (currentPos + 1) % g.sampleQueueInfo.maxSize;
    nextEntry = &g.sampleQueueInfo.entries[nextPos];
    
    // Process the sample value from the queue. If it uses the same
    // task type as the next one, don't bother recomputing schedules
    // as the next one will just recompute the schedule again.

    if (currentEntry->ttInfo == nextEntry->ttInfo)
      AddSampleValue(currentEntry->cycles,
                     currentEntry->ttInfo,
                     &currentEntry->ttgFile->spinLock);
    else
      AddSampleValueAndRecomputeSchedule(currentEntry->cycles,
                                         currentEntry->ttInfo,
                                         &currentEntry->ttgFile->spinLock);

    // The last thing to do is dereference the task type group file.
    
    DereferenceTTGFile(currentEntry->ttgFile);

    // Reduce the size of the queue by one since we've processed this
    // entry.

    InterlockedDecrement(&g.sampleQueueInfo.size);

    // Advance to the next entry.

    currentPos = nextPos;
    currentEntry = nextEntry;
    --numToProcess;
  }

  // Now, process the final entry.

  AddSampleValueAndRecomputeSchedule(currentEntry->cycles,
                                     currentEntry->ttInfo,
                                     &currentEntry->ttgFile->spinLock);
  DereferenceTTGFile(currentEntry->ttgFile);
  InterlockedDecrement(&g.sampleQueueInfo.size);
}
