// This file contains routines related to simple schedules.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "SimpleSched.h"
#include "MyFloat.h"
#include "Globals.h"

/////////////////////////////////////////////////////////////////////////////
// MACROS
/////////////////////////////////////////////////////////////////////////////

// This macro computes the minimum of two numbers:
#define MIN(a,b) ((a)<(b)?(a):(b))

/////////////////////////////////////////////////////////////////////////////
// SCHEDULE MODIFICATION
/////////////////////////////////////////////////////////////////////////////

// This function extends the given speed schedule to use the maximum
// available speed at the end.

void AddPostDeadlinePart (PSPEED_SCHEDULE schedule)
{
  PSPEED_SCHEDULE_PHASE lastPhase = &schedule->phase[schedule->numPhases - 1];

  if (lastPhase->speedIndex == g.registryInfo.maxAvailableSpeedIndex) {
    // If the last phase already uses the maximum available speed,
    // just make that phase last forever.

    lastPhase->intervals = 0;
  }
  else {
    // Otherwise, create an additional phase at the end that uses the
    // maximum available speed and that lasts forever.

    ++lastPhase;
    ++schedule->numPhases;
    lastPhase->speedIndex = g.registryInfo.maxAvailableSpeedIndex;
    lastPhase->intervals = 0;
  }
}

// This function adds an additional phase at the end of the given
// schedule with the given speed index and number of intervals. If
// this phase can be incorporated into the last existing phase, that
// is used instead. It assumes the schedule is not complete, i.e., the
// schedule does not end with an indefinitely long phase.

void AddPhase (PSPEED_SCHEDULE schedule, ULONG speedIndex, long intervals)
{
  PSPEED_SCHEDULE_PHASE lastPhase = &schedule->phase[schedule->numPhases - 1];

  if (lastPhase >= schedule->phase && lastPhase->speedIndex == speedIndex) {
    // If the last phase already uses the speed index we're adding, then
    // just extend that phase by the given number of intervals.

    lastPhase->intervals += intervals;
  }
  else {
    ++lastPhase;
    ++schedule->numPhases;
    lastPhase->speedIndex = speedIndex;
    lastPhase->intervals = intervals;
  }
}

/////////////////////////////////////////////////////////////////////////////
// SIMPLE SCHEDULE CREATORS
/////////////////////////////////////////////////////////////////////////////

// This function computes a Flat/Chan style speed schedule and puts it
// in the location pointed to by 'schedule'. This routine assumes that
// schedule has space for at least three phases. The other parameters
// indicate the average speed that this schedule should attain before
// the deadline (in cycles per interval) and the deadline (in
// intervals).

void ComputeFlatSchedule (
  PSPEED_SCHEDULE schedule,
  long deadlineI,
  double desiredCpi
  )
{
  long leftSetting, rightSetting, leftI, rightI;
  double leftCpi, rightCpi;

  // If the speed needed is no higher than the minimum worthwhile
  // speed, just use that constant speed until the deadline.

  if (desiredCpi <= g.registryInfo.minWorthwhileCpi) {
    schedule->numPhases = 1;
    schedule->phase[0].speedIndex = g.registryInfo.minWorthwhileSpeedIndex;
    schedule->phase[0].intervals = deadlineI;
    AddPostDeadlinePart(schedule);
    return;
  }

  // If the speed needed is no lower than the maximum worthwhile
  // speed, just use that constant speed until the deadline.

  if (desiredCpi >= g.registryInfo.maxWorthwhileCpi) {
    schedule->numPhases = 1;
    schedule->phase[0].speedIndex = g.registryInfo.maxWorthwhileSpeedIndex;
    schedule->phase[0].intervals = deadlineI;
    AddPostDeadlinePart(schedule);
    return;
  }

  // Find the two consecutive worthwhile settings 'leftSetting' and
  // 'rightSetting' such that the left setting has a speed less than
  // or equal to that desired and the right setting has a speed
  // greater than that desired.

  rightSetting = 1;
  while (CPI_OF_SETTING(rightSetting) <= desiredCpi)
    ++rightSetting;
  leftSetting = rightSetting - 1;

  // Determine how many intervals should be spent at the 'left' and
  // 'right' speeds. We want to solve the equations:
  //
  // (1) leftI + rightI = deadlineI
  // (2) leftCpi * leftI + rightCpi * rightI = desiredCpi * deadlineI
  //
  // The solution to this is:
  //
  // leftI = deadlineI * f
  // rightI = deadlineI * (1 - f)
  //
  // where
  //
  // f = (rightCpi - desiredCpi) / (rightCpi - leftCpi).
  //
  // To make the intervals both integral, we have to round. We round
  // the 'left' intervals down because these intervals use the slower
  // speed, and this ensures that the actual speed attained is at
  // least as high as the desired speed. (However, we add 0.0001
  // before rounding down so that if there is an exact answer we can
  // find it without worrying about accumulated errors.)

  leftCpi = CPI_OF_SETTING(leftSetting);
  rightCpi = CPI_OF_SETTING(rightSetting);
  leftI = RoundDown(deadlineI * (rightCpi - desiredCpi) /
                    (rightCpi - leftCpi) + 0.0001);
  rightI = deadlineI - leftI;

  // Create the schedule using the determined numbers of intervals.
  // Don't create a schedule phase that's 0 intervals, though.

  if (rightI <= 0) {
    schedule->numPhases = 1;
    schedule->phase[0].speedIndex = INDEX_OF_SETTING(leftSetting);
    schedule->phase[0].intervals = deadlineI;
  }
  else if (leftI <= 0) {
    schedule->numPhases = 1;
    schedule->phase[0].speedIndex = INDEX_OF_SETTING(rightSetting);
    schedule->phase[0].intervals = deadlineI;
  }
  else {
    schedule->numPhases = 2;
    schedule->phase[0].speedIndex = INDEX_OF_SETTING(leftSetting);
    schedule->phase[0].intervals = leftI;
    schedule->phase[1].speedIndex = INDEX_OF_SETTING(rightSetting);
    schedule->phase[1].intervals = rightI;
  }

  // Finally, extend the schedule to cover the post-deadline part.

  AddPostDeadlinePart(schedule);
}

// This function computes a LongShort/Chan style speed schedule and
// puts it in the location pointed to by 'schedule'. This routine
// assumes that schedule has space for a number of phases at least
// equal to the number of available speeds. The other parameters
// indicate the LongShort/Chan update period to simulate (in
// intervals) and the deadline (in intervals).

void ComputeLongShortSchedule (
  PSPEED_SCHEDULE schedule,
  long deadlineI,
  long updatePeriodI
  )
{
  long currentSetting = 0, intervalsRemaining, intervals;
  int utilizations[12] = { 0, 3, 6, 9, 10, 11, 12, 13, 14, 15, 16, 17 };
  int i;
  double desiredCpi;

  // Initialize the schedule to be empty. We'll add phases for each
  // update period.

  schedule->numPhases = 0;

  // Loop through the 12 update periods, setting the speed for each
  // of them. Stop when we finish all 12 update periods or we reach
  // the deadline. Keep track of the intervals remaining before the
  // deadline so we can stop if that reaches 0.

  intervalsRemaining = deadlineI;
  for (i = 0; i < 12 && intervalsRemaining; ++i) {
    // The desired cycles per interval for an update period is based on
    // the average utilization observed as a fraction of the maximum.
    // The utilizations are given as numerators of fractions where 18
    // is the denominator.
    
    desiredCpi = g.registryInfo.maxWorthwhileCpi * utilizations[i] / 18.0;

    // Search for the first setting that has speed at least as high
    // as the needed one.
    
    while (desiredCpi > CPI_OF_SETTING(currentSetting))
      ++currentSetting;

    // Add this update period to the schedule. Make sure not to add
    // more intervals than there are remaining.

    intervals = MIN(updatePeriodI, intervalsRemaining);
    AddPhase(schedule, INDEX_OF_SETTING(currentSetting), intervals);

    // Update the intervals remaining before the deadline.

    intervalsRemaining -= intervals;
  }

  // Finally, add the post-deadline part of the schedule.
  
  AddPostDeadlinePart(schedule);
}

// This function computes a Past/Peg style speed schedule and puts it
// in the location pointed to by 'schedule'. This routine assumes that
// schedule has space for at least three phases. The other parameters
// indicate the Past/Peg update period to simulate (in intervals) and
// the deadline (in intervals).

void ComputePastPegSchedule (
  PSPEED_SCHEDULE schedule,
  long deadlineI,
  long updatePeriodI
  )
{
  // Start with a phase using the slowest worthwhile speed and lasting
  // for the update period (but no longer than the deadline).

  schedule->numPhases = 1;
  schedule->phase[0].speedIndex = g.registryInfo.minWorthwhileSpeedIndex;
  schedule->phase[0].intervals = MIN(updatePeriodI, deadlineI);

  // Finally, add the post-deadline part of the schedule.
  
  AddPostDeadlinePart(schedule);
}

// This function fills in the given schedule with a LongRun-like speed
// schedule. It assumes there's enough room in this schedule for one
// phase per available speed.

void ComputeLongRunSchedule (
  PSPEED_SCHEDULE schedule
  )
{
  short i;

  // There is one phase for each available speed. The first phase takes
  // 7 milliseconds, the next ones 5 milliseconds, and the last one
  // should be of indefinite length.

  for (i = 0; i < g.registryInfo.numAvailableSpeeds; ++i) {
    schedule->phase[i].speedIndex = i;
    if (i == 0)
      schedule->phase[i].intervals = 7;
    else if (i == g.registryInfo.numAvailableSpeeds - 1)
      schedule->phase[i].intervals = 0;
    else
      schedule->phase[i].intervals = 5;
  }

  schedule->numPhases = g.registryInfo.numAvailableSpeeds;
}

/////////////////////////////////////////////////////////////////////////////
// SCHEDULE PROPERTIES
/////////////////////////////////////////////////////////////////////////////

// This function computes the PDC of the given schedule assuming the
// given deadline (in intervals).

double GetPdcOfSchedule (
  PSPEED_SCHEDULE schedule,
  long deadlineI
  )
{
  long intervalsRemaining = deadlineI, intervals;
  PSPEED_SCHEDULE_PHASE currentPhase = schedule->phase;
  double pdc = 0.0;

  // Loop through the schedule's phases to count the PDC. Keep track
  // of the intervals remaining before the deadline.

  for (currentPhase = schedule->phase; intervalsRemaining; ++currentPhase) {
    // The intervals to consider for this phase are the intervals of
    // this phase, unless they are more than the intervals remaining.
    // A special case is if there are '0' intervals in this phase,
    // meaning this phase is indefinitely long.

    if (currentPhase->intervals == 0 ||
        currentPhase->intervals > intervalsRemaining)
      intervals = intervalsRemaining;
    else
      intervals = currentPhase->intervals;

    // Increment the PDC and decrement the intervals remaining.

    pdc += CPI_OF_INDEX(currentPhase->speedIndex) * intervals;
    intervalsRemaining -= intervals;
  }

  // Return the computed PDC.

  return pdc;
}

// This function returns a boolean indicating whether the given speed
// schedule is valid.

BOOLEAN IsSpeedScheduleValid (
  PSPEED_SCHEDULE schedule
  )
{
  short i;

  // Make sure the number of phases is valid. It must be positive, no
  // greater than the maximum number of phases for that schedule, and
  // no greater than the maximum number of schedule phases allowable.

  if (schedule->numPhases < 1 ||
      schedule->numPhases > schedule->maxNumPhases ||
      schedule->numPhases > MAX_NUM_SCHEDULE_PHASES) {
    DebugPrint(("RSTask: WARNING - Schedule has an invalid number of phases "
                "(%d).\n", schedule->numPhases));
    return FALSE;
  }

  // Next, make sure that each phase has a valid speed index, i.e.,
  // one that is less than the number of available speeds. Also, make
  // sure that each phase has a valid number of intervals: the last
  // phase should have 0 intervals while the other phases should have
  // a positive number of intervals.

  for (i = 0; i < schedule->numPhases; ++i) {
    if (schedule->phase[i].speedIndex >=
        (ULONG) g.registryInfo.numAvailableSpeeds) {
      DebugPrint(("RSTask: WARNING - Invalid speed index (%lu) in "
                  "schedule.\n", schedule->phase[i].speedIndex));
      return FALSE;
    }
    if (i < schedule->numPhases - 1) {
      if (schedule->phase[i].intervals < 1) {
        DebugPrint(("RSTask: WARNING - Nonpositive # of intervals (%ld) in "
                    "schedule.\n", schedule->phase[i].intervals));
        return FALSE;
      }
    }
    else {
      if (schedule->phase[i].intervals != 0) {
        DebugPrint(("RSTask: WARNING - Nonzero # of intervals in last phase "
                    "of schedule.\n"));
        return FALSE;
      }
    }
  }

  // If everything's checked out OK, return TRUE.

  return TRUE;
}
