// This file contains routines for controlling the speed of the CPU.

#ifndef _RSTASK_SPEED_CONTROL_H
#define _RSTASK_SPEED_CONTROL_H

#include "../Common/Schedule.h"

/////////////////////////////////////////////////////////////////////////////
// MACROS
/////////////////////////////////////////////////////////////////////////////

// These are the possible states of the speed controller:

#define SPEED_CONTROLLER_STATE_IDLE                  0
#define SPEED_CONTROLLER_STATE_MAXIMUM               1
#define SPEED_CONTROLLER_STATE_SCHEDULE              2
#define SPEED_CONTROLLER_STATE_SCHEDULE_PAUSED       3
#define SPEED_CONTROLLER_STATE_SCHEDULE_RESUMING     4

/////////////////////////////////////////////////////////////////////////////
// STRUCTURES
/////////////////////////////////////////////////////////////////////////////

typedef struct _speed_control_info {

  // This is the spin lock controlling access to this structure:
  KSPIN_LOCK spinLock;

  // This is the timer used for scheduling speed changes:
  KTIMER timer;

  // This is the DPC used for advancing to the next phase of the
  // current speed schedule:
  KDPC phaseAdvanceDpc;

  // This is the DPC used to restore the default speed policy after
  // a short period spent idle:
  KDPC restorePolicyDpc;

  // This is the current speed index (we keep track of it so
  // that we don't pointlessly issue a request to change it):
  ULONG currentSpeedIndex;

  // This is the state of the speed controller:
  int state;

  ////////////////////////////////////////////////////////////////////////
  // The following field is only valid when the state is 'schedule-
  // paused'.
  ////////////////////////////////////////////////////////////////////////

  // This is the timestamp counter recorded when the schedule was most
  // recently paused. (It's only valid if we're in a schedule-paused
  // state.)
  LONGLONG pauseTSC;

  /////////////////////////////////////////////////////////////////////////
  // The following fields are only valid when the state is 'schedule'.
  /////////////////////////////////////////////////////////////////////////

  // This is the timestamp counter recorded when the current schedule
  // started:
  LONGLONG scheduleStartTSC;

  // This is the timestamp counter recorded when the speed was last
  // changed during the current schedule:
  LONGLONG lastSpeedChangeTSC;

  // This is a buffer that can hold up to MAX_NUM_SCHEDULE_PHASES
  // phase. It's used to hold the phases of the current schedule.
  SPEED_SCHEDULE_PHASE phaseBuffer[MAX_NUM_SCHEDULE_PHASES];

  // This is the schedule that's currently active:
  SPEED_SCHEDULE_HEADER schedule;

  // This is the ID of the current schedule. It's incremented each
  // time the schedule changes:
  LONG currentScheduleId;

  // This is the number of hundreds of nanoseconds by which we still
  // should reduce this schedule's length due to time spent performing
  // I/O. It may be negative, since we might have overcorrected when
  // reducing a schedule's length:
  LONG accruedTimeToReduceHns;

} SPEED_CONTROLLER_INFO, *PSPEED_CONTROLLER_INFO;

/////////////////////////////////////////////////////////////////////////////
// FUNCTION PROTOTYPES
/////////////////////////////////////////////////////////////////////////////

void InitializeSpeedController (PSPEED_CONTROLLER_INFO info);
void CloseSpeedController (PSPEED_CONTROLLER_INFO info);

void UseMaximumSpeed (void);
void UseIdleSpeed (void);

void StartSpeedSchedule (PSPEED_SCHEDULE schedule, PKSPIN_LOCK spinLock);
LONGLONG GetCyclesSinceScheduleStart (void);

void PauseSchedule (void);
void ResumeSchedule (void);

#endif
