// This file contains definitions for task type group files.

#ifndef _TTG_FILE_H
#define _TTG_FILE_H

#include "../Common/TTG.h"

/////////////////////////////////////////////////////////////////////////////
// MACROS
/////////////////////////////////////////////////////////////////////////////

// This is the number of buckets to use in the hash table for accessing
// a task type group file by name:
#define NUM_TTG_FILE_HASH_BUCKETS                257

/////////////////////////////////////////////////////////////////////////////
// STRUCTURES
/////////////////////////////////////////////////////////////////////////////

// This structure contains the hash table data for an open task
// type group file:

typedef struct _ttg_file_hash_info {

  // This boolean indicates if the file is in the hash table, i.e., if
  // the subsequent fields are valid.
  BOOLEAN valid;

  // This is a string containing the name of the file in uppercase.
  // Note that name.Buffer points to a buffer allocated by
  // RtlUpcaseUnicodeString that must eventually be freed by a call to
  // RtlFreeUnicodeString when the file is removed from the hash
  // table.
  UNICODE_STRING name;

  // This is the pointer to the next task type group file in the hash.
  struct _ttg_file *next;

  // This is the address of the pointer that's pointing to this
  // structure. This pointer may be some other structure's 'next'
  // element, or the head pointer of a hash bucket.
  struct _ttg_file **prev;

} TTG_FILE_HASH_INFO, *PTTG_FILE_HASH_INFO;

// This structure contains information about an open task type group
// file, i.e., one loaded into memory.

typedef struct _ttg_file {

  // This field is the part of the hash table associated with this
  // file. It can be considered to be part of the hash table.
  TTG_FILE_HASH_INFO hashTableInfo;

  // This is the number of entities referencing this file in memory.
  // If it is zero, that means this file is being opened or closed
  // by some thread, and that thread has exclusive access to all the
  // remaining fields in this structure.
  LONG referenceCount;

  // This spin lock protects access to the rest of this structure.
  // Note that it does not protect access to the reference count or
  // the hash table information; those parts are protected by the spin
  // lock that protects the lists of open files.
  KSPIN_LOCK spinLock;

  // This is the location in memory where the contents of the file
  // were copied. It is 0 if we were unable to allocate space for
  // the contents of the file:
  PTTG_INFO contents;

  // This is the handle we have open to the real file on disk. If we
  // don't have such an open file handle, this will be NULL:
  HANDLE fileHandle;

  // This boolean indicates whether we should ever write back any
  // changes we make to the in-memory copy to the disk file:
  BOOLEAN writeBack;

  // This boolean indicates whether there is a flush of this file
  // to disk in progress. It acts as a lock on the on-disk file:
  // before flushing, one makes sure this is false, then sets it
  // to true for as long as it takes to flush the file.
  BOOLEAN flushInProgress;

} TTG_FILE, *PTTG_FILE, **PPTTG_FILE;

// This structure is used to keep track of open task type group files:

typedef struct _open_ttg_files_info {

  // This spin lock protects access to this structure as well as to
  // the reference count and hash table information fields of all task
  // type group file structures.
  KSPIN_LOCK spinLock;

  // This lookaside list is used for allocating and freeing task type
  // group file structures:
  NPAGED_LOOKASIDE_LIST lookasideList;

  // This hash table contains pointers to all the open task type group
  // files indexed by a hash of their file names. (It does not include
  // files that have been disassociated with the corresponding file
  // on disk, however.)
  PTTG_FILE hashBucketHead[NUM_TTG_FILE_HASH_BUCKETS];

  // This counter keeps track of how many files are open.
  long numOpenFiles;

  // This handle to the task type group directory remains open to
  // allow us to easily open files in it. When not open, this value
  // is NULL.
  HANDLE                       ttgDirectoryHandle;

  // This boolean indicates whether we're uninstalling RightSpeed.
  // If so, we shouldn't open any further files.
  BOOLEAN                      uninstalling;

} OPEN_TTG_FILES_INFO, *POPEN_TTG_FILES_INFO;

/////////////////////////////////////////////////////////////////////////////
// FUNCTION PROTOTYPES
/////////////////////////////////////////////////////////////////////////////

NTSTATUS InitializeOpenTTGFilesInfo (POPEN_TTG_FILES_INFO info);
void CloseOpenTTGFilesInfo (POPEN_TTG_FILES_INFO info);
NTSTATUS PrepareTTGFilesInfoForUninstall (POPEN_TTG_FILES_INFO info);

NTSTATUS OpenTTGFile (PFILE_OBJECT fileObject);
void CloseTTGFile (PTTG_FILE ttgFile);
NTSTATUS FlushTTGFile (PTTG_FILE ttgFile, PVOID buf);
void FlushAllTTGFiles (void);
NTSTATUS AllowFileModificationTTGFile (PTTG_FILE ttgFile);

void ReferenceTTGFile (PTTG_FILE ttgFile);
void DereferenceTTGFile (PTTG_FILE ttgFile);

#endif
