// This file contains routines for handling the set of ongoing tasks.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/TTG.h"
#include "../Common/ErrCodes.h"
#include "Tasks.h"
#include "Globals.h"
#include "Log.h"
#include "SpeedControl.h"
#include "SampleQ.h"

// This function initializes the structure that manages ongoing tasks.

void InitializeOngoingTasks (PONGOING_TASKS_INFO info)
{
  KeInitializeSpinLock(&info->spinLock);
  info->lastTaskId = 0;
  g.sharedState->numOngoingTasks = 0;
  info->ttInfoApplies = FALSE;
}

// This function releases all resources associated with the structure
// managing ongoing tasks. This involves dereferencing any task type
// group file it references.

void CloseOngoingTasks (PONGOING_TASKS_INFO info)
{
  KIRQL oldirql;
  PTTG_FILE ttgFile = 0;

  // Acquire the spin lock so that we can read the task type group
  // file we have a reference to (if any) and modify the structure so
  // that it has no ongoing tasks and therefore references no TTG
  // file.

  KeAcquireSpinLock(&info->spinLock, &oldirql);
  if (info->ttInfoApplies) {
    ttgFile = info->ttgFile;
    info->ttInfoApplies = FALSE;
  }
  g.sharedState->numOngoingTasks = 0;
  KeReleaseSpinLock(&info->spinLock, oldirql);

  // If we were referencing a TTG file, dereference it.

  if (ttgFile)
    DereferenceTTGFile(ttgFile);
}

// This function begins a new task. It is passed pointers to the task
// type information (ttInfo) and the task type group file to which
// this task type belongs (ttgFile). It puts the new task ID in the
// given location (taskIdPtr). It returns a status value indicating
// whether it successfully began the new task.

NTSTATUS BeginTask (PTTG_FILE ttgFile, PTT_INFO ttInfo, long *taskIdPtr)
{
  KIRQL oldirql;
  PTTG_FILE ttgFileToDereference;

  DebugPrint(("RSTask: Beginning task at %x, %x\n", ttgFile, ttInfo));

  // Acquire the spin lock for the ongoing tasks information structure
  // before reading and modifying it.

  KeAcquireSpinLock(&g.ongoingTasksInfo.spinLock, &oldirql);

  // If there are no ongoing tasks, this task will become the new
  // single ongoing task.

  if (g.sharedState->numOngoingTasks == 0) {
    // Use as the task ID the last task ID assigned to any task plus 2.
    // We only use even task ID's so that 0xFFFFFFFF signifies an invalid
    // task ID.

    g.ongoingTasksInfo.lastTaskId += 2;
    *taskIdPtr = g.ongoingTasksInfo.lastTaskId;

    // Put the new task ID into the list of ongoing tasks.
    
    g.ongoingTasksInfo.ongoingTaskIds[0] = *taskIdPtr;
    g.sharedState->numOngoingTasks = 1;

    // Initialize the parts of the structure dealing with the single
    // ongoing task.
    
    g.ongoingTasksInfo.ttInfoApplies = TRUE;
    g.ongoingTasksInfo.ttInfo = ttInfo;
    g.ongoingTasksInfo.ioOccurredDuringTask = FALSE;
    g.ongoingTasksInfo.ttgFile = ttgFile;
    ReferenceTTGFile(ttgFile);  // this structure references the TTG file

    // Release the ongoing tasks info spin lock now that we're done
    // updating it.

    KeReleaseSpinLock(&g.ongoingTasksInfo.spinLock, oldirql);

    // Log the task's beginning if we're logging tasks.

    DebugPrint(("RSTask: Beginning task %ld when when no other task is "
                "ongoing.\n", *taskIdPtr));
    if (LOGGING_TASKS)
      LogAnyEventTypeAndLong(EVENT_TYPE_BEGIN_TASK, *taskIdPtr);

    // Tell the speed controller to begin scheduling the CPU according
    // to the schedule for this task type.

    StartSpeedSchedule(&ttInfo->schedule, &ttgFile->spinLock);

    // Signal the idle thread so that it can end this task when the
    // system next goes idle.

    KeSetEvent(&g.idleThreadInfo.threadWaitEvent, IO_NO_INCREMENT, FALSE);

    // Return success.

    return STATUS_SUCCESS;
  }

  // If 'ttInfoApplies' is true, it should no longer be now that we're
  // starting another task while that single ongoing task is ongoing.

  if (g.ongoingTasksInfo.ttInfoApplies) {
    // Use as the task ID the last task ID assigned to any task plus 2.
    // We only use even task ID's so that 0xFFFFFFFF signifies an invalid
    // task ID.

    g.ongoingTasksInfo.lastTaskId += 2;
    *taskIdPtr = g.ongoingTasksInfo.lastTaskId;

    // Add the new task to the list of ongoing tasks.
    
    g.ongoingTasksInfo.ongoingTaskIds[1] = *taskIdPtr;
    g.sharedState->numOngoingTasks = 2;

    // Invalidate the 'ttInfoApplies' fields and record the TTG file
    // those fields were referencing so that we can dereference it.
    // (We don't dereference it now, because dereferencing a TTG file
    // must be done at passive level, when a spin lock is not held.
    // This is because dereferencing a TTG file can trigger closing
    // that file.)
    
    g.ongoingTasksInfo.ttInfoApplies = FALSE;
    ttgFileToDereference = g.ongoingTasksInfo.ttgFile;

    // Now that we're done with the ongoing tasks list, we can
    // release the spin lock.
    
    KeReleaseSpinLock(&g.ongoingTasksInfo.spinLock, oldirql);

    // If we're logging tasks, record this one starting.

    DebugPrint(("RSTask: Going to maximum speed because a second task "
                "(id %ld) has become ongoing.\n", *taskIdPtr));
    if (LOGGING_TASKS)
      LogAnyEventTypeAndLong(EVENT_TYPE_BEGIN_TASK, *taskIdPtr);

    // Since there are multiple tasks, switch to "maximum CPU speed"
    // mode.
    
    UseMaximumSpeed();

    // Now we can dereference the TTG file.

    DereferenceTTGFile(ttgFileToDereference);

    // Return success.

    return STATUS_SUCCESS;
  }

  // If the list of ongoing tasks is full, we can't begin this new task.
  // Just release the spin lock and return failure.

  if (g.sharedState->numOngoingTasks >= MAX_NUM_ONGOING_TASKS) {
    KeReleaseSpinLock(&g.ongoingTasksInfo.spinLock, oldirql);
    DebugPrint(("RSTask: WARNING - Too many ongoing tasks when a new one "
                "arrives.\n"));
    if (LOGGING_WARNINGS)
      LogError(RS_TASK_WARNING_TOO_MANY_ONGOING_TASKS);
    return RIGHTSPEED_ERROR_TOO_MANY_ONGOING_TASKS;
  }

  // At this point, we know there's nothing special going on; we are
  // just adding a task to a list of ongoing tasks. We compute the
  // next task ID, add it to the list of ongoing tasks, and release
  // the spin lock.

  g.ongoingTasksInfo.lastTaskId += 2;
  *taskIdPtr = g.ongoingTasksInfo.lastTaskId;
  g.ongoingTasksInfo.ongoingTaskIds[g.sharedState->numOngoingTasks] =
    *taskIdPtr;
  ++g.sharedState->numOngoingTasks;
  KeReleaseSpinLock(&g.ongoingTasksInfo.spinLock, oldirql);

  // If we're logging tasks, log the beginning of this one.

  DebugPrint(("RSTask: Starting task %ld when several tasks are "
              "ongoing.\n", *taskIdPtr));
  if (LOGGING_TASKS)
    LogAnyEventTypeAndLong(EVENT_TYPE_BEGIN_TASK, *taskIdPtr);

  // Return success.
  
  return STATUS_SUCCESS;
}

// This function ends a task of a given task ID, and returns a status
// code.

NTSTATUS EndTask (long taskId)
{
  KIRQL oldirql;
  PTTG_FILE ttgFile;
  PTT_INFO ttInfo;
  LONGLONG cycles;
  long i, numOngoingTasks;
  BOOLEAN useSample;

  DebugPrint(("RSTask: Ending task %ld\n", taskId));

  // Acquire the spin lock controlling access to the ongoing tasks
  // info structure.

  KeAcquireSpinLock(&g.ongoingTasksInfo.spinLock, &oldirql);

  // If there is a single ongoing task for which all the extra fields
  // apply, handling the end of this task involves two things.
  // First, computing how long it took so we can record this in the
  // sample and recompute the schedule. Second, ending the task and
  // changing the speed controller state to idle.
  
  if (g.ongoingTasksInfo.ttInfoApplies) {

    // Make sure this is the right ID for the single ongoing task. If
    // not, fail.
    
    if (taskId != g.ongoingTasksInfo.ongoingTaskIds[0]) {
      KeReleaseSpinLock(&g.ongoingTasksInfo.spinLock, oldirql);
      DebugPrint(("RSTask: WARNING - Tried to end the single task using "
                  "the wrong id (%ld instead of %ld).\n", taskId,
                  g.ongoingTasksInfo.ongoingTaskIds[0]));
      return RIGHTSPEED_ERROR_INVALID_TASK_ID;
    }

    // Save any information about this task that we will need to
    // complete its processing: the task type information, the task
    // type group file, and whether to add this run of the task to the
    // sample. (We add it only if this task type uses samples, but not
    // if this run had I/O and this task type does not sample runs
    // with I/O.)

    ttInfo = g.ongoingTasksInfo.ttInfo;
    ttgFile = g.ongoingTasksInfo.ttgFile;
    useSample = ttInfo->useSample &&
                (ttInfo->sampleIfIo ||
                 !g.ongoingTasksInfo.ioOccurredDuringTask);

    // Update the ongoing tasks info structure to reflect the fact
    // that there are now no ongoing tasks. At this point, the
    // ongoing tasks info structure will no longer be referencing
    // 'ttgFile'. However, *we* are now referencing it, so there's
    // no need to dereference and rereference it.
    
    g.sharedState->numOngoingTasks = 0;
    g.ongoingTasksInfo.ttInfoApplies = FALSE;

    // Release the spin lock controlling access to the ongoing tasks
    // info structure.
    
    KeReleaseSpinLock(&g.ongoingTasksInfo.spinLock, oldirql);

    // If we're logging information about tasks, now is a good time
    // to log the end of this task.

    if (LOGGING_TASKS)
      LogAnyEventTypeAndLong(EVENT_TYPE_END_TASK, taskId);

    if (!useSample) {
      // If we're not sampling this run, just switch the speed controller
      // state to idle and dereference the TTG File.
      
      UseIdleSpeed();
      DereferenceTTGFile(ttgFile);
    }
    else {
      // If we're sampling this run, compute how many cycles this run
      // was and then switch the speed controller state to idle.
      
      cycles = GetCyclesSinceScheduleStart();
      UseIdleSpeed();

      // Add the sample value to the queue for later processing, and
      // signal the idle thread so that it wakes up and processes
      // this entry. Note that adding the sample value to the queue
      // makes the queue reference the TTG file, so now we no longer
      // need to dereference it.

      AddSampleValueToQueue(cycles, ttInfo, ttgFile);
      KeSetEvent(&g.idleThreadInfo.threadWaitEvent, IO_NO_INCREMENT, FALSE);

      // If we're logging information, log the length of this run.
      
      if (LOGGING_INFORMATION)
        LogAnyEventTypeAndBytes(EVENT_TYPE_SAMPLE_VALUE,
                                &cycles,
                                sizeof(LONGLONG));
      DebugPrint(("RSTask: Adding sample value %lx%08lx for single task "
                  "found.\n",
                  ((LARGE_INTEGER *) &cycles)->HighPart,
                  ((LARGE_INTEGER *) &cycles)->LowPart));
    }

    // We're done ending this single ongoing task, so return success.
    
    return STATUS_SUCCESS;
  }

  // At this point, we know there is not a single ongoing task, so
  // ending the task of the given ID is simply a matter of finding it
  // in the list and removing it.

  for (i = 0; i < g.sharedState->numOngoingTasks; ++i) {
    if (taskId == g.ongoingTasksInfo.ongoingTaskIds[i]) {
      // We've found the task ID in the list. Since the list is kept
      // in contiguous order, removing it involves moving all
      // later-arriving task ID's down to fill the space left by the
      // one we remove.

      while (i < g.sharedState->numOngoingTasks - 1) {
        g.ongoingTasksInfo.ongoingTaskIds[i] =
          g.ongoingTasksInfo.ongoingTaskIds[i+1];
        ++i;
      }

      // Reduce the number of ongoing tasks.

      numOngoingTasks = --g.sharedState->numOngoingTasks;

      // Now we're done with the ongoing tasks info structure.
      
      KeReleaseSpinLock(&g.ongoingTasksInfo.spinLock, oldirql);

      // If we reduced the number of ongoing tasks to zero, then go
      // to the idle state.

      if (numOngoingTasks == 0)
        UseIdleSpeed();

      // If we're logging task information, log the end of this task.

      DebugPrint(("RSTask: Removed task id %ld from list, so now only %ld "
                  "remain.\n", taskId, numOngoingTasks));
      if (LOGGING_TASKS)
        LogAnyEventTypeAndLong(EVENT_TYPE_END_TASK, taskId);

      // Now we're done, so return success.
      
      return STATUS_SUCCESS;
    }
  }

  // We didn't find any ongoing task of this ID, so just release the
  // spin lock and return an error code.

  KeReleaseSpinLock(&g.ongoingTasksInfo.spinLock, oldirql);
  DebugPrint(("RSTask: WARNING - Tried to end task %ld that was not "
              "on the ongoing tasks list.\n", taskId));
  return RIGHTSPEED_ERROR_INVALID_TASK_ID;
}

// This function ends all ongoing tasks. It's called by the idle
// thread.

void EndAllTasks (void)
{
  KIRQL oldirql;
  PTTG_FILE ttgFile;
  PTT_INFO ttInfo;
  LONGLONG cycles;
  long numTasksEnded;
  BOOLEAN useSample;

  DebugPrint(("RSTask: Ending all tasks.\n"));

  // Acquire the spin lock controlling access to the ongoing tasks info
  // structure.

  KeAcquireSpinLock(&g.ongoingTasksInfo.spinLock, &oldirql);

  // If there is a single ongoing task to which all the task type
  // information applies, there are two things we must do. First, add
  // this run of the task to the task type sample and recompute that
  // task type's schedule. Second, ending the task and changing the
  // speed controller state to idle.
  
  if (g.ongoingTasksInfo.ttInfoApplies) {

    // Save any information about this task that we will need to
    // complete its processing: the task type information, the task
    // type group file, and whether to add this run of the task to the
    // sample. (We add it only if this task type uses samples, but not
    // if this run had I/O and this task type does not sample runs
    // with I/O.)
    
    ttInfo = g.ongoingTasksInfo.ttInfo;
    ttgFile = g.ongoingTasksInfo.ttgFile;
    useSample = ttInfo->useSample &&
                (ttInfo->sampleIfIo ||
                 !g.ongoingTasksInfo.ioOccurredDuringTask);

    // Update the ongoing tasks info structure to reflect the fact
    // that there are now no ongoing tasks. At this point, the
    // ongoing tasks info structure will no longer be referencing
    // 'ttgFile'. However, *we* are now referencing it, so there's
    // no need to dereference and rereference it.
    
    g.sharedState->numOngoingTasks = 0;
    g.ongoingTasksInfo.ttInfoApplies = FALSE;

    // Release the spin lock controlling access to the ongoing tasks
    // info structure.

    KeReleaseSpinLock(&g.ongoingTasksInfo.spinLock, oldirql);

    // If we're logging information about tasks, now is a good time
    // to log the end of this task.

    if (LOGGING_TASKS)
      LogAnyEventType(EVENT_TYPE_END_ALL_TASKS);

    if (!useSample) {
      // If we're not sampling this run, just switch the speed controller
      // state to idle and dereference the TTG File.
      
      UseIdleSpeed();
      DereferenceTTGFile(ttgFile);
    }
    else {
      // If we're sampling this run, compute how many cycles this run
      // was and then switch the speed controller state to idle.
      
      cycles = GetCyclesSinceScheduleStart();
      UseIdleSpeed();

      if (LOGGING_INFORMATION)
        LogAnyEventTypeAndBytes(EVENT_TYPE_SAMPLE_VALUE,
                                &cycles,
                                sizeof(LONGLONG));
      DebugPrint(("RSTask: Adding sample value %lx%08lx for single task "
                  "found.\n",
                  ((LARGE_INTEGER *) &cycles)->HighPart,
                  ((LARGE_INTEGER *) &cycles)->LowPart));

      // Add the sample value to the queue for later processing. Note
      // that adding the sample value to the queue makes the queue
      // reference the TTG file, so now we no longer need to
      // dereference it. There's no need to signal the idle thread
      // to wake up and process this entry since this function is
      // called by the idle thread just before it processes all
      // entries in the queue.
    
      AddSampleValueToQueue(cycles, ttInfo, ttgFile);
    }
  }
  else {
    // At this point, we know there is no single ongoing task to which
    // the task type information applies, so ending all tasks is just
    // a matter of setting the number of ongoing tasks to 0.

    numTasksEnded = g.sharedState->numOngoingTasks;
    g.sharedState->numOngoingTasks = 0;

    // We're now done with the ongoing tasks information structure.
    
    KeReleaseSpinLock(&g.ongoingTasksInfo.spinLock, oldirql);

    // If we ended any tasks, switch to the idle speed. This is also
    // a good time to log the end of the tasks if we are logging task
    // information.
    
    if (numTasksEnded > 0) {
      UseIdleSpeed();

      if (LOGGING_TASKS)
        LogAnyEventType(EVENT_TYPE_END_ALL_TASKS);
      DebugPrint(("RSTask: Removed all %ld ongoing tasks from the list.\n",
                  numTasksEnded));
    }
  }
}
