// This file contains routines for manipulating the system timer
// resolution.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "FindFunc.h"
#include "TimerRes.h"
#include "Globals.h"
#include "Log.h"

/////////////////////////////////////////////////////////////////////////////
// INTERNAL TYPES
/////////////////////////////////////////////////////////////////////////////

// This type describes a function pointer that points to a
// ZwQueryTimerResolution function:

typedef
NTSYSAPI
NTSTATUS
(*ZwQueryTimerResolutionFunction)(
  OUT PULONG CoarsestResolution,
  OUT PULONG FinestResolution,
  OUT PULONG ActualResolution
  );

// This type describes a function pointer that points to a
// ZwSetTimerResolution function:

typedef
NTSYSAPI
NTSTATUS
(*ZwSetTimerResolutionFunction)(
  IN ULONG RequestedResolution,
  IN BOOLEAN Set,
  OUT PULONG ActualResolution
  );

/////////////////////////////////////////////////////////////////////////////
// TIMER RESOLUTION
/////////////////////////////////////////////////////////////////////////////

// This function initializes the system timer resolution by setting
// it to the finest setting attainable.

NTSTATUS InitializeTimerResolution (PTIMER_RESOLUTION_INFO info)
{
  ULONG coarsestRes = 0, finestRes = 0, actualRes = 0;
  ZwQueryTimerResolutionFunction queryTimerResolutionFunction;
  ZwSetTimerResolutionFunction setTimerResolutionFunction;
  NTSTATUS status;
  PVOID ntdllBase;

  // Find the base address of NTDLL.DLL.

  status = FindNtdllBase(&ntdllBase);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Cannot find base of NTDLL.DLL (%x).\n",
                status));
    return status;
  }

  // Use this base address to get the function pointers to query and
  // set the timer resolution information.

  status = FindNtdllFunction("ZwQueryTimerResolution",
                             ntdllBase,
                             (PVOID *) &queryTimerResolutionFunction);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not find ZwQueryTimerResolution "
                "function (%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_FIND_QUERY_TIMER_RESOLUTION_FUNCTION,
                 status);
    return status;
  }
  status = FindNtdllFunction("ZwSetTimerResolution",
                             ntdllBase,
                             (PVOID *) &setTimerResolutionFunction);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not find ZwSetTimerResolution "
                "function (%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_FIND_SET_TIMER_RESOLUTION_FUNCTION,
                 status);
    return status;
  }

  // Query the timer resolution information.

  status = (*queryTimerResolutionFunction)(&coarsestRes,
                                           &finestRes,
                                           &actualRes);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - ZwQueryTimerResolution failed (%x).\n",
                status));
    BothLogShort(RS_TASK_ERROR_CANT_QUERY_TIMER_RESOLUTION,
                 status);
    return status;
  }
  DebugPrint(("RSTask: Found resolution info - finest %lu, coarsest %lu, "
              "current %lu.\n", finestRes, coarsestRes, actualRes));

  // Make sure the finest resolution available is a reasonable number.

  if (finestRes < 1) {
    DebugPrint(("RSTask: ERROR - Finest resolution available is not "
                "positive.\n"));
    BothLogShort(RS_TASK_ERROR_SCHEDULING_INTERVAL_NOT_POSITIVE,
                 STATUS_UNSUCCESSFUL);
    return STATUS_UNSUCCESSFUL;
  }

  // Set the timer resolution to the finest available.

  status = (*setTimerResolutionFunction)(finestRes,
                                         TRUE,  // set rather than un-set
                                         &actualRes);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - ZwSetTimerResolution failed (%x).\n",
                status));
    BothLogOneParameter(RS_TASK_ERROR_CANT_SET_TIMER_RESOLUTION,
                        status,
                        finestRes);
    return status;
  }

  // Make sure the consequent resolution is the one requested.

  if (actualRes != finestRes) {
    DebugPrint(("RSTask: ERROR - ZwSetTimerResolution failed: resolution "
                "is %lu instead of %lu\n", actualRes, finestRes));
    BothLogTwoParameters(RS_TASK_ERROR_SET_TIMER_RESOLUTION_FAILED,
                          status,
                          finestRes,
                          actualRes);
    return STATUS_UNSUCCESSFUL;
  }
  DebugPrint(("RSTask: Set resolution to %lu.\n", finestRes));

  // Record the scheduling interval information, including not only
  // the interval length but also its negative and its half.

  info->schedulingIntervalHns = actualRes;
  info->negSchedulingIntervalHns = actualRes;
  info->negSchedulingIntervalHns = -info->negSchedulingIntervalHns;
  info->halfSchedulingIntervalHns = actualRes / 2;

  // Return success.

  return STATUS_SUCCESS;
}

// This function restores the timer resolution to what it was before
// we set it.

NTSTATUS RestoreTimerResolution (PTIMER_RESOLUTION_INFO info)
{
  ZwSetTimerResolutionFunction setTimerResolutionFunction;
  ULONG newResolution;
  NTSTATUS status;
  PVOID ntdllBase;

  // Find the base address of NTDLL.DLL.

  status = FindNtdllBase(&ntdllBase);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Cannot find base of NTDLL.DLL (%x).\n",
                status));
    return status;
  }

  // Use this base address to get the function pointer to set the
  // timer resolution information.

  status = FindNtdllFunction("ZwSetTimerResolution",
                             ntdllBase,
                             (PVOID *) &setTimerResolutionFunction);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not find ZwSetTimerResolution "
                "function (%x).\n", status));
    BothLogShortRepeat(RS_TASK_ERROR_CANT_FIND_SET_TIMER_RESOLUTION_FUNCTION,
                       status,
                       1);
    return status;
  }

  // Unset the timer resolution information that we set earlier.

  status = (*setTimerResolutionFunction)(info->schedulingIntervalHns,
                                         FALSE,  // un-set rather than set
                                         &newResolution);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not restore timer resolution (%x).\n",
                status));
    BothLogOneParameter(RS_TASK_ERROR_CANT_RESTORE_TIMER_RESOLUTION,
                        status,
                        info->schedulingIntervalHns);
    return status;
  }
  DebugPrint(("RSTask: Restored resolution to %lu\n", newResolution));

  // Return success.

  return STATUS_SUCCESS;
}

