// This file contains routines for updating the information in a task
// type.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/Algorithm.h"
#include "../Common/TTG.h"
#include "Convert.h"
#include "Gamma.h"
#include "IdleThread.h"
#include "Sample.h"
#include "SimpleSched.h"
#include "TTGFile.h"
#include "Update.h"

// This function computes the initial schedule for a given task type.
// This schedule is based on the algorithm type and the parameters of
// that algorithm.

void ComputeInitialSchedule (PTT_INFO ttInfo)
{
  // Do different things based on the algorithm type.

  switch (ttInfo->algorithmType) {
  case ALGORITHM_TYPE_FLAT :
    {
      PALGORITHM_DATA_FLAT algorithmData =
        GET_ALGORITHM_DATA(ttInfo, PALGORITHM_DATA_FLAT);

      ComputeFlatSchedule(&ttInfo->schedule,
                          HnsToIntervals(ttInfo->deadlineHns),
                          CycPerMsToCpi(algorithmData->cycPerMs));
    }
    break;

  case ALGORITHM_TYPE_FLAT_AG_BASIC :
    {
      PALGORITHM_DATA_FLAT_AG_BASIC algorithmData =
        GET_ALGORITHM_DATA(ttInfo, PALGORITHM_DATA_FLAT_AG_BASIC);
      PAG_BASIC_DATA agBasicData = &algorithmData->agBasicData;

      ComputeFlatSchedule(&ttInfo->schedule,
                          agBasicData->deadlineI,
                          CycPerMsToCpi(algorithmData->cycPerMs));
      agBasicData->pdc = GetPdcOfSchedule(&ttInfo->schedule,
                                          agBasicData->deadlineI);
    }
    break;

  case ALGORITHM_TYPE_LONGSHORT :
    {
      PALGORITHM_DATA_LONGSHORT algorithmData =
        GET_ALGORITHM_DATA(ttInfo, PALGORITHM_DATA_LONGSHORT);

      ComputeLongShortSchedule(&ttInfo->schedule,
                               HnsToIntervals(ttInfo->deadlineHns),
                               HnsToIntervals(algorithmData->updatePeriodHns));
    }
    break;

  case ALGORITHM_TYPE_LONGSHORT_AG_BASIC :
    {
      PALGORITHM_DATA_LONGSHORT_AG_BASIC algorithmData =
        GET_ALGORITHM_DATA(ttInfo, PALGORITHM_DATA_LONGSHORT_AG_BASIC);
      PAG_BASIC_DATA agBasicData = &algorithmData->agBasicData;

      ComputeLongShortSchedule(&ttInfo->schedule,
                               agBasicData->deadlineI,
                               HnsToIntervals(algorithmData->updatePeriodHns));
      agBasicData->pdc = GetPdcOfSchedule(&ttInfo->schedule,
                                          agBasicData->deadlineI);
    }
    break;

  case ALGORITHM_TYPE_PASTPEG :
    {
      PALGORITHM_DATA_PASTPEG algorithmData =
        GET_ALGORITHM_DATA(ttInfo, PALGORITHM_DATA_PASTPEG);

      ComputePastPegSchedule(&ttInfo->schedule,
                             HnsToIntervals(ttInfo->deadlineHns),
                             HnsToIntervals(algorithmData->updatePeriodHns));
    }
    break;

  case ALGORITHM_TYPE_PASTPEG_AG_BASIC :
    {
      PALGORITHM_DATA_PASTPEG_AG_BASIC algorithmData =
        GET_ALGORITHM_DATA(ttInfo, PALGORITHM_DATA_PASTPEG_AG_BASIC);
      PAG_BASIC_DATA agBasicData = &algorithmData->agBasicData;

      ComputePastPegSchedule(&ttInfo->schedule,
                             agBasicData->deadlineI,
                             HnsToIntervals(algorithmData->updatePeriodHns));
      agBasicData->pdc = GetPdcOfSchedule(&ttInfo->schedule,
                                          agBasicData->deadlineI);
    }
    break;

  case ALGORITHM_TYPE_LONGRUN :
    ComputeLongRunSchedule(&ttInfo->schedule);
    break;

  case ALGORITHM_TYPE_LONGRUN_AG_BASIC :
    {
      PALGORITHM_DATA_LONGRUN_AG_BASIC algorithmData =
        GET_ALGORITHM_DATA(ttInfo, PALGORITHM_DATA_LONGRUN_AG_BASIC);
      PAG_BASIC_DATA agBasicData = &algorithmData->agBasicData;

      ComputeLongRunSchedule(&ttInfo->schedule);
      agBasicData->pdc = GetPdcOfSchedule(&ttInfo->schedule,
                                          agBasicData->deadlineI);
    }
    break;
  }
}

// This function adds a sample value, given as a number of cycles, to
// the task type given. The other parameter is the task type group file
// containing this task type; this is used for locking the task type
// information.

void AddSampleValue (
  ULONGLONG cycles,
  PTT_INFO ttInfo,
  PKSPIN_LOCK spinLock
  )
{
  KIRQL oldirql;
  long autoScheduleType = AUTO_SCHEDULE_OF_ALGORITHM(ttInfo->algorithmType);

  // Do different things depending on the automatic scheduling type
  // of the algorithm. Note that the automatic scheduling data is always
  // the first part of the algorithm data, so we can use the algorithm
  // data pointer as the automatic scheduling data pointer.

  switch (autoScheduleType) {
  case AUTO_SCHEDULE_TYPE_AG_BASIC :
    {
      PAG_BASIC_DATA agBasicData = GET_ALGORITHM_DATA(ttInfo, PAG_BASIC_DATA);

      KeAcquireSpinLock(spinLock, &oldirql);
      AddAgedSampleValue(&agBasicData->sample,
                         cycles);
      KeReleaseSpinLock(spinLock, oldirql);
    }
    break;

  }
}

// This function adds a sample value, given as a number of cycles, to
// the task type given. It then recomputes the schedule for this task
// type based on the new sample. The other parameter is the task type
// group file containing this task type; this is used for locking the
// task type information.

void AddSampleValueAndRecomputeSchedule (
  ULONGLONG cycles,
  PTT_INFO ttInfo,
  PKSPIN_LOCK spinLock
  )
{
  KIRQL oldirql;
  long autoScheduleType = AUTO_SCHEDULE_OF_ALGORITHM(ttInfo->algorithmType);
  LARGE_INTEGER hns;

  // Do different things depending on the automatic scheduling type
  // of the algorithm. Note that the automatic scheduling data is always
  // the first part of the algorithm data, so we can use the algorithm
  // data pointer as the automatic scheduling data pointer.

  switch (autoScheduleType) {
  case AUTO_SCHEDULE_TYPE_AG_BASIC :
    {
      // This is the procedure for the aged-gamma basic automatic
      // scheduling type.

      PAG_BASIC_DATA agBasicData = GET_ALGORITHM_DATA(ttInfo, PAG_BASIC_DATA);
      double mean, variance;
      unsigned long count;

      // Acquire the spin lock for this task type group so that we
      // have undisturbed access to the sample.

      KeAcquireSpinLock(spinLock, &oldirql);

      // Add the sample value.

      count = AddAgedSampleValue(&agBasicData->sample,
                                 cycles);

      // If the new sample value count is at least
      // 'sampleValuesBeforeFirstReschedule' and if the number of
      // sample values beyond that is an integer multiple of
      // 'sampleValuesBetweenReschedules', we'll go on to also
      // recompute the schedule. Otherwise, we just release the spin
      // lock and return.

      if (count < agBasicData->sampleValuesBeforeFirstReschedule ||
          ((count - agBasicData->sampleValuesBeforeFirstReschedule) %
           agBasicData->sampleValuesBetweenReschedules) != 0) {
        KeReleaseSpinLock(spinLock, oldirql);
        return;
      }

      // While we still have the spin lock, get the mean and variance
      // from the aged sample data.
      
      GetAgedSampleMeanAndVariance(&agBasicData->sample,
                                   &mean,
                                   &variance);

      // Release the spin lock, since computing the gamma schedule
      // may take a while.
      
      KeReleaseSpinLock(spinLock, oldirql);

      // Compute the gamma schedule using the mean and variance we
      // obtained from the sample data.

      ComputeGammaSchedule(&ttInfo->schedule,
                           mean,
                           variance,
                           agBasicData->pdc,
                           agBasicData->deadlineI,
                           spinLock);
    }
    break;

  }
}

// This function resets the sample for a given task type. It assumes
// the spin lock is held that protects access to the task type
// information.

void ResetSample (PTT_INFO ttInfo)
{
  long autoScheduleType = AUTO_SCHEDULE_OF_ALGORITHM(ttInfo->algorithmType);

  // Do different things depending on the automatic scheduling type
  // of the algorithm. Note that the automatic scheduling data is always
  // the first part of the algorithm data, so we can use the algorithm
  // data pointer as the automatic scheduling data pointer.

  switch (autoScheduleType) {
  case AUTO_SCHEDULE_TYPE_AG_BASIC :
    {
      // This is the procedure for the aged-gamma basic automatic
      // scheduling type.

      PAG_BASIC_DATA agBasicData = GET_ALGORITHM_DATA(ttInfo, PAG_BASIC_DATA);

      // Reset the sample, then compute an initial schedule for the
      // task type.

      ResetAgedSample(&agBasicData->sample);
      ComputeInitialSchedule(ttInfo);
    }
    break;
  }
}

// This function changes the deadline for a given task type to the
// given new deadline in hundreds of nanoseconds. It assumes the spin
// lock is held that protects access to the task type information.

void ChangeDeadline (
  PTT_INFO ttInfo,
  ULONG newDeadlineHns
  )
{
  long autoScheduleType = AUTO_SCHEDULE_OF_ALGORITHM(ttInfo->algorithmType);

  // Change the deadline in the task type info structure.

  ttInfo->deadlineHns = newDeadlineHns;

  // Do different things depending on the automatic scheduling type
  // of the algorithm. Note that the automatic scheduling data is always
  // the first part of the algorithm data, so we can use the algorithm
  // data pointer as the automatic scheduling data pointer.

  switch (autoScheduleType) {
  case AUTO_SCHEDULE_TYPE_AG_BASIC :
    {
      // This is the procedure for the aged-gamma basic automatic
      // scheduling type.

      PAG_BASIC_DATA agBasicData = GET_ALGORITHM_DATA(ttInfo, PAG_BASIC_DATA);

      agBasicData->deadlineI = HnsToIntervals(ttInfo->deadlineHns);
    }
    break;
  }

  // Compute a new initial schedule for this task type.

  ComputeInitialSchedule(ttInfo);
}
