// This file contains the main functionality for the RSTask driver.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/KernelUser.h"
#include "Globals.h"
#include "GetState.h"
#include "Ioctls.h"
#include "RegInfo.h"
#include "Unicode.h"
#include "TTGFile.h"
#include "IdleThread.h"
#include "Flush.h"
#include "Log.h"
#include "SampleQ.h"
#include "FastIo.h"
#include "Tasks.h"
#include "IoCount.h"

////////////////////////////////////////////////////////////////////////////
//                               GLOBALS
////////////////////////////////////////////////////////////////////////////

RS_TASK_GLOBALS g;

////////////////////////////////////////////////////////////////////////////
//                           FORWARD DEFINES
////////////////////////////////////////////////////////////////////////////

NTSTATUS DriverEntry (IN PDRIVER_OBJECT  DriverObject,
                      IN PUNICODE_STRING RegistryPath);
void ReleaseResources (void);

NTSTATUS DispatchDeviceControl (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS DispatchCreate (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS DispatchClose (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS DispatchGeneral (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS DispatchShutdown (IN PDEVICE_OBJECT devObj, IN PIRP irp);
void Unload (PDRIVER_OBJECT driverObject);

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#endif

////////////////////////////////////////////////////////////////////////////
//                             DRIVER ENTRY
////////////////////////////////////////////////////////////////////////////

// DriverEntry is the installable driver initialization.  Here we set
// up the driver, create our one logger device, and do all sorts of
// miscellaneous initialization.

NTSTATUS DriverEntry (IN PDRIVER_OBJECT  DriverObject,
                      IN PUNICODE_STRING RegistryPath)
{
  NTSTATUS status = STATUS_SUCCESS;
  UNICODE_STRING deviceNameUnicodeString, miscUnicodeString;
  KIRQL oldirql;
  unsigned int i;
  KFLOATING_SAVE floatingPointSave;

  DebugPrint(("RSTask: Entering DriverEntry.\n"));

  // Initialize "resources obtained" to 0.

  g.resourcesObtained = 0;

  // Initialize driver object for error messages.

  g.ourDriver = DriverObject;

  // Get a pointer to the shared state from RSInit. If we can't get
  // it, RightSpeed is deactivated, and we should exit. (Or something
  // even more wrong is going on, in which case we should also exit.)

  status = GetSharedState();
  if (!NT_SUCCESS(status)) {
    ReleaseResources();
    return status;
  }

  // Initialize timer resolution.

  DebugPrint(("RSTask: Initializing timer resolution.\n"));
  status = InitializeTimerResolution(&g.timerResolutionInfo);
  if (!NT_SUCCESS(status)) {
    ReleaseResources();
    return status;
  }
  g.resourcesObtained |= TIMER_RESOLUTION_RESOURCE_FLAG;

  // Read registry data.

  DebugPrint(("RSTask: Reading registry data.\n"));
  KeSaveFloatingPointState(&floatingPointSave);
  status = ReadRegistryInfo(&g.registryInfo);
  KeRestoreFloatingPointState(&floatingPointSave);
  if (!NT_SUCCESS(status)) {
    ReleaseResources();
    return status;
  }

  // Initialize the sample queue.

  DebugPrint(("RSTask: Initializing sample queue.\n"));
  status = InitializeSampleQueue(&g.sampleQueueInfo);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Can't initialize sample queue (%x).\n",
                status));
    ReleaseResources();
    return status;
  }
  g.resourcesObtained |= SAMPLE_QUEUE_RESOURCE_FLAG;

  // Initialize open TTG file information.

  DebugPrint(("RSTask: Initializing open task type group files info.\n"));
  status = InitializeOpenTTGFilesInfo(&g.openTTGFilesInfo);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Can't initialize open TTG files info.\n"));
    ReleaseResources();
    return status;
  }
  g.resourcesObtained |= OPEN_TTG_FILES_INFO_RESOURCE_FLAG;

  // Initialize the speed controller.

  DebugPrint(("RSTask: Initializing speed controller.\n"));
  InitializeSpeedController(&g.speedControllerInfo);
  g.resourcesObtained |= SPEED_CONTROLLER_RESOURCE_FLAG;

  // Initialize ongoing tasks.

  DebugPrint(("RSTask: Initializing ongoing tasks list.\n"));
  InitializeOngoingTasks(&g.ongoingTasksInfo);
  g.resourcesObtained |= ONGOING_TASKS_RESOURCE_FLAG;

  // Initialize idle thread.

  DebugPrint(("RSTask: Initializing idle thread.\n"));
  status = InitializeIdleThread(&g.idleThreadInfo);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not initialize idle thread "
                "(%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_INITIALIZE_IDLE_THREAD,
                 status);
    ReleaseResources();
    return status;
  }
  g.resourcesObtained |= IDLE_THREAD_RESOURCE_FLAG;

  // Initialize and start periodic flush.

  DebugPrint(("RSTask: Initializing and starting periodic flush.\n"));
  InitializePeriodicFlush(&g.periodicFlushInfo);
  StartPeriodicFlush(&g.periodicFlushInfo);
  g.resourcesObtained |= PERIODIC_FLUSH_RESOURCE_FLAG;

  // Initialize shared-state hooks.

  DebugPrint(("RSTask: Initializing shared-state hooks.\n"));
  KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
  g.replacedNotifyIosCompleteFunc = g.sharedState->notifyIosCompleteFunc;
  g.sharedState->notifyIosCompleteFunc = NotifyIosComplete;
  KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);
  g.resourcesObtained |= SHARED_STATE_HOOKS_RESOURCE_FLAG;

  // Set up the device driver entry points.

  DebugPrint(("RSTask: Setting up driver entry points.\n"));
  for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++)
    DriverObject->MajorFunction[i] = DispatchGeneral;
  DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] =
                                     DispatchDeviceControl;
  DriverObject->MajorFunction[IRP_MJ_CREATE] = DispatchCreate;
  DriverObject->MajorFunction[IRP_MJ_CLOSE] = DispatchClose;
  DriverObject->MajorFunction[IRP_MJ_SHUTDOWN] = DispatchShutdown;

  DriverObject->FastIoDispatch = &FastIoDispatchTable;
  DriverObject->DriverUnload = Unload;

  // Create the one and only RSTask device.

  DebugPrint(("RSTask: Creating interface device.\n"));
  RtlInitUnicodeString(&deviceNameUnicodeString, L"\\Device\\RSTask");
  status = IoCreateDevice(DriverObject,
                          0,  // size of device extension
                          &deviceNameUnicodeString,
                          FILE_DEVICE_RS_TASK,
                          0,  // device characteristics
                          FALSE, // exclusive?
                          &g.interfaceDeviceObject);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not create interface device (%x).\n",
                status));
    BothLogShort(RS_TASK_ERROR_CANT_CREATE_INTERFACE_DEVICE,
                 status);
    ReleaseResources();
    return status;
  }
  g.resourcesObtained |= INTERFACE_DEVICE_RESOURCE_FLAG;

  DebugPrint(("RSTask: Created interface device.\n"));

  // Create a link to this device so it can be accessed from user level

  DebugPrint(("RSTask: Creating symbolic link.\n"));
  RtlInitUnicodeString(&miscUnicodeString, L"\\DosDevices\\RSTask");
  status = IoCreateSymbolicLink(&miscUnicodeString,
                                &deviceNameUnicodeString);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not create symbolic link to "
                "interface device (%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_CREATE_INTERFACE_SYMBOLIC_LINK,
                 status);
    ReleaseResources();
    return status;
  }
  g.resourcesObtained |= SYMBOLIC_LINK_RESOURCE_FLAG;

  // Arrange to be notified on shutdown.

  DebugPrint(("RSTask: Registering for shutdown notification.\n"));
  status = IoRegisterShutdownNotification(g.interfaceDeviceObject);
  if (NT_SUCCESS(status))
    g.resourcesObtained |= SHUTDOWN_NOTIFICATION_RESOURCE_FLAG;

  // Return success.

  DebugPrint(("RSTask: Initialized successfully.\n"));
  return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////
//                          DISPATCH ROUTINES
////////////////////////////////////////////////////////////////////////////

// DispatchDeviceControl is the dispatch routine called to handle any
// device control IRP sent to our RSTask device. The main thing to
// do here is check what kind of device control request it is and handle
// it accordingly.

NTSTATUS DispatchDeviceControl (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation(irp);
  PFILE_OBJECT fileObject = irpStack->FileObject;
  ULONG ioControlCode;
  NTSTATUS status;

  // Set the request up as returning no information, by default.

  irp->IoStatus.Information = 0;

  // Get the control code (what kind of request this is).

  ioControlCode = irpStack->Parameters.DeviceIoControl.IoControlCode;

  // Do the appropriate thing, depending on the control code. Set 'status'
  // to the status code generated by that operation.

  switch (ioControlCode) {
  case RS_TASK_IOCTL_END_TASK_AND_BEGIN_ANOTHER :
    status = IoctlEndTaskAndBeginAnother(
      (PTTG_FILE) fileObject->FsContext,
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.InputBufferLength,
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.OutputBufferLength,
      &irp->IoStatus.Information
      );
    break;

  case RS_TASK_IOCTL_BEGIN_TASK :
    status = IoctlBeginTask(
      (PTTG_FILE) fileObject->FsContext,
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.InputBufferLength,
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.OutputBufferLength,
      &irp->IoStatus.Information
      );
    break;

  case RS_TASK_IOCTL_END_TASK :
    status = IoctlEndTask(
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.InputBufferLength
      );
    break;

  case RS_TASK_IOCTL_SET_GLOBAL_POLICY :
    status = IoctlSetGlobalPolicy(
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.InputBufferLength
      );
    break;

  case RS_TASK_IOCTL_QUERY_DEADLINE :
    status = IoctlQueryDeadline(
      (PTTG_FILE) fileObject->FsContext,
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.InputBufferLength,
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.OutputBufferLength,
      &irp->IoStatus.Information
      );
    break;

  case RS_TASK_IOCTL_CHANGE_DEADLINE :
    status = IoctlChangeDeadline(
      (PTTG_FILE) fileObject->FsContext,
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.InputBufferLength
      );
    break;

  case RS_TASK_IOCTL_RESET_SAMPLE :
    status = IoctlResetSample(
      (PTTG_FILE) fileObject->FsContext,
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.InputBufferLength
      );
    break;

  case RS_TASK_IOCTL_RESET_ALL_SAMPLES :
    status = IoctlResetAllSamples(
      (PTTG_FILE) fileObject->FsContext
      );
    break;

  case RS_TASK_IOCTL_ALLOW_FILE_MODIFICATION :
    status = IoctlAllowFileModification(
      (PTTG_FILE) fileObject->FsContext
      );
    break;

  case RS_TASK_IOCTL_FLUSH :
    status = IoctlFlush(
      (PTTG_FILE) fileObject->FsContext
      );
    break;

  case RS_TASK_IOCTL_FLUSH_ALL :
    status = IoctlFlushAllTTGFiles();
    break;

  case RS_TASK_IOCTL_QUERY_TTG_FILE_PREAMBLE :
    status = IoctlQueryTTGFilePreamble(
      (PTTG_FILE) fileObject->FsContext,
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.OutputBufferLength,
      &irp->IoStatus.Information
      );
    break;

  case RS_TASK_IOCTL_QUERY_NUM_SPEEDS :
    status = IoctlQueryNumSpeeds(
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.OutputBufferLength,
      &irp->IoStatus.Information
      );
    break;

  case RS_TASK_IOCTL_QUERY_TTG_DIRECTORY :
    status = IoctlQueryTTGDirectory(
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.OutputBufferLength,
      &irp->IoStatus.Information
      );
    break;

  case RS_TASK_IOCTL_PREPARE_FOR_UNINSTALL :
    status = IoctlPrepareForUninstall();
    break;

  case RS_TASK_IOCTL_GET_FUNCTION_POINTER :
    status = IoctlGetFunctionPointer(
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.InputBufferLength,
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.OutputBufferLength,
      &irp->IoStatus.Information
      );
    break;

  case RS_TASK_IOCTL_QUERY_AVAILABLE_SPEEDS_KHZ :
    status = IoctlQueryAvailableSpeedsKHz(
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.OutputBufferLength,
      &irp->IoStatus.Information
      );
    break;

  case RS_TASK_IOCTL_QUERY_IDLE_THREAD_TIME :
    status = IoctlQueryIdleThreadTime(
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.OutputBufferLength,
      &irp->IoStatus.Information
      );
    break;

  default :
    DebugPrint(("RSTask: WARNING - Unrecognized I/O control code %x.\n",
                ioControlCode));
    status = STATUS_NOT_IMPLEMENTED;
    break;
  }

  // Fill in the IRP with the status value and complete the IRP.

  irp->IoStatus.Status = status;
  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return status;
}

// DispatchCreate is the dispatch routine called to handle any create
// IRP sent to our RSTask device. The main thing to do here is
// initialize the FsContext field to the TTG file pointer of the
// corresponding TTG file. If we cannot, we fail the create.

NTSTATUS DispatchCreate (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  PIO_STACK_LOCATION currentIrpStack = IoGetCurrentIrpStackLocation(irp);
  PFILE_OBJECT fileObject = currentIrpStack->FileObject;
  NTSTATUS status;

  DebugPrint(("RSTask: Opening TTG file for new file object at %x.\n",
              fileObject));

  // Open the TTG file referred to by the file object, and return
  // the status value in the IRP.

  irp->IoStatus.Status = status = OpenTTGFile(fileObject);
  irp->IoStatus.Information = FILE_OPENED;
    
  // Complete the IRP.

  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return status;
}

// DispatchClose is the dispatch routine called to handle any
// close IRP sent to our RSTask device. Here we dereference the TTG
// file referenced by this file object.

NTSTATUS DispatchClose (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  PIO_STACK_LOCATION currentIrpStack = IoGetCurrentIrpStackLocation(irp);
  PFILE_OBJECT fileObject = currentIrpStack->FileObject;
  PTTG_FILE ttgFile = fileObject->FsContext;

  // Dereference the TTG file this file object corresponds to.

  if (ttgFile) {
    fileObject->FsContext = 0;
    DereferenceTTGFile(ttgFile);
  }

  // Set the request up as successful.

  irp->IoStatus.Status = STATUS_SUCCESS;
  irp->IoStatus.Information = 0;
    
  // Complete the IRP.

  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

// DispatchGeneral is the dispatch routine called to handle any
// miscellaneous IRP sent to our RSTask device. We just complete it.

NTSTATUS DispatchGeneral (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  // Set the request up as successful

  irp->IoStatus.Status = STATUS_SUCCESS;
  irp->IoStatus.Information = 0;
    
  // Complete the IRP

  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

// DispatchShutdown is called when the machine is about to shut down.

NTSTATUS DispatchShutdown (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  // Stop the current speed schedule and restore the standard policy.

//  StopCurrentSpeedSchedule();
//  RestoreDefaultSpeedPolicy();

  // Set the request up as successful and complete it.

  irp->IoStatus.Status = STATUS_SUCCESS;
  irp->IoStatus.Information = 0;
  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

/////////////////////////////////////////////////////////////////////////////
// RELEASING RESOURCES
/////////////////////////////////////////////////////////////////////////////

// This routine releases any resources the driver obtained, in
// preparation for unloading.

void ReleaseResources (void)
{
  UNICODE_STRING deviceLinkUnicodeString;
  KIRQL oldirql;

  DebugPrint(("RSTask: Releasing resources.\n"));

  if (g.resourcesObtained & SHUTDOWN_NOTIFICATION_RESOURCE_FLAG) {
    DebugPrint(("RSTask: Deregistering for shutdown notification.\n"));
    IoUnregisterShutdownNotification(g.interfaceDeviceObject);
    g.resourcesObtained &= ~SHUTDOWN_NOTIFICATION_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & SYMBOLIC_LINK_RESOURCE_FLAG) {
    DebugPrint(("RSTask: Deleting symbolic link.\n"));
    RtlInitUnicodeString(&deviceLinkUnicodeString, L"\\DosDevices\\RSTask");
    IoDeleteSymbolicLink(&deviceLinkUnicodeString);
    g.resourcesObtained &= ~SYMBOLIC_LINK_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & INTERFACE_DEVICE_RESOURCE_FLAG) {
    DebugPrint(("RSTask: Deleting interface device.\n"));
    IoDeleteDevice(g.interfaceDeviceObject);
    g.resourcesObtained &= ~INTERFACE_DEVICE_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & SHARED_STATE_HOOKS_RESOURCE_FLAG) {
    DebugPrint(("RSTask: Removing shared state hooks.\n"));
    KeAcquireSpinLock(&g.sharedState->spinLock, &oldirql);
    g.sharedState->notifyIosCompleteFunc = g.replacedNotifyIosCompleteFunc;
    KeReleaseSpinLock(&g.sharedState->spinLock, oldirql);
    g.resourcesObtained &= ~SHARED_STATE_HOOKS_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & PERIODIC_FLUSH_RESOURCE_FLAG) {
    DebugPrint(("RSTask: Stopping periodic flush.\n"));
    StopPeriodicFlush(&g.periodicFlushInfo);
    g.resourcesObtained &= ~PERIODIC_FLUSH_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & IDLE_THREAD_RESOURCE_FLAG) {
    DebugPrint(("RSTask: Terminating idle thread.\n"));
    CloseIdleThread(&g.idleThreadInfo);
    g.resourcesObtained &= ~IDLE_THREAD_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & ONGOING_TASKS_RESOURCE_FLAG) {
    DebugPrint(("RSTask: Closing ongoing tasks info.\n"));
    CloseOngoingTasks(&g.ongoingTasksInfo);
    g.resourcesObtained &= ~ONGOING_TASKS_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & SPEED_CONTROLLER_RESOURCE_FLAG) {
    DebugPrint(("RSTask: Closing speed controller.\n"));
    CloseSpeedController(&g.speedControllerInfo);
    g.resourcesObtained &= ~SPEED_CONTROLLER_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & OPEN_TTG_FILES_INFO_RESOURCE_FLAG) {
    DebugPrint(("RSTask: Closing open task type group files info.\n"));
    CloseOpenTTGFilesInfo(&g.openTTGFilesInfo);
    g.resourcesObtained &= ~OPEN_TTG_FILES_INFO_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & SAMPLE_QUEUE_RESOURCE_FLAG) {
    DebugPrint(("RSTask: Closing sample queue info.\n"));
    CloseSampleQueue(&g.sampleQueueInfo);
    g.resourcesObtained &= ~SAMPLE_QUEUE_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & SPEED_CURVES_RESOURCE_FLAG) {
    DebugPrint(("RSTask: Freeing speed curves.\n"));
    ExFreePool(g.registryInfo.speedCurves);
    g.resourcesObtained &= ~SPEED_CURVES_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & WORTHWHILE_SETTINGS_INFO_RESOURCE_FLAG) {
    DebugPrint(("RSTask: Freeing worthwhile settings info.\n"));
    ExFreePool(g.registryInfo.worthwhileSettingsInfo);
    g.resourcesObtained &= ~WORTHWHILE_SETTINGS_INFO_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & AVAILABLE_SPEEDS_INFO_RESOURCE_FLAG) {
    DebugPrint(("RSTask: Freeing available speeds info.\n"));
    ExFreePool(g.registryInfo.availableSpeedsInfo);
    g.resourcesObtained &= ~AVAILABLE_SPEEDS_INFO_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & TTG_DIRECTORY_RESOURCE_FLAG) {
    DebugPrint(("RSTask: Freeing TTG directory string.\n"));
    MyFreeUnicodeString(&g.registryInfo.ttgDirectory);
    g.resourcesObtained &= ~TTG_DIRECTORY_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & TIMER_RESOLUTION_RESOURCE_FLAG) {
    DebugPrint(("RSTask: Restoring old timer resolution.\n"));
    RestoreTimerResolution(&g.timerResolutionInfo);
    g.resourcesObtained &= ~TIMER_RESOLUTION_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & RSINIT_FILE_OBJECT_RESOURCE_FLAG) {
    DebugPrint(("RSTask: Dereferencing RSInit.\n"));
    ObDereferenceObject(g.RSInitFileObject);
    g.resourcesObtained &= ~RSINIT_FILE_OBJECT_RESOURCE_FLAG;
  }
}

// This is the unload routine for our driver. It releases any resources
// the driver obtained in preparation for unloading.

void Unload (PDRIVER_OBJECT driverObject)
{
  DebugPrint(("RSTask: Unloading.\n"));

  ReleaseResources();

  DebugPrint(("RSTask: Completed unload.\n"));
}
