// This file contains functions for identifying the automatic task
// types of messages.

#include <windows.h>
#include "../Common/ErrCodes.h"
#include "AutoTask.h"
#include "Globals.h"

namespace {

  ///////////////////////////////////////////////////////////////////////////
  // CONSTANTS
  ///////////////////////////////////////////////////////////////////////////

  // The task types fall into four categories: key-down, key-up, misc
  // key, and mouse-click. These are the base task type values for
  // each such category:

  const DWORD BASE_TASK_TYPE_KEY_DOWN = 0;
  const DWORD BASE_TASK_TYPE_KEY_UP = 15;
  const DWORD BASE_TASK_TYPE_MOUSE_CLICK = 30;

  // The key-down and key-up task types are computed by adding the
  // base task type value for that category to the key type for the
  // particular type of key it relates to. These are the types of keys
  // we use:

  const DWORD KEY_TYPE_MODIFIER = 0;
  const DWORD KEY_TYPE_LETTER = 1;
  const DWORD KEY_TYPE_ARROW = 2;
  const DWORD KEY_TYPE_SPACEBAR = 3;
  const DWORD KEY_TYPE_EASTERN_LANGUAGE = 4;
  const DWORD KEY_TYPE_BACKSPACE = 5;
  const DWORD KEY_TYPE_CTRL_LETTER = 6;
  const DWORD KEY_TYPE_PUNCTUATION = 7;
  const DWORD KEY_TYPE_NUMBER = 8;
  const DWORD KEY_TYPE_ENTER = 9;
  const DWORD KEY_TYPE_DEL = 10;
  const DWORD KEY_TYPE_TAB = 11;
  const DWORD KEY_TYPE_PAGE = 12;
  const DWORD KEY_TYPE_FKEY = 13;
  const DWORD KEY_TYPE_OTHER = 14;

  // The mouse-click task types are computed by adding the base task
  // type value for the mouse click type to the click type for the
  // particular type of mouse click it relates to. These are the types
  // of keys we use:

  const DWORD CLICK_TYPE_LDOWN = 0;
  const DWORD CLICK_TYPE_LUP = 1;
  const DWORD CLICK_TYPE_LDOUBLE = 2;
  const DWORD CLICK_TYPE_RDOWN = 3;
  const DWORD CLICK_TYPE_RUP = 4;
  const DWORD CLICK_TYPE_OTHER = 5;

}

/////////////////////////////////////////////////////////////////////////////
// IDENTIFICATION OF TASK TYPES
/////////////////////////////////////////////////////////////////////////////

// This function determines the automatic application task type
// associated with a message of the given type 'messageType'. The
// 'wParam' parameter indicates the wParam parameter of the message;
// this corresponds to the virtual key code of the key the message
// concerns if it's a key message. The function returns the automatic
// task type, or RIGHTSPEED_INVALID_TASK_TYPE if there is no
// associated task type.

DWORD GetAutoTaskType (UINT messageType, WPARAM wParam)
{
  DWORD baseTaskType;

  switch (messageType) {
  case WM_KEYDOWN :
  case WM_KEYUP :
    // Determine the base task type for this message type.

    baseTaskType = (messageType == WM_KEYDOWN) ? BASE_TASK_TYPE_KEY_DOWN :
                                                 BASE_TASK_TYPE_KEY_UP;

    // If the key is shift, control, or menu, consider this a modifier
    // type.

    if (wParam == VK_SHIFT || wParam == VK_CONTROL || wParam == VK_MENU)
      return baseTaskType + KEY_TYPE_MODIFIER;

    // If the key is a letter, consider its key type to be
    // KEY_TYPE_CONTROL_MODIFIED_KEY, KEY_TYPE_CAPITAL_LETTER, or
    // KEY_TYPE_UNMODIFIED_LETTER, depending on the modifiers that are
    // currently down.

    if (wParam >= 'A' && wParam <= 'Z') {
      if (systemwideGlobals.ctrlKeyDown)
        return baseTaskType + KEY_TYPE_CTRL_LETTER;
      else
        return baseTaskType + KEY_TYPE_LETTER;
    }

    // If the key is an arrow, e.g., the right-arrow key, and no
    // modifiers are pressed, consider its type to be KEY_TYPE_ARROW.

    if (wParam >= VK_LEFT && wParam <= VK_DOWN) {
      if (systemwideGlobals.shiftKeyDown || systemwideGlobals.ctrlKeyDown)
        return baseTaskType + KEY_TYPE_OTHER;
      else
        return baseTaskType + KEY_TYPE_ARROW;
    }

    // If the key is the spacebar, and the ctrl key is not down,
    // consider its type to be KEY_TYPE_SPACEBAR.
    
    if (wParam == VK_SPACE) {
      if (systemwideGlobals.ctrlKeyDown)
        return baseTaskType + KEY_TYPE_OTHER;
      else
        return baseTaskType + KEY_TYPE_SPACEBAR;
    }

    // If this key is the one used for Eastern-language characters,
    // consider its type to be KEY_TYPE_EASTERN_LANGUAGE as long as
    // the ctrl key is not down.

    if (wParam == VK_PROCESSKEY) {
      if (systemwideGlobals.ctrlKeyDown)
        return baseTaskType + KEY_TYPE_OTHER;
      else
        return baseTaskType + KEY_TYPE_EASTERN_LANGUAGE;
    }

    // If the key is backspace, consider its type to be
    // KEY_TYPE_BACKSPACE as long as ctrl is not pressed.
    
    if (wParam == VK_BACK) {
      if (systemwideGlobals.ctrlKeyDown)
        return baseTaskType = KEY_TYPE_OTHER;
      else
        return baseTaskType = KEY_TYPE_BACKSPACE;
    }

    // If the key is a number, consider its key type to be a number,
    // punctuation, or other, depending on the modifiers that are
    // currently down.
    
    if (wParam >= '0' && wParam <= '9') {
      if (systemwideGlobals.shiftKeyDown)
        return baseTaskType + KEY_TYPE_PUNCTUATION;
      else if (systemwideGlobals.ctrlKeyDown)
        return baseTaskType + KEY_TYPE_OTHER;
      else
        return baseTaskType + KEY_TYPE_NUMBER;
    }

    // If the key is a punctuation mark such as the comma, consider it
    // puncutation regardless of whether the shift key is down, since
    // shifted punctuation is still punctuation.  The ctrl key does
    // make it something other than normal punctuation, so we consider
    // it other in this case.

    if ((wParam >= VK_MULTIPLY && wParam <= VK_DIVIDE) ||
        (wParam >= 186 && wParam <= 192) ||
        (wParam >= 219 && wParam <= 222)) {
      if (systemwideGlobals.ctrlKeyDown)
        return baseTaskType + KEY_TYPE_OTHER;
      else
        return baseTaskType + KEY_TYPE_PUNCTUATION;
    }
      

    // If the key is enter and no modifiers are pressed, consider its
    // type to be KEY_TYPE_ENTER.

    if (wParam == VK_RETURN) {
      if (systemwideGlobals.shiftKeyDown || systemwideGlobals.ctrlKeyDown)
        return baseTaskType + KEY_TYPE_OTHER;
      else
        return baseTaskType + KEY_TYPE_ENTER;
    }

    // If the key is the delete key, consider its type to be KEY_TYPE_DEL
    // as long as no modifiers are pressed.

    if (wParam == VK_DELETE) {
      if (systemwideGlobals.shiftKeyDown || systemwideGlobals.ctrlKeyDown)
        return baseTaskType + KEY_TYPE_OTHER;
      else
        return baseTaskType + KEY_TYPE_DEL;
    }

    // If the key is the tab key, consider its type to be KEY_TYPE_TAB
    // as long as no modifiers are pressed.

    if (wParam == VK_TAB) {
      if (systemwideGlobals.shiftKeyDown || systemwideGlobals.ctrlKeyDown)
        return baseTaskType + KEY_TYPE_OTHER;
      else
        return baseTaskType + KEY_TYPE_TAB;
    }

    // If the key is page-up or page-down, consider its type to be
    // KEY_TYPE_PAGE as long as no modifiers are pressed.
    
    if (wParam == VK_PRIOR || wParam == VK_NEXT) {
      if (systemwideGlobals.shiftKeyDown || systemwideGlobals.ctrlKeyDown)
        return baseTaskType + KEY_TYPE_OTHER;
      else
        return baseTaskType + KEY_TYPE_PAGE;
    }

    // If the key is one of the F-keys, e.g., F4 or F10, consider its
    // type to be KEY_TYPE_FKEY as long as no modifiers are pressed.

    if (wParam >= VK_F1 && wParam <= VK_F24) {
      if (systemwideGlobals.shiftKeyDown || systemwideGlobals.ctrlKeyDown)
        return baseTaskType + KEY_TYPE_OTHER;
      else
        return baseTaskType + KEY_TYPE_FKEY;
    }

    // At this point, we know the key is some miscellaneous key, so
    // just consider it to be of type KEY_TYPE_OTHER.
    
    return baseTaskType + KEY_TYPE_OTHER;
    break;

  // The next several types correspond to various mouse-click types:

  case WM_LBUTTONDOWN :
  case WM_NCLBUTTONDOWN :
    return BASE_TASK_TYPE_MOUSE_CLICK + CLICK_TYPE_LDOWN;
    break;

  case WM_LBUTTONUP :
  case WM_NCLBUTTONUP :
    return BASE_TASK_TYPE_MOUSE_CLICK + CLICK_TYPE_LUP;
    break;
    
  case WM_LBUTTONDBLCLK :
  case WM_NCLBUTTONDBLCLK :
    return BASE_TASK_TYPE_MOUSE_CLICK + CLICK_TYPE_LDOUBLE;
    break;

  case WM_RBUTTONDOWN :
  case WM_NCRBUTTONDOWN :
    return BASE_TASK_TYPE_MOUSE_CLICK + CLICK_TYPE_RDOWN;
    break;

  case WM_RBUTTONUP :
  case WM_NCRBUTTONUP :
    return BASE_TASK_TYPE_MOUSE_CLICK + CLICK_TYPE_RUP;
    break;

  case WM_RBUTTONDBLCLK :
  case WM_MBUTTONDOWN :  case WM_MBUTTONUP :  case WM_MBUTTONDBLCLK :
  case WM_NCRBUTTONDBLCLK :
  case WM_NCMBUTTONDOWN :  case WM_NCMBUTTONUP :  case WM_NCMBUTTONDBLCLK :
    return BASE_TASK_TYPE_MOUSE_CLICK + CLICK_TYPE_OTHER;
    break;

  // The next two types correspond to key types corresponding to
  // alt-modified keys:

  case WM_SYSKEYDOWN :
    // If the key is shift, control, or menu, consider this a modifier
    // type.

    if (wParam == VK_SHIFT || wParam == VK_CONTROL || wParam == VK_MENU)
      return BASE_TASK_TYPE_KEY_DOWN + KEY_TYPE_MODIFIER;
    else
      return BASE_TASK_TYPE_KEY_DOWN + KEY_TYPE_OTHER;
    break;

  case WM_SYSKEYUP :
    // If the key is shift, control, or menu, consider this a modifier
    // type.

    if (wParam == VK_SHIFT || wParam == VK_CONTROL || wParam == VK_MENU)
      return BASE_TASK_TYPE_KEY_UP + KEY_TYPE_MODIFIER;
    else
      return BASE_TASK_TYPE_KEY_UP + KEY_TYPE_OTHER;
    break;

  // The next type is a miscellaneous key type:

  case WM_HOTKEY :
    return BASE_TASK_TYPE_KEY_DOWN + KEY_TYPE_OTHER;
    break;

  // Any other type of message doesn't correspond to an automatic task
  // type.

  default :
    return RIGHTSPEED_INVALID_TASK_TYPE;
  }
}
