// This file implements the RSInitHandle class, which represents an
// open handle to the RSInit device.

#include <windows.h>
#include <winioctl.h>
#include <iostream>
#include <winsvc.h>
#include "RSInitHandle.h"

//////////////////////////////////////////////////////////////////////////////
// INITIALIZATION / TERMINATION
//////////////////////////////////////////////////////////////////////////////

// This function initializes the object to represent an unopened
// handle.

RSInitHandle::RSInitHandle() :
  h(INVALID_HANDLE_VALUE),
  sharedState(0)
{
}

// This function initializes the handle by opening a handle to the
// RSInit device.

bool RSInitHandle::Initialize ()
{
  DWORD returnSize;

  // Open the handle by opening the file \\.\RSInit.

  h = CreateFile("\\\\.\\RSInit",
                 GENERIC_READ | GENERIC_WRITE,
                 0,     // share mode
                 NULL,  // security attributes
                 OPEN_EXISTING,
                 FILE_ATTRIBUTE_NORMAL,
                 NULL); // template
  if (h == INVALID_HANDLE_VALUE)
    return false;

  // Send an I/O control request to get the pointer to the kernel-user
  // shared state. If we fail, close the handle.
  
  if (!DeviceIoControl(h,
                       RS_INIT_IOCTL_GET_KERNEL_USER_SHARED_STATE,
                       0,                   // input buffer
                       0,                   // input buffer length
                       &sharedState,        // output buffer
                       sizeof(sharedState), // output buffer length
                       &returnSize,
                       NULL)) {             // overlapped structure
    Close();
    return false;
  }

  // Return success.

  return true;
}

// This function closes the handle, if it is open.

void RSInitHandle::Close ()
{
  // Invalidate the shared-state pointer, since it's only valid as long
  // as we hold the handle open.

  sharedState = 0;

  // Close the handle if it's open.
  
  if (h != INVALID_HANDLE_VALUE) {
    CloseHandle(h);
    h = INVALID_HANDLE_VALUE;
  }
}

//////////////////////////////////////////////////////////////////////////////
// LOGGING
//////////////////////////////////////////////////////////////////////////////

// This function logs an event. The parameters are the event type, a
// pointer to the event data, and the length of the event data.

void RSInitHandle::LogBytes (
  short eventType,
  LPCVOID data,
  DWORD dataLength)
{
  struct {
    long eventType;
    char data[MAX_LOG_ENTRY_DATA_LENGTH];
  } sendBuf;
  DWORD returnSize;

  // Make sure there isn't too much data.

  if (dataLength > MAX_LOG_ENTRY_DATA_LENGTH)
    return;

  // Fill in the buffer to send to the device.

  sendBuf.eventType = eventType;
  memcpy(sendBuf.data, data, dataLength);

  // Send the log-event request.

  DeviceIoControl(h,
                  RS_INIT_IOCTL_LOG_EVENT,
                  &sendBuf,         // input buffer
                  dataLength + 4,  // input buffer length
                  0,               // output buffer
                  0,               // output buffer length
                  &returnSize,
                  NULL);           // overlapped structure
}

// This function logs a string message, passed as the parameter 's'.

void RSInitHandle::LogString (const string &s)
{
  LogBytes(EVENT_TYPE_STRING, s.c_str(), s.size());
}

// This function logs the initiation of an automatic system task.

void RSInitHandle::LogAutoSystemTask ()
{
  struct {
    long eventType;
    DWORD threadId;
  } sendBuf;
  DWORD returnSize;

  // Fill in the buffer to send to the device.

  sendBuf.eventType = EVENT_TYPE_AUTO_SYSTEM_TASK;
  sendBuf.threadId = GetCurrentThreadId();

  // Send the log-event request.

  DeviceIoControl(h,
                  RS_INIT_IOCTL_LOG_EVENT,
                  &sendBuf,         // input buffer
                  sizeof(sendBuf), // input buffer length
                  0,               // output buffer
                  0,               // output buffer length
                  &returnSize,
                  NULL);           // overlapped structure
}

// This function logs the initiation of an automatic application task
// of the given task type.

void RSInitHandle::LogAutoAppTask (DWORD taskType)
{
  struct {
    long eventType;
    DWORD taskType;
    DWORD threadId;
  } sendBuf;
  DWORD returnSize;

  // Fill in the buffer to send to the device.

  sendBuf.eventType = EVENT_TYPE_AUTO_APP_TASK;
  sendBuf.taskType = taskType;
  sendBuf.threadId = GetCurrentThreadId();

  // Send the log-event request.

  DeviceIoControl(h,
                  RS_INIT_IOCTL_LOG_EVENT,
                  &sendBuf,         // input buffer
                  sizeof(sendBuf), // input buffer length
                  0,               // output buffer
                  0,               // output buffer length
                  &returnSize,
                  NULL);           // overlapped structure
}

// This function logs that hooking started on the process of the given
// name.

void RSInitHandle::LogProcessHooked (char *processName)
{
  struct {
    long eventType;
    DWORD processId;
    char processName[MAX_LOG_ENTRY_DATA_LENGTH - 4];
  } sendBuf;
  DWORD length, returnSize;

  // Compute the length of the process name; truncate this length if
  // it won't fit in the event data.

  length = strlen(processName);
  if (length > MAX_LOG_ENTRY_DATA_LENGTH - 4)
    length = MAX_LOG_ENTRY_DATA_LENGTH - 4;

  // Fill in the buffer to send to the device.

  sendBuf.eventType = EVENT_TYPE_PROCESS_HOOKED;
  sendBuf.processId = GetCurrentProcessId();
  memcpy(sendBuf.processName, processName, length);

  // Send the log-event request.

  DeviceIoControl(h,
                  RS_INIT_IOCTL_LOG_EVENT,
                  &sendBuf,         // input buffer
                  length + 8,      // input buffer length
                  0,               // output buffer
                  0,               // output buffer length
                  &returnSize,
                  NULL);           // overlapped structure
}

// This function logs that hooking started on the thread with the
// given ID.

void RSInitHandle::LogThreadHooked (DWORD threadId)
{
  struct {
    long eventType;
    DWORD threadId;
    DWORD processId;
  } sendBuf;
  DWORD returnSize;

  // Fill in the buffer to send to the device.

  sendBuf.eventType = EVENT_TYPE_THREAD_HOOKED;
  sendBuf.threadId = threadId;
  sendBuf.processId = GetCurrentProcessId();

  // Send the log-event request.

  DeviceIoControl(h,
                  RS_INIT_IOCTL_LOG_EVENT,
                  &sendBuf,         // input buffer
                  sizeof(sendBuf), // input buffer length
                  0,               // output buffer
                  0,               // output buffer length
                  &returnSize,
                  NULL);           // overlapped structure
}

// This function logs the event that hooking was stopped for a thread.
// The thread ID is passed as a parameter.

void RSInitHandle::LogThreadUnhooked (DWORD threadId)
{
  struct {
    long eventType;
    DWORD threadId;
  } sendBuf;
  DWORD returnSize;

  // Fill in the buffer to send to the device.

  sendBuf.eventType = EVENT_TYPE_THREAD_UNHOOKED;
  sendBuf.threadId = GetCurrentThreadId();

  // Send the log-event request.

  DeviceIoControl(h,
                  RS_INIT_IOCTL_LOG_EVENT,
                  &sendBuf,         // input buffer
                  sizeof(sendBuf), // input buffer length
                  0,               // output buffer
                  0,               // output buffer length
                  &returnSize,
                  NULL);           // overlapped structure
}

// This function logs the occurrence of an error with the given code.
// Some error codes correspond to something other than an error, such
// as information or a warning.

void RSInitHandle::LogError (DWORD errorCode)
{
  struct {
    long eventType;
    DWORD errorCode;
  } sendBuf;
  DWORD returnSize;

  // Fill in the buffer to send to the device.

  sendBuf.eventType = EVENT_TYPE_ERROR;
  sendBuf.errorCode = errorCode;

  // Send the log-event request.

  DeviceIoControl(h,
                  RS_INIT_IOCTL_LOG_EVENT,
                  &sendBuf,        // input buffer
                  sizeof(sendBuf), // input buffer length
                  NULL, // output buffer
                  0,    // output buffer length
                  &returnSize,
                  NULL); // overlapped
}

// This function logs the occurrence of an error with the given code
// and given extra parameter. Some error codes correspond to
// something other than an error, such as information or a warning.

void RSInitHandle::LogErrorOneParameter (DWORD errorCode, DWORD param)
{
  struct {
    long eventType;
    DWORD errorCode;
    DWORD param;
  } sendBuf;
  DWORD returnSize;

  // Fill in the buffer to send to the device.

  sendBuf.eventType = EVENT_TYPE_ERROR;
  sendBuf.errorCode = errorCode;
  sendBuf.param = param;

  // Send the log-event request.

  DeviceIoControl(h,
                  RS_INIT_IOCTL_LOG_EVENT,
                  &sendBuf,        // input buffer
                  sizeof(sendBuf), // input buffer length
                  NULL, // output buffer
                  0,    // output buffer length
                  &returnSize,
                  NULL); // overlapped
}

// This function logs the occurrence of an error with the given code
// and given extra two parameters. Some error codes correspond to
// something other than an error, such as information or a warning.

void RSInitHandle::LogErrorTwoParameters (DWORD errorCode, DWORD param1,
                                          DWORD param2)
{
  struct {
    long eventType;
    DWORD errorCode;
    DWORD params[2];
  } sendBuf;
  DWORD returnSize;

  // Fill in the buffer to send to the device.

  sendBuf.eventType = EVENT_TYPE_ERROR;
  sendBuf.errorCode = errorCode;
  sendBuf.params[0] = param1;
  sendBuf.params[1] = param2;

  // Send the log-event request.

  DeviceIoControl(h,
                  RS_INIT_IOCTL_LOG_EVENT,
                  &sendBuf, // input buffer
                  sizeof(sendBuf), // input buffer length
                  NULL, // output buffer
                  0,    // output buffer length
                  &returnSize,
                  NULL); // overlapped
}

// This function logs the beginning of some type of action, given as a
// parameter.

void RSInitHandle::LogActionStarted (DWORD actionType)
{
  struct {
    long eventType;
    DWORD actionType;
  } sendBuf;
  DWORD returnSize;

  // Fill in the buffer to send to the device.

  sendBuf.eventType = EVENT_TYPE_TIMED_ACTION_STARTED;
  sendBuf.actionType = actionType;

  // Send the log-event request.

  DeviceIoControl(h,
                  RS_INIT_IOCTL_LOG_EVENT,
                  &sendBuf,         // input buffer
                  sizeof(sendBuf), // input buffer length
                  NULL, // output buffer
                  0,    // output buffer length
                  &returnSize,
                  NULL); // overlapped
}

// This function logs the ending of some type of action, given as a
// parameter.

void RSInitHandle::LogActionStopped (DWORD actionType)
{
  struct {
    long eventType;
    DWORD actionType;
  } sendBuf;
  DWORD returnSize;

  // Fill in the buffer to send to the device.

  sendBuf.eventType = EVENT_TYPE_TIMED_ACTION_STOPPED;
  sendBuf.actionType = actionType;

  // Send the log-event request.

  DeviceIoControl(h,
                  RS_INIT_IOCTL_LOG_EVENT,
                  &sendBuf,         // input buffer
                  sizeof(sendBuf), // input buffer length
                  NULL, // output buffer
                  0,    // output buffer length
                  &returnSize,
                  NULL); // overlapped
}
