// This file contains the implementation for the class representing an
// open handle to the RSTask device. Most such handles correspond to
// task type group files.

#include <windows.h>
#include <winsvc.h>
#include <winioctl.h>
#include "../Common/KernelUser.h"
#include "../Common/ErrCodes.h"
#include "Globals.h"
#include "Log.h"
#include "RSTaskHandle.h"

/////////////////////////////////////////////////////////////////////////////
// CONVERSION
/////////////////////////////////////////////////////////////////////////////

namespace {

  // This is the number of hundreds of nanoseconds per millisecond:

  const long HNS_PER_MS = 10000;

  // This function rounds an assumed-nonnegative floating-point number
  // to the nearest integer.

  unsigned long RoundToNearest (double x)
  {
    return (unsigned long)(x + 0.5);
  }

}

/////////////////////////////////////////////////////////////////////////////
// INITIALIZING / CLOSING
/////////////////////////////////////////////////////////////////////////////

// This function initializes an RSTaskHandle to be an invalid handle.

RSTaskHandle::RSTaskHandle () :
  h(INVALID_HANDLE_VALUE)
{
}

// This function initializes an RSTaskHandle by opening the task type
// group file of the given name via RSTask.

RSTaskHandle::RSTaskHandle(const char *fileName) :
  h(INVALID_HANDLE_VALUE)
{
  Open(fileName);
}

// This function initializes an RSTaskHandle to represent a handle to
// RSTask that's already open.

RSTaskHandle::RSTaskHandle(const HANDLE& hVal) :
  h(hVal)
{
}

// This function attempts to open a task type group file of the given
// name. It returns a boolean indicating if it was successful. It
// assumes that this object currently is closed.

BOOL RSTaskHandle::Open (const char *fileName)
{
  h = CreateFile(fileName,
                 GENERIC_READ | GENERIC_WRITE,
                 0,              // share mode
                 NULL,           // security attributes
                 OPEN_EXISTING,
                 FILE_ATTRIBUTE_NORMAL,
                 NULL);          // template
  return (h != INVALID_HANDLE_VALUE);
}

// This function closes the handle this object is associated with and
// associates this object with no handle.

void RSTaskHandle::Close ()
{
  CloseHandle(h);
  h = INVALID_HANDLE_VALUE;
}

/////////////////////////////////////////////////////////////////////////////
// FILE OPERATIONS
/////////////////////////////////////////////////////////////////////////////

// This function makes an I/O control request asking that the opened
// TTG file referred to by this object be allowed to be modified on
// disk. It returns a boolean indicating if it was successful.

BOOL RSTaskHandle::AllowFileModification () 
{
  DWORD returnSize;

  return DeviceIoControl(h,
                         RS_TASK_IOCTL_ALLOW_FILE_MODIFICATION,
                         NULL,    // input buffer
                         0,       // input buffer length
                         NULL,    // output buffer
                         0,       // output buffer length
                         &returnSize,
                         NULL);   // overlapped
}

// This function makes an I/O control request asking that the opened
// TTG file referred to by this object be flushed to disk. Note that
// this flush will happen asynchronously, so the flush may occur after
// this function returns. This function returns a boolean indicating
// if it was successful.

BOOL RSTaskHandle::Flush ()
{
  DWORD returnSize;

  return DeviceIoControl(h,
                         RS_TASK_IOCTL_FLUSH,
                         NULL,    // input buffer
                         0,       // input buffer length
                         NULL,    // output buffer
                         0,       // output buffer length
                         &returnSize,
                         NULL);   // overlapped
}

// This function reads the preamble (beginning) of the task type group
// file opened by this object's handle. It puts this preamble in the
// location pointed to by 'ttgInfo'. It returns a boolean indicating
// if it was successful.

BOOL RSTaskHandle::QueryTTGInfoPreamble (PTTG_INFO ttgInfo)
{
  DWORD returnSize, err, expectedSize = TTG_INFO_PREAMBLE_SIZE;

  // Make the I/O control request asking for TTG_INFO_PREAMBLE_SIZE
  // bytes. Make sure the number of bytes returned is that number.

  if (!DeviceIoControl(h,
                       RS_TASK_IOCTL_QUERY_TTG_FILE_PREAMBLE,
                       NULL,  // input buffer
                       0,     // input buffer length
                       ttgInfo,  // output buffer
                       expectedSize,  // output buffer length
                       &returnSize,
                       NULL)) {
    err = GetLastError();
    if (err != ERROR_INVALID_HANDLE && (LOGGING_ERRORS))
      LOG_ERROR_ONE_PARAMETER(RS_LIB_ERROR_CANT_QUERY_TTG_INFO_PREAMBLE,
                              err);
    return FALSE;
  }
  else if (returnSize != expectedSize) {
    if (LOGGING_ERRORS)
      LOG_ERROR_ONE_PARAMETER(RS_LIB_ERROR_BAD_BYTE_COUNT_IN_QUERIED_PREAMBLE,
                              returnSize);
    return FALSE;
  }

  // Return success.
  
  return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// TASK OPERATIONS
/////////////////////////////////////////////////////////////////////////////

// This function indicates the beginning of a task of a given type.
// The task belongs to the task type group whose file is opened by
// this object. This function returns a task ID representing the new
// task, or RIGHTSPEED_INVALID_TASK_ID if it does not succeed at
// creating a new task.

DWORD RSTaskHandle::BeginTask (DWORD TaskType)
{
  DWORD returnSize, taskId;

  // If tasks are disabled systemwide, don't do anything but return the
  // special error value.

  if (systemwideGlobals.disableTasks)
    return RIGHTSPEED_INVALID_TASK_ID;

  // Send the I/O control code. If it fails, return the special error
  // value.

  if (!DeviceIoControl(h,
                       RS_TASK_IOCTL_BEGIN_TASK,
                       &TaskType,          // input buffer
                       sizeof(DWORD),      // input buffer length
                       &taskId,            // output buffer
                       sizeof(DWORD),      // output buffer length
                       &returnSize,
                       NULL))            // overlapped
    return RIGHTSPEED_INVALID_TASK_ID;

  // We succeeded, so return the task ID returned.

  return taskId;
}

// This function ends the task of the given ID. It returns a boolean
// indicating if it succeeded.

BOOL RSTaskHandle::EndTask (DWORD TaskId)
{
  DWORD returnSize;

  return DeviceIoControl(h,
                        RS_TASK_IOCTL_END_TASK,
                        &TaskId,            // input buffer
                        sizeof(DWORD),      // input buffer length
                        NULL,               // output buffer
                        0,                  // output buffer length
                        &returnSize,
                        NULL);              // overlapped
}

// This function ends one task and begins another. The task to end is
// given as a task ID. The task to begin is given as a task type,
// assumed to belong to the task type group whose file is opened by
// this object's handle. This function returns a task ID for the
// newly begun task, or RIGHTSPEED_INVALID_TASK_ID if it could not
// begin a new task. For efficiency, this function does not merely
// call the EndTask and BeginTask methods, but instead makes a single
// I/O control call.

DWORD RSTaskHandle::EndTaskAndBeginAnother (DWORD TaskId, DWORD TaskType)
{
  DWORD returnSize, newTaskId;
  struct {
    DWORD taskId;
    DWORD taskType;
  } sendBuffer;

  // If tasks are disabled systemwide, we shouldn't do anything but
  // return the error value.

  if (systemwideGlobals.disableTasks)
    return RIGHTSPEED_INVALID_TASK_ID;

  // Fill in the buffer with I/O control data.

  sendBuffer.taskId = TaskId;
  sendBuffer.taskType = TaskType;

  // Send the I/O control request. If it fails, return the error
  // value.

  if (!DeviceIoControl(h,
                       RS_TASK_IOCTL_END_TASK_AND_BEGIN_ANOTHER,
                       &sendBuffer, // input buffer
                       sizeof(sendBuffer), // input buffer length
                       &newTaskId, // output buffer
                       sizeof(DWORD), // output buffer length
                       &returnSize,
                       NULL)) // overlapped
    return RIGHTSPEED_INVALID_TASK_ID;

  // Return the ID of the newly begun task.

  return newTaskId;
}

// This function queries the deadline for a certain given task type
// belonging to the task type group opened by this object's handle.
// This deadline in milliseconds is put in 'DeadlinePtr'. This
// function returns a boolean indicating if it was successful.

BOOL RSTaskHandle::QueryDeadline (DWORD TaskType, double *DeadlinePtr)
{
  DWORD returnSize, deadlineHns;

  // Send an I/O control request asking for the deadline in hundreds
  // of nanoseconds. If it fails, return FALSE.

  if (!DeviceIoControl(h,
                       RS_TASK_IOCTL_QUERY_DEADLINE,
                       &TaskType,  // input buffer
                       sizeof(DWORD), // input buffer length
                       &deadlineHns, // output buffer
                       sizeof(DWORD), // output buffer length
                       &returnSize,
                       NULL) || // overlapped
      returnSize != sizeof(DWORD))
    return FALSE;

  // Convert the deadline in hundreds of nanoseconds to one in
  // milliseconds and put that in 'DeadlinePtr'. Return success.

  *DeadlinePtr = deadlineHns * 1.0 / HNS_PER_MS;
  return TRUE;
}

// This function changes the deadline for a given task type belonging
// to the task type group whose file is opened by this object's handle.
// The new deadline, in milliseconds, is passed as a parameter. This
// function returns a boolean indicating if it was successful.

BOOL RSTaskHandle::ChangeDeadline (DWORD TaskType, double Deadline)
{
  struct {
    DWORD taskType;
    unsigned long deadlineHns;
  } sendBuffer;
  DWORD returnSize;

  // Create the data for the I/O control request. This involves
  // converting the deadline in milliseconds to one in hundreds of
  // nanoseconds, as expected by the I/O control request.

  sendBuffer.taskType = TaskType;
  sendBuffer.deadlineHns = RoundToNearest(Deadline * HNS_PER_MS);

  // Make the I/O control request.

  return DeviceIoControl(h,
                         RS_TASK_IOCTL_CHANGE_DEADLINE,
                         &sendBuffer,  // input buffer
                         sizeof(sendBuffer), // input buffer length
                         NULL, // output buffer
                         0, // output buffer length
                         &returnSize,
                         NULL); // overlapped
}

// This function resets the sample of task work amounts for a given
// task type belonging to the task type group whose file is opened by
// this object's handle. It returns a boolean indicating if it was
// successful.

BOOL RSTaskHandle::ResetSample (DWORD TaskType)
{
  DWORD returnSize;

  return DeviceIoControl(h,
                         RS_TASK_IOCTL_RESET_SAMPLE,
                         &TaskType,  // input buffer
                         sizeof(DWORD), // input buffer length
                         NULL, // output buffer
                         0, // output buffer length
                         &returnSize,
                         NULL);
}

// This function resets the sample of task work amounts for each task
// type belonging to the task type group whose file is opened by this
// object's handle. It returns a boolean indicating if it was
// successful.

BOOL RSTaskHandle::ResetAllSamples ()
{
  DWORD returnSize;

  return DeviceIoControl(h,
                         RS_TASK_IOCTL_RESET_ALL_SAMPLES,
                         NULL,  // input buffer
                         0, // input buffer length
                         NULL, // output buffer
                         0, // output buffer length
                         &returnSize,
                         NULL);
}

//////////////////////////////////////////////////////////////////////////////
// MISCELLANEOUS INFORMATION RETRIEVAL
//////////////////////////////////////////////////////////////////////////////

// This function queries RSTask, using this object's handle, to determine
// the number of available speeds on this CPU. It puts this number in
// 'NumSpeedsPtr'. It returns a boolean indicating if it succeeded.

BOOL RSTaskHandle::QueryNumSpeeds (DWORD *NumSpeedsPtr)
{
  DWORD returnSize;

  // Send the I/O control request.

  if (!DeviceIoControl(h,
                       RS_TASK_IOCTL_QUERY_NUM_SPEEDS,
                       NULL,               // input buffer
                       0,                  // input buffer length
                       NumSpeedsPtr,       // output buffer
                       sizeof(DWORD),      // output buffer length
                       &returnSize,
                       NULL)) {            // overlapped
    if (LOGGING_ERRORS)
      LOG_ERROR_ONE_PARAMETER(RS_LIB_ERROR_CANT_QUERY_NUMBER_OF_SPEEDS,
                              GetLastError());
    return FALSE;
  }

  // Check that the number of speeds is a valid number.

  if (*NumSpeedsPtr < 1 || *NumSpeedsPtr > MAX_NUM_AVAILABLE_SPEEDS) {
    if (LOGGING_ERRORS)
      LOG_ERROR_ONE_PARAMETER(RS_LIB_ERROR_INVALID_NUMBER_OF_SPEEDS,
                              *NumSpeedsPtr);
    return FALSE;
  }

  // Return success.

  return TRUE;
}

// This function queries RSTask, using this object's handle, to
// determine the full path of the task type group directory, including
// a \??\ at the beginning and a trailing null character. It puts this
// directory in the buffer 'buf' of length 'length' and returns the
// number of bytes placed there not including the trailing null. It
// returns 0 if it fails for any reason, including the buffer not
// being long enough.

DWORD RSTaskHandle::QueryTTGDirectory (char *buf, DWORD length)
{
  DWORD returnSize;

  // Make the I/O control request. Even if it succeeds, consider
  // it to have failed if it returned fewer than 5 bytes, since
  // there should be at least five bytes including the initial
  // \??\.

  if (!DeviceIoControl(h,
                       RS_TASK_IOCTL_QUERY_TTG_DIRECTORY,
                       NULL,               // input buffer
                       0,                  // input buffer length
                       buf,                // output buffer
                       length,             // output buffer length
                       &returnSize,
                       NULL) ||    // overlapped
      returnSize < 5) {
    if (LOGGING_ERRORS)
      LOG_ERROR_ONE_PARAMETER(RS_LIB_ERROR_CANT_QUERY_TTG_DIRECTORY,
                              GetLastError());
    return 0;
  }

  // Return the number of bytes returned, not including the trailing
  // null character.

  return (returnSize - 1);
}

// This function makes an I/O control request of RSTask using this
// object's handle asking it to change its global policy. The given
// policy type describes which policy to change and the given policy
// value describes the new value to associate with that policy. This
// function returns a boolean indicating if it was successful.

BOOL RSTaskHandle::SetGlobalPolicy (DWORD policyType, DWORD policyValue)
{
  DWORD returnSize;
  struct {
    DWORD policyType;
    DWORD policyValue;
  } sendBuffer;

  // Fill in the I/O control request.

  sendBuffer.policyType = policyType;
  sendBuffer.policyValue = policyValue;

  // Make the I/O control request.
  
  return DeviceIoControl(h,
                         RS_TASK_IOCTL_SET_GLOBAL_POLICY,
                         &sendBuffer,
                         sizeof(sendBuffer),
                         NULL,  // output buffer
                         0, // output buffer length
                         &returnSize,
                         NULL);
}

// This function makes an I/O control request of RSTask using this
// object's handle asking it for  a list of all the available speeds
// in KHz.  Those speeds should go in 'AvailableSpeedsPtr', an array
// of 'NumSpeeds' DWORD's.  This routine returns a boolean indicating
// whether the I/O control request completed successfully and filled
// in all 'NumSpeeds' entries.

BOOL RSTaskHandle::QueryAvailableSpeedsKHz (DWORD *AvailableSpeedsPtr,
                                            DWORD NumSpeeds)
{
  DWORD bufSize = NumSpeeds * sizeof(DWORD);
  DWORD returnSize;

  // Make the I/O control request.
  
  return (DeviceIoControl(h,
                          RS_TASK_IOCTL_QUERY_AVAILABLE_SPEEDS_KHZ,
                          NULL, // input buffer
                          0, // input buffer length
                          AvailableSpeedsPtr,
                          bufSize,
                          &returnSize,
                          NULL) &&
          returnSize == bufSize);
}
