// This file contains functions for opening and creating task type group
// files.

#include <windows.h>
#include <vector>
#include "../Common/TTG.h"
#include "../Common/Parameters.h"
#include "../Common/ErrCodes.h"
#include "Array.h"
#include "Globals.h"
#include "Log.h"
#include "RSTaskHandle.h"
#include "TaskType.h"

using namespace std;

namespace {

  //////////////////////////////////////////////////////////////////////////
  // UTILITY FUNCTIONS
  //////////////////////////////////////////////////////////////////////////

  // This function rounds up the given number 'n' to the nearest integer
  // multiple of the given base 'base'.

  int RoundUpToMultiple (int n, int base)
  {
    int remainder = n % base;
    if (remainder)
      return n + base - remainder;
    else
      return n;
  }

  // This function gets various information from RSTask: the number of
  // speeds available on this CPU and the task type group directory.
  // It puts the number of speeds in 'numSpeedsPtr', the maximum speed
  // in 'maxSpeedPtr', and the directory name in the buffer
  // 'ttgDirectoryFullPathBuf' of length 'bufLength'. It returns a
  // boolean indicating if it was successful.

  bool GetRSTaskInfo (
    DWORD *numSpeedsPtr,
    DWORD *maxSpeedPtr,
    char *ttgDirectoryFullPathBuf,
    DWORD bufLength
    )
  {
    // Open a handle to the RSTask device that doesn't correspond to any
    // actual task type group, just to get information from RSTask.

    RSTaskHandle h("\\\\.\\RSTask");
    if (h == INVALID_HANDLE_VALUE) {
      if (LOGGING_ERRORS)
        LOG_ERROR_ONE_PARAMETER(RS_LIB_ERROR_CANT_OPEN_MAIN_RS_TASK_DEVICE,
                                GetLastError());
      return false;
    }

    // Read the number of speeds and the TTG directory.

    if (!h.QueryNumSpeeds(numSpeedsPtr) ||
        !h.QueryTTGDirectory(ttgDirectoryFullPathBuf, bufLength)) {
      h.Close();
      return false;
    }

    // Read the available speeds to determine the maximum available one.
    // Then, convert it from KHz to MHz and put it in maxSpeedPtr.

    DWORD *availableSpeeds = new DWORD[*numSpeedsPtr];
    if (!h.QueryAvailableSpeedsKHz(availableSpeeds, *numSpeedsPtr)) {
      delete availableSpeeds;
      h.Close();
      return false;
    }
    *maxSpeedPtr = (availableSpeeds[*numSpeedsPtr - 1] + 500) / 1000;
    delete availableSpeeds;

    // We're done getting information from RSTask, so close the handle and
    // return success.

    h.Close();
    return true;
  }

}

////////////////////////////////////////////////////////////////////////////
// CREATING / OPENING
////////////////////////////////////////////////////////////////////////////

// This function creates a new task type group file. It assumes that
// it has the proper permissions to modify the file on disk. The file
// name, assumed to have been checked for reasonableness, is passed as
// 'TTGFileName'. The number of task types to have in this new file is
// passed as 'NumTaskTypes'. 'TaskTypeDescriptions' is an array of
// descriptions, one per task type. 'TaskTypeDescriptions' can be
// NULL, meaning to use the default task type parameters for all task
// types; even if 'TaskTypeDescriptions' is not null, some of its
// entries may be NULL strings meaning to use the default task type
// parameters for those task types. This function returns a boolean
// indicating if it was successful.

bool CreateNewTTGFile (LPCSTR TTGFileName,
                       DWORD NumTaskTypes,
                       LPCSTR TaskTypeDescriptions[])
{
  // First, make sure the number of task types requested is a reasonable
  // number.

  if (NumTaskTypes < 1 || NumTaskTypes > MAX_NUM_TASK_TYPES) {
    SetLastError(RIGHTSPEED_ERROR_BAD_NUMBER_OF_TASK_TYPES);
    return false;
  }

  // Read various information from RSTask: the number of speeds available
  // on this CPU, and the full path of the TTG directory.

  DWORD numSpeeds, maxSpeed;
  char ttgDirectoryFullPath[MAX_PATH];
  if (!GetRSTaskInfo(&numSpeeds, &maxSpeed, ttgDirectoryFullPath, MAX_PATH))
    return false;

  // Build up the full path of the task type group file name one
  // directory at a time, creating any intermediate directories as we
  // go along.
  //
  // First, start 'fullPath' as just the task type group directory
  // (as given by RSTask, but without the first 4 bytes, \??\), and
  // start 'nextFilePart' as the task type group file name passed to
  // this function.
  //
  // Each time through the loop, we'll find the first directory part
  // from 'nextFilePart' and append it to 'fullPath', move the
  // 'nextFilePart' pointer beyond that directory part, and create the
  // directory now given in 'fullPath'.
  //
  // When we can't find any more directory parts, we'll just append
  // the contents of 'nextFilePart' to 'fullPath', making 'fullPath'
  // end as the full path of the file we must create.

  string fullPath = string(ttgDirectoryFullPath + 4) + "\\";
  const char *nextFilePart = TTGFileName, *backslashPos;
  while (1) {
    // Find the backslash that ends the first directory part.

    backslashPos = strchr(nextFilePart, '\\');
    if (backslashPos) {
      // If there is a directory part, append it to 'fullPath' and move
      // 'nextFilePart' beyond that directory part.

      fullPath += string(nextFilePart, backslashPos - nextFilePart + 1);
      nextFilePart = backslashPos + 1;

      // Try to create the directory now in 'fullPath'. If we can't, and
      // it isn't because the directory doesn't exist, fail.

      if (!CreateDirectory(fullPath.c_str(), NULL) &&
          GetLastError() != ERROR_ALREADY_EXISTS) {
        if (LOGGING_ERRORS)
          LOG_ERROR_ONE_PARAMETER(RS_LIB_ERROR_CANT_CREATE_DIRECTORY,
                                  GetLastError());
        return false;
      }
    }
    else {
      // If we didn't find any directory part, just append 'nextFilePart'
      // to 'fullPath' and break out of the loop now that 'fullPath'
      // is the full path of the file we must create.

      fullPath += nextFilePart;
      break;
    }
  }

  // Determine how many phases to use in each schedule. This is equal
  // to the number of speeds, or MAX_NUM_SCHEDULE_PHASES, whichever is
  // less.

  int maxNumSchedulePhases;
  if (numSpeeds > MAX_NUM_SCHEDULE_PHASES)
    maxNumSchedulePhases = MAX_NUM_SCHEDULE_PHASES;
  else
    maxNumSchedulePhases = numSpeeds;

  // From this deduce the size of each task type information's
  // preamble, which includes everything up to and including the
  // schedule. In other words, it's the size of the task type
  // information structure, minus the dummy speed schedule phase at
  // the end, plus the number of speed schedule phases we need. We
  // then round this size up to the nearest multiple of 8 so the
  // next structure can start at an 8-byte boundary.
  
  int ttInfoPreambleSize =
    RoundUpToMultiple(sizeof(TT_INFO) - sizeof(SPEED_SCHEDULE_PHASE) +
                      maxNumSchedulePhases * sizeof(SPEED_SCHEDULE_PHASE),
                      8);

  // Loop through all the task type descriptors to obtain the following
  // pieces of information:
  //
  // 1) 'bufSize' : how big a buffer is needed to hold the longest
  //                task type information structure
  // 2) 'ttgInfoSize' : the total size of all task type information
  //                    structures
  // 3) 'taskTypes' : a vector containing one TaskType object for
  //                  each task type that describes that task type

  int bufSize = 0;
  int ttgInfoSize = 0, ttInfoSize;
  vector<TaskType> taskTypes(NumTaskTypes);
  const char *description;
  DWORD i;

  for (i = 0; i < NumTaskTypes; ++i) {
    // The description for the i-th task type is TaskTypeDescriptions[i],
    // unless TaskTypeDescriptions is NULL.

    description = TaskTypeDescriptions ? TaskTypeDescriptions[i] : NULL;

    // Initialize the task type for this descriptor.

    if (!taskTypes[i].ApplyDescription(description, maxSpeed)) {
      SetLastError(RIGHTSPEED_ERROR_BAD_TASK_TYPE_DESCRIPTION);
      return false;
    }

    // Determine the task type information structure size by adding
    // the preamble size to the algorithm data size and rounding up to
    // the nearest multiple of 8.

    ttInfoSize = ttInfoPreambleSize +
                 RoundUpToMultiple(taskTypes[i].GetAlgorithmDataSize(), 8);

    // Use the task type information structure size to update 'bufSize'
    // and 'ttgInfoSize'.

    if (ttInfoSize > bufSize)
      bufSize = ttInfoSize;
    ttgInfoSize += ttInfoSize;
  }

  // Now, compute the size of the task type group "extended"
  // preamble. That's the size of the preamble plus room for one
  // unsigned long per task type to hold the array of task type
  // information offsets. We also round up to the nearest multiple of
  // 8 so the next structure can start at an 8-byte boundary.

  int ttgInfoExtendedPreambleSize =
    RoundUpToMultiple(TTG_INFO_PREAMBLE_SIZE +
                      NumTaskTypes * sizeof(PTT_INFO), 8);

  // Use this task type group preamble size to update 'bufSize' to be
  // the maximum size of buffer needed for either a task type
  // information structure or the task type group preamble.  Also use
  // it to update 'ttgInfoSize' to be the total size of the task type
  // group info, including not only the task type information
  // structures but also the task type group preamble.
     
  ttgInfoSize += ttgInfoExtendedPreambleSize;
  if (ttgInfoExtendedPreambleSize > bufSize)
    bufSize = ttgInfoExtendedPreambleSize;

  // Allocate an array of size 'bufSize' to hold task type information
  // structures and the task type group preamble before we write those
  // things to the file. Use a "safe" array, i.e., one that automatically
  // gets freed when it goes out of scope.

  jrl::SafeArray<char> buf(bufSize);
  if (!buf) {
    if (LOGGING_ERRORS)
      LOG_ERROR_NO_PARAMETERS(RS_LIB_ERROR_CANT_ALLOCATE_TT_INFO_BUFFER_SPACE);
    SetLastError(ERROR_NOT_ENOUGH_MEMORY);
    return false;
  }

  // Create the task type group file. Use CREATE_ALWAYS so that we
  // create it even if it already exists.

  HANDLE fileHandle = CreateFile(fullPath.c_str(),
                                 GENERIC_READ | GENERIC_WRITE,
                                 0,     // sharing
                                 NULL,  // security attributes
                                 CREATE_ALWAYS,
                                 FILE_ATTRIBUTE_NORMAL |
                                 FILE_FLAG_SEQUENTIAL_SCAN,
                                 NULL);  // template
  if (fileHandle == INVALID_HANDLE_VALUE) {
    if (LOGGING_ERRORS)
      LOG_ERROR_ONE_PARAMETER(RS_LIB_ERROR_CANT_OPEN_TTG_FILE_FOR_MAKING,
                              GetLastError());
    return false;
  }

  // Consider the buffer to be a pointer to a task type group preamble,
  // and fill in that preamble.
  
  PTTG_INFO ttgInfo = (PTTG_INFO) (char *) buf;
  ttgInfo->signature = TTG_INFO_SIGNATURE;
  ttgInfo->version = TTG_INFO_VERSION;
  ttgInfo->size = ttgInfoSize;
  ttgInfo->numTaskTypes = NumTaskTypes;
  // Set the next two fields to 0 to indicate this file was never checked.
  ttgInfo->timeLastChecked.QuadPart = 0;
  ttgInfo->schedulingIntervalHnsWhenLastChecked = 0;

  // Fill in the task type information offset array. The offset for
  // each task type is the previous task type's offset plus the
  // previous task type's size.

  int ttInfoOffset = ttgInfoExtendedPreambleSize;
  for (i = 0; i < NumTaskTypes; ++i) {
    ttgInfo->ttInfoOffsets[i] = ttInfoOffset;
    ttInfoOffset += (ttInfoPreambleSize +
                     RoundUpToMultiple(taskTypes[i].GetAlgorithmDataSize(),8));
  }

  // Write the extended preamble in 'buf' to the file.

  DWORD bytesWritten, err;
  if (!WriteFile(fileHandle,
                 buf,
                 ttgInfoExtendedPreambleSize,
                 &bytesWritten,
                 NULL) || // overlapped
      bytesWritten != ttgInfoExtendedPreambleSize) {
    err = GetLastError();
    if (LOGGING_ERRORS)
      LOG_ERROR_ONE_PARAMETER(RS_LIB_ERROR_CANT_WRITE_PREAMBLE_BYTES,
                              err);
    CloseHandle(fileHandle);
    DeleteFile(fullPath.c_str()); // delete the file since it's not valid
    SetLastError(err);
    return false;
  }

  // Loop through all the task types. For each one, write its
  // information to the buffer, treating the buffer as a pointer to a
  // task type information structure. Then, write that buffer to the
  // file.

  PTT_INFO ttInfo = (PTT_INFO) (char *) buf;
  TaskType *taskType;
  for (i = 0; i < NumTaskTypes; ++i) {
    // Get a pointer to the object holding information about this task
    // type. Use that to determine the task type information size.

    taskType = &taskTypes[i];
    ttInfoSize = ttInfoPreambleSize +
                 RoundUpToMultiple(taskType->GetAlgorithmDataSize(), 8);

    // Write the task type information preamble information to the buffer.

    ttInfo->deadlineHns = taskType->GetDeadlineHns();
    ttInfo->algorithmType = taskType->GetAlgorithmType();
    ttInfo->algorithmDataOffset = ttInfoPreambleSize;
    ttInfo->useSample = taskType->GetUseSample();
    ttInfo->sampleIfIo = taskType->GetSampleIfIo();
    ttInfo->schedule.numPhases = 0; // initialize the schedule to 'empty'
    ttInfo->schedule.maxNumPhases = maxNumSchedulePhases;

    // Write the task type algorithm data to the buffer.

    taskType->InitializeAlgorithmData(ttInfoPreambleSize + (char *) ttInfo);

    // Write the buffer to the file.

    if (!WriteFile(fileHandle,
                   buf,
                   ttInfoSize,
                   &bytesWritten,
                   NULL) || // overlapped
        bytesWritten != ttInfoSize) {
      err = GetLastError();
      if (LOGGING_ERRORS)
        LOG_ERROR_ONE_PARAMETER(RS_LIB_ERROR_CANT_WRITE_TT_INFO_BYTES,
                                err);
      CloseHandle(fileHandle);
      DeleteFile(fullPath.c_str()); // delete the file since it's not valid
      SetLastError(err);
      return false;
    }
  }

  // We're done with the file handle, so close it and return success.
  
  CloseHandle(fileHandle);
  return true;
}

// This function creates or opens a task type group file. The file
// name is given as 'TTGFileName' and is assumed to have been checked
// for reasonableness. The creation disposition 'CreationDisposition'
// can be one of CREATE_ALWAYS, CREATE_NEW, OPEN_ALWAYS, or
// OPEN_EXISTING, with the same meanings as for the Windows
// CreateFile() call. The remaining two parameters, 'NumTaskTypes'
// and 'TaskTypeDescriptions', are only used if this function actually
// creates (instead of just opening) a file. If so, 'NumTaskTypes'
// gives the number of task types the file should have, and
// 'TaskTypeDescriptions' is an array of descriptions, one per task
// type. 'TaskTypeDescriptions' can be NULL, meaning to use the
// default task type parameters for all task types; even if
// 'TaskTypeDescriptions' is not null, some of its entries may be NULL
// strings meaning to use the default task type parameters for those
// task types. This function returns a handle to the open file, which
// can be closed with CloseHandle(); if it fails, it returns
// INVALID_HANDLE_VALUE after having set the last error with
// SetLastError().

HANDLE CreateTTGFile (
  LPCSTR TTGFileName,
  DWORD CreationDisposition,
  DWORD NumTaskTypes,
  LPCSTR TaskTypeDescriptions[]
  )
{
  RSTaskHandle h;
  DWORD err;

  // Note that in this function we sometimes seem to return
  // INVALID_HANDLE_VALUE without setting the last error. This is
  // because the RSTaskHandle object sets the last error when it
  // fails to perform an operation.

  // First, try to open it as a normal task type group file via
  // RSTask. Form the full path by prepending \\.\RSTask.

  string fullPath = string("\\\\.\\RSTask\\") + TTGFileName;
  if (h.Open(fullPath.c_str())) {
    // If we successfully opened it, proceed depending on the creation
    // disposition.

    if (CreationDisposition == CREATE_NEW) {
      // If we were supposed to create a new file where one did not exist
      // before, fail. In other words, close the handle so that it becomes
      // INVALID_HANDLE_VALUE and set the error to ERROR_FILE_EXISTS.

      h.Close();
      SetLastError(ERROR_FILE_EXISTS);
    }
    else if (CreationDisposition == CREATE_ALWAYS) {
      // If we are supposed to create a new file even if one already exists,
      // we must tell RSTask to let us modify the file on disk, create a
      // new file on disk, then open that one.

      if (!h.AllowFileModification())
        h.Close();
      else {
        h.Close();
        if (CreateNewTTGFile(TTGFileName, NumTaskTypes, TaskTypeDescriptions))
          h.Open(fullPath.c_str());
      }
    }
  }
  else {
    // If we weren't able to open the file...

    err = GetLastError();
    if ((err == ERROR_PATH_NOT_FOUND || err == ERROR_FILE_NOT_FOUND) &&
        CreationDisposition != OPEN_EXISTING) {
      // If the reason we couldn't open it was that it didn't exist
      // and if the creation disposition allows us to create the file
      // if it doesn't exist (any disposition other than OPEN_EXISTING),
      // then create a new file and open that.

      if (CreateNewTTGFile(TTGFileName, NumTaskTypes, TaskTypeDescriptions))
        h.Open(fullPath.c_str());
    }
  }

  // Return the handle. This may be INVALID_HANDLE_VALUE, in which case
  // we've already set the last error with SetLastError().

  return h;
}
