// This file contains the implementation of the class representing a task
// type group file.

#include "stdafx.h"
#include <windows.h>
#include <iomanip>
#include <sstream>
#include <math.h>
#include "../Common/Algorithm.h"
#include "stdafx.h"
#include "exceptions.h"
#include "globals.h"
#include "Helpers.h"
#include "TTGFile.h"
#include "typeconv.h"

/////////////////////////////////////////////////////////////////////////////
// Unnamed namespace
/////////////////////////////////////////////////////////////////////////////

namespace {

  // These are the minimum and maximum versions of TTG file supported
  // by this program:

  static const unsigned long MIN_VERSION_SUPPORTED = 0x000007;
  static const unsigned long MAX_VERSION_SUPPORTED = 0x000007;

  // This function prints a FILETIME.

  ostream& operator<< (ostream& os, const FILETIME& fileTime)
  {
    SYSTEMTIME systemTime;

    FileTimeToSystemTime(&fileTime, &systemTime);
    os.fill('0');
    os << systemTime.wMonth
       << '/' << setw(2) << systemTime.wDay
       << '/' << systemTime.wYear
       << ' ' << systemTime.wHour
       << ':' << setw(2) << systemTime.wMinute
       << ':' << setw(2) << systemTime.wSecond;
	return os;
  }

  // This function prints a speed schedule given as a pointer.

  ostream& operator<< (ostream& os, const SPEED_SCHEDULE *schedule)
  {
    for (int i = 0; i < schedule->numPhases; ++i) {
      os << "speed " << schedule->phase[i].speedIndex;
      if (i < schedule->numPhases - 1)
        os << " for " << schedule->phase[i].intervals << " intervals, then ";
    }
    os << " (room for " << schedule->maxNumPhases - schedule->numPhases
       << " more phases)";

    return os;
  }

  // This function prints a speed schedule.

  ostream& operator<< (ostream& os, const SPEED_SCHEDULE& schedule)
  {
    return os << &schedule;
  }

  // This function prints an aged sample.

  ostream& operator<< (ostream& os, const AGED_SAMPLE& sample)
  {
    os << "  Aging factor: " << sample.agingFactor << endl;
    os << "  Count: " << sample.n << endl;
    os << "  Weighted count: " << sample.weightedCount << endl;
    os << "  Weighted sum: " << sample.weightedSum << endl;
    os << "  Weighted sum sq: " << sample.weightedSumSq << endl;

    if (sample.n > 0) {
      double mean = sample.weightedSum / sample.weightedCount;
      os << "  Mean: " << mean << endl;
      if (sample.n > 1){
        double variance = sample.weightedSumSq / sample.weightedCount -
                          mean * mean;
        os << "  Variance: " << variance << endl;
        os << "  Standard deviation: " << sqrt(variance) << endl;
        os << "  Alpha: " << mean * mean / variance << endl;
        os << "  Beta: " << variance / mean << endl;
      }
    }
    
    return os;
  }

  // This function prints an AG_BASIC_DATA structure.

  ostream& operator<< (ostream& os,
                       const AG_BASIC_DATA& agBasicData)
  {
    os << "Sample:" << endl << agBasicData.sample;
    os << "Deadline: " << agBasicData.deadlineI << " intervals" << endl;
    os << "Samples values before first reschedule: "
       << agBasicData.sampleValuesBeforeFirstReschedule << endl;
    os << "Samples values between reschedules: "
       << agBasicData.sampleValuesBetweenReschedules << endl;
    os << "PDC: " << agBasicData.pdc << endl;
    return os;
  }

  // This function prints an ALGORITHM_DATA_FLAT structure.

  ostream& operator<< (ostream& os,
                       const ALGORITHM_DATA_FLAT& algorithmData)
  {
    os << " Speed: " << algorithmData.cycPerMs << " cyc per ms" << endl;
    return os;
  }

  // This function prints an ALGORITHM_DATA_FLAT_AG_BASIC structure.

  ostream& operator<< (ostream& os,
                       const ALGORITHM_DATA_FLAT_AG_BASIC algorithmData)
  {
    os << " Speed: " << algorithmData.cycPerMs << " cyc per ms" << endl;
    os << " Aged-gamma basic data:" << endl << algorithmData.agBasicData;
    return os;
  }

  // This function prints an ALGORITHM_DATA_LONGSHORT structure.

  ostream& operator<< (ostream& os,
                       const ALGORITHM_DATA_LONGSHORT& algorithmData)
  {
    os << " Update period: " << algorithmData.updatePeriodHns << "00 ns"
       << endl;
    return os;
  }

  // This function prints an ALGORITHM_DATA_LONGSHORT_AG_BASIC structure.

  ostream& operator<< (ostream& os,
                       const ALGORITHM_DATA_LONGSHORT_AG_BASIC algorithmData)
  {
    os << " Update period: " << algorithmData.updatePeriodHns << "00 ns"
       << endl;
    os << " Aged-gamma basic data:" << endl << algorithmData.agBasicData;
    return os;
  }

  // This function prints an ALGORITHM_DATA_PASTPEG structure.

  ostream& operator<< (ostream& os,
                       const ALGORITHM_DATA_PASTPEG& algorithmData)
  {
    os << " Update period: " << algorithmData.updatePeriodHns << "00 ns"
       << endl;
    return os;
  }

  // This function prints an ALGORITHM_DATA_PASTPEG_AG_BASIC structure.

  ostream& operator<< (ostream& os,
                       const ALGORITHM_DATA_PASTPEG_AG_BASIC algorithmData)
  {
    os << " Update period: " << algorithmData.updatePeriodHns << "00 ns"
       << endl;
    os << " Aged-gamma basic data:" << endl << algorithmData.agBasicData;
    return os;
  }

  // This function prints an ALGORITHM_DATA_LONGRUN_AG_BASIC structure.

  ostream& operator<< (ostream& os,
                       const ALGORITHM_DATA_LONGRUN_AG_BASIC algorithmData)
  {
    os << " Aged-gamma basic data:" << endl << algorithmData.agBasicData;
    return os;
  }
  
}

//////////////////////////////////////////////////////////////////////////////
//                         TTGFile members
//////////////////////////////////////////////////////////////////////////////

// This function initializes the members of the object to reflect an
// opened file.

TTGFile::TTGFile() :
  hFile(INVALID_HANDLE_VALUE),
  hMap(0),
  ttgInfo(0)
{
}

// This function opens the file of the given name. It maps it into
// memory for ease of access.

void TTGFile::Open (string& fileName)
{
  // Open the file.

  hFile = CreateFile(fileName.c_str(),
                     GENERIC_READ,
                     FILE_SHARE_READ | FILE_SHARE_WRITE,
                     NULL, // security attributes
                     OPEN_EXISTING,
                     FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
                     NULL); // template
  if (hFile == INVALID_HANDLE_VALUE) {
    DisplayError("Could not open task type group file " + fileName);
    throw Exceptions::OpenFailed();
  }

  // Determine the file size so that we can map it.

  fileSize = GetFileSize(hFile, NULL);
  if (fileSize < sizeof(TTG_INFO)) {
    cerr << "ERROR - Task type group file " << fileName << " too small."
         << endl;
    throw Exceptions::InvalidTTGFileFormat();
  }

  // Create a mapping of this file into this address space.

  hMap = CreateFileMapping(hFile,
                           NULL, // security attributes
                           PAGE_READONLY,
                           0, // maximum size high word (use default)
                           0, // maximum size low word (use default)
                           NULL); // name of mapping object (none)
  if (hMap == NULL) {
    DisplayError("Could not create file mapping for " + fileName);
    throw Exceptions::OpenFailed();
  }

  // Map the file into memory.

  ttgInfo = (PTTG_INFO) MapViewOfFile(hMap,
                                      FILE_MAP_READ,
                                      0, // file offset high word
                                      0, // file offset low word
                                      0); // number of bytes to map (default)
  if (!ttgInfo) {
    DisplayError("Could not map view of file " + fileName);
    throw Exceptions::OpenFailed();
  }

  // Check the preamble's signature and version.

  if (ttgInfo->signature != TTG_INFO_SIGNATURE) {
    cerr << "ERROR - File signature is incorrect." << endl;
    throw Exceptions::InvalidTTGFileFormat();
  }
  if (ttgInfo->version < MIN_VERSION_SUPPORTED ||
      ttgInfo->version > MAX_VERSION_SUPPORTED) {
    cerr << "ERROR - Version " << ttgInfo->version << " not supported."
         << endl;
    throw Exceptions::VersionNotSupported();
  }
}

// This function closes the task type group file if it is open. It undoes
// any mappings that are active for the file.

void TTGFile::Close ()
{
  // If the file is mapped into memory, unmap it.

  if (ttgInfo) {
    if (!UnmapViewOfFile(ttgInfo)) {
      // If the unmap fails, set hMap to NULL and hFile to
      // INVALID_HANDLE_VALUE so that we never try to close the file
      // mapping and file handles.

      DisplayError("Could not unmap view of file");
      hMap = NULL;
      hFile = INVALID_HANDLE_VALUE;
    }
  }

  // If there is a mapping handle open, close it.

  if (hMap && CloseHandle(hMap))
    hMap = NULL;

  // If there is a file handle open, close it.

  if (hFile != INVALID_HANDLE_VALUE && CloseHandle(hFile))
    hFile = INVALID_HANDLE_VALUE;
}

// TTGFile::PrintPreamble reads and prints the TTG file preamble.

void TTGFile::PrintPreamble (ostream& os)
{
  os << "Version: " << ttgInfo->version / 0x10000 << "."
     << (ttgInfo->version % 0x10000) / 0x100 << "."
     << (ttgInfo->version % 0x100) << endl;

  os << "Size: " << ttgInfo->size << endl;
  os << "Number of task types: " << ttgInfo->numTaskTypes << endl;
  os << "Time last checked: " << * (FILETIME *) &ttgInfo->timeLastChecked
     << endl;
  os << "Scheduling interval when last checked: "
     << ttgInfo->schedulingIntervalHnsWhenLastChecked << "00 ns"
     << endl;
}

// TTGFile::Print prints a description of the TTG file to the given
// output stream.

void TTGFile::Print (ostream& os)
{
  PrintPreamble(os);
  os << endl;

  for (int i = 0; i < ttgInfo->numTaskTypes; ++i)
    PrintTaskTypeInformation(os, i);
}

// TTGFile::PrintTaskTypeInformation reads and prints the given task
// type information in the file.

void TTGFile::PrintTaskTypeInformation (ostream& os, int index)
{
  os << "Task type " << index << ":" << endl;

  PTT_INFO ttInfo = (PTT_INFO)(ttgInfo->ttInfoOffsets[index] +
                               (PCHAR) ttgInfo);

  os << "Deadline: " << ttInfo->deadlineHns << "00 ns" << endl;
  os << "Use sample: " << (ttInfo->useSample ? "yes" : "no") << endl;
  os << "Sample if I/O: " << (ttInfo->sampleIfIo ? "yes" : "no") << endl;
  os << "Schedule: " << ttInfo->schedule << endl;

  switch (ttInfo->algorithmType) {
  case ALGORITHM_TYPE_FLAT :
    {
      PALGORITHM_DATA_FLAT algorithmData =
        GET_ALGORITHM_DATA(ttInfo, PALGORITHM_DATA_FLAT);

      os << "Flat algorithm parameters:" << endl;
      os << *algorithmData;
    }
    break;

  case ALGORITHM_TYPE_FLAT_AG_BASIC :
    {
      PALGORITHM_DATA_FLAT_AG_BASIC algorithmData =
        GET_ALGORITHM_DATA(ttInfo, PALGORITHM_DATA_FLAT_AG_BASIC);

      os << "Flat aged-gamma basic algorithm parameters:" << endl;
      os << *algorithmData;
    }
    break;

  case ALGORITHM_TYPE_LONGSHORT :
    {
      PALGORITHM_DATA_LONGSHORT algorithmData =
        GET_ALGORITHM_DATA(ttInfo, PALGORITHM_DATA_LONGSHORT);

      os << "Longshort algorithm parameters:" << endl;
      os << *algorithmData;
    }
    break;

  case ALGORITHM_TYPE_LONGSHORT_AG_BASIC :
    {
      PALGORITHM_DATA_LONGSHORT_AG_BASIC algorithmData =
        GET_ALGORITHM_DATA(ttInfo, PALGORITHM_DATA_LONGSHORT_AG_BASIC);

      os << "Longshort aged-gamma basic algorithm parameters:" << endl;
      os << *algorithmData;
    }
    break;

  case ALGORITHM_TYPE_PASTPEG :
    {
      PALGORITHM_DATA_PASTPEG algorithmData =
        GET_ALGORITHM_DATA(ttInfo, PALGORITHM_DATA_PASTPEG);

      os << "Past/Peg algorithm parameters:" << endl;
      os << *algorithmData;
    }
    break;

  case ALGORITHM_TYPE_PASTPEG_AG_BASIC :
    {
      PALGORITHM_DATA_PASTPEG_AG_BASIC algorithmData =
        GET_ALGORITHM_DATA(ttInfo, PALGORITHM_DATA_PASTPEG_AG_BASIC);

      os << "Past/Peg aged-gamma basic algorithm parameters:" << endl;
      os << *algorithmData;
    }
    break;

  case ALGORITHM_TYPE_LONGRUN :
    os << "Longrun algorithm (no parameters)" << endl;
    break;

  case ALGORITHM_TYPE_LONGRUN_AG_BASIC :
    {
      PALGORITHM_DATA_LONGRUN_AG_BASIC algorithmData =
        GET_ALGORITHM_DATA(ttInfo, PALGORITHM_DATA_LONGRUN_AG_BASIC);

      os << "LongRun aged-gamma basic algorithm parameters:" << endl;
      os << *algorithmData;
    }
    break;

  default :
    os << "Algorithm type " << ttInfo->algorithmType << " not supported."
       << endl;
    break;
  }

  os << endl;
}
