// This file contains definitions for scheduling algorithms.

#ifndef _RIGHTSPEED_COMMON_ALGORITHM_H
#define _RIGHTSPEED_COMMON_ALGORITHM_H

///////////////////////////////////////////////////////////////////////////
// Algorithm types
///////////////////////////////////////////////////////////////////////////

// These are the types of automatic scheduling an algorithm can use:

#define AUTO_SCHEDULE_TYPE_NONE                          0x0
#define AUTO_SCHEDULE_TYPE_AG_BASIC                      0x1

// These are the algorithm types. Their least-significant 16 bits
// indicate the automatic scheduling system they use:

#define ALGORITHM_TYPE_FLAT                              0x00010000
#define ALGORITHM_TYPE_FLAT_AG_BASIC                     0x00010001

#define ALGORITHM_TYPE_LONGSHORT                         0x00020000
#define ALGORITHM_TYPE_LONGSHORT_AG_BASIC                0x00020001

#define ALGORITHM_TYPE_PASTPEG                           0x00030000
#define ALGORITHM_TYPE_PASTPEG_AG_BASIC                  0x00030001

#define ALGORITHM_TYPE_LONGRUN                           0x00040000
#define ALGORITHM_TYPE_LONGRUN_AG_BASIC                  0x00040001

// This macro converts an algorithm type to its auto schedule type:
#define AUTO_SCHEDULE_OF_ALGORITHM(a)                    (a & 0xFFFF)

// This is the default algorithm used when none is specified:
#define DEFAULT_ALGORITHM_TYPE         ALGORITHM_TYPE_FLAT_AG_BASIC

// This is a special value that indicates a nonexistent algorithm:
#define INVALID_ALGORITHM_TYPE         0

///////////////////////////////////////////////////////////////////////////
// Algorithm data structures
///////////////////////////////////////////////////////////////////////////

// This structure records an aged sample:

typedef struct _aged_sample {
  double agingFactor;     // the amount by which each sample value's weight
                          //   gets multiplied when a new sample value arrives
  double weightedCount;   // the weighted sum of 1 for each value
  double weightedSum;     // the weighted sum of all values
  double weightedSumSq;   // the weighted sum of the squares of all values
  unsigned long n;        // the number of sample values
} AGED_SAMPLE, *PAGED_SAMPLE;

// This structure contains data needed by the aged-gamma basic
// strategy to compute a schedule. It contains the aged sample, the
// target PDC, and the deadline in intervals.

typedef struct _aged_gamma_data {
  AGED_SAMPLE sample;
  unsigned long deadlineI;
  unsigned long sampleValuesBeforeFirstReschedule;
  unsigned long sampleValuesBetweenReschedules;
  double pdc;
} AG_BASIC_DATA, *PAG_BASIC_DATA;

// The following structures describe the algorithm data for each
// algorithm type. The algorithm data for algorithm type
// ALGORITHM_TYPE_FOO has the structure ALGORITHM_DATA_FOO. Note
// that the first component of each structure is the part needed
// by the automatic scheduling aspect of the algorithm:

typedef struct _algorithm_data_flat {
  unsigned long cycPerMs;         // pre-deadline speed in cycles per millisec
} ALGORITHM_DATA_FLAT, *PALGORITHM_DATA_FLAT;

typedef struct _algorithm_data_flat_ag_basic {
  AG_BASIC_DATA agBasicData;
  unsigned long cycPerMs;         // pre-deadline speed in cycles per millisec
} ALGORITHM_DATA_FLAT_AG_BASIC, *PALGORITHM_DATA_FLAT_AG_BASIC;

typedef struct _algorithm_data_longshort {
  unsigned long updatePeriodHns;  // update period in hundreds of nanosec
} ALGORITHM_DATA_LONGSHORT, *PALGORITHM_DATA_LONGSHORT;

typedef struct _algorithm_data_longshort_ag_basic {
  AG_BASIC_DATA agBasicData;
  unsigned long updatePeriodHns;  // update period in hundreds of nanosec
} ALGORITHM_DATA_LONGSHORT_AG_BASIC, *PALGORITHM_DATA_LONGSHORT_AG_BASIC;

typedef struct _algorithm_data_pastpeg {
  unsigned long updatePeriodHns;  // update period in hundreds of nanosec
} ALGORITHM_DATA_PASTPEG, *PALGORITHM_DATA_PASTPEG;

typedef struct _algorithm_data_pastpeg_ag_basic {
  AG_BASIC_DATA agBasicData;
  unsigned long updatePeriodHns;  // update period in hundreds of nanosec
} ALGORITHM_DATA_PASTPEG_AG_BASIC, *PALGORITHM_DATA_PASTPEG_AG_BASIC;

// ALGORITHM_DATA_LONGRUN is not defined, since there's no data for
// the ALGORITHM_TYPE_LONGRUN algorithm.

typedef struct _algorithm_data_longrun_ag_basic {
  AG_BASIC_DATA agBasicData;
} ALGORITHM_DATA_LONGRUN_AG_BASIC, *PALGORITHM_DATA_LONGRUN_AG_BASIC;

#endif
