// This file contains definitions related to RightSpeed logging, both
// to RSLog and to the system logger.

#ifndef _RIGHTSPEED_COMMON_LOGGING_H
#define _RIGHTSPEED_COMMON_LOGGING_H

////////////////////////////////////////////////////////////////////////////
// SIZES
////////////////////////////////////////////////////////////////////////////

// The following describe the maximum size for the data part of a
// log entry as well as the maximum size for the entire log entry.

#define MAX_LOG_ENTRY_DATA_LENGTH   255
#define MAX_LOG_ENTRY_LENGTH        (MAX_LOG_ENTRY_DATA_LENGTH + 8 + 2 + 1)

////////////////////////////////////////////////////////////////////////////
// CATEGORIES
////////////////////////////////////////////////////////////////////////////

// The following are bit fields describing categories of log entries.

#define RS_EVENT_CATEGORY_ERRORS                   1
#define RS_EVENT_CATEGORY_WARNINGS                 2
#define RS_EVENT_CATEGORY_INFORMATION              4
#define RS_EVENT_CATEGORY_SPEEDS                   8
#define RS_EVENT_CATEGORY_TASKS                    16
#define RS_EVENT_CATEGORY_IO                       32
#define RS_EVENT_CATEGORY_SCHEDULES                64

// The RS_EVENT_CATEGORY_SET structure represents a set of event
// categories as a bit vector:

typedef unsigned long RS_EVENT_CATEGORY_SET;

////////////////////////////////////////////////////////////////////////////
// TYPES
////////////////////////////////////////////////////////////////////////////

// This enumeration contains all the types of logged event supported:

typedef enum _rs_event_type {
  EVENT_TYPE_STRING,
  EVENT_TYPE_SPEED_CHANGED,
  EVENT_TYPE_BUFFER_OVERFLOWED,
  EVENT_TYPE_CONTEXT_SWITCH,
  EVENT_TYPE_PROCESS_CREATE,
  EVENT_TYPE_PROCESS_DESTROY,
  EVENT_TYPE_THREAD_CREATE,
  EVENT_TYPE_THREAD_DESTROY,
  EVENT_TYPE_BEGIN_TASK,
  EVENT_TYPE_END_TASK,
  EVENT_TYPE_SET_SCHEDULE,
  EVENT_TYPE_LONG,
  EVENT_TYPE_LONG_AND_LONG,
  EVENT_TYPE_LONG_AND_DOUBLE,
  EVENT_TYPE_END_ALL_TASKS,
  EVENT_TYPE_IO_WAIT_STARTS,
  EVENT_TYPE_IO_WAIT_STOPS,
  EVENT_TYPE_SAMPLE_VALUE,
  EVENT_TYPE_ERROR,
  EVENT_TYPE_AUTO_SYSTEM_TASK,
  EVENT_TYPE_AUTO_APP_TASK,
  EVENT_TYPE_PROCESS_HOOKED,
  EVENT_TYPE_PROCESS_UNHOOKED,
  EVENT_TYPE_THREAD_HOOKED,
  EVENT_TYPE_THREAD_UNHOOKED,
  EVENT_TYPE_TIMED_ACTION_STARTED,
  EVENT_TYPE_TIMED_ACTION_STOPPED
} RS_EVENT_TYPE;

////////////////////////////////////////////////////////////////////////////
// PREAMBLE
////////////////////////////////////////////////////////////////////////////

// This is the structure of the file preamble:

typedef struct _log_file_preamble {

  // The version of RightSpeed that generated this file:
  unsigned long version;

  // The processor type this version of RightSpeed is tailored to:
  unsigned long cpuType;

  // The maximum speed, in cycles per ms:
  unsigned long maxCycPerMs;

  // The start time in various forms:
  LARGE_INTEGER startTSC;         // as a timestamp count
  LARGE_INTEGER startGMTTime;     // as the number of 100-ns intervals since
                                  //   Jan 1, 1601 GMT
  LARGE_INTEGER startLocalTime;   // as the number of 100-ns intervals since
                                  //   Jan 1, 1601 local time

} LOG_FILE_PREAMBLE;

////////////////////////////////////////////////////////////////////////////
// LOG ERROR CODES
////////////////////////////////////////////////////////////////////////////

// The following are the "error codes" used in the RightSpeed logger and
// the system logger.
//
// The first two bits indicate severity:
//         00 Success; 01 Information; 10 Warning; 11 Error
// The next bit should be 1 to indicate it's a "customer code"
// The next 13 bits are the facility, in this case RS, 0x83E
// The final 16 bits are the error code

//////////// RSTask ////////////

#define RS_TASK_ERROR_CANT_GET_SHARED_STATE                        0xE83E0000
#define RS_TASK_ERROR_CANT_OPEN_MAIN_REGISTRY_KEY                  0xE83E0001
#define RS_TASK_ERROR_CANT_READ_MAIN_DIRECTORY                     0xE83E0002
#define RS_TASK_ERROR_CANT_INITIALIZE_TTG_DIRECTORY_STRING         0xE83E0003
#define RS_TASK_ERROR_CANT_FILL_IN_TTG_DIRECTORY_STRING            0xE83E0004
#define RS_TASK_ERROR_CANT_INITIALIZE_SPEED_SCHEDULER              0xE83E0007
#define RS_TASK_ERROR_CANT_INITIALIZE_FLOATING_VARIABLES           0xE83E0008
#define RS_TASK_ERROR_SCHEDULING_INTERVAL_NOT_POSITIVE             0xE83E0009
#define RS_TASK_ERROR_CANT_INITIALIZE_OPEN_TTG_FILES_INFO          0xE83E000A
#define RS_TASK_ERROR_CANT_INITIALIZE_ONGOING_TASKS                0xE83E000B
#define RS_TASK_ERROR_CANT_INITIALIZE_IDLE_THREAD                  0xE83E000C
#define RS_TASK_ERROR_CANT_INITIALIZE_PERIODIC_FLUSH_TIMER         0xE83E000D
#define RS_TASK_ERROR_CANT_CREATE_INTERFACE_DEVICE                 0xE83E000E
#define RS_TASK_ERROR_CANT_CREATE_INTERFACE_SYMBOLIC_LINK          0xE83E000F
#define RS_TASK_ERROR_MAIN_DIRECTORY_IS_NOT_STRING                 0xE83E0010
#define RS_TASK_ERROR_INVALID_FLAT_PDC_FRACTION                    0xE83E0011
#define RS_TASK_ERROR_INVALID_AGING_FACTOR                         0xE83E0012
#define RS_TASK_ERROR_AGED_SAMPLE_EMPTY_BUT_INVALID                0xE83E0013
#define RS_TASK_ERROR_AGED_SAMPLE_WEIGHTED_COUNT_WRONG             0xE83E0014
#define RS_TASK_ERROR_AGED_SAMPLE_NONEMPTY_AND_INVALID             0xE83E0015
#define RS_TASK_ERROR_MORE_SAMPLE_VALUES_THAN_WINDOW_SIZE          0xE83E0016
#define RS_TASK_ERROR_SAMPLE_WINDOW_LAST_ADDED_POSITION_TOO_HIGH   0xE83E0017
#define RS_TASK_ERROR_SAMPLE_WINDOW_SENTINEL_NOT_ZERO              0xE83E0018
#define RS_TASK_ERROR_SAMPLE_WINDOW_HAS_TOO_FEW_SORTED_VALUES      0xE83E0019
#define RS_TASK_ERROR_SAMPLE_WINDOW_HAS_TOO_MANY_SORTED_VALUES     0xE83E001A
#define RS_TASK_ERROR_SAMPLE_WINDOW_HAS_INVALID_PREVIOUS_PTR       0xE83E001B
#define RS_TASK_ERROR_SAMPLE_WINDOW_OUT_OF_SORTED_ORDER            0xE83E001C
#define RS_TASK_ERROR_SAMPLE_WINDOW_HAS_INVALID_SENTINEL_PREV      0xE83E001D
#define RS_TASK_ERROR_FEWER_SCHEDULE_PHASES_THAN_SPEEDS            0xE83E001E
#define RS_TASK_ERROR_TOO_MANY_SPEEDS_FOR_PACE                     0xE83E001F
#define RS_TASK_ERROR_INVALID_DEADLINE                             0xE83E0020
#define RS_TASK_ERROR_ALGORITHM_NEEDS_UNRESTRICTED_SPEEDS          0xE83E0021
#define RS_TASK_ERROR_ALGORITHM_NEEDS_RESTRICTED_SPEEDS            0xE83E0022
#define RS_TASK_ERROR_ALGORITHM_CANT_CHECK_ALGORITHM_TYPE          0xE83E0023
#define RS_TASK_ERROR_ATTEMPT_TO_CHANGE_DEADLINE_TO_NONPOSITIVE    0xE83E0024
#define RS_TASK_ERROR_CANT_REINITIALIZE_AFTER_DEADLINE_CHANGE      0xE83E0025
#define RS_TASK_ERROR_CHANGE_DEADLINE_UNIMPLEMENTED                0xE83E0026
#define RS_TASK_ERROR_INVALID_WINDOW_SIZE                          0xE83E0027
#define RS_TASK_ERROR_NEXT_IN_SORT_ARRAY_STARTS_TOO_SOON           0xE83E0028
#define RS_TASK_ERROR_PREV_IN_SORT_ARRAY_STARTS_TOO_SOON           0xE83E0029
#define RS_TASK_ERROR_RECENT_SAMPLE_DATA_PAST_FILE_END             0xE83E002A
#define RS_TASK_ERROR_INVALID_ALGORITHM_TYPE                       0xE83E002B
#define RS_TASK_ERROR_TASK_TYPE_INFO_HAS_TOO_EARLY_OFFSE           0xE83E002C
#define RS_TASK_ERROR_TASK_TYPE_INFO_GOES_PAST_FILE_END            0xE83E002D
#define RS_TASK_ERROR_TASK_TYPE_INFO_HAS_TOO_FEW_PHASES            0xE83E002E
#define RS_TASK_ERROR_TASK_TYPE_INFO_HAS_TOO_MANY_PHASES           0xE83E002F
#define RS_TASK_ERROR_TASK_TYPE_INFO_SCHEDULE_PAST_FILE_END        0xE83E0030
#define RS_TASK_ERROR_TASK_TYPE_INFO_ALGORITHM_DATA_EARLY          0xE83E0031
#define RS_TASK_ERROR_TTG_INFO_TOO_SMALL                           0xE83E0032
#define RS_TASK_ERROR_TTG_SIGNATURE_WRONG                          0xE83E0033
#define RS_TASK_ERROR_INVALID_TTG_VERSION                          0xE83E0034
#define RS_TASK_ERROR_INVALID_TTG_SIZE_FIELD                       0xE83E0035
#define RS_TASK_ERROR_TTG_INFO_TOO_MANY_TASK_TYPES                 0xE83E0036
#define RS_TASK_ERROR_TTG_INFO_PREAMBLE_GOES_PAST_FILE_END         0xE83E0037
#define RS_TASK_ERROR_TASK_TYPE_INFO_HAS_TOO_EARLY_OFFSET          0xE83E0038
#define RS_TASK_ERROR_CANT_QUERY_SYSTEM_MODULE_INFO_SIZE           0xE83E0039
#define RS_TASK_ERROR_CANT_ALLOCATE_SYSTEM_MODULE_INFO_BUF         0xE83E003A
#define RS_TASK_ERROR_CANT_QUERY_SYSTEM_MODULE_INFO                0xE83E003B
#define RS_TASK_ERROR_INVALID_SYSTEM_MODULE_INFO_SIZE              0xE83E003C
#define RS_TASK_ERROR_CANT_FIND_NTDLL_MODULE                       0xE83E003D
#define RS_TASK_ERROR_CANT_FIND_FUNCTION_WITH_NULL_NTDLL           0xE83E003E
#define RS_TASK_ERROR_CANT_FIND_FUNCTION_IN_NTDLL                  0xE83E003F
#define RS_TASK_ERROR_CANT_GET_RSINIT_DEVICE_POINTER               0xE83E0040
#define RS_TASK_ERROR_CANT_ALLOCATE_IRP_FOR_SHARED_STATE           0xE83E0041
#define RS_TASK_ERROR_CANT_ALLOCATE_SPACE_TO_FLUSH_TTG_INFO        0xE83E0042
#define RS_TASK_ERROR_CANT_CONVERT_TTG_DIRECTORY_TO_ANSI           0xE83E0043
#define RS_TASK_ERROR_CANT_ALLOCATE_PERIODIC_FLUSH_WORK_ITEM       0xE83E0044
#define RS_TASK_ERROR_CANT_INITIALIZE_TIMER_RESOULUTION            0xE83E0045
#define RS_TASK_ERROR_CANT_READ_AVAILABLE_SPEEDS_REGISTRY_KEY      0xE83E0046
#define RS_TASK_ERROR_AVAILABLE_SPEEDS_REGISTRY_KEY_NOT_BINARY     0xE83E0047
#define RS_TASK_ERROR_AVAILABLE_SPEEDS_REGISTRY_KEY_TOO_SMALL      0xE83E0048
#define RS_TASK_ERROR_AVAILABLE_SPEEDS_REGISTRY_KEY_BAD_SIZE       0xE83E0049
#define RS_TASK_ERROR_UNRESTRICTED_SPEEDS_NOT_SUPPORTED            0xE83E004A
#define RS_TASK_ERROR_INVALID_NUMBER_OF_AVAILABLE_SPEEDS           0xE83E004B
#define RS_TASK_ERROR_MINIMUM_SPEED_NOT_POSITIVE                   0xE83E004C
#define RS_TASK_ERROR_AVAILABLE_SPEEDS_OUT_OF_ORDER                0xE83E004D
#define RS_TASK_ERROR_CANT_ALLOCATE_AVAILABLE_SPEED_INFO           0xE83E004E
#define RS_TASK_ERROR_CANT_ALLOCATE_SPEED_INFO_ARRAY               0xE83E004F
#define RS_TASK_ERROR_CANT_READ_SPEED_CODES                        0xE83E0050
#define RS_TASK_ERROR_SPEED_CODES_NOT_BINARY                       0xE83E0051
#define RS_TASK_ERROR_SPEED_CODES_WRONG_SIZE                       0xE83E0052
#define RS_TASK_ERROR_CANT_CREATE_IDLE_THREAD                      0xE83E0053
#define RS_TASK_ERROR_CANT_DISABLE_THREAD_PRIORITY_BOOST           0xE83E0054
#define RS_TASK_ERROR_CANT_SET_THREAD_PRIORITY                     0xE83E0055
#define RS_TASK_ERROR_CANT_GET_THREAD_OBJECT_POINTER               0xE83E0056
#define RS_TASK_ERROR_CANT_CLOSE_THREAD_HANDLE                     0xE83E0057
#define RS_TASK_ERROR_CANT_ALLOCATE_TTG_FILE                       0xE83E0058
#define RS_TASK_ERROR_CANT_OPEN_TTG_FILE                           0xE83E0059
#define RS_TASK_ERROR_CANT_GET_TTG_FILE_SIZE                       0xE83E005A
#define RS_TASK_ERROR_TTG_FILE_SIZE_DOESNT_FIT_IN_32_BITS          0xE83E005B
#define RS_TASK_ERROR_TTG_FILE_SIZE_TOO_BIG                        0xE83E005C
#define RS_TASK_ERROR_CANT_ALLOCATE_TTG_INFO_SPACE                 0xE83E005D
#define RS_TASK_ERROR_CANT_READ_TTG_FILE                           0xE83E005E
#define RS_TASK_ERROR_READ_WRONG_NUMBER_OF_TTG_FILE_BYTES          0xE83E005F
#define RS_TASK_ERROR_CANT_GET_FILE_OBJECT_POINTER                 0xE83E0060
#define RS_TASK_ERROR_CANT_SET_TTG_FILE_POSITION_FOR_WRITING       0xE83E0061
#define RS_TASK_ERROR_CANT_WRITE_TTG_FILE                          0xE83E0062
#define RS_TASK_ERROR_WROTE_WRONG_NUMBER_OF_TTG_FILE_BYTES         0xE83E0063
#define RS_TASK_ERROR_CANT_APPEND_TTG_DIRECTORY_TO_TTG_FILE        0xE83E0064
#define RS_TASK_ERROR_CANT_APPEND_FILE_NAME_TO_TTG_FILE            0xE83E0065
#define RS_TASK_ERROR_CLOSING_TTG_FILE_LISTS_WHEN_SOME_ARE_OPEN    0xE83E0066
#define RS_TASK_ERROR_CANT_CONVERT_FILE_NAME_TO_UPPERCASE          0xE83E0067
#define RS_TASK_ERROR_CANT_CLOSE_TTG_FILE                          0xE83E0068
#define RS_TASK_ERROR_CANT_REALLOCATE_BUFFER                       0xE83E0069
#define RS_TASK_ERROR_ALGORITHM_DATA_GOES_PAST_EOF                 0xE83E006A
#define RS_TASK_ERROR_DPC_ROUTINE_CALLED_DURING_INFINITE_PHASE     0xE83E006B
#define RS_TASK_ERROR_DPC_ROUTINE_CALLED_DURING_PAUSE              0xE83E006C
#define RS_TASK_ERROR_CANT_FIND_QUERY_TIMER_RESOLUTION_FUNCTION    0xE83E006D
#define RS_TASK_ERROR_CANT_FIND_SET_TIMER_RESOLUTION_FUNCTION      0xE83E006E
#define RS_TASK_ERROR_CANT_QUERY_TIMER_RESOLUTION                  0xE83E006F
#define RS_TASK_ERROR_CANT_SET_TIMER_RESOLUTION                    0xE83E0070
#define RS_TASK_ERROR_SET_TIMER_RESOLUTION_FAILED                  0xE83E0071
#define RS_TASK_ERROR_CANT_RESTORE_TIMER_RESOLUTION                0xE83E0072
#define RS_TASK_ERROR_CANT_ALLOCATE_SAMPLE_QUEUE                   0xE83E0073
#define RS_TASK_ERROR_IDLE_THREAD_WAIT_FAILED                      0xE83E0074
#define RS_TASK_ERROR_INVALID_PDC_FRACTION                         0xE83E0075
#define RS_TASK_ERROR_INVALID_INTERVAL_LENGTH                      0xE83E0076
#define RS_TASK_ERROR_INSUFFICIENT_SCHEDULE_PHASES                 0xE83E0077
#define RS_TASK_ERROR_ATTEMPT_TO_WRITE_MSR_AT_HIGH_LEVEL           0xE83E007A
#define RS_TASK_ERROR_GENERATED_INVALID_SCHEDULE                   0xE83E007B
#define RS_TASK_ERROR_CANT_READ_TSC_TICKS_PER_MS                   0xE83E007C
#define RS_TASK_ERROR_TSC_TICKS_PER_MS_NOT_DWORD                   0xE83E007D
#define RS_TASK_ERROR_CANT_READ_WORTHWHILE_SETTINGS_REGISTRY_KEY   0xE83E007E
#define RS_TASK_ERROR_WORTHWHILE_SETTINGS_REGISTRY_KEY_NOT_BINARY  0xE83E007F
#define RS_TASK_ERROR_WORTHWHILE_SETTINGS_REGISTRY_KEY_TOO_SMALL   0xE83E0080
#define RS_TASK_ERROR_WORTHWHILE_SETTINGS_REGISTRY_KEY_BAD_SIZE    0xE83E0081
#define RS_TASK_ERROR_WORTHWHILE_SETTINGS_OUT_OF_ORDER             0xE83E0082
#define RS_TASK_ERROR_WORTHWHILE_SETTINGS_GO_BEYOND_AVAILABLE      0xE83E0083
#define RS_TASK_ERROR_CANT_ALLOCATE_WORTHWHILE_SETTINGS_INFO       0xE83E0084
#define RS_TASK_ERROR_CANT_READ_CPU_MODIFICATION_TIME              0xE83E0085
#define RS_TASK_ERROR_CPU_MODIFICATION_TIME_INVALID                0xE83E0086
#define RS_TASK_ERROR_CANT_OPEN_TTG_DIRECTORY                      0xE83E0087
#define RS_TASK_ERROR_UNINSTALL_DURING_OPEN_OR_CLOSE               0xE83E0088
#define RS_TASK_ERROR_CANT_ALLOCATE_SPEED_CURVES                   0xE83E0089
#define RS_TASK_ERROR_CANT_MAKE_SPEED_CURVES_FILE_NAME             0xE83E008A
#define RS_TASK_ERROR_CANT_OPEN_SPEED_CURVES_FILE                  0xE83E008B
#define RS_TASK_ERROR_CANT_QUERY_SIZE_SPEED_CURVES_FILE            0xE83E008C
#define RS_TASK_ERROR_INVALID_SPEED_CURVES_FILE_SIZE               0xE83E008D
#define RS_TASK_ERROR_CANT_READ_SPEED_CURVES                       0xE83E008E
#define RS_TASK_ERROR_BAD_ORDER_IN_SPEED_CURVES                    0xE83E008F
#define RS_TASK_ERROR_BAD_ENTRY_IN_SPEED_CURVES                    0xE83E0090
#define RS_TASK_ERROR_MAIN_DIRECTORY_TOO_BIG                       0xE83E0091

#define RS_TASK_WARNING_ATTEMPT_TO_OPEN_FILE_DURING_OPEN_OR_CLOSE  0xA83E0000
#define RS_TASK_WARNING_CANT_SET_THREAD_PRIORITY_TO_NORMAL         0xA83E0001
#define RS_TASK_WARNING_NEGATIVE_INTERVALS_AFTER_PAUSE             0xA83E0002
#define RS_TASK_WARNING_PUSHING_PHASE_WITH_HIGH_SPEED              0xA83E0003
#define RS_TASK_WARNING_SAMPLE_QUEUE_FULL                          0xA83E0005
#define RS_TASK_WARNING_TOO_MANY_ONGOING_TASKS                     0xA83E0006

//////////// RSIoCnt ////////////

#define RS_IOCNT_ERROR_CANT_GET_SHARED_STATE                       0xE83E1000
#define RS_IOCNT_ERROR_CANT_CREATE_MAIN_DEVICE                     0xE83E1001
#define RS_IOCNT_ERROR_CANT_REGISTER_FOR_REGISTRATION_CHANGES      0xE83E1002
#define RS_IOCNT_ERROR_CANT_GET_RSINIT_DEVICE_POINTER              0xE83E1003
#define RS_IOCNT_ERROR_CANT_ALLOCATE_IRP_FOR_SHARED_STATE          0xE83E1004
#define RS_IOCNT_ERROR_CANT_CREATE_RECOGNIZER_FILTER_DEVICE        0xE83E1005
#define RS_IOCNT_ERROR_CANT_ATTACH_TO_RECOGNIZER                   0xE83E1006
#define RS_IOCNT_ERROR_WRONG_TYPE_ATTACHED_TO_RECOGNIZER           0xE83E1007
#define RS_IOCNT_ERROR_CANT_ATTACH_TO_NULL_FILE_SYSTEM             0xE83E1008
#define RS_IOCNT_ERROR_CANT_CREATE_FILE_SYSTEM_FILTER_DEVICE       0xE83E1009
#define RS_IOCNT_ERROR_CANT_ATTACH_TO_FILE_SYSTEM                  0xE83E100A
#define RS_IOCNT_ERROR_CANT_REFERENCE_DRIVE_FILE_HANDLE            0xE83E100B
#define RS_IOCNT_ERROR_CANT_GET_DRIVE_RELATED_DEVICE_OBJECT        0xE83E100C
#define RS_IOCNT_ERROR_RECEIVED_FAST_DETACH_FOR_UNRELATED_DEVICE   0xE83E100D
#define RS_IOCNT_ERROR_CANT_ATTACH_TO_DEVICE                       0xE83E100E

#define RS_IOCNT_WARNING_CANT_CREATE_MAIN_DEVICE_SYMBOLIC_LINK     0xA83E1000
#define RS_IOCNT_WARNING_CANT_GET_RAWDISK_ADDRESS                  0xA83E1001
#define RS_IOCNT_WARNING_DELETING_WHILE_ATTACHED                   0xA83E1002
#define RS_IOCNT_WARNING_ATTEMPT_TO_ATTACH_TWICE                   0xA83E1005
#define RS_IOCNT_WARNING_MOUNT_COMPLETION_CALLED_ON_ERROR          0xA83E1006
#define RS_IOCNT_WARNING_NOT_FILTERING_WHEN_PASSING_IRP_NORMALLY   0xA83E1007
#define RS_IOCNT_WARNING_NOT_FILTERING_WHEN_PASSING_COUNTED_IRP    0xA83E1008
#define RS_IOCNT_WARNING_STACK_SPACE_INSUFFICIENT_FOR_COUNTING     0xA83E1009
#define RS_IOCNT_WARNING_NOT_FILTERING_WHEN_FORWARDING_IRP         0xA83E100A
#define RS_IOCNT_WARNING_NOT_FILTERING_WHEN_MOUNT_REQUESTED        0xA83E100B
#define RS_IOCNT_WARNING_STACK_SPACE_INSUFFICIENT_TO_PASS_MOUNT    0xA83E100C

//////////// RSInit ////////////

#define RS_INIT_ERROR_CANT_ALLOCATE_SHARED_STATE                   0xE83E2000
#define RS_INIT_ERROR_CANT_CREATE_INTERFACE_DEVICE                 0xE83E2001
#define RS_INIT_ERROR_CANT_CREATE_INTERFACE_SYMBOLIC_LINK          0xE83E2002
#define RS_INIT_ERROR_CANT_ALLOCATE_MDL_FOR_STATE                  0xE83E2003
#define RS_INIT_ERROR_CANT_MAP_SHARED_STATE_TO_USER                0xE83E2004

#define RS_INIT_WARNING_CANT_SHARE_STATE_WITH_NULL_FILE_OBJECT     0xA83E2000
#define RS_INIT_WARNING_CANT_SHARE_STATE_BECAUSE_ACTIVE_MAPPING    0xA83E2001

//////////// RSLog ////////////

#define RS_LOG_ERROR_CANT_WRITE_NEXT_LOG_FILE_NUMBER               0xE38E3000
#define RS_LOG_ERROR_CANT_READ_MAIN_DIRECTORY                      0xE38E3001
#define RS_LOG_ERROR_MAIN_DIRECTORY_IS_NOT_STRING                  0xE38E3002
#define RS_LOG_ERROR_CANT_FILL_IN_LOG_FILE_NAME                    0xE38E3003
#define RS_LOG_ERROR_CANT_CREATE_LOG_FILE                          0xE38E3004
#define RS_LOG_ERROR_CANT_ALLOCATE_LOG_BUFFER                      0xE38E3005
#define RS_LOG_ERROR_CANT_CREATE_INTERFACE_DEVICE                  0xE38E3006
#define RS_LOG_ERROR_CANT_CREATE_INTERFACE_SYMBOLIC_LINK           0xE38E3007
#define RS_LOG_ERROR_CANT_GET_SHARED_STATE                         0xE38E3008
#define RS_LOG_ERROR_CANT_OPEN_MAIN_REGISTRY_KEY                   0xE38E3009
#define RS_LOG_ERROR_CANT_ALLOCATE_FLUSH_WORK_ITEM                 0xE38E300A
#define RS_LOG_ERROR_CANT_GET_RSINIT_DEVICE_POINTER                0xE38E300B
#define RS_LOG_ERROR_CANT_ALLOCATE_IRP_FOR_SHARED_STATE            0xE38E300C
#define RS_LOG_ERROR_CANT_CREATE_LOG_FILE_THREAD                   0xE38E300D
#define RS_LOG_ERROR_CANT_GET_THREAD_OBJECT_POINTER                0xE38E300E
#define RS_LOG_ERROR_CANT_CLOSE_THREAD_HANDLE                      0xE38E300F
#define RS_LOG_ERROR_CANT_INITIALIZE_MAIN_DIRECTORY_STRING         0xE38E3010
#define RS_LOG_ERROR_LOG_FILE_THREAD_WAIT_FAILED                   0xE38E3011
#define RS_LOG_ERROR_FLUSH_FAILED                                  0xE38E3012
#define RS_LOG_ERROR_CANT_WRITE_LOG_FILE_PREAMBLE                  0xE38E3013
#define RS_LOG_ERROR_WAIT_FOR_THREAD_FAILED                        0xE38E3014

//////////// RSLib ////////////

#define RS_LIB_ERROR_CANT_CREATE_SYSTEM_TTG                        0xE38E4000
#define RS_LIB_ERROR_CANT_GET_SYSTEM_TTG_INFO_PREAMBLE             0xE38E4001
#define RS_LIB_ERROR_SYSTEM_TTG_HAS_WRONG_NUM_TASK_TYPES           0xE38E4002
#define RS_LIB_ERROR_CANT_CREATE_APP_TTG                           0xE38E4003
#define RS_LIB_ERROR_CANT_GET_APP_TTG_INFO_PREAMBLE                0xE38E4004
#define RS_LIB_ERROR_APP_TTG_HAS_WRONG_NUM_TASK_TYPES              0xE38E4005
#define RS_LIB_ERROR_CANT_OPEN_MAIN_RS_TASK_DEVICE                 0xE38E4006
#define RS_LIB_ERROR_CANT_CREATE_DIRECTORY                         0xE38E4007
#define RS_LIB_ERROR_CANT_ALLOCATE_TT_INFO_BUFFER_SPACE            0xE38E4008
#define RS_LIB_ERROR_CANT_OPEN_TTG_FILE_FOR_MAKING                 0xE38E4009
#define RS_LIB_ERROR_CANT_WRITE_PREAMBLE_BYTES                     0xE38E400A
#define RS_LIB_ERROR_CANT_WRITE_TT_INFO_BYTES                      0xE38E400B
#define RS_LIB_ERROR_CANT_QUERY_NUMBER_OF_SPEEDS                   0xE38E400C
#define RS_LIB_ERROR_CANT_QUERY_TTG_DIRECTORY                      0xE38E400D
#define RS_LIB_ERROR_CANT_QUERY_TTG_INFO_PREAMBLE                  0xE38E400E
#define RS_LIB_ERROR_BAD_BYTE_COUNT_IN_QUERIED_PREAMBLE            0xE38E400F
#define RS_LIB_ERROR_CANT_OPEN_RSINIT_DEVICE                       0xE38E4010
#define RS_LIB_ERROR_CANT_GET_SHARED_STATE                         0xE38E4011
#define RS_LIB_ERROR_INVALID_NUMBER_OF_SPEEDS                      0xE38E4012

#define RS_LIB_WARNING_CANT_BEGIN_AUTO_SYSTEM_TASK                 0xA38E4000
#define RS_LIB_WARNING_CANT_BEGIN_AUTO_APP_TASK                    0xA38E4001
#define RS_LIB_WARNING_MESSAGE_TYPE_NOT_KEY                        0xA38E4002

#define RS_LIB_SUCCESS_INITIALIZED_OK                              0x238E4000

#endif
