//
// InstallRoutines.cpp contains routines for installing RightSpeed in stages.
//
// For 79-character width, truncate the window after this sentence is finished.

#include "stdafx.h"
#include "Cpu.h"
#include "Install.h"
#include "InstallDlg.h"
#include "InstallRoutines.h"
#include "../Common/Parameters.h"
#include "../Common/Version.h"
#include <winioctl.h>
#include <winsvc.h>
#include <shlobj.h>

////////////////////////////////////////////////////////////////////////////
//                                MACROS
////////////////////////////////////////////////////////////////////////////

#define INITIAL_GUESS_SERVICE_CONFIG_SIZE       256

////////////////////////////////////////////////////////////////////////////
//                          FORWARD DEFINITIONS
////////////////////////////////////////////////////////////////////////////

void DoInstall (HWND hWnd, CString& mainDirectory, DWORD previousVersion);

BOOL IsItAGoodNTVersion (void);
BOOL DoNeededSystemFilesExist (char *systemRoot);
BOOL IsItAGoodProcessor (void);
BOOL DoesAnyConflictingFileExist (char *systemRoot);
BOOL CreateMainDirectoryAndSubdirectories (CString& mainDirectory);
BOOL CreateRightSpeedRegistryKeys (HWND hWnd, CString& mainDirectory);
BOOL InstallDriver (char *driverName, char *dependencies, DWORD startTime,
                    char *group, BOOL isFileSystem, DWORD *tagPtr,
                    SC_HANDLE schSCManager, char *systemRoot);
BOOL InstallDLL (char *systemRoot);
BOOL InstallUtilities (CString& mainDirectory, DWORD previousVersion);
BOOL MyRegSetValue (HKEY hKey, char *name, BYTE *valuePtr, DWORD type,
                    DWORD length, char *errorDescription);
BOOL GetTagAndGroup (SC_HANDLE schService, DWORD *tagPtr, char *groupName,
                     DWORD maxGroupSize);
BOOL RemoveTagFromGroup (DWORD tag, char *groupName);
BOOL AddTagsToGroup (char *groupName, DWORD *tags, DWORD numNewTags);

////////////////////////////////////////////////////////////////////////////
//                            ERROR DISPLAY
////////////////////////////////////////////////////////////////////////////

// DisplayError displays, in a message box, the error message given
// in the string generalReason as well as the error message implied
// by the system error state.

void DisplayError (char *generalReason)
{
  char *lpMsgBuf;
  CString str;

  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                NULL, GetLastError(),
                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf,
                0, NULL);
  str.Format("%s because %s", generalReason, lpMsgBuf);
  MessageBox(NULL, str, "Error", MB_OK | MB_ICONEXCLAMATION);

  LocalFree(lpMsgBuf);
}

////////////////////////////////////////////////////////////////////////////
//                            FREE DISK SPACE
////////////////////////////////////////////////////////////////////////////

// GetFreeMB gets the number of MB free in the given directory (directoryName).
// If the directory doesn't exist, it tries to remove subdirectory names from
// the end of its name until it reaches one that does exist.  It puts the
// number of free MB in the given pointer (freeMBPtr) and returns a boolean
// indicating whether it was successful.

BOOL GetFreeMB (CString& directoryName, UINT *freeMBPtr)
{
  ULARGE_INTEGER freeBytes, dummy1, dummy2;

  // Check how much free space is in the directory

  if (GetDiskFreeSpaceEx(directoryName, &freeBytes, &dummy1, &dummy2)) {
    *freeMBPtr = (UINT) (freeBytes.QuadPart / 1024 / 1024);
    return TRUE;
  }

  // If we didn't succeed, try subdirectories

  char smallerDirectoryName[MAX_PATH];
  int endPos;

  // First, copy the directory name into smallerDirectoryName so we can play
  // around with it, and put its end position in endPos

  strcpy(smallerDirectoryName, directoryName);
  endPos = strlen(smallerDirectoryName);

  // Keep looping as long as the last call to GetDiskFreeSpaceEx returned
  // a path (or file) not found error

  while (GetLastError() == ERROR_PATH_NOT_FOUND ||
         GetLastError() == ERROR_FILE_NOT_FOUND) {
    // Each time through the loop, scan backward from the end for the last \.
    // If there isn't one, break out of the loop and fail.  If there is such
    // a \, replace it with a null character and try getting the disk space
    // again.
    for (endPos--; endPos > 1 && smallerDirectoryName[endPos] != '\\';
         endPos--)
      ;
    if (endPos < 2) break;
    smallerDirectoryName[endPos] = '\0';
    if (GetDiskFreeSpaceEx(smallerDirectoryName, &freeBytes, &dummy1,
                           &dummy2)) {
      *freeMBPtr = (UINT) (freeBytes.QuadPart / 1024 / 1024);
      return TRUE;
    }
  }

  // We failed, so just return failure

  return FALSE;
}

////////////////////////////////////////////////////////////////////////////
//                    HIGH-LEVEL INSTALLATION ROUTINES
////////////////////////////////////////////////////////////////////////////

// InstallAsynchronously installs RightSpeed asynchronously.  It takes
// as its single parameter an LPVOID which can be cast into a pointer
// to an INSTALL_PARAMETERS structure containing all the information
// needed for the install.  Its strange definition style is so that it
// can be called as a thread routine.

UINT InstallAsynchronously (LPVOID param)
{
  INSTALL_PARAMETERS *installParameters = (INSTALL_PARAMETERS *) param;

  DoInstall(installParameters->hWnd, installParameters->mainDirectory,
            installParameters->previousVersion);
  PostMessage(installParameters->hWnd, WM_INSTALL_COMPLETE, 0, 0);
  return 0;
}

// DoInstall installs RightSpeed. It takes as parameters a window
// handle (hWnd) that it uses for changing the status message, a
// directory name (mainDirectory) that it uses as the place to install
// VTrace, a personal name (personalName) that the user entered, a
// DWORD indicating the currently installed version (previousVersion,
// 0 if none installed).

void DoInstall (HWND hWnd, CString& mainDirectory, DWORD previousVersion)
{
  SC_HANDLE schSCManager;
  char systemRoot[MAX_PATH];
  DWORD tag[NUM_TAGS_NEEDED];

  // Make sure this version of NT can run VTrace

  PostMessage(hWnd, WM_CHANGE_STATUS,
              (WPARAM) "Checking that we can run on this version of Windows",
              0);
  if (!IsItAGoodNTVersion()) return;
  if (GetEnvironmentVariable("SystemRoot", systemRoot, 128) == 0) {
    DisplayError("Could not obtain the system root directory environment "
                 "variable");
    return;
  }
  if (!DoNeededSystemFilesExist(systemRoot)) return;
  if (!IsItAGoodProcessor()) return;
  if (previousVersion == 0 && DoesAnyConflictingFileExist(systemRoot)) return;

  // Create main RightSpeed directory and its traces subdirectory

  PostMessage(hWnd, WM_CHANGE_STATUS,
              (WPARAM) "Creating RightSpeed directories",
              0);
  if (!CreateMainDirectoryAndSubdirectories(mainDirectory)) return;

  // Create and set the registry information for RightSpeed

  PostMessage(hWnd, WM_CHANGE_STATUS,
              (WPARAM) "Setting RightSpeed registry information", 0);
  if (!CreateRightSpeedRegistryKeys(hWnd, mainDirectory)) return;

  PostMessage(hWnd, WM_CHANGE_STATUS, (WPARAM) "Installing drivers", 0);

  // Get handle to service control manager

  schSCManager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
  if (schSCManager == NULL) {
    DisplayError("Could not open the service control manager");
    return;
  }

  // Install drivers. The difference between NT and 2000 are as
  // follows. In NT, we use the "Filter" group, while in 2000 we use
  // the "System Bus Extender" group (the same group as the disks) so
  // that RSIoCnt doesn't get "moved up" ahead of RSInit in the
  // actual load order.

#if TARGET_NT_VERSION == 4

  if (!InstallDriver("RSInit",
                     "", // dependencies
                     SERVICE_BOOT_START,
                     "Filter", // group
                     FALSE, // file system?
                     &tag[0],
                     schSCManager,
                     systemRoot) ||
      !InstallDriver("RSLog",
                     "", // dependencies
                     SERVICE_AUTO_START,
                     "", // group
                     FALSE, // file system?
                     NULL, // tag
                     schSCManager,
                     systemRoot) ||
      !InstallDriver("RSIoCnt",
                     "", // dependencies
                     SERVICE_BOOT_START,
                     "Filter", // group
                     TRUE, // file system?
                     &tag[1],
                     schSCManager,
                     systemRoot) ||
      !InstallDriver("RSTask",
                     "", // dependencies
                     SERVICE_AUTO_START,
                     "", // group
                     TRUE, // file system?
                     NULL, // tag
                     schSCManager,
                     systemRoot)) {
    CloseServiceHandle(schSCManager);
    return;
  }

  // Add tags to the Filter group.

  if (!AddTagsToGroup("Filter", tag, NUM_TAGS_NEEDED)) {
    CloseServiceHandle(schSCManager);
    return;
  }

#else

  if (!InstallDriver("RSInit",
                     "", // dependencies
                     SERVICE_BOOT_START,
                     "System Bus Extender", // group
                     FALSE, // file system?
                     &tag[0],
                     schSCManager,
                     systemRoot) ||
      !InstallDriver("RSLog",
                     "", // dependencies
                     SERVICE_AUTO_START,
                     "", // group
                     FALSE, // file system?
                     NULL, // tag
                     schSCManager,
                     systemRoot) ||
      !InstallDriver("RSIoCnt",
                     "", // dependencies
                     SERVICE_BOOT_START,
                     "System Bus Extender", // group
                     TRUE, // file system?
                     &tag[1],
                     schSCManager,
                     systemRoot) ||
      !InstallDriver("RSTask",
                     "", // dependencies
                     SERVICE_AUTO_START,
                     "", // group
                     TRUE, // file system?
                     NULL, // tag
                     schSCManager,
                     systemRoot)) {
    CloseServiceHandle(schSCManager);
    return;
  }

  // Add tags to the System Bus Extender group.

  if (!AddTagsToGroup("System Bus Extender", tag, NUM_TAGS_NEEDED)) {
    CloseServiceHandle(schSCManager);
    return;
  }

#endif

  // Release service control manager handle
  
  CloseServiceHandle(schSCManager);

  // Now that we've installed the drivers, use the RSInit driver to
  // read CPU information.

  if (!AddCpuInformationToRegistry(hWnd, mainDirectory))
    return;

  // Install RightSpeed DLL

  PostMessage(hWnd, WM_CHANGE_STATUS, (WPARAM) "Installing DLL", 0);
  if (!InstallDLL(systemRoot)) return;

  // Install utilities in main RightSpeed directory and put shortcuts to them
  // in the Start menu under a new directory called RightSpeed.

  PostMessage(hWnd, WM_CHANGE_STATUS,
              (WPARAM) "Installing utilities in Start menu under RightSpeed",
              0);
  if (!InstallUtilities(mainDirectory, previousVersion)) return;

  // Display success message

  PostMessage(hWnd, WM_CHANGE_STATUS, (WPARAM) "Installation complete", 0);
  MessageBox(NULL,
     "Install performed successfully.\n"
     "You must reboot the system for some aspects of the installation "
     "to be complete.\n",
             "Success",
             MB_OK);
}

////////////////////////////////////////////////////////////////////////////
//                    FILE/DIRECTORY BASIC OPERATIONS
////////////////////////////////////////////////////////////////////////////

// RecursiveCreateDirectory creates a directory with the given name (name)
// and security attributes (security).  Unlike CreateDirectory, which just
// fails if a subdirectory of the desired directory does not exist, this
// routine actually creates any necessary subdirectories.  For example,
// if the desired name was C:\FOO\BAR\ZIP, and C:\FOO did not exist, then
// this routine would create C:\FOO, then C:\FOO\BAR, then C:\FOO\BAR\ZIP.

BOOL RecursiveCreateDirectory (char *name, LPSECURITY_ATTRIBUTES security)
{
  char *p;
  BOOL result;
  DWORD err;

  // First, try to create the directory straight on with CreateDirectory.
  // Usually, this will succeed.  (That's why we do it first -- make the
  // common case fast.)

  result = CreateDirectory(name, security);
  if (result) return TRUE;
  err = GetLastError();
  if (err == ERROR_ALREADY_EXISTS) return TRUE;

  // If it failed because the path didn't exist, call RecursiveCreateDirectory
  // on the subdirectory of the given directory, then try CreateDirectory
  // again.

  if (err == ERROR_PATH_NOT_FOUND) {
    p = strrchr(name, '\\');
    if (p != NULL) {
      *p = '\0';
      result = RecursiveCreateDirectory(name, security);
      *p = '\\';
      if (result)
        result = CreateDirectory(name, security);
    }
  }

  return result;
}

// CreateShortcut creates a shortcut with a given name (linkName) pointing
// to an existing file or directory of another given name (targetName).  It
// assumes that CoInitialize() has already been called to initialize COM,
// which is used by this routine.

BOOL CreateShortcut (char *linkName, char *targetName)
{
  HRESULT hres;
  IShellLink* psl;
  IPersistFile* ppf;
  WORD linkNameAsUnicode[MAX_PATH];

  // Get a pointer to the IShellLink interface
  hres = CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER,
                          IID_IShellLink, (void **) &psl);
  if (!SUCCEEDED(hres))
    return FALSE;

  // Query IShellLink for the IPersistFile interface for saving the
  // shell link in persistent storage
  hres = psl->QueryInterface(IID_IPersistFile, (void **) &ppf);
  if (!SUCCEEDED(hres)) {
    psl->Release();
    return FALSE;
  }

  // Set the path to the shell link target
  hres = psl->SetPath(targetName);
  if (!SUCCEEDED(hres)) {
    ppf->Release();
    psl->Release();
    return FALSE;
  }

  // Convert the link name ANSI string to Unicode
  if (MultiByteToWideChar(CP_ACP, 0, linkName, -1, linkNameAsUnicode,
                          MAX_PATH) == 0) {
    ppf->Release();
    psl->Release();
    return FALSE;
  }

  // Save the link via the IPersistFile::Save method
  hres = ppf->Save(linkNameAsUnicode, TRUE);
    
  // Release pointers to IPersistFile and IShellLink
  ppf->Release();
  psl->Release();

  return SUCCEEDED(hres);
}

// This function attempts to determine the location of the Program Files
// directory.  If it can, it puts it in 'dir' and returns TRUE; otherwise,
// it returns FALSE.

BOOL GetProgramFilesDirectory (CString& dir)
{
  DWORD err, type, dirNameLength;
  HKEY hKey;
  char dirName[MAX_PATH];

  // Open the registry key where the information is.

  err = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                     "Software\\Microsoft\\Windows\\CurrentVersion",
                     NULL,
                     KEY_READ,
                     &hKey);
  if (err != ERROR_SUCCESS)
    return FALSE;

  // Read the string, presumed to be of type REG_SZ, in the value
  // ProgramFilesDir.

  dirNameLength = MAX_PATH;
  err = RegQueryValueEx(hKey,
                        "ProgramFilesDir",
                        0,
                        &type,
                        (LPBYTE) dirName,
                        &dirNameLength);
  if (err != ERROR_SUCCESS || type != REG_SZ)
    return FALSE;

  // We succeeded, so put the directory name in 'dir' and return TRUE.

  dir = dirName;
  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
//                            STRING OPERATIONS
////////////////////////////////////////////////////////////////////////////

// ContainsRightSpeedDLL searches for the RightSpeed DLL name in a
// string which is a space-separated list of DLL names.  The
// RightSpeed DLL name must be in all caps for it to match here.  It
// returns a boolean indicating whether the name was found.

BOOL ContainsRightSpeedDLL (char *str)
{
  char *p;

  // Search starting at the beginning of the string.  Each time
  // through the loop, look for the first occurrence of the substring
  // "RIGHTSPD.DLL". If it isn't found, break out of the loop and
  // return FALSE.  Otherwise, if the substring represents the full
  // name of a DLL (i.e., it is bounded on the left by a space or
  // beginning-of-string, and bounded on the right by a space or
  // end-of-string), then return TRUE. Otherwise, advance the search
  // start point by 12, since the next occurrence of RIGHTSPD.DLL must
  // be at least 12 characters further in the string.

  for (p = str; 1; p+=12) {
    p = strstr(p, "RIGHTSPD.DLL");
    if (p == NULL) break;
    if ((p == str || *(p-1) == ' ') && (*(p+12) == ' ' || *(p+12) == '\0'))
      return TRUE;
  }
  return FALSE;
}

// ConvertVersionNumberToString converts a RightSpeed version number
// (version) to a string (str).  For instance, it converts 0x010203 to
// "1.2.3".

void ConvertVersionNumberToString (CString& str, DWORD version)
{
  if (version & 0xFF)
    str.Format("%u.%u.%u", version >> 16, (version >> 8) & 0xFF,
               version & 0xFF);
  else
    str.Format("%u.%u", version >> 16, (version >> 8) & 0xFF);
}

// FindStringInMultiSz finds the position of the given full string in
// the given REG_MULTI_SZ value of the given length. If the string
// isn't found, it returns -1.

int FindStringInMultiSz (char *multiSzVal, int multiSzLength,
                         char *searchString)
{
  for (int possibleMatchPos = 0;
       possibleMatchPos < multiSzLength && multiSzVal[possibleMatchPos];
       possibleMatchPos += (strlen(&multiSzVal[possibleMatchPos]) + 1))
    if (!stricmp(&multiSzVal[possibleMatchPos], searchString))
      return possibleMatchPos;

  return -1;
}

////////////////////////////////////////////////////////////////////////////
//                        LOAD ORDER GROUP TAGS
////////////////////////////////////////////////////////////////////////////

// GetTagAndGroup determines the tag and group of a given service.
// If such a tag and group exists, it puts them in tagPtr and groupName and
// returns TRUE; otherwise, it returns FALSE.

BOOL GetTagAndGroup (SC_HANDLE schService, DWORD *tagPtr, char *groupName,
                     DWORD maxGroupSize)
{
  QUERY_SERVICE_CONFIG *serviceConfigInfo;
  DWORD sizeNeeded;

  // Read the configuration information for the service. The first guess
  // as to the needed size for this information could be wrong, so try
  // again if the return value indicates a bigger size is needed.

  serviceConfigInfo = (QUERY_SERVICE_CONFIG *)
    malloc(INITIAL_GUESS_SERVICE_CONFIG_SIZE);
  if (serviceConfigInfo == NULL) return FALSE;
  if (!QueryServiceConfig(schService, serviceConfigInfo,
                          INITIAL_GUESS_SERVICE_CONFIG_SIZE, &sizeNeeded)) {
    if (GetLastError() != ERROR_INSUFFICIENT_BUFFER) {
      free(serviceConfigInfo);
      return FALSE;
    }
    else {
      free(serviceConfigInfo);
      serviceConfigInfo = (QUERY_SERVICE_CONFIG *) malloc(sizeNeeded);
      if (serviceConfigInfo == NULL) return FALSE;
      if (!QueryServiceConfig(schService, serviceConfigInfo, sizeNeeded,
                              &sizeNeeded)) {
        free(serviceConfigInfo);
        return FALSE;
      }
    }
  }

  // Extract the tag and group information if it exists.

  if (serviceConfigInfo->lpLoadOrderGroup &&
      serviceConfigInfo->lpLoadOrderGroup[0] &&
      strlen(serviceConfigInfo->lpLoadOrderGroup) < maxGroupSize) {
    strcpy(groupName, serviceConfigInfo->lpLoadOrderGroup);
    *tagPtr = serviceConfigInfo->dwTagId;
    free(serviceConfigInfo);
    return TRUE;
  }
  else {
    free(serviceConfigInfo);
    return FALSE;
  }
}

// RemoveTagFromGroup removes the given tag number from the given
// group. Actually, it removes the tag from the GroupOrderList
// registry entry that indicates the order in which that group
// services should load.

BOOL RemoveTagFromGroup (DWORD tag, char *groupName)
{
  DWORD err, type, length, lengthNeeded, i, j, *groupOrderList;
  HKEY hKey;

  // Open the group order list registry key.

  err = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                     "System\\CurrentControlSet\\Control\\GroupOrderList",
                     NULL, KEY_READ | KEY_WRITE, &hKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not open registry key for service group order list");
    return FALSE;
  }

  // Read the value named after the group to get the actual group order
  // list. First we read the value with no buffer to put it into, to find
  // out how long a buffer we need, if any. If there is no such key, we're
  // done. Otherwise, we read the value from the registry. Note that the
  // structure for the value is a sequence of DWORD's, the first of which
  // is the number of tags n, and the next n of which are those tags in the
  // order in which they should be started.

  lengthNeeded = 0;
  err = RegQueryValueEx(hKey, groupName, NULL, &type, NULL, &lengthNeeded);
  if (err == ERROR_FILE_NOT_FOUND) {
    RegCloseKey(hKey);
    return TRUE;
  }
  else if (err == ERROR_SUCCESS) {
    groupOrderList = (DWORD *) malloc(lengthNeeded);
    length = lengthNeeded;
    err = RegQueryValueEx(hKey, groupName, NULL, &type,
                          (LPBYTE) groupOrderList, &length);
    if (err != ERROR_SUCCESS) {
      SetLastError(err);
      DisplayError("Could not read group order list value");
      RegCloseKey(hKey);
      return FALSE;
    }
    if (type != REG_BINARY || lengthNeeded != length ||
        (groupOrderList[0]+1) * sizeof(DWORD) != length) {
      MessageBox(NULL, "Read invalid group order list value", "Error",
                 MB_OK | MB_ICONEXCLAMATION);
      RegCloseKey(hKey);
      return FALSE;
    }
  }
  else {
    SetLastError(err);
    DisplayError("Could not get size of group order list value");
    RegCloseKey(hKey);
    return FALSE;
  }

  // Search for the tag in the group order list. If it's there, remove it.

  for (i = 1; i <= groupOrderList[0]; i++) {
    if (groupOrderList[i] == tag) {
      for (j = i; j < groupOrderList[0]; j++)
        groupOrderList[j] = groupOrderList[j+1];
      groupOrderList[0]--;
      break;
    }
  }

  // Set the new group order list in the registry.

  err = RegSetValueEx(hKey, groupName, 0, REG_BINARY, (LPBYTE) groupOrderList,
                      (groupOrderList[0] + 1) * sizeof(DWORD));
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not set the group order list in the registry");
    RegCloseKey(hKey);
    return FALSE;
  }

  // Release resources used and return a status value.

  free(groupOrderList);
  RegCloseKey(hKey);

  return TRUE;
}

// AddTagsToGroup adds a given set of tags to the load order of a given
// service group. The name of the group is given in groupName; the set of
// tags is given in the array tags of length numNewTags. The order the tags
// appear in the array match the order in which they should apppear in the
// group order list. If a tag already exists in the group order list, it
// should not be added.

BOOL AddTagsToGroup (char *groupName, DWORD *tags, DWORD numNewTags)
{
  DWORD err, type, length, lengthNeeded, i, j, *groupOrderList;
  BOOL found, retVal;
  HKEY hKey;

  // Open the group order list registry key.

  err = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                     "System\\CurrentControlSet\\Control\\GroupOrderList",
                     NULL, KEY_READ | KEY_WRITE, &hKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not open registry key for service group order list");
    return FALSE;
  }

  // Read the value named after the group to get the actual group
  // order list. First we read the value with no buffer to put it
  // into, to find out how long a buffer we need, if any. If there is
  // no such key, the group does not exist, so we create our own value
  // with no tags. Otherwise, we read the value from the registry. In
  // either case, we allocate extra space for the value to accommodate
  // the extra numNewTags tags. Note that the structure for the value
  // is a sequence of DWORD's, the first of which is the number of
  // tags n, and the next n of which are those tags in the order in
  // which they should be started.

  lengthNeeded = 0;
  err = RegQueryValueEx(hKey, groupName, NULL, &type, NULL, &lengthNeeded);
  if (err == ERROR_FILE_NOT_FOUND) {
    groupOrderList = (DWORD *) malloc((numNewTags+1) * sizeof(DWORD));
    groupOrderList[0] = 0;
  }
  else if (err == ERROR_SUCCESS) {
    groupOrderList = (DWORD *) malloc(lengthNeeded +
                                      numNewTags * sizeof(DWORD));
    length = lengthNeeded;
    err = RegQueryValueEx(hKey, groupName, NULL, &type,
                          (LPBYTE) groupOrderList, &length);
    if (err != ERROR_SUCCESS) {
      SetLastError(err);
      DisplayError("Could not read group order list value");
      RegCloseKey(hKey);
      return FALSE;
    }
    if (type != REG_BINARY || lengthNeeded != length ||
        (groupOrderList[0]+1) * sizeof(DWORD) != length) {
      MessageBox(NULL, "Read invalid group order list value", "Error",
                 MB_OK | MB_ICONEXCLAMATION);
      RegCloseKey(hKey);
      return FALSE;
    }
  }
  else {
    SetLastError(err);
    DisplayError("Could not read size of group order list value");
    RegCloseKey(hKey);
    return FALSE;
  }

  // Add the array of tags to the list of tags in 'groupOrderList'. Ignore any
  // tags that are already found in the array.

  for (i = 0; i < numNewTags; i++) {
    found = FALSE;
    for (j = 1; j <= groupOrderList[0]; j++) {
      if (groupOrderList[j] == tags[i]) {
        found = TRUE;
        break;
      }
    }
    if (!found) {
      groupOrderList[0]++;
      groupOrderList[groupOrderList[0]] = tags[i];
    }
  }

  // Change the group order list to the new group order list.

  retVal = MyRegSetValue(hKey, groupName, (LPBYTE) groupOrderList, REG_BINARY,
                         (groupOrderList[0] + 1) * sizeof(DWORD),
                         "new group order list");

  // Release resources used and return a status value.

  free(groupOrderList);
  RegCloseKey(hKey);

  return retVal;
}

////////////////////////////////////////////////////////////////////////////
//                          REGISTRY OPERATIONS
////////////////////////////////////////////////////////////////////////////

// MyRegSetValue is like RegSetValueEx, except it handles error conditions.
// It takes as parameters the key to change (hKey), the name of the value
// to change (name), a pointer to the new value (valuePtr), the new type
// (type), the length of the new value (length), and a description of the
// registry value to be used in generating error messages (errorDescription).

BOOL MyRegSetValue (HKEY hKey, char *name, BYTE *valuePtr, DWORD type,
                    DWORD length, char *errorDescription)
{
  LONG status;

  status = RegSetValueEx(hKey, name, 0, type, valuePtr, length);
  if (status != ERROR_SUCCESS) {
    char errorString[256];
    sprintf(errorString, "Could not set the %s in the registry",
            errorDescription);
    SetLastError(status);
    DisplayError(errorString);
    return FALSE;
  }
  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
//                      LOW-LEVEL INSTALLATION ROUTINES
////////////////////////////////////////////////////////////////////////////

// IsItAGoodNTVersion returns whether this is a version of Windows NT that
// VTrace can run on.

BOOL IsItAGoodNTVersion (void)
{
  OSVERSIONINFO versionInfo;

  // Read the version information

  versionInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
  if (!GetVersionEx(&versionInfo)) {
    DisplayError("Could not get Windows NT version information");
    return FALSE;
  }

  // Check to see if it's the right version of Windows NT

  if (versionInfo.dwPlatformId != VER_PLATFORM_WIN32_NT || 
      versionInfo.dwMajorVersion != TARGET_NT_VERSION) {
#if TARGET_NT_VERSION == 4
      MessageBox(NULL, "This is not Windows NT 4.0. Cannot install.", "Error",
                 MB_OK | MB_ICONEXCLAMATION);
#else
      MessageBox(NULL, "This is not Windows 2000. Cannot install.", "Error",
                 MB_OK | MB_ICONEXCLAMATION);
#endif
    return FALSE;
  }

  return TRUE;
}

// IsItAGoodProcessor checks whether it is running on a uniprocessor
// x86 that is Pentium or better. It returns a boolean indicating if
// that is the case.

BOOL IsItAGoodProcessor (void)
{
  SYSTEM_INFO sysInfo;

  GetSystemInfo(&sysInfo);
  if (sysInfo.dwNumberOfProcessors > 1) {
    MessageBox(NULL,
               "RightSpeed does not work on a multiprocessor. Cannot install.",
               "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }
  if (sysInfo.wProcessorArchitecture != PROCESSOR_ARCHITECTURE_INTEL) {
    MessageBox(NULL,
               "RightSpeed can only work on an Intel processor. "
               "Cannot install.",
               "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }
  if (sysInfo.wProcessorLevel < 5) {
    MessageBox(NULL,
               "RightSpeed needs a Pentium or better processor. "
               "Cannot install.",
               "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }

  return TRUE;
}

// DoNeededSystemFilesExist checks whether system files we need exist in the
// system directory, specified by the given string (systemRoot).

BOOL DoNeededSystemFilesExist (char *systemRoot)
{
  char fileName[MAX_PATH];
  HANDLE hfile;

  // Check for OLE32.DLL, which is needed for creating shortcuts to
  // utilities in the Start menu.

  sprintf(fileName, "%s\\SYSTEM32\\OLE32.DLL", systemRoot);
  hfile = CreateFile(fileName, GENERIC_READ, FILE_SHARE_READ, NULL,
                     OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (hfile == INVALID_HANDLE_VALUE) {
    DisplayError("The system file OLE32.DLL is missing");
    return FALSE;
  }
  CloseHandle(hfile);

  return TRUE;
}

// DoesConflictingFileExist returns a BOOL indicating whether a file
// in the system directory exists.  The parameters are the system root
// directory (systemRoot), and the file name (fileName).  The file name
// may contain backslashes.

BOOL DoesConflictingFileExist (char *systemRoot, char *fileName)
{
  char fullPath[MAX_PATH];
  HANDLE fileHandle;

  // Put the full path of the file in fullPath

  sprintf(fullPath, "%s\\%s", systemRoot, fileName);

  // Try opening the file.  If it succeeds, we know the file exists.

  fileHandle = CreateFile(fullPath, GENERIC_READ, FILE_SHARE_READ |
                          FILE_SHARE_WRITE | FILE_SHARE_DELETE, NULL,
                          OPEN_EXISTING, 0, NULL);
  if (fileHandle != INVALID_HANDLE_VALUE) {
    char errorString[MAX_PATH + 240];
    CloseHandle(fileHandle);
    sprintf(errorString,
            "RightSpeed cannot be installed.\n"
            "Apparently, some other software created a file with the same "
            "name as one that RightSpeed needs to create.\n"
            "In case you think this file is unnecessary and want to try "
            "deleting it,\n"
            "its name is \"%s\".", fullPath);
    MessageBox(NULL, errorString, "Error", MB_OK | MB_ICONEXCLAMATION);
    return TRUE;
  }

  return FALSE;
}

// DoesAnyConflictingFileExist returns a BOOL indicating whether a
// conficting file exists, like a RIGHTSPD.DLL that some other program is
// using. It takes as a parameter the name of the system root directory.

BOOL DoesAnyConflictingFileExist (char *systemRoot)
{
  if (DoesConflictingFileExist(systemRoot, "System32\\RightSpd.dll") ||
      DoesConflictingFileExist(systemRoot, "System32\\Drivers\\RSIoCnt.sys") ||
      DoesConflictingFileExist(systemRoot, "System32\\Drivers\\RSTask.sys") ||
      DoesConflictingFileExist(systemRoot, "System32\\Drivers\\RSLog.sys") ||
      DoesConflictingFileExist(systemRoot, "System32\\Drivers\\RSInit.sys"))
    return TRUE;

  return FALSE;
}

// CreateMainDirectoryAndSubdirectories creates the main RightSpeed
// directory, named by the parameter mainDirectory, as well as
// subdirectories of that named Logs and TTGroups.

BOOL CreateMainDirectoryAndSubdirectories (CString& mainDirectory)
{
  char dirName[MAX_PATH];

  // Create the main directory

  strcpy(dirName, mainDirectory);
  if (!RecursiveCreateDirectory(dirName, NULL)) {
    DisplayError("Could not create main directory");
    return FALSE;
  }

  // Create the Logs subdirectory

  sprintf(dirName, "%s\\Logs", mainDirectory);
  if (!CreateDirectory(dirName, NULL) &&
      GetLastError() != ERROR_ALREADY_EXISTS) {
    DisplayError("Could not create Logs subdirectory in RightSpeed main "
                 "directory");
    return FALSE;
  }

  // Create the TTG subdirectory

  sprintf(dirName, "%s\\TTG", mainDirectory);
  if (!CreateDirectory(dirName, NULL) &&
      GetLastError() != ERROR_ALREADY_EXISTS) {
    DisplayError("Could not create TTG subdirectory in RightSpeed main "
                 "directory");
    return FALSE;
  }

  return TRUE;
}

// CreateRightSpeedRegistryKeys creates the RightSpeed registry key.
// Some of the pieces of data it must put there are the location of
// its main directory and the personal name given by the user; those
// are given as parameters to this function.

BOOL CreateRightSpeedRegistryKeys (HWND hWnd, CString& mainDirectory)
{
  HKEY hKey, oldKey;
  LONG err;
  DWORD n;

  // Create a subkey "Jay Lorch" under software in the registry.  This
  // is the "company" of which RightSpeed is a "product".

  err = RegCreateKey(HKEY_LOCAL_MACHINE, "Software\\Jay Lorch", &hKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not create a registry key for this software");
    return FALSE;
  }

  // Create a subkey "RightSpeed" under "Jay Lorch" in the registry.  This is
  // where we put all the registry information for RightSpeed.

  oldKey = hKey;
  err = RegCreateKey(oldKey, "RightSpeed", &hKey);
  RegCloseKey(oldKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not create a registry key for this software");
    return FALSE;
  }

  // Create a subkey "CurrentVersion" under "RightSpeed" in the
  // registry.  This is where we hold all the information for the
  // current version of RightSpeed.

  oldKey = hKey;
  err = RegCreateKey(oldKey, "CurrentVersion", &hKey);
  RegCloseKey(oldKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not create a registry key for this version of the "
                 "software");
    return FALSE;
  }

  // Set various parameters in the registry

  n = RIGHTSPEED_VERSION;
  if (!MyRegSetValue(hKey, "Version", (BYTE *) &n, REG_DWORD, 4,
                     "version number")) {
    RegCloseKey(hKey);
    return FALSE;
  }

  if (!MyRegSetValue(hKey, "MainDirectory", (BYTE *) (LPCTSTR) mainDirectory,
                     REG_SZ, mainDirectory.GetLength() + 1,
                     "RightSpeed main directory")) {
    RegCloseKey(hKey);
    return FALSE;
  }

  // Close the registry key

  RegCloseKey(hKey);

  return TRUE;
}

// InstallDriver installs a given driver (named driverName), dependent on zero
// or more other drivers (specified in dependencies), that starts at a certain
// time (given by startTime), in a certain group (named in group).  To do this,
// it needs an open handle to the service control manager (schSCManager) and
// the name of the system root directory (systemRoot).  It returns a boolean
// indicating whether it was successful.

BOOL InstallDriver (char *driverName, char *dependencies, DWORD startTime,
                    char *group, BOOL isFileSystem, DWORD *tagPtr,
                    SC_HANDLE schSCManager, char *systemRoot)
{
  SC_HANDLE schService;
  char targetFileLocation[MAX_PATH], fileToInstall[MAX_PATH],
       errString[MAX_PATH];
  char oldGroupName[128];
  DWORD maxGroupSize = 128, oldTag;

  // Copy the driver file (ending in .SYS) to the drivers subdirectory of
  // the system32 subdirectory of the system root directory.  This is a good
  // place for it, because it is guaranteed to be available for reading at the
  // time that drivers are loaded (as opposed to the RightSpeed main directory,
  // which might be on a remote file system).

  sprintf(fileToInstall, "%s.SYS", driverName);
  sprintf(targetFileLocation, "%s\\SYSTEM32\\DRIVERS\\%s.SYS", systemRoot,
          driverName);
  if (!CopyFile(fileToInstall, targetFileLocation, FALSE)) {
    sprintf(errString, "Could not copy %s.sys to the drivers directory",
            driverName);
    DisplayError(errString);
    return FALSE;
  }

  // Set targetFileLocation to contain the name of the new driver location, in
  // general terms (using %systemroot% to denote the system root)

  sprintf(targetFileLocation, "%%systemroot%%\\system32\\drivers\\%s.sys",
          driverName);

  // Create the service for this driver.

  schService = OpenService(schSCManager, driverName, SERVICE_ALL_ACCESS);
  if (schService == NULL)
    schService = CreateService(schSCManager, driverName, driverName,
                               SERVICE_ALL_ACCESS,
                               isFileSystem ? SERVICE_FILE_SYSTEM_DRIVER
                                            : SERVICE_KERNEL_DRIVER,
                               startTime,
                               SERVICE_ERROR_NORMAL, targetFileLocation,
                               group[0] ? group : NULL,
                               tagPtr,
                               dependencies[0] ? dependencies : NULL,
                               NULL, NULL);
  else {
    // If the driver belongs to a different group than we now are adding
    // it to, then remove its tag from that group.

    if (GetTagAndGroup(schService, &oldTag, oldGroupName, maxGroupSize) &&
        strcmp(oldGroupName, group))
      RemoveTagFromGroup(oldTag, oldGroupName);

    // Change the service parameters to the appropriate new ones.

    if (!ChangeServiceConfig(schService,
                             isFileSystem ? SERVICE_FILE_SYSTEM_DRIVER
                                          : SERVICE_KERNEL_DRIVER,
                             startTime, SERVICE_ERROR_NORMAL, NULL, group,
                             tagPtr, dependencies, NULL, NULL, NULL)) {
      sprintf(errString, "Could not change configuration of %s service",
              driverName);
      DisplayError(errString);
      return FALSE;
    }
  }

  if (schService == NULL && GetLastError() != ERROR_SERVICE_EXISTS) {
    sprintf(errString, "Could not create %s service", driverName);
    DisplayError(errString);
    return FALSE;
  }

  // Creating the service opens a handle to it, which we don't need, so close
  // that handle

  if (schService != NULL)
    CloseServiceHandle(schService);

  return TRUE;
}

// InstallDLL installs the RightSpeed DLL.  To do this, it needs the
// name of the system root directory (given in systemRoot).

BOOL InstallDLL (char *systemRoot)
{
  LONG err;
  HKEY hKey;
  DWORD type, length;
  char appInitDLLs[MAX_PATH];
  char realTargetFileName[MAX_PATH], temporaryTargetFileName[MAX_PATH];

  // Copy RIGHTSPD.DLL to the system32 subdirectory of the system root
  // directory.  Because RIGHTSPD.DLL might already exist and be open
  // (e.g. if this is an upgrade to a running version of RightSpeed),
  // we cannot simply overwrite it.  Instead, we copy RIGHTSPD.DLL to
  // a temporary location in the System32 directory and arrange for it
  // to be moved to the proper location at reboot.  We put the
  // temporary location in the System32 directory, since that location
  // will always be available at reboot.

  sprintf(realTargetFileName, "%s\\SYSTEM32\\RIGHTSPD.DLL", systemRoot);
  sprintf(temporaryTargetFileName, "%s\\SYSTEM32\\TempRSFileRIGHTSPD.DLL",
          systemRoot);
  if (!CopyFile("RIGHTSPD.DLL", temporaryTargetFileName, FALSE)) {
    DisplayError("Could not copy RIGHTSPD.DLL to the System32 directory");
    return FALSE;
  }
  if (!MoveFileEx(temporaryTargetFileName, realTargetFileName,
                  MOVEFILE_DELAY_UNTIL_REBOOT | MOVEFILE_REPLACE_EXISTING)) {
    DisplayError("Could not install RIGHTSPD.DLL in the System32 directory");
    DeleteFile(temporaryTargetFileName);
    return FALSE;
  }

  // Open the Windows software registry key

  err = RegOpenKeyEx
          (HKEY_LOCAL_MACHINE,
           "Software\\Microsoft\\Windows NT\\CurrentVersion\\Windows", 0,
           KEY_QUERY_VALUE | KEY_SET_VALUE, &hKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not open registry key for Windows software");
    return FALSE;
  }

  // Query the value of "AppInit_DLLs", which returns a space-separated list
  // of all the DLL's that are automatically loaded into each application

  length = sizeof(appInitDLLs);
  err = RegQueryValueEx(hKey, "AppInit_DLLs", 0, &type, (BYTE *) appInitDLLs,
                        &length);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not get AppInit_DLLs registry value");
    RegCloseKey(hKey);
    return FALSE;
  }
  if (type != REG_SZ) {
    MessageBox(NULL, "AppInit_DLLs registry value is not a string", "Error",
               MB_OK | MB_ICONEXCLAMATION);
    RegCloseKey(hKey);
    return FALSE;
  }

  // If RIGHTSPD.DLL does not already appear on that list, append it
  // to the end of the list and change the "AppInit_DLLs" value to be
  // the new list

  if (!ContainsRightSpeedDLL(appInitDLLs)) {
    if (appInitDLLs[0] == '\0')
      strcpy(appInitDLLs, "RIGHTSPD.DLL");
    else
      strcat(appInitDLLs, " RIGHTSPD.DLL");
    err = RegSetValueEx(hKey, "AppInit_DLLs", 0, REG_SZ, (BYTE *) appInitDLLs,
                        strlen(appInitDLLs) + 1); // includes trailing null
    if (err != ERROR_SUCCESS) {
      SetLastError(err);
      DisplayError("Could not set AppInit_DLLs registry value");
      RegCloseKey(hKey);
      return FALSE;
    }
  }

  // Close the Windows software registry key

  RegCloseKey(hKey);

  return TRUE;
}

// InstallUtilities installs various utility programs in the
// RightSpeed main directory (mainDirectory).  For those of them that
// the user might want to run, it puts shortcuts to them in the Start
// menu, under a new subdirectory RightSpeed <version>.  It returns a
// boolean indicating whether it was successful.  It takes an
// additional parameter, previousVersion, that indicates the previous
// version of RightSpeed that was installed, so that if this value is
// non-zero the old program folder for that version can be deleted.

BOOL InstallUtilities (CString& mainDirectory, DWORD previousVersion)
{
  char sourceFile[MAX_PATH], destFile[MAX_PATH],
       topLevelProgramFolder[MAX_PATH];
  CString programFolder, versionAsString;
  LPITEMIDLIST idList;

  // Find the program group directory name

  if (SHGetSpecialFolderLocation(NULL, CSIDL_COMMON_PROGRAMS, &idList) !=
      NO_ERROR) {
    DisplayError("Could not find Start menu directory");
    return FALSE;
  }
  if (!SHGetPathFromIDList(idList, topLevelProgramFolder)) {
    DisplayError("Could not get Start menu directory name");
    return FALSE;
  }

  // Add \RightSpeed {version number} to the end of the directory name
  // to get the program folder name for RightSpeed.

  ConvertVersionNumberToString(versionAsString, RIGHTSPEED_VERSION);
  programFolder = topLevelProgramFolder;
  programFolder += "\\RightSpeed ";
  programFolder += versionAsString;

  // If there was a previous version installed, try to rename the old program
  // folder to the new name

  if (previousVersion != 0 && previousVersion != RIGHTSPEED_VERSION) {
    CString oldProgramFolder;
    ConvertVersionNumberToString(versionAsString, previousVersion);
    oldProgramFolder = topLevelProgramFolder;
    oldProgramFolder += "\\RightSpeed ";
    oldProgramFolder += versionAsString;
    MoveFile(oldProgramFolder, programFolder);
  }

  // Create the new program folder for RightSpeed.

  if (!CreateDirectory(programFolder, NULL) &&
      GetLastError() != ERROR_ALREADY_EXISTS) {
    DisplayError("Could not create RightSpeed subdirectory in Start menu");
    return FALSE;
  }

  // Initialize COM, to be used for creating shortcuts.

  if (!SUCCEEDED(CoInitialize(NULL))) {
    DisplayError("Could not initialize the Component Object Model (COM)");
    return FALSE;
  }

  // Copy the Readme file to the main RightSpeed directory, and create
  // a shortcut to it in the RightSpeed program folder

  sprintf(sourceFile, "Readme.txt");
  sprintf(destFile, "%s\\Readme.txt", mainDirectory);
  if (!CopyFile(sourceFile, destFile, FALSE)) {
    DisplayError("Could not copy Readme text file to RightSpeed directory");
    CoUninitialize();
    return FALSE;
  }
  sprintf(sourceFile, "%s\\Readme file.lnk", programFolder);
  if (!CreateShortcut(sourceFile, destFile)) {
    DisplayError("Could not create shortcut to Readme file in the Start menu");
    CoUninitialize();
    return FALSE;
  }

  // Copy the Uninstall program to the main VTrace directory, and create a
  // shortcut to it in the VTrace program folder

  sprintf(sourceFile, "Uninstal.exe");
  sprintf(destFile, "%s\\Uninstal.exe", mainDirectory);
  if (!CopyFile(sourceFile, destFile, FALSE)) {
    DisplayError("Could not copy Uninstall program to RightSpeed directory");
    CoUninitialize();
    return FALSE;
  }
  sprintf(sourceFile, "%s\\Uninstall RightSpeed.lnk", programFolder);
  if (!CreateShortcut(sourceFile, destFile)) {
    DisplayError("Could not create shortcut to Uninstall program in the Start "
                 "menu");
    CoUninitialize();
    return FALSE;
  }

  // Close COM now that we're done with it

  CoUninitialize();

  return TRUE;
}
