// This file implements the RSInitHandle class, which represents an
// open handle to the RSInit device.

#include "stdafx.h"
#include <windows.h>
#include <winioctl.h>
#include <iostream>
#include <winsvc.h>
#include "InstallRoutines.h"
#include "RSInitHandle.h"

//////////////////////////////////////////////////////////////////////////////
// INITIALIZATION / TERMINATION
//////////////////////////////////////////////////////////////////////////////

// This function initializes the object to represent an unopened
// handle.

RSInitHandle::RSInitHandle() :
  h(INVALID_HANDLE_VALUE),
  scManagerHandle(NULL),
  serviceHandle(NULL),
  mustStopRSInit(false)
{
}

// This function initializes the handle by opening a handle to the
// RSInit device.

bool RSInitHandle::Initialize ()
{
  // Get a handle to the service control manager.

  scManagerHandle = OpenSCManager(NULL, // which machine: default
                                  NULL, // which database: default
                                  SC_MANAGER_ALL_ACCESS);
  if (scManagerHandle == NULL) {
    DisplayError("Could not open the service control manager");
    return false;
  }

  // Get a handle to the RSInit service.

  serviceHandle = OpenService(scManagerHandle,
                              "RSInit",
                              SERVICE_ALL_ACCESS);
  if (serviceHandle == NULL) {
    DisplayError("Could not open a handle to the RSInit service");
    return false;
  }

  // Start the RSInit service if it isn't started already.

  if (StartService(serviceHandle,
                   NULL, // service arguments
                   NULL)) { // argument vectors
    mustStopRSInit = true;
  }
  else if (GetLastError() != ERROR_SERVICE_ALREADY_RUNNING) {
    DisplayError("Could not start RSInit service");
    return false;
  }

  // Open the handle by opening the file \\.\RSInit.

  h = CreateFile("\\\\.\\RSInit",
                 GENERIC_READ | GENERIC_WRITE,
                 0,     // share mode
                 NULL,  // security attributes
                 OPEN_EXISTING,
                 FILE_ATTRIBUTE_NORMAL,
                 NULL); // template
  if (h == INVALID_HANDLE_VALUE)
    return false;

  // Return success.

  return true;
}

// This function closes the handle, if it is open. It also closes any
// other resources that we have open.

void RSInitHandle::Close ()
{
  // Close the handle if it's open.
  
  if (h != INVALID_HANDLE_VALUE) {
    CloseHandle(h);
    h = INVALID_HANDLE_VALUE;
  }

  // If we have to stop RSInit, do so.

  if (mustStopRSInit && serviceHandle != NULL) {
    SERVICE_STATUS serviceStatus;
    ControlService(serviceHandle,
                   SERVICE_CONTROL_STOP,
                   &serviceStatus);
    mustStopRSInit = false;
  }

  // If we have an open handle to RSInit, close it.

  if (serviceHandle != NULL) {
    CloseServiceHandle(serviceHandle);
    serviceHandle = NULL;
  }

  // If we have an open handle to the service control manager, close it.

  if (scManagerHandle != NULL) {
    CloseServiceHandle(scManagerHandle);
    scManagerHandle = NULL;
  }
}

// This function writes the speed MSR with the given two DWORD's.

void RSInitHandle::WriteSpeedMsr (DWORD longRunCode1, DWORD longRunCode2)
{
  struct {
    DWORD longRunCode1;
    DWORD longRunCode2;
  } sendBuf;
  DWORD returnSize;

  // Fill in the buffer to send to the device.

  sendBuf.longRunCode1 = longRunCode1;
  sendBuf.longRunCode2 = longRunCode2;

  // Send the write-speed-MSR request.

  DeviceIoControl(h,
                  RS_INIT_IOCTL_WRITE_SPEED_MSR,
                  &sendBuf,         // input buffer
                  sizeof(sendBuf), // input buffer length
                  0,               // output buffer
                  0,               // output buffer length
                  &returnSize,
                  NULL);           // overlapped structure
}
