// This file contains functions for dealing with CPU speeds
// available on this CPU.

#include "stdafx.h"
#include <vector>
#include "../Common/Parameters.h"
#include "Cpu.h"
#include "InstallRoutines.h"
#include "RSInitHandle.h"
#include "Speeds.h"

/////////////////////////////////////////////////////////////////////////////
// MACROS
/////////////////////////////////////////////////////////////////////////////

// This is the number of times to iterate through the CPU loop:
#define CPU_LOOP_ITERATIONS           1000000000

/////////////////////////////////////////////////////////////////////////////
// SPEEDS INFO
/////////////////////////////////////////////////////////////////////////////

// This function determines information about the current speed. It
// determines what the LongRun code of the current speed is and puts
// it in 'codePtr'. It determines what the reported speed is in MHz
// and puts that in 'reportedMHzPtr'. It determines what the reported voltage
// is in mV and puts that in mvPtr.

void GetCurrentSpeedInformation (DWORD *codePtr, DWORD *reportedMHzPtr,
                                 DWORD *mvPtr)
{
  *** ELIDED DUE TO CONFIDENTIALITY AGREEMENTS ***
}

// This function performs a busy-wait loop for a given number of
// iterations.

long CpuLoop (long numIterations)
{
  long i, j = 0;

  for (i = 0; i < numIterations; ++i)
    j += i;

  return j;
}

// This function determines what speeds are available on this CPU and
// fills in the list 'availableSpeeds' with them. It returns a boolean
// indicating if it was successful.

BOOL GetAvailableSpeeds (list<AvailableSpeedInfo>& availableSpeeds,
                         DWORD tscTicksPerMs)
{
  *** ELIDED DUE TO CONFIDENTIALITY AGREEMENTS ***

  // Make sure there aren't too many available speeds.

  if (availableSpeeds.size() > MAX_NUM_AVAILABLE_SPEEDS) {
    MessageBox(NULL, "There are too many available speeds.", "Error",
               MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }

  // Now, we use the cycle count for each speed to compute a measured
  // speed (as opposed to the CPU-reported speed) for each speed.
  // Presuming that the maximum speed we found has a timestamp tick
  // per cycle, we set 'cpuLoopCycles' to be the timestamp count
  // observed when running the CPU loop at the maximum speed. We
  // use this value to compute actual speeds.

  double cpuLoopCycles = availableSpeeds.back().cpuLoopTSC;
  list<AvailableSpeedInfo>::iterator it;
  DWORD ratio, prevKHz = 0;
  for (it = availableSpeeds.begin(); it != availableSpeeds.end(); ++it) {
    // To compute the speed in cycles per millisecond, take the number
    // of cycles in the CPU loop, divide it by the timestamp counter
    // ticks to run that CPU loop at this speed, then multiply by
    // the number of timestamp counter ticks per millisecond.
    it->khz = (DWORD) (cpuLoopCycles / it->cpuLoopTSC * tscTicksPerMs + 0.5);

    // Make sure the computed speed is within 5% of the speed reported
    // by the processor; if it's not, use the reported speed. The speed
    // in cycles per millisecond should be about 1000 times the speed
    // in MHz, so make sure that the ratio is between 950 and 1050.

    ratio = (it->khz / it->reportedMHz);
    if (ratio < 950 || ratio > 1050)
      it->khz = it->reportedMHz * 1000;

    if (it->khz < prevKHz) {
      MessageBox(NULL, "The available speeds are out of order.", "Error",
                 MB_OK | MB_ICONEXCLAMATION);
      return FALSE;
    }
    else
      prevKHz = it->khz;
  }

  // Return success.

  return TRUE;
}

// This function computes the worthwhile settings given a list of
// available speeds, and puts those worthwhile settings in
// 'worthwhileSettings'.

BOOL GetWorthwhileSettings (list<WorthwhileSettingInfo>& worthwhileSettings,
                            list<AvailableSpeedInfo>& availableSpeeds)
{
  // First, convert the available speeds into arrays so that we can
  // easily access entries by index.

  short numAvailableSpeeds = availableSpeeds.size();
  DWORD *avail_khz = new DWORD[numAvailableSpeeds];
  DWORD *avail_mv = new DWORD[numAvailableSpeeds];
  list<AvailableSpeedInfo>::iterator it;
  short i;
  for (i = 0, it = availableSpeeds.begin(); it != availableSpeeds.end();
       ++i, ++it) {
    avail_khz[i] = it->khz;
    avail_mv[i] = it->mv;
  }

  // Compute the first worthwhile setting as the one with the lowest
  // voltage and thus the lowest energy consumption.  Among those
  // tied, prefer ones with higher speeds.

  short best = 0;
  for (i = 1; i < numAvailableSpeeds; ++i)
    if (avail_mv[i] <= avail_mv[best])
      best = i;

  worthwhileSettings.push_back(WorthwhileSettingInfo(best,
                                                     avail_khz[best],
                                                     avail_mv[best]));

  // Loop through the rest of the speeds, each time setting 'last' to
  // the last worthwhile setting and computing the next 'best' worthwhile
  // setting.  This is the one with a higher voltage than the last one
  // which minimizes the slope of the power-vs-speed line passing between
  // it and the last one.  In other words, it's the one that minimizes
  // (P_i - P_last) / (s_i - s_last), preferring ones with higher speeds
  // in case of ties.

  short last;
  double ratio, bestRatio;
  while (1) {
    last = best;

    // Our first guess for the best one is the smallest successor speed
    // that has higher energy consumption.

    for (best = last + 1; best < numAvailableSpeeds; ++best)
      if (avail_mv[best] > avail_mv[last])
        break;

    // If there is no such successor, we're done, so break out of the
    // loop.

    if (best >= numAvailableSpeeds)
      break;

    // Compute the ratio (P_i - P_last) / (s_i - s_last) for this initial
    // guess at the best i.

    bestRatio = (1.0 * avail_mv[best] * avail_mv[best] * avail_khz[best] -
                 1.0 * avail_mv[last] * avail_mv[last] * avail_khz[last]) /
                 (avail_khz[best] - avail_khz[last]);

    // Loop through the remaining speeds looking for one with a better
    // ratio (or an equal ratio, considering that we prefer higher speeds
    // when ratios are tied).

    for (i = best + 1; i < numAvailableSpeeds; ++i) {
      if (avail_mv[i] > avail_mv[last]) {
        ratio = (1.0 * avail_mv[i] * avail_mv[i] * avail_khz[i] -
                 1.0 * avail_mv[last] * avail_mv[last] * avail_khz[last]) /
                 (avail_khz[i] - avail_khz[last]);
        if (ratio <= bestRatio) {
          best = i;
          bestRatio = ratio;
        }
      }
    }

    // Add the new worthwhile setting to the set of worthwhile settings.

    worthwhileSettings.push_back(WorthwhileSettingInfo(best,
                                                       avail_khz[best],
                                                       avail_mv[best]));
  }

  // Free the arrays we allocated and return success.

  delete avail_khz;
  delete avail_mv;
  return TRUE;
}
