// This file contains routines for determining if RightSpeed should run
// this boot.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "Activate.h"

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, ShouldActivateRightSpeed)
#endif

//////////////////////////////////////////////////////////////////////////
// ACTIVATION
//////////////////////////////////////////////////////////////////////////

// ShouldActivateRightSpeed returns a boolean indicating whether
// RightSpeed should be activated this time. It returns TRUE if it
// determines that we are not using the LastKnownGood configuration.

BOOLEAN ShouldActivateRightSpeed (void)
{
  NTSTATUS status;
  UNICODE_STRING unicodeString;
  OBJECT_ATTRIBUTES objectAttributes;
  HANDLE keyHandle;
  struct {
    ULONG   TitleIndex;
    ULONG   Type;
    ULONG   DataLength;
    WCHAR   Data[2];
  } keyValueInfo;
  ULONG keyValueInfoLength;
  ULONG current, lastKnownGood;

  // Open the registry key for selecting which configuration to start
  // the system with.

  RtlInitUnicodeString(&unicodeString, L"\\Registry\\Machine\\System\\Select");
  InitializeObjectAttributes(&objectAttributes,
                             &unicodeString,
                             OBJ_CASE_INSENSITIVE,
                             NULL, // root directory
                             NULL); // security descriptor
  status = ZwOpenKey(&keyHandle,
                     KEY_QUERY_VALUE | KEY_SET_VALUE,
                     &objectAttributes);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSInit: ERROR - Cannot open the configuration key (%x).\n",
                status));
    return FALSE;
  }

  // Read which configuration the current one is.

  RtlInitUnicodeString(&unicodeString, L"Current");
  status = ZwQueryValueKey(keyHandle,
                           &unicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSInit: ERROR - Cannot determine the current configuration "
                "(%x).\n", status));
    ZwClose(keyHandle);
    return FALSE;
  }
  if (keyValueInfo.Type != REG_DWORD) {
    DebugPrint(("RSInit: ERROR - Current configuration is type %d, not "
                "REG_DWORD", keyValueInfo.Type));
    ZwClose(keyHandle);
    return FALSE;
  }
  current = * (ULONG *) keyValueInfo.Data;

  // Read which configuration the last-known-good one is.

  RtlInitUnicodeString(&unicodeString, L"LastKnownGood");
  status = ZwQueryValueKey(keyHandle,
                           &unicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSInit: ERROR - Cannot determine the last known good "
                "configuration number (%x).\n", status));
    ZwClose(keyHandle);
    return FALSE;
  }
  if (keyValueInfo.Type != REG_DWORD) {
    DebugPrint(("RSInit: ERROR - Last known good configuration is type %d, "
                "not REG_DWORD", keyValueInfo.Type));
    ZwClose(keyHandle);
    return FALSE;
  }
  lastKnownGood = * (ULONG *) keyValueInfo.Data;

  // Close the registry key.

  ZwClose(keyHandle);

  // Activate RightSpeed if and only if the current configuration is not
  // the last-known-good one.

  if (current == lastKnownGood) {
    DebugPrint(("RSInit: Deactivating RightSpeed because the LastKnownGood "
                "configuration is in use.\n"));
    return FALSE;
  }
  else {
    DebugPrint(("RSInit: Activating RightSpeed because the LastKnownGood "
                "configuration is not in use.\n"));
    return TRUE;
  }
}
