// This file contains routines for performing I/O control requests
// for RSInit.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/MSR.h"
#include "Globals.h"
#include "Log.h"

////////////////////////////////////////////////////////////////////////////
// I/O CONTROL ROUTINES
////////////////////////////////////////////////////////////////////////////

// This function returns the pointer to the global shared state. This
// I/O control call is made from kernel level, so we just copy the
// pointer directly to the output buffer.

NTSTATUS IoctlGetSharedState (
  PCHAR outBuf,
  ULONG outBufLength,
  PULONG returnedBytesPtr
  )
{
  DebugPrint(("RSInit: Responding to get-shared-state request.\n"));
  if (outBufLength < sizeof(SHARED_STATE *))
    return STATUS_INVALID_PARAMETER;

  * (SHARED_STATE **) outBuf = g.sharedState;
  *returnedBytesPtr = sizeof(SHARED_STATE *);
  return STATUS_SUCCESS;
}

// This function maps the kernel-user shared state into the address space
// of the calling process and returns a pointer to it. It records in
// the associated file object enough information to free this mapping
// when the file object is closed.

NTSTATUS IoctlGetKernelUserSharedState (
  PFILE_OBJECT fileObject,
  PCHAR outBuf,
  ULONG outBufLength,
  PULONG returnedBytesPtr
  )
{
  PVOID address;
  PMDL mdl;

  // Make sure there's enough room allocated for the pointer to be
  // returned.

  if (outBufLength < sizeof(PVOID)) {
    DebugPrint(("RSInit: ERROR - Request for user-shared state had "
                "too short an output buffer.\n"));
    return STATUS_INVALID_PARAMETER;
  }

  // Make sure there's an associated file object where we can store
  // the information necessary to undo this mapping when it closes.

  if (!fileObject) {
    DebugPrint(("RSInit: ERROR - Can't share state with user level "
                "because file object is null.\n"));
    if (LOGGING_WARNINGS)
      LogError(RS_INIT_WARNING_CANT_SHARE_STATE_WITH_NULL_FILE_OBJECT);
    return STATUS_UNSUCCESSFUL;
  }

  // If we've already created a mapping for a region of memory using this
  // file object, fail this call.

  if (fileObject->FsContext || fileObject->FsContext2) {
    DebugPrint(("RSInit: ERROR - Can't share state with user level because "
                "we're already mapping memory with this file object.\n"));
    if (LOGGING_WARNINGS)
      LogError(RS_INIT_WARNING_CANT_SHARE_STATE_BECAUSE_ACTIVE_MAPPING);
    return STATUS_UNSUCCESSFUL;
  }

  // Allocate an MDL big enough to map the kernel-user shared state.

  mdl = IoAllocateMdl((PVOID) &g.sharedState->kernelUserSharedState,
                      sizeof(KERNEL_USER_SHARED_STATE),
                      FALSE,  // secondary buffer
                      FALSE,  // charge quota
                      NULL);  // associated IRP
  if (!mdl) {
    DebugPrint(("RSInit: ERROR - Can't get user shared state because "
                "MDL initialization failed.\n"));
    BothLogShort(RS_INIT_ERROR_CANT_ALLOCATE_MDL_FOR_STATE,
                 STATUS_INSUFFICIENT_RESOURCES);
    return STATUS_INSUFFICIENT_RESOURCES;
  }

  // Build the MDL for the kernel-user shared state.

  MmBuildMdlForNonPagedPool(mdl);

  // Map the locked pages into user mode and get the address.
  
  address = MmMapLockedPages(mdl, UserMode);
  if (!address) {
    DebugPrint(("RSInit: ERROR - Can't get user shared state "
                "because MmMapLockedPages failed.\n"));
    BothLogShort(RS_INIT_ERROR_CANT_MAP_SHARED_STATE_TO_USER,
                 STATUS_UNSUCCESSFUL);
    IoFreeMdl(mdl);
    return STATUS_UNSUCCESSFUL;
  }

  // Record the MDL and address in the file object so that the address
  // can be unmapped when the file object is closed.

  fileObject->FsContext = address;
  fileObject->FsContext2 = mdl;

  // If this is NT 4.0, it might be Service Pack 3 or earlier, in which
  // case we have been given the base address rather than the actual
  // address of the kernel-user shared state. So, for NT 4.0, we convert
  // to an address in the proper location on the page.

#if TARGET_NT_VERSION == 4
  address = (PVOID) ((ULONG)PAGE_ALIGN(address) + MmGetMdlByteOffset(mdl));
#endif

  DebugPrint(("RSInit: Mapped user shared state to %x\n", address));

  // Return the address in the output buffer.

  * (PVOID *) outBuf = address;
  *returnedBytesPtr = sizeof(address);

  return STATUS_SUCCESS;
}

// This function logs the event passed in the input buffer. The first
// four bytes are interpreted as the event type.

NTSTATUS IoctlLogEvent (
  PCHAR inBuf,
  ULONG inBufLength
  )
{
  if (inBufLength < 4) {
    DebugPrint(("RSInit: Received invalid log-event request.\n"));
    return STATUS_INVALID_PARAMETER;
  }

  LogAnyEventTypeAndBytes(* (short *) inBuf, &inBuf[4], inBufLength - 4);
  return STATUS_SUCCESS;
}

// This function writes the speed MSR with the 8 bytes passed in the
// input buffer.

NTSTATUS IoctlWriteSpeedMsr (
  PCHAR inBuf,
  ULONG inBufLength
  )
{
  ULONG value1, value2;

  if (inBufLength < 8) {
    DebugPrint(("RSInit: Received invalid write-speed-MSR request.\n"));
    return STATUS_INVALID_PARAMETER;
  }

  value1 = * (ULONG *) inBuf;
  value2 = * (ULONG *) &inBuf[4];

  __asm {
    mov ecx, TMTA_MSR_SET_SPEED
    mov eax, value1
    mov edx, value2
    wrmsr
  }

  return STATUS_SUCCESS;
}
