// This file contains routines for detaching from devices.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "Globals.h"
#include "Detach.h"
#include "Log.h"

///////////////////////////////////////////////////////////////////////////
// SAFE ROUTINES
///////////////////////////////////////////////////////////////////////////

// This function safely detaches us from the device it's filtering. It
// returns a boolean indicating if it successfully does so.

BOOLEAN SafeDetachDevice (PDEVICE_OBJECT ourDevice)
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = ourDevice->DeviceExtension;
  BOOLEAN success = FALSE;

  // If there is no filtered device, just return failure.

  if (!ourDeviceExtension->filteredDevice)
    return FALSE;

  // Do the detachment while holding the device filtering lock.

  ExAcquireResourceExclusiveLite(&g.deviceFilteringLock, TRUE);
  if (ourDeviceExtension->filteredDevice) {
    DebugPrint(("RSIoCnt: Detaching device %x from %x\n", ourDevice,
                ourDeviceExtension->filteredDevice));
    IoDetachDevice(ourDeviceExtension->filteredDevice);
    ourDeviceExtension->filteredDevice = NULL;
    success = TRUE;
  }
  ExReleaseResourceLite(&g.deviceFilteringLock);

  // Return an indication of whether we actually detached.

  return success;
}

// This function deletes one of our devices as long as it isn't attached
// to by some other device. It returns a boolean indicating if it
// actually deleted it.

BOOLEAN SafeDeleteDevice (PDEVICE_OBJECT ourDevice)
{
  POUR_DEVICE_EXTENSION ourDeviceExtension;

  if (ourDevice->AttachedDevice) {
    DebugPrint(("RSIoCnt: WARNING - Can't delete device %x while "
                "attached to.\n", ourDevice));
    ourDeviceExtension = ourDevice->DeviceExtension;
    BothLogOneParameter(RS_IOCNT_WARNING_DELETING_WHILE_ATTACHED,
                        STATUS_UNSUCCESSFUL,
                        ourDeviceExtension->ourDeviceType);
    return FALSE;
  }

  IoDeleteDevice(ourDevice);
  return TRUE;
}

///////////////////////////////////////////////////////////////////////////
// DETACH ROUTINES
///////////////////////////////////////////////////////////////////////////

// This function detaches every device we have that is attached to some
// other device. Typically, it is called as a prelude to unloading.

void DetachAllDevices (void)
{
  PDEVICE_OBJECT ourDevice, nextDevice;

  DebugPrint(("RSIoCnt: Detaching from all devices.\n"));

  // Loop through all the devices associated with our driver. For each,
  // detach it if it's attached, then delete it if we successfully
  // detached.

  for (ourDevice = g.ourDriver->DeviceObject;
       ourDevice != NULL;
       ourDevice = nextDevice) {
    nextDevice = ourDevice->NextDevice;
    if (SafeDetachDevice(ourDevice))
      SafeDeleteDevice(ourDevice);
  }
}

// This function detaches any of our devices that are filtering file
// systems, then deletes those now-useless devices.

void DetachFromAllFileSystems (void)
{
  PDEVICE_OBJECT ourDevice, nextDevice;
  POUR_DEVICE_EXTENSION ourDeviceExtension;

  DebugPrint(("RSIoCnt: Detaching from all file systems.\n"));

  // Loop through all the devices. For any that are file system filters,
  // detach them and, if successful, delete them.

  for (ourDevice = g.ourDriver->DeviceObject;
       ourDevice != NULL;
       ourDevice = nextDevice) {
    nextDevice = ourDevice->NextDevice;
    ourDeviceExtension = ourDevice->DeviceExtension;
    if (ourDeviceExtension->ourDeviceType == OUR_DEVICE_TYPE_FS_FILTER) {
      if (SafeDetachDevice(ourDevice))
        SafeDeleteDevice(ourDevice);
    }
  }
}
