// This file contains routines for performing I/O control operations.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "Ioctls.h"
#include "Filter.h"
#include "Globals.h"
#include "Log.h"

// This function turns on filtering for the drive passed. The given
// input buffer should contain one character consisting of the drive
// letter to filter.

NTSTATUS IoctlFilterDrive (PCHAR inBuf, ULONG inBufLength)
{
  WCHAR fileSystemName[] = L"\\DosDevices\\A:\\";
  UNICODE_STRING fileSystemNameUnicode;
  char driveLetter;
  OBJECT_ATTRIBUTES objectAttributes;
  IO_STATUS_BLOCK ioStatus;
  HANDLE fileHandle;
  PFILE_OBJECT fileObject;
  PDEVICE_OBJECT fsDevice;
  NTSTATUS status;

  // Make sure there's exactly one character in the drive letter
  // specification and record it in 'driveLetter'.

  if (inBufLength != 1) {
    DebugPrint(("RSIoCnt: ERROR - Drive letter passed to filter-drive "
                "I/O control has length %d instead of 1.\n", inBufLength));
    return STATUS_INVALID_PARAMETER;
  }
  driveLetter = *inBuf;

  // Make sure the drive letter is a letter.

  if ((driveLetter >= 'a' && driveLetter <= 'z') ||
      (driveLetter >= 'A' && driveLetter <= 'Z')) {
    DebugPrint(("RSIoCnt: Received request to filter %c:.\n", driveLetter));
  }
  else {
    DebugPrint(("RSIoCnt: ERROR - Received invalid request to filter %c:.\n",
                driveLetter));
    return STATUS_INVALID_PARAMETER;
  }

  // Change the 'A' in fileSystemName, which starts out as \DosDevices\A:\,
  // to the drive letter.

  fileSystemName[12] = driveLetter;

  // Open the root directory of this drive.

  RtlInitUnicodeString(&fileSystemNameUnicode, fileSystemName);
  InitializeObjectAttributes(&objectAttributes,
                             &fileSystemNameUnicode,
                             OBJ_CASE_INSENSITIVE,
                             NULL,     // root directory
                             NULL);    // security descriptor
  status = ZwCreateFile(&fileHandle,
                        SYNCHRONIZE | FILE_ANY_ACCESS,
                        &objectAttributes,
                        &ioStatus,
                        NULL,   // allocation size
                        0,      // file attributes
                        FILE_SHARE_READ | FILE_SHARE_WRITE,
                        FILE_OPEN,
                        FILE_SYNCHRONOUS_IO_NONALERT | FILE_DIRECTORY_FILE,
                        NULL,   // extended-attributes buffer
                        0);     // extended-attributes length
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSIoCnt: ERROR - Could not open drive %c.\n", driveLetter));
    return status;
  }

  // Convert the file handle into a file object and reference that
  // object.

  status = ObReferenceObjectByHandle(fileHandle,
                                     FILE_READ_DATA,
                                     NULL,        // object type
                                     KernelMode,
                                     &fileObject,
                                     NULL);       // handle information
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSIoCnt: ERROR - Could not reference file handle (%x).\n",
                status));
    BothLogShort(RS_IOCNT_ERROR_CANT_REFERENCE_DRIVE_FILE_HANDLE,
                 status);
    ZwClose(fileHandle);
    return status;
  }

  // Get the device object for this file object.

  fsDevice = IoGetRelatedDeviceObject(fileObject);
  if (!fsDevice) {
    DebugPrint(("RSIoCnt: Could not get related device object.\n"));
    BothLogShort(RS_IOCNT_ERROR_CANT_GET_DRIVE_RELATED_DEVICE_OBJECT,
                 STATUS_UNSUCCESSFUL);
    ObDereferenceObject(fileObject);
    ZwClose(fileHandle);
    return STATUS_UNSUCCESSFUL;
  }

  // Begin filtering for the file system represented by that device.

  status = FilterFileSystem(fsDevice);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSIoCnt: Failed to filter file system for drive (%x).\n",
                status));
    ObDereferenceObject(fileObject);
    ZwClose(fileHandle);
    return status;
  }

  // Free resources we've obtained and return success.

  ObDereferenceObject(fileObject);
  ZwClose(fileHandle);
  return STATUS_SUCCESS;
}
