// This file contains the main functionality for RSIoCnt, the driver
// that is responsible for counting synchronous I/O's.
//
// (Uses some code by Mark Russinovich and Bryce Cogswell)
//

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "Detach.h"
#include "FastIo.h"
#include "Filter.h"
#include "GetState.h"
#include "Globals.h"
#include "IoCount.h"
#include "Ioctls.h"
#include "Log.h"
#include "Sleep.h"

///////////////////////////////////////////////////////////////////////////
// MACROS
///////////////////////////////////////////////////////////////////////////

// This macro can be used as if it were a function to indicate whether
// a given IRP is synchronous:

#define IsIrpSynchronous(irp)  ((irp)->Flags & (IRP_SYNCHRONOUS_API | \
                                                IRP_SYNCHRONOUS_PAGING_IO))

// This macro gives the number of milliseconds to wait after detaching
// all devices before being confident that all ongoing operations on
// those devices are complete.

#define MILLISECONDS_TO_WAIT_AFTER_DETACHING_ALL_DEVICES    5000

//////////////////////////////////////////////////////////////////////////
// IFS KIT DEFINITIONS
//////////////////////////////////////////////////////////////////////////

typedef
VOID
(*PDRIVER_FS_NOTIFICATION) (
    IN struct _DEVICE_OBJECT *DeviceObject,
    IN BOOLEAN FsActive
    );

NTKERNELAPI
NTSTATUS
IoRegisterFsRegistrationChange(
    IN PDRIVER_OBJECT DriverObject,
    IN PDRIVER_FS_NOTIFICATION DriverNotificationRoutine
    );

NTKERNELAPI
VOID
IoUnregisterFsRegistrationChange(
    IN PDRIVER_OBJECT DriverObject,
    IN PDRIVER_FS_NOTIFICATION DriverNotificationRoutine
    );

///////////////////////////////////////////////////////////////////////////
// FORWARD DEFININTIONS
///////////////////////////////////////////////////////////////////////////

NTSTATUS DriverEntry (IN PDRIVER_OBJECT DriverObject,
                      IN PUNICODE_STRING RegistryPath);

NTSTATUS DispatchCreate (PDEVICE_OBJECT ourDevice, PIRP irp);
NTSTATUS DispatchCleanupClose (PDEVICE_OBJECT ourDevice, PIRP irp);
NTSTATUS DispatchGeneral (PDEVICE_OBJECT ourDevice, PIRP irp);
NTSTATUS DispatchDeviceControl (PDEVICE_OBJECT ourDevice, PIRP irp);
NTSTATUS DispatchFSControl (PDEVICE_OBJECT ourDevice, PIRP irp);
NTSTATUS DispatchPnp (PDEVICE_OBJECT ourDevice, PIRP irp);
NTSTATUS DispatchPower (PDEVICE_OBJECT ourDevice, PIRP irp);

NTSTATUS PassOnNormally (PDEVICE_OBJECT ourDevice, PIRP irp);
NTSTATUS PassOnToDecrementIoCountWhenComplete (PDEVICE_OBJECT ourDevice,
                                               PIRP irp);
void ReleaseResources (void);
void Unload (PDRIVER_OBJECT ourDriverObject);

////////////////////////////////////////////////////////////////////////////
// SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#endif

///////////////////////////////////////////////////////////////////////////
// GLOBALS
///////////////////////////////////////////////////////////////////////////

RS_IOCNT_GLOBALS g;

///////////////////////////////////////////////////////////////////////////
// DRIVER ENTRY
///////////////////////////////////////////////////////////////////////////

// DriverEntry is the installable driver initialization.

NTSTATUS DriverEntry (IN PDRIVER_OBJECT DriverObject,
                      IN PUNICODE_STRING RegistryPath)
{
  NTSTATUS status;
  UNICODE_STRING deviceNameUnicodeString, miscUnicodeString;
  PFILE_OBJECT rawFileObject;
  PDEVICE_OBJECT rawDeviceObject;
  POUR_DEVICE_EXTENSION ourDeviceExtension;
  int i;

  DebugPrint(("RSIoCnt: Entering DriverEntry\n"));

  // Initialize "resources obtained" and our driver object, since these
  // may be needed to release resources if there is a failure.

  g.resourcesObtained = 0;
  g.ourDriver = DriverObject;

  // Get shared state information from RSInit. If we can't get it, fail.

  status = GetSharedState();
  if (!NT_SUCCESS(status)) {
    ReleaseResources();
    return status;
  }

  // Initialize the ERESOURCE that functions as a lock on all our
  // device object information.

  ExInitializeResource(&g.deviceFilteringLock);

  // Create dispatch points for all routines that must be handled.
  // All entry points are registered since we might filter a file
  // system that processes all of them.

  for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++)
    DriverObject->MajorFunction[i] = DispatchGeneral;

  DriverObject->MajorFunction[IRP_MJ_CREATE] = DispatchCreate;
  DriverObject->MajorFunction[IRP_MJ_FILE_SYSTEM_CONTROL] =
                                                  DispatchFSControl;
  DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] =
                                                  DispatchDeviceControl;
  DriverObject->MajorFunction[IRP_MJ_PNP] = DispatchPnp;
  DriverObject->MajorFunction[IRP_MJ_POWER] = DispatchPower;
  DriverObject->MajorFunction[IRP_MJ_CLEANUP] = DispatchCleanupClose;
  DriverObject->MajorFunction[IRP_MJ_CLOSE] = DispatchCleanupClose;

  // Set up the fast I/O dispatch table and unload routine.

  DriverObject->FastIoDispatch = &FastIoDispatchTable;
  DriverObject->DriverUnload = Unload;

  // Create a main device object.

  RtlInitUnicodeString(&deviceNameUnicodeString, L"\\Device\\RSIoCnt");
  status = IoCreateDevice(DriverObject,
                          sizeof(OUR_DEVICE_EXTENSION),
                          &deviceNameUnicodeString,
                          FILE_DEVICE_RS_IOCNT,
                          0,      // device characteristics
                          FALSE,  // exclusive?
                          &g.mainDevice);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSIoCnt: ERROR - Could not create main device (%x).\n",
                status));
    BothLogShort(RS_IOCNT_ERROR_CANT_CREATE_MAIN_DEVICE,
                 status);
    ReleaseResources();
    return status;
  }
  g.resourcesObtained |= MAIN_DEVICE_RESOURCE_FLAG;

  // Fill in the device extension of our main device.

  ourDeviceExtension = g.mainDevice->DeviceExtension;
  ourDeviceExtension->ourDeviceType = OUR_DEVICE_TYPE_MAIN;
  ourDeviceExtension->filteredDevice = NULL;

  DebugPrint(("RSIoCnt: Successfully created main device object.\n"));

  // Create a symbolic link to the main device object.

  RtlInitUnicodeString(&miscUnicodeString, L"\\DosDevices\\RSIoCnt");
  status = IoCreateSymbolicLink(&miscUnicodeString,
                                &deviceNameUnicodeString);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSIoCnt: WARNING - Could not create symbolic link to "
                "main device (%x).\n", status));
    BothLogShort(RS_IOCNT_WARNING_CANT_CREATE_MAIN_DEVICE_SYMBOLIC_LINK,
                 status);
  }
  else
    g.resourcesObtained |= SYMBOLIC_LINK_RESOURCE_FLAG;

  // Register for file system registration changes.

  status = IoRegisterFsRegistrationChange(DriverObject,
                                          NotifyFsRegistrationChange);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSIoCnt: ERROR - Could not register for file system "
                "registration changes (%x).\n", status));
    BothLogShort(RS_IOCNT_ERROR_CANT_REGISTER_FOR_REGISTRATION_CHANGES,
                 status);
    ReleaseResources();
    return status;
  }
  g.resourcesObtained |= REGISTER_FS_CHANGE_RESOURCE_FLAG;

  DebugPrint(("RSIoCnt: Successfully registered for file system info.\n"));

  // Attempt to open the RAW device object since this driver has already
  // been started by system initialization and will not make it through
  // the normal file system notification procedures otherwise. Then,
  // simulate being notified of this device registering itself as a
  // file system by calling NotifyFsRegistrationChange.

  RtlInitUnicodeString(&deviceNameUnicodeString, L"\\Device\\RawDisk" );
  status = IoGetDeviceObjectPointer(&deviceNameUnicodeString,
                                    FILE_READ_ATTRIBUTES,
                                    &rawFileObject,
                                    &rawDeviceObject);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSIoCnt: WARNING - Could not find address of "
                "\\Device\\RawDisk.\n", status));
    BothLogShort(RS_IOCNT_WARNING_CANT_GET_RAWDISK_ADDRESS,
                 status);
  }
  else {
    DebugPrint(("RSIoCnt: Discovered that \\Device\\RawDisk is at %x.\n",
                rawDeviceObject));
    NotifyFsRegistrationChange(rawDeviceObject, TRUE);
    ObDereferenceObject(rawFileObject);
  }

  return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////
// REQUEST COMPLETION ROUTINES
////////////////////////////////////////////////////////////////////////////

// This is our completion routine for decrementing the I/O count.

NTSTATUS DecrementIoCountCompletionRoutine (
  PDEVICE_OBJECT DeviceObject,
  PIRP irp,
  PVOID Context
  )
{
  DecrementIoCount();
  
  // We must do the following in a completion routine.

  if (irp->PendingReturned) {
    IoMarkIrpPending(irp);
  }

  return irp->IoStatus.Status;
}

// This is our completion routine for IRP's that we sent synchronously
// to a filtered device.

NTSTATUS SynchronousIrpCompletionRoutine (
  PDEVICE_OBJECT DeviceObject,
  PIRP irp,
  PVOID Context
  )
{
  PKEVENT Event = (PKEVENT) Context;

  // We must do the following in a completion routine.

  if (irp->PendingReturned) {
    IoMarkIrpPending(irp);
  }

  // Set the event that we put in the context field to indicate that
  // the IRP is complete.

  KeSetEvent(Event, IO_NO_INCREMENT, FALSE);

  // Return STATUS_MORE_PROCESSING_REQUIRED since we will handle the
  // completion of this IRP elsewhere.

  return STATUS_MORE_PROCESSING_REQUIRED;
}

// This is our completion routine for successful mount requests. It
// creates a new RSIoCnt device to filter the file system that just
// got mounted. The context passed to this routine is the "real
// device" of the VPB involved in the mount request.

NTSTATUS MountCompletionRoutine (PDEVICE_OBJECT DeviceObject, PIRP irp,
                                 PVOID Context)
{
  PDEVICE_OBJECT realDevice = (PDEVICE_OBJECT) Context;
  NTSTATUS status = irp->IoStatus.Status;

  // Do a paranoid check to make sure the IRP is returning success. If it
  // is, filter the file system.

  if (NT_SUCCESS(status)) {
    DebugPrint(("RSIoCnt: About to filter device object %x of VPB %x of "
                "real device %x.\n",
                realDevice->Vpb->DeviceObject, realDevice->Vpb, realDevice));
    FilterFileSystem(realDevice->Vpb->DeviceObject);
  }
  else {
    DebugPrint(("RSIoCnt: WARNING - MountCompletionRoutine called even though "
                "mount failed with status %x.\n", status));
    if (LOGGING_WARNINGS)
      LogErrorOneParameter(RS_IOCNT_WARNING_MOUNT_COMPLETION_CALLED_ON_ERROR,
                           status);
  }

  // We must do the following in a completion routine.

  if (irp->PendingReturned) {
    IoMarkIrpPending(irp);
  }

  // Return the status copied from the IoStatus field of the IRP.

  return status;
}

////////////////////////////////////////////////////////////////////////////
// PASSING IRP ON TO NEXT LOWER DEVICE
////////////////////////////////////////////////////////////////////////////

// PassOnNormally passes on the given IRP to the next lower device
// (also given). It makes no effort to hook the completion of the IRP.

NTSTATUS PassOnNormally (PDEVICE_OBJECT ourDevice, IN PIRP irp)
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = ourDevice->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;

  // Make sure there is a device being filtered before doing this.

  if (!filteredDevice) {
    DebugPrint(("RSIoCnt: WARNING - Tried to pass an IRP on normally when "
                "there was no device filtered by %x.\n", ourDevice));
    if (LOGGING_WARNINGS)
      LogError(RS_IOCNT_WARNING_NOT_FILTERING_WHEN_PASSING_IRP_NORMALLY);
    irp->IoStatus.Status = STATUS_INVALID_DEVICE_REQUEST;
    irp->IoStatus.Information = 0;
    IoCompleteRequest(irp, IO_NO_INCREMENT);
    return STATUS_INVALID_DEVICE_REQUEST;
  }

  // Skip the current stack location and call the next lower driver.

  IoSkipCurrentIrpStackLocation(irp);
  return IoCallDriver(filteredDevice, irp);
}

// PassOnAsCountedIo increments the ongoing I/O count, then passes on
// the given IRP to the next lower device in such a way that the I/O
// count gets decremented when the IRP completes.

NTSTATUS PassOnAsCountedIo (PDEVICE_OBJECT ourDevice, PIRP irp)
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = ourDevice->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;

  if (!filteredDevice) {
    DebugPrint(("RSIoCnt: WARNING - Received a counted I/O request when "
                "there was no device filtered by %x.\n", ourDevice));
    if (LOGGING_WARNINGS)
      LogError(RS_IOCNT_WARNING_NOT_FILTERING_WHEN_PASSING_COUNTED_IRP);

    irp->IoStatus.Status = STATUS_INVALID_DEVICE_REQUEST;
    irp->IoStatus.Information = 0;
    IoCompleteRequest(irp, IO_NO_INCREMENT);
    return STATUS_INVALID_DEVICE_REQUEST;
  }

  // If there isn't enough room for another stack location in the IRP,
  // we can't put in a new completion routine, so we just pass the IRP
  // on normally.

  if (irp->CurrentLocation < ourDevice->StackSize) {
    DebugPrint(("RSIoCnt: WARNING - Not enough stack space to pass on "
                "IRP from device %x to device %x.\n",
                ourDevice, filteredDevice));
    if (LOGGING_WARNINGS)
      LogError(RS_IOCNT_WARNING_STACK_SPACE_INSUFFICIENT_FOR_COUNTING);

    // Skip the current stack location and call the next lower driver.

    IoSkipCurrentIrpStackLocation(irp);
    return IoCallDriver(filteredDevice, irp);
  }

  // Copy parameters down to next level in the stack for the driver
  // below us and set the completion routine.

  IoCopyCurrentIrpStackLocationToNext(irp);
  IoSetCompletionRoutine(irp,
                         DecrementIoCountCompletionRoutine,
                         NULL,    // context
                         TRUE,    // invoke on success
                         TRUE,    // invoke on error
                         TRUE);   // invoke on cancel

  // Increment the I/O count just before passing it on to the
  // lower-level driver.

  IncrementIoCount();
  return IoCallDriver(ourDeviceExtension->filteredDevice, irp);
}

// This function passes on the given IRP to the given filtered device
// synchronously. In other words, it waits for the IRP to be complete
// before returning. It returns the status value for that IRP.

NTSTATUS ForwardIrpSynchronous (PDEVICE_OBJECT filteredDevice, PIRP irp)
{
  KEVENT event;
  NTSTATUS status;

  // First, make sure there really is a device being filtered.

  if (!filteredDevice) {
    DebugPrint(("RSIoCnt: WARNING - Tried to forward an IRP synchronously "
                "when there was no filtered device.\n"));
    if (LOGGING_WARNINGS)
      LogError(RS_IOCNT_WARNING_NOT_FILTERING_WHEN_FORWARDING_IRP);

    irp->IoStatus.Status = STATUS_INVALID_DEVICE_REQUEST;
    irp->IoStatus.Information = 0;
    IoCompleteRequest(irp, IO_NO_INCREMENT);
    return STATUS_INVALID_DEVICE_REQUEST;
  }

  // Create an event to put in the IRP so we can wait for it to complete.

  KeInitializeEvent(&event,
                    NotificationEvent,
                    FALSE); // initial signalled state of event

  // Copy the IRP stack for the next device and set a completion routine.

  IoCopyCurrentIrpStackLocationToNext(irp);
  IoSetCompletionRoutine(irp,
                         SynchronousIrpCompletionRoutine,
                         &event,  // context
                         TRUE,    // invoke on success
                         TRUE,    // invoke on error
                         TRUE);   // invoke on cancel

  // Call the next lower device.

  status = IoCallDriver(filteredDevice, irp);

  // Wait for the actual completion.

  if (status == STATUS_PENDING) {
    KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);
    status = irp->IoStatus.Status;
  }

  // Return the status value for the IRP.
  
  return status;
}

///////////////////////////////////////////////////////////////////////////
// DISPATCH SUBROUTINES
///////////////////////////////////////////////////////////////////////////

// This is the dispatch routine for remove-device requests.

NTSTATUS DispatchRemoveDevice (PDEVICE_OBJECT ourDevice, PIRP irp)
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = ourDevice->DeviceExtension;
  NTSTATUS status;

  DebugPrint(("RSIoCnt: Removing device %x of type %d.\n",
              ourDevice, ourDeviceExtension->ourDeviceType));

  // If we're filtering a device, forward this IRP on to that device,
  // then detach from it.

  if (ourDeviceExtension->filteredDevice) {
    status = ForwardIrpSynchronous(ourDeviceExtension->filteredDevice, irp);
    SafeDetachDevice(ourDevice);
  }
  else
    status = STATUS_SUCCESS;

  // Delete our device.

  IoDeleteDevice(ourDevice);

  // Complete the IRP.

  irp->IoStatus.Status = status;
  irp->IoStatus.Information = 0;
  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return status;
}

// This is the dispatch routine for device-control requests passed to
// the main device.

NTSTATUS MainDeviceDispatchDeviceControl (PDEVICE_OBJECT ourDevice, PIRP irp)
{
  PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation(irp);
  ULONG ioControlCode;
  NTSTATUS status;

  DebugPrint(("RSIoCnt: Received device control request to main device.\n"));

  // Set the request up as returning no information, by default.

  irp->IoStatus.Information = 0;

  // Get the control code (what kind of request this is)

  ioControlCode = irpStack->Parameters.DeviceIoControl.IoControlCode;

  // Do the appropriate thing, depending on the control code.

  switch (ioControlCode) {
  case RS_IOCNT_IOCTL_FILTER_DRIVE :
    status = IoctlFilterDrive(
      irp->AssociatedIrp.SystemBuffer,
      irpStack->Parameters.DeviceIoControl.InputBufferLength
      );
    break;

  case RS_IOCNT_IOCTL_DETACH_FROM_FILE_SYSTEMS :
    DetachFromAllFileSystems();
    status = STATUS_SUCCESS;
    break;

  default :
    DebugPrint(("RSIoCnt: WARNING - Unrecognized I/O control code %x.\n",
                ioControlCode));
    status = STATUS_INVALID_DEVICE_REQUEST;
    break;
  }

  // Complete the IRP.

  irp->IoStatus.Status = status;
  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return status;
}

// This is the dispatch routine for file-system-control requests passed to
// a recognizer filter device.

NTSTATUS FSRecognizerFilterDispatchFSControl (PDEVICE_OBJECT ourDevice,
                                              PIRP irp)
{
  PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation(irp);
  POUR_DEVICE_EXTENSION ourDeviceExtension;
  PDEVICE_OBJECT filteredDevice;
  NTSTATUS status;

  // If this isn't a mount request, just pass it on normally.

  if (irpStack->MinorFunction != IRP_MN_MOUNT_VOLUME)
    return PassOnNormally(ourDevice, irp);

  DebugPrint(("RSIoCnt: File system recognizer filter device %x "
              "received mount request: IRP %x\n", ourDevice, irp));

  // If there is no filtered device, just complete the IRP normally.

  ourDeviceExtension = ourDevice->DeviceExtension;
  filteredDevice = ourDeviceExtension->filteredDevice;
  if (!filteredDevice) {
    DebugPrint(("RSIoCnt: WARNING - Received mount request but there "
                "is no device filtered by %x.\n", ourDevice));
    if (LOGGING_WARNINGS)
      LogError(RS_IOCNT_WARNING_NOT_FILTERING_WHEN_MOUNT_REQUESTED);
    return PassOnNormally(ourDevice, irp);
  }

  // If there isn't room on the stack to pass on the IRP on with a
  // completion routine, just pass it on normally.
    
  if (irp->CurrentLocation < ourDevice->StackSize) {
    DebugPrint(("RSIoCnt: WARNING - Cannot pass on mount request with "
                "another stack location from %x to %x because there "
                "isn't room on the stack.\n", ourDevice, filteredDevice));
    if (LOGGING_WARNINGS)
      LogError(RS_IOCNT_WARNING_STACK_SPACE_INSUFFICIENT_TO_PASS_MOUNT);
    return PassOnNormally(ourDevice, irp);
  }

  // Copy the current stack location to the next one, set the
  // mount completion routine to be called in case of successful
  // completion, and pass on the IRP to the filtered device. Use
  // the "real device" of the VPB of the mounted volume as the
  // context passed to the completion routine.

  IoCopyCurrentIrpStackLocationToNext(irp);
  IoSetCompletionRoutine(irp,
                         MountCompletionRoutine,
                         irpStack->Parameters.MountVolume.Vpb->RealDevice,
                                   // context
                         TRUE,     // invoke on success
                         FALSE,    // invoke on error
                         FALSE);   // invoke on cancel
  return IoCallDriver(filteredDevice, irp);
}

///////////////////////////////////////////////////////////////////////////
// DISPATCH ROUTINES
///////////////////////////////////////////////////////////////////////////

// This is the dispatch routine called for any create IRP.

NTSTATUS DispatchCreate (PDEVICE_OBJECT ourDevice, PIRP irp)
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = ourDevice->DeviceExtension;
  
  switch (ourDeviceExtension->ourDeviceType) {
  case OUR_DEVICE_TYPE_FS_FILTER :
    // If this is a file system filter, pass this IRP on and count its
    // start and end since it's synchronous.

    if (g.sharedState->numOngoingTasks)
      return PassOnAsCountedIo(ourDevice, irp);
    else
      return PassOnNormally(ourDevice, irp);
    break;

  case OUR_DEVICE_TYPE_MAIN :
    // If this is the main device object, just return success and
    // indicate the "file" was opened.

    irp->IoStatus.Status = STATUS_SUCCESS;
    irp->IoStatus.Information = FILE_OPENED;
    IoCompleteRequest(irp, IO_DISK_INCREMENT);
    return STATUS_SUCCESS;
    break;

  default:
    // If this is a file system recognizer filter (or anything else),
    // just pass it on.

    return PassOnNormally(ourDevice, irp);
    break;
  }
}

// This is the dispatch routine called for any cleanup or close IRP.

NTSTATUS DispatchCleanupClose (PDEVICE_OBJECT ourDevice, PIRP irp)
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = ourDevice->DeviceExtension;
  
  switch (ourDeviceExtension->ourDeviceType) {
  case OUR_DEVICE_TYPE_FS_FILTER :
    // If this is a file system filter, pass it on as a counted I/O
    // operation if and only if it's synchronous.

    if (IsIrpSynchronous(irp) && g.sharedState->numOngoingTasks)
      return PassOnAsCountedIo(ourDevice, irp);
    else
      return PassOnNormally(ourDevice, irp);
    break;

  case OUR_DEVICE_TYPE_MAIN :
    // If this is the main device object, just return success.

    irp->IoStatus.Status = STATUS_SUCCESS;
    irp->IoStatus.Information = 0;
    IoCompleteRequest(irp, IO_DISK_INCREMENT);
    return STATUS_SUCCESS;
    break;

  default :
    // If this is a recognizer filter (or anything else), just pass it
    // on normally.

    return PassOnNormally(ourDevice, irp);
    break;
  }
}

// This is the dispatch routine called for a general I/O request.

NTSTATUS DispatchGeneral (PDEVICE_OBJECT ourDevice, PIRP irp)
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = ourDevice->DeviceExtension;
  
  switch (ourDeviceExtension->ourDeviceType) {
  case OUR_DEVICE_TYPE_FS_FILTER :
    // If this is a file system filter, pass it on as a counted I/O
    // operation if and only if it's synchronous.

    if (IsIrpSynchronous(irp) && g.sharedState->numOngoingTasks)
      return PassOnAsCountedIo(ourDevice, irp);
    else
      return PassOnNormally(ourDevice, irp);
    break;

  case OUR_DEVICE_TYPE_FS_RECOGNIZER_FILTER :
    // If this is a recognizer filter, just pass it on normally.

    return PassOnNormally(ourDevice, irp);
    break;

  default :
    // If this is the main device object (or anything else), return
    // failure as this is an invalid request for non-filtering
    // devices.

    irp->IoStatus.Status = STATUS_INVALID_DEVICE_REQUEST;
    irp->IoStatus.Information = 0;
    IoCompleteRequest(irp, IO_NO_INCREMENT);
    return STATUS_INVALID_DEVICE_REQUEST;
    break;
  }
}

// This is the dispatch routine called for any device-control IRP.

NTSTATUS DispatchDeviceControl (PDEVICE_OBJECT ourDevice, PIRP irp)
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = ourDevice->DeviceExtension;
  
  switch (ourDeviceExtension->ourDeviceType) {
  case OUR_DEVICE_TYPE_FS_FILTER :
    // If this is a file system filter, pass it on as a counted I/O
    // operation if and only if it's synchronous.

    if (IsIrpSynchronous(irp) && g.sharedState->numOngoingTasks)
      return PassOnAsCountedIo(ourDevice, irp);
    else
      return PassOnNormally(ourDevice, irp);
    break;

  case OUR_DEVICE_TYPE_MAIN :
    // If this is the main device object (or anything else), use the
    // dispatch routine for device-control requests to the main device.

    return MainDeviceDispatchDeviceControl(ourDevice, irp);
    break;

  default :
    // If this is a recognizer filter (or anything else), just pass it
    // on normally.

    return PassOnNormally(ourDevice, irp);
    break;
  }
}

// This is the dispatch routine called for any file-system-control IRP.

NTSTATUS DispatchFSControl (PDEVICE_OBJECT ourDevice, PIRP irp)
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = ourDevice->DeviceExtension;
  
  switch (ourDeviceExtension->ourDeviceType) {
  case OUR_DEVICE_TYPE_FS_FILTER :
    // If this is a file system filter, pass it on as a counted I/O
    // operation if and only if it's synchronous.

    if (IsIrpSynchronous(irp) && g.sharedState->numOngoingTasks)
      return PassOnAsCountedIo(ourDevice, irp);
    else
      return PassOnNormally(ourDevice, irp);
    break;

  case OUR_DEVICE_TYPE_FS_RECOGNIZER_FILTER :
    // If this is a recognizer filter, pass it on to the file-system-
    // recognizer-filter dispatch routine for file-system-control
    // requests;
    return FSRecognizerFilterDispatchFSControl(ourDevice, irp);
    break;

  default :
    // If this is the main device object (or anything else), fail
    // the request.

    irp->IoStatus.Status = STATUS_INVALID_DEVICE_REQUEST;
    irp->IoStatus.Information = 0;
    IoCompleteRequest(irp, IO_NO_INCREMENT);
    return STATUS_INVALID_DEVICE_REQUEST;
    break;
  }
}

// This is the dispatch routine for PnP requests.

NTSTATUS DispatchPnp (PDEVICE_OBJECT ourDevice, PIRP irp)
{
  PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation(irp);
  POUR_DEVICE_EXTENSION ourDeviceExtension = ourDevice->DeviceExtension;

  // If this is a remove-device request, handle it specially using the
  // remove-device dispatch routine.

  if (irpStack->MinorFunction == IRP_MN_REMOVE_DEVICE)
    return DispatchRemoveDevice(ourDevice, irp);

  // Any other PnP request is handled relatively normally:
  
  switch (ourDeviceExtension->ourDeviceType) {
  case OUR_DEVICE_TYPE_FS_FILTER :
    // If this is a file system filter, pass it on as a counted I/O
    // operation if and only if it's synchronous.

    if (IsIrpSynchronous(irp) && g.sharedState->numOngoingTasks)
      return PassOnAsCountedIo(ourDevice, irp);
    else
      return PassOnNormally(ourDevice, irp);
    break;

  case OUR_DEVICE_TYPE_MAIN :
    // If this is the main device object, just return success.

    irp->IoStatus.Status = STATUS_SUCCESS;
    irp->IoStatus.Information = 0;
    IoCompleteRequest(irp, IO_DISK_INCREMENT);
    return STATUS_SUCCESS;
    break;

  default :
    // If this is a recognizer filter (or anything else), just pass it
    // on normally.

    return PassOnNormally(ourDevice, irp);
    break;
  }
}

// This is the dispatch routine for power requests.

NTSTATUS DispatchPower (PDEVICE_OBJECT ourDevice, PIRP irp)
{
  POUR_DEVICE_EXTENSION ourDeviceExtension = ourDevice->DeviceExtension;
  PDEVICE_OBJECT filteredDevice = ourDeviceExtension->filteredDevice;

  // Do different things depending on whether we're filtering another
  // device.

  if (filteredDevice) {
    // If we're filtering another device, pass the IRP on using the
    // special power routines.
    
    PoStartNextPowerIrp(irp);
    IoSkipCurrentIrpStackLocation(irp);
    return PoCallDriver(filteredDevice, irp);
  }
  else {
    // If we're not filtering, just complete the IRP successfully.
    
    irp->IoStatus.Status = STATUS_SUCCESS;
    irp->IoStatus.Information = 0;
    IoCompleteRequest(irp, IO_NO_INCREMENT);
    return STATUS_SUCCESS;
  }
}

/////////////////////////////////////////////////////////////////////////////
// UNLOADING
/////////////////////////////////////////////////////////////////////////////

// This routine releases any resources obtained by the driver since it
// started.

void ReleaseResources ()
{
  UNICODE_STRING deviceLinkUnicodeString;

  DebugPrint(("RSIoCnt: Releasing resources.\n"));

  if (g.resourcesObtained & SYMBOLIC_LINK_RESOURCE_FLAG) {
    RtlInitUnicodeString(&deviceLinkUnicodeString, L"\\DosDevices\\RSIoCnt");
    IoDeleteSymbolicLink(&deviceLinkUnicodeString);
    g.resourcesObtained &= ~SYMBOLIC_LINK_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & MAIN_DEVICE_RESOURCE_FLAG) {
    IoDeleteDevice(g.mainDevice);
    g.resourcesObtained &= ~MAIN_DEVICE_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & REGISTER_FS_CHANGE_RESOURCE_FLAG) {
    IoUnregisterFsRegistrationChange(g.ourDriver,
                                     NotifyFsRegistrationChange);
    g.resourcesObtained &= ~REGISTER_FS_CHANGE_RESOURCE_FLAG;
  }
  if (g.resourcesObtained & RSINIT_FILE_OBJECT_RESOURCE_FLAG) {
    ObDereferenceObject(g.RSInitFileObject);
    g.resourcesObtained &= ~RSINIT_FILE_OBJECT_RESOURCE_FLAG;
  }
}

// This is the unload routine for our driver.

void Unload (PDRIVER_OBJECT ourDriver)
{
  DebugPrint(("RSIoCnt: Preparing for unload.\n"));

  // Detach from all devices we're filtering, then wait a little bit
  // for all outstanding I/O's to complete.

  DetachAllDevices();
  DebugPrint(("RSIoCnt: Waiting a short while for outstanding IRP's to "
              "complete.\n"));
  Sleep(MILLISECONDS_TO_WAIT_AFTER_DETACHING_ALL_DEVICES);

  // Release any resources obtained by the driver.

  ReleaseResources();

  DebugPrint(("RSIoCnt: Finished unloading.\n"));
}
