// This file contains routines for performing floating-point operations
// that are otherwise unsupported by the DDK.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "MyFloat.h"

/////////////////////////////////////////////////////////////////////////////
// MACROS
/////////////////////////////////////////////////////////////////////////////

// This is the bit mask describing the set of bits in the floating-point
// state that specify the rounding mode:

#define ROUNDING_MODE_BIT_MASK          0xc

// This is the inverse of the above bit mask:

#define ROUNDING_MODE_INVERSE_BIT_MASK  0xf3

/////////////////////////////////////////////////////////////////////////////
// GLOBAL VARIABLES
/////////////////////////////////////////////////////////////////////////////

// Defining _fltused allows floating-point operations to occur:

int _fltused;

/////////////////////////////////////////////////////////////////////////////
// ROUNDING MODE
/////////////////////////////////////////////////////////////////////////////

// This function returns the active rounding mode.

char GetDefaultRoundingMode (void)
{
  short controlWord;

  __asm {
    fstcw   controlWord    ; get the control word
  }

  return (controlWord >> 8) & ROUNDING_MODE_BIT_MASK;
}

// This function sets the active rounding mode to the given one.

void SetDefaultRoundingMode (char mode)
{
  short newControlWord;

  __asm {
    fstsw   ax                   ; get the control word in ax
    and     ah, ROUNDING_MODE_INVERSE_BIT_MASK ; remove rounding mode bits
    or      ah, mode             ; add desired rounding mode bits
    mov     newControlWord, ax   ; move new control word to memory
    fldcw   newControlWord       ; set the new control word from memory
  }
}

/////////////////////////////////////////////////////////////////////////////
// ROUNDING
/////////////////////////////////////////////////////////////////////////////

// This function rounds the given floating-point number toward zero.

long RoundTowardZero (double x)
{
  long n;
  short savedControlWord, desiredControlWord;

  __asm {
     fld     x                        ; push x on the floating-point stack
     fstcw   savedControlWord         ; save control word
     mov     ax, savedControlWord     ; move control word to ax
     or      ah, ROUNDING_MODE_BIT_MASK  ; set bits 10 and 11 to 0x11
     mov     desiredControlWord, ax   ; move new control word to memory
     fldcw   desiredControlWord       ; set control word from memory
     fistp   dword ptr n              ; pop x as a long into n
     fldcw   savedControlWord         ; restore save control word
  }

  return n;
}

// This function rounds the given floating-point number to the nearest
// integer.

long RoundToNearest (double x)
{
  long n;
  short savedControlWord, desiredControlWord;

  __asm {
     fld     x                        ; push x on the floating-point stack
     fstcw   savedControlWord         ; save control word
     mov     ax, savedControlWord     ; move control word to ax
     and     ah, ROUNDING_MODE_INVERSE_BIT_MASK  ; set bits 10 and 11 to 0x00
     mov     desiredControlWord, ax   ; move new control word to memory
     fldcw   desiredControlWord       ; set control word from memory
     fistp   dword ptr n              ; pop x as a long into n
     fldcw   savedControlWord         ; restore saved control word
  }

  return n;
}

// This function rounds the given floating-point number down.

long RoundDown (double x)
{
  long n;
  short savedControlWord, desiredControlWord;

  __asm {
     fld     x                        ; push x on the floating-point stack
     fstcw   savedControlWord         ; save control word
     mov     ax, savedControlWord     ; move control word to ax
     and     ah, ROUNDING_MODE_INVERSE_BIT_MASK  ; remove rounding mode bits
     or      ah, ROUNDING_MODE_DOWN   ; add desired rounding mode bits
     mov     desiredControlWord, ax   ; move new control word to memory
     fldcw   desiredControlWord       ; set control word from memory
     fistp   dword ptr n              ; pop x as a long into n
     fldcw   savedControlWord         ; restore saved control word
  }

  return n;
}

// This function rounds the given floating-point number up.

long RoundUp (double x)
{
  long n;
  short savedControlWord, desiredControlWord;

  __asm {
     fld     x                        ; push x on the floating-point stack
     fstcw   savedControlWord         ; save control word
     mov     ax, savedControlWord     ; move control word to ax
     and     ah, ROUNDING_MODE_INVERSE_BIT_MASK  ; remove rounding mode bits
     or      ah, ROUNDING_MODE_UP     ; add desired rounding mode bits
     mov     desiredControlWord, ax   ; move new control word to memory
     fldcw   desiredControlWord       ; set control word from memory
     fistp   dword ptr n              ; pop x as a long into n
     fldcw   savedControlWord         ; restore old control word
  }

  return n;
}

// This routine rounds in an unspecified manner, depending on the
// current state of the FPU control word.

long RoundDefault (double x)
{
  long n;

  __asm {
     fld     x                 ; push x on the floating-point stack
     fistp   dword ptr n       ; pop x as a long into n
  }

  return n;
}

// This function computes and returns the square root of x.

double SquareRoot (double x)
{
  __asm {
     fld     x                 ; push x on the floating-point stack
     fsqrt                     ; perform square-root operation
     fstp    x                 ; pop square root and put in x
  }

  return x;
}

// This function computes two to the power of the given number.

double TwoToThePowerOf (double x)
{
  long n;
  double excess, result;

  // Round x to an integer. It doesn't matter what the rounding mode
  // is, since all we care about is finding an integer no more than
  // 1.0 away in either direction, and any of the rounding modes will
  // do that. Then, compute the excess as x - n.

  n = RoundDefault(x);
  excess = x - n;

  // Now, x = excess + n, and excess is in the range -1.0 to 1.0.
  // This makes excess a viable candidate for the x86 instruction
  // f2xm1, allowing us to compute result = 2^(excess) - 1.

  __asm {
    fld      excess            ; push the excess on the floating-point stack
    f2xm1                      ; compute 2^(excess) - 1
    fld1                       ; push 1 on the stack
    fadd                       ; add 1 to get 2^(excess)
    fscale                     ; scale the result by 2^n
    fstp     result            ; store this in result
  }

  return result;
}

// This function computes and returns 'base' to the 'exponent' power.

double Exponentiate (double base, double exponent)
{
  __asm {
    fld      exponent          ; push exponent on the floating-point stack
    fld      base              ; push base on the floating-point stack
    fyl2x                      ; compute exponent * log_2(base)
    fstp     base              ; pop the result into base
  }

  return TwoToThePowerOf(base);
}

void DoFloatingPointTests (void)
{
  double z;
  long n;

  z = Exponentiate(3.0, 4.0);
  n = RoundDefault(z * 1000);
  DebugPrint(("3.0^4.0 = %lu/1000\n", n));

  z = Exponentiate(3.0, 4.2);
  n = RoundDefault(z * 1000);
  DebugPrint(("3.0^4.2 = %lu/1000\n", n));

  z = Exponentiate(0.3, 2.2);
  n = RoundDefault(z * 1000);
  DebugPrint(("0.3^2.2 = %lu/1000\n", n));

  z = Exponentiate(1.6, 8.7);
  n = RoundDefault(z * 1000);
  DebugPrint(("1.6^8.7 = %lu/1000\n", n));
}
