// This file contains routines for reading configuration information
// from the registry.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/DrvShare.h"
#include "../Common/ErrCodes.h"
#include "Globals.h"
#include "Log.h"
#include "RegInfo.h"
#include "Unicode.h"

// This function reads the configuration information for RightSpeed from
// the registry and puts it in the structure passed.

NTSTATUS ReadRegistryInfo (REGISTRY_INFO *info)
{
  NTSTATUS status;
  OBJECT_ATTRIBUTES objectAttributes;
  struct {
    ULONG   TitleIndex;
    ULONG   Type;
    ULONG   DataLength;
    WCHAR   Data[MAX_RIGHTSPEED_DIRECTORY_LENGTH];
  } keyValueInfo;
  HANDLE keyHandle;
  ULONG keyValueInfoLength;
  UNICODE_STRING miscUnicodeString;

  // Open the registry key for RightSpeed.

  RtlInitUnicodeString(&miscUnicodeString,
    L"\\Registry\\Machine\\Software\\Jay Lorch\\RightSpeed\\CurrentVersion");
  InitializeObjectAttributes(&objectAttributes,
                             &miscUnicodeString,
                             OBJ_CASE_INSENSITIVE,
                             NULL,  // root directory
                             NULL); // security descriptor
  status = ZwOpenKey(&keyHandle,
                     KEY_QUERY_VALUE | KEY_SET_VALUE,
                     &objectAttributes);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: ERROR - Could not open main registry key (%x).\n",
                status));
    BothLogShort(RS_LOG_ERROR_CANT_OPEN_MAIN_REGISTRY_KEY,
                 status);
    return status;
  }

  // Read the version.

  RtlInitUnicodeString(&miscUnicodeString, L"Version");
  status = ZwQueryValueKey(keyHandle,
                           &miscUnicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (status == STATUS_SUCCESS && keyValueInfo.Type == REG_DWORD)
    info->version = * (ULONG *) keyValueInfo.Data;
  else
    info->version = 0;

  // Read the timestamp count ticks per millisecond:

  RtlInitUnicodeString(&miscUnicodeString, L"TscTicksPerMs");
  status = ZwQueryValueKey(keyHandle,
                           &miscUnicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (status == STATUS_SUCCESS && keyValueInfo.Type == REG_DWORD)
    info->tscTicksPerMs = * (ULONG *) keyValueInfo.Data;
  else
    info->tscTicksPerMs = 0;

  // Read the size of the log buffer, in KB; multiply by 1024 to
  // convert from KB to bytes.

  RtlInitUnicodeString(&miscUnicodeString, L"LogBufferKB");
  status = ZwQueryValueKey(keyHandle, &miscUnicodeString,
                           KeyValuePartialInformation, &keyValueInfo,
                           sizeof(keyValueInfo), &keyValueInfoLength);
  if (status == STATUS_SUCCESS && keyValueInfo.Type == REG_DWORD) {
    info->logBufferSize = * (ULONG *) keyValueInfo.Data;
    if (info->logBufferSize < MIN_LOG_BUFFER_SIZE ||
        info->logBufferSize > MAX_LOG_BUFFER_SIZE)
      info->logBufferSize = DEFAULT_LOG_BUFFER_SIZE;
  }
  else
    info->logBufferSize = DEFAULT_LOG_BUFFER_SIZE;
  info->logBufferSize *= 1024;

  // Read the logged event category set.

  RtlInitUnicodeString(&miscUnicodeString, L"LoggedEventCategorySet");
  status = ZwQueryValueKey(keyHandle,
                           &miscUnicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (NT_SUCCESS(status) && keyValueInfo.Type == REG_DWORD)
    info->loggedEventCategorySet = * (ULONG *) keyValueInfo.Data;
  else
    info->loggedEventCategorySet = DEFAULT_RS_EVENT_CATEGORY_SET;

  // Read the main directory of RightSpeed.

  RtlInitUnicodeString(&miscUnicodeString, L"MainDirectory");
  status = ZwQueryValueKey(keyHandle,
                           &miscUnicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: ERROR - Could not read main directory registry key "
                " (%x).\n", status));
    BothLogShort(RS_LOG_ERROR_CANT_READ_MAIN_DIRECTORY,
                 status);
    ZwClose(keyHandle);
    return status;
  }
  if (keyValueInfo.Type != REG_SZ) {
    DebugPrint(("RSLog: ERROR - Main directory registry entry is %lu, "
                "not REG_SZ\n", keyValueInfo.Type));
    BothLogShort(RS_LOG_ERROR_MAIN_DIRECTORY_IS_NOT_STRING,
                 RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA);
    ZwClose(keyHandle);
    return RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA;
  }

  // Allocate the main directory Unicode string.

  status = MyAllocateUnicodeString(&info->mainDirectory,
                                   (USHORT) keyValueInfo.DataLength);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: ERROR - Could not initialize log file name string "
                "(%x).\n", status));
    BothLogShort(RS_LOG_ERROR_CANT_INITIALIZE_MAIN_DIRECTORY_STRING,
                 status);
    ZwClose(keyHandle);
    return status;
  }

  // Copy the main directory string to the main directory Unicode string.

  status = RtlAppendUnicodeToString(&info->mainDirectory, keyValueInfo.Data);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: ERROR - Could not copy main directory bytes to "
                "main directory Unicode string (%x).\n", status));
    BothLogShortRepeat(RS_LOG_ERROR_CANT_INITIALIZE_MAIN_DIRECTORY_STRING,
                       status,
                       1);
    ZwClose(keyHandle);
    return status;
  }

  // We no longer need the registry key, so close the handle.

  ZwClose(keyHandle);

  // Return success.

  return STATUS_SUCCESS;
}

// This function frees resources associated with the given registry
// information.

void FreeRegistryInfo (REGISTRY_INFO *info)
{
  MyFreeUnicodeString(&info->mainDirectory);
}

// This function reloads the registry information for configuring
// RSLog. It updates the set of logged event categories accordingly.

NTSTATUS ReloadRegistryInfo (REGISTRY_INFO *info)
{
  NTSTATUS status;
  OBJECT_ATTRIBUTES objectAttributes;
  struct {
    ULONG   TitleIndex;
    ULONG   Type;
    ULONG   DataLength;
    ULONG   DataDword;
  } keyValueInfo;
  HANDLE keyHandle;
  ULONG keyValueInfoLength;
  UNICODE_STRING miscUnicodeString;

  // Open the registry key for RightSpeed.

  RtlInitUnicodeString(&miscUnicodeString,
    L"\\Registry\\Machine\\Software\\Jay Lorch\\RightSpeed\\CurrentVersion");
  InitializeObjectAttributes(&objectAttributes,
                             &miscUnicodeString,
                             OBJ_CASE_INSENSITIVE,
                             NULL,  // root directory
                             NULL); // security descriptor
  status = ZwOpenKey(&keyHandle,
                     KEY_QUERY_VALUE | KEY_SET_VALUE,
                     &objectAttributes);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSLog: ERROR - Could not open main registry key (%x).\n",
                status));
    BothLogShortRepeat(RS_LOG_ERROR_CANT_OPEN_MAIN_REGISTRY_KEY,
                       status,
                       1);
    return status;
  }

  // Read the version.

  RtlInitUnicodeString(&miscUnicodeString, L"Version");
  status = ZwQueryValueKey(keyHandle,
                           &miscUnicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (status == STATUS_SUCCESS && keyValueInfo.Type == REG_DWORD)
    info->version = keyValueInfo.DataDword;
  else
    info->version = 0;

  // Read the timestamp count ticks per millisecond:

  RtlInitUnicodeString(&miscUnicodeString, L"TscTicksPerMs");
  status = ZwQueryValueKey(keyHandle,
                           &miscUnicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (status == STATUS_SUCCESS && keyValueInfo.Type == REG_DWORD)
    info->tscTicksPerMs = keyValueInfo.DataDword;
  else
    info->tscTicksPerMs = 0;

  // Read the logged event category set.

  RtlInitUnicodeString(&miscUnicodeString, L"LoggedEventCategorySet");
  status = ZwQueryValueKey(keyHandle,
                           &miscUnicodeString,
                           KeyValuePartialInformation,
                           &keyValueInfo,
                           sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (NT_SUCCESS(status) && keyValueInfo.Type == REG_DWORD)
    info->loggedEventCategorySet = keyValueInfo.DataDword;
  else
    info->loggedEventCategorySet = DEFAULT_RS_EVENT_CATEGORY_SET;

  // If logging is enabled, indicate the current logged event category
  // set in the shared state.

  if (!g.whyLoggingDisabled)
    g.sharedState->kernelUserSharedState.loggedEventCategorySet =
      info->loggedEventCategorySet;

  // We no longer need the registry key, so close the handle.

  ZwClose(keyHandle);

  // Return success.

  return STATUS_SUCCESS;
}
