#include "stdafx.h"
#include <windows.h>
#include <iomanip>
#include <sstream>
#include "../Common/Logging.h"
#include "../Common/Version.h"
#include "Exceptions.h"
#include "FileTime.h"
#include "Times.h"
#include "LogFile.h"
#include "TypeConv.h"

/////////////////////////////////////////////////////////////////////////////
//                            Unnamed namespace
/////////////////////////////////////////////////////////////////////////////

namespace {

  // The number of 100-ns units per second:

  const static long HNS_PER_SEC = 10000000;

  // The maximum version of RSLog supported by this program:

  static const unsigned long MIN_VERSION_SUPPORTED = 0x000103;
  static const unsigned long MAX_VERSION_SUPPORTED = 0x000103;

  // PaddedHex is a class representing an unsigned long that, when printed
  // to an output stream, is written as an 8-character hex value padded on
  // the left by zeroes.

  class PaddedHex {
  public:
    PaddedHex(unsigned long nVal) : n(nVal) { }
    ~PaddedHex() { }
    unsigned long n;
  };

  ostream& operator<< (ostream& os, PaddedHex h)
  {
    os.fill('0');
    os.width(8);
    return os << hex << h.n << dec;
  }

  // Hex<T> is a class representing a T that, when printed to an output
  // stream, is written in hex.

  template <class T> class Hex {
  public:
    Hex(T nVal) : n(nVal) { }
    ~Hex() { }
    T n;
  };

  template <class T> ostream& operator<< (ostream& os, Hex<T> h)
  {
    return os << hex << h.n << dec;
  }

  // PrintNTDeviceType outputs to the given stream 'os' a description of
  // the given NT device type.
  
  void PrintNTDeviceType (ostream& os, unsigned long deviceType)
  {
    switch (deviceType) {
    case 3 : os << "CD file system"; break;
    case 8 : os << "Disk file system"; break;
    case 9 : os << "Generic file system"; break;
    case 0x12 : os << "Network"; break;
    case 0x14 : os << "Network file system"; break;
    case 0x20 : os << "Tape file system"; break;
    case 0x21 : os << "Transport"; break;
    default : os << "Unknown device type"; break;
    }
  }

}

//////////////////////////////////////////////////////////////////////////////
//                           LogFile members
//////////////////////////////////////////////////////////////////////////////

// LogFile::LogFile initializes the members of the object.

LogFile::LogFile(const Parameters& paramsVal) :
  params(paramsVal),
  version(0),
  tscTicksPerSec(631146000)
{
}

// LogFile::PrintPreamble reads and prints the log preamble.

void LogFile::PrintPreamble (ostream& os)
{
  try {
    LOG_FILE_PREAMBLE preamble;
    ReadBytes((char *) &preamble, sizeof(preamble));
    version = preamble.version;
    os << "Version:    "
       << version / 0x10000 << "."
       << (version % 0x10000) / 0x100 << "."
       << (version % 0x100) << endl;
    if (version > MAX_VERSION_SUPPORTED || version < MIN_VERSION_SUPPORTED)
      throw Exceptions::VersionNotSupported();

    os << "CPU type:   ";
    switch (preamble.cpuType) {
    case RIGHTSPEED_CPU_TYPE_TMTA :
      os << "Transmeta" << endl;
      break;
    case RIGHTSPEED_CPU_TYPE_AMD :
      os << "AMD" << endl;
      break;
    default :
      os << "unknown" << endl;
      break;
    }

    tscTicksPerSec = preamble.maxCycPerMs * 1000;
    os << "TSC ticks per second:  " << tscTicksPerSec << endl;

    Windows::FileTime startLocalFileTime(preamble.startLocalTime.QuadPart, 0,
                                         false);
    Windows::FileTime startGMTFileTime(preamble.startGMTTime.QuadPart, 0,
                                       false);
    os << "Started:    " << startLocalFileTime << " (local)" << endl;
    os << "            " << startGMTFileTime << " (GMT)" << endl;

    jrl::Nanosec nsSince1970 = params.useGMT ? startGMTFileTime.GetTime() :
                               startLocalFileTime.GetTime();
    timeOffset = jrl::Time((unsigned long) (nsSince1970 / 1000000000),
                           nsSince1970 % 1000000000) -
                 jrl::Time(preamble.startTSC.QuadPart, tscTicksPerSec);
  }
  catch (jrl::GenIoExceptions::Eof) {
    throw Exceptions::InvalidLogFileFormat();
  }
}

// LogFile::Print prints a description of the log file to the given
// output stream.

void LogFile::Print (ostream& os)
{
  os << setprecision(10);
  PrintPreamble(os);
  os << endl;

  try {
    while (1) {
      PrintNextEntry(os);
    }
  }
  catch (jrl::GenIoExceptions::Eof) {
  }
}

// LogFile::PrintNextEntry reads and prints the next entry in the log.

void LogFile::PrintNextEntry (ostream& os)
{
  ULONGLONG cycles = ReadUlonglong();
  int type = ReadShort();
  int length = ReadUchar();
  ReadBytes(data, length);
  int i;

  if (params.skipContextSwitches && type == EVENT_TYPE_CONTEXT_SWITCH)
    return;

  os << jrl::Time(cycles, tscTicksPerSec, timeOffset) << "  ";

  switch (type) {
  case EVENT_TYPE_STRING :
    data[length] = '\0';
    os << data;
    break;

  case EVENT_TYPE_SPEED_CHANGED :
    os << "Speed changed to " << jrl::ConvertToLong(data);
    break;

  case EVENT_TYPE_BUFFER_OVERFLOWED :
    os << "Buffer overflowed";
    break;

  case EVENT_TYPE_CONTEXT_SWITCH :
    os << "Context switch to thread " << jrl::ConvertToUlong(data);
    break;

  case EVENT_TYPE_PROCESS_CREATE :
    data[length] = '\0';
    os << "Created process " << jrl::ConvertToUlong(data) << ", parent "
       << jrl::ConvertToUlong(&data[4]) << ", name " << &data[8];
    break;

  case EVENT_TYPE_PROCESS_DESTROY :
    os << "Destroyed process " << jrl::ConvertToUlong(data);
    break;

  case EVENT_TYPE_THREAD_CREATE :
    os << "Created thread " << jrl::ConvertToUlong(data) << " of process "
       << jrl::ConvertToUlong(&data[4]);
    break;

  case EVENT_TYPE_THREAD_DESTROY :
    os << "Created thread " << jrl::ConvertToUlong(data) << " of process "
       << jrl::ConvertToUlong(&data[4]);
    break;

  case EVENT_TYPE_BEGIN_TASK :
    os << "Began task " << jrl::ConvertToUlong(data);
    break;

  case EVENT_TYPE_END_TASK :
    os << "Ended task " << jrl::ConvertToUlong(data);
    break;

  case EVENT_TYPE_SET_SCHEDULE :
    os << "Set schedule: ";
    for (i = 0; i < length / 4; i += 2) {
      os << "speed " << jrl::ConvertToUlong(&data[i*4]);
      os << " for " << jrl::ConvertToLong(&data[i*4+4]) << " intervals";
      if (i + 2 < (length / 4))
        os << "; ";
    }
    break;

  case EVENT_TYPE_LONG_AND_LONG :
    os << "Long and long: " << jrl::ConvertToLong(data) << ", "
       << jrl::ConvertToLong(&data[4]);
    break;

  case EVENT_TYPE_LONG_AND_DOUBLE :
    os << "Long and double: " << jrl::ConvertToLong(data) << ", "
       << * (double *) &data[4];
    break;

  case EVENT_TYPE_END_ALL_TASKS :
    os << "Ended all tasks";
    break;

  case EVENT_TYPE_IO_WAIT_STARTS :
    os << "Started waiting for I/O";
    break;

  case EVENT_TYPE_IO_WAIT_STOPS :
    os << "Stopped waiting for I/O";
    break;

  case EVENT_TYPE_SAMPLE_VALUE :
    os << "Found sample value " << jrl::ConvertToLong(&data[4])
       << ", " << jrl::ConvertToUlong(data);
    break;

  case EVENT_TYPE_ERROR :
    {
      unsigned long errorCode = jrl::ConvertToUlong(data);
      int severity = (errorCode & 0xC0000000) >> 30;
      int component = (errorCode & 0x0000F000) >> 12;
      int code = (errorCode & 0x00000FFF);
      int dataElements = (length - 4) / 4;

      switch (component) {
      case 0 : os << "RSTask: "; break;
      case 1 : os << "RSIoCnt: "; break;
      case 2 : os << "RSInit: "; break;
      case 3 : os << "RSLog: "; break;
      case 4 : os << "RSLib: "; break;
      default : os << "RightSpeed: "; break;
      }
      switch (severity) {
      case 0 : os << "Success code "; break;
      case 1 : os << "Information code "; break;
      case 2 : os << "WARNING code "; break;
      case 3 : os << "ERROR code "; break;
      }
      os << hex << code;
      if (dataElements) {
        os << " (";
        for (i = 0; i < dataElements; ++i) {
          if (i > 0)
            os << ", ";
          os << jrl::ConvertToUlong(&data[i*4 + 4]);
        }
        os << ")";
      }
      os << dec;
    }
    break;

  case EVENT_TYPE_AUTO_SYSTEM_TASK :
    os << "Auto system task by thread " << jrl::ConvertToUlong(data);
    break;

  case EVENT_TYPE_AUTO_APP_TASK :
    os << "Auto app task " << jrl::ConvertToUlong(data) << " by thread "
       << jrl::ConvertToUlong(&data[4]);
    break;

  case EVENT_TYPE_PROCESS_HOOKED :
    data[length] = '\0';
    os << "Hooked process " << jrl::ConvertToUlong(data) << ": "
       << &data[4];
    break;

  case EVENT_TYPE_THREAD_HOOKED :
    os << "Hooked thread " << jrl::ConvertToUlong(data) << " in process "
       << jrl::ConvertToUlong(&data[4]);
    break;

  case EVENT_TYPE_THREAD_UNHOOKED :
    os << "Unhooked thread " << jrl::ConvertToUlong(data);
    break;

  case EVENT_TYPE_TIMED_ACTION_STARTED :
    os << "Timed action " << jrl::ConvertToUlong(data) << " started";
    break;

  case EVENT_TYPE_TIMED_ACTION_STOPPED :
    os << "Timed action " << jrl::ConvertToUlong(data) << " stopped";
    break;

  default :
    os << "Unknown entry type " << type;
    break;
  }

  os << endl;
}
