// This file contains functions implementing the classes in genio.h.

#include "stdafx.h"
#include <algorithm>
#include <iostream>
#include "array.h"
#include "genio.h"

using jrl::WindowsHandle;
using jrl::InputFile;
using jrl::InputCompressedFile;
using jrl::OutputFile;
using jrl::OutputCompressedFile;
using jrl::GeneralInput;
using jrl::GeneralOutput;
using jrl::RawInput;
using jrl::RawBufferedInput;
using jrl::RawOutput;
using jrl::RawBufferedOutput;
using jrl::RawBufferedOutputFile;
using jrl::RawBufferedOutputCompressedFile;
using jrl::RawBufferedGeneralOutput;

//////////////////////////////////////////////////////////////////////////////
// The local namespace for this file
//////////////////////////////////////////////////////////////////////////////

namespace {

  // This class represents a handle that should be closed when this object
  // goes out of scope.

  class TemporaryHandle {
  public:
    TemporaryHandle() : h(INVALID_HANDLE_VALUE) { }
    TemporaryHandle(HANDLE hVal) : h(hVal) { }
    ~TemporaryHandle() { Close(); }

    TemporaryHandle& operator= (HANDLE hVal) { Close(); h = hVal; }
    operator HANDLE () { return h; }
    HANDLE * operator & () { return &h; }

    // This function returns the handle, and sets this temporary handle
    // to invalid so that the handle returned doesn't get closed when this
    // object goes out of scope.

    HANDLE MakePermanent ()
    {
      HANDLE retVal = h; 
      h = INVALID_HANDLE_VALUE;
      return retVal;
    }

    // This function closes this handle.

    BOOL Close ()
    {
      if (h != INVALID_HANDLE_VALUE) {
        BOOL retVal = CloseHandle(h);
        h = INVALID_HANDLE_VALUE;
        return retVal;
      }
      else
        return TRUE;
    }

  private:
    HANDLE h;     // the handle
  };

  // This class represents a change to one of the standard handles (stdin,
  // stdout, etc.) that should not last past the time this object goes out
  // of scope.

  class TemporaryStdHandleChange {
  public:
    TemporaryStdHandleChange() : changed(false) { }
    ~TemporaryStdHandleChange() { Restore(); }

    // This function makes a change to a standard handle.

    void Change (DWORD whichStdVal, HANDLE newHandle)
    {
      if (changed) {
	cerr << "Error: Can't change standard handle twice" << endl;
	throw jrl::GenIoExceptions::Open();
      }

      hSaveHandle = GetStdHandle(whichStd);
      if (SetStdHandle(whichStdVal, newHandle)) {
        changed = true;
        whichStd = whichStdVal;
      }
      else
        throw jrl::GenIoExceptions::Open();
    }

    // This function restores any change this object performed on a
    // standard handle.

    void Restore ()
    {
      if (changed) {
        SetStdHandle(whichStd, hSaveHandle);
        changed = false;
      }
    }

  private:
    bool changed;        // whether a change to a standard handle occurred
    DWORD whichStd;      // which standard handle was changed
    HANDLE hSaveHandle;  // what the standard handle used to be before change
  };

  // This function opens the file with name 'fileName' to be read,
  // using the given security attributes 'sa'. It returns an open file
  // handle to be used for reading from the file.

  HANDLE OpenFileForReading (string fileName, SECURITY_ATTRIBUTES *sa)
  {
    HANDLE fileHandle = CreateFile(fileName.c_str(), GENERIC_READ,
                                   FILE_SHARE_READ | FILE_SHARE_WRITE, sa,
                                   OPEN_EXISTING,
                                   FILE_ATTRIBUTE_NORMAL, NULL);
    if (fileHandle == INVALID_HANDLE_VALUE)
      throw jrl::GenIoExceptions::Open();
    else
      return fileHandle;
  }

  // This function opens the file with name 'fileName' to be written
  // using the given security attributes 'sa'. The parameter
  // 'deleteIfExists' indicates whether a file of that name that
  // already exists should be deleted. The parameter 'append'
  // indicates whether a file of that name that already exists should
  // be appended to. (If they're both true, 'append' wins.) It returns
  // an open file handle to be used for writing to the file.

  HANDLE OpenFileForWriting (string fileName, SECURITY_ATTRIBUTES *sa,
                             bool deleteIfExists, bool append)
  {
    // Set up flags for the open depending on the parameters.

    DWORD creationDisposition;
    if (append)
      creationDisposition = OPEN_ALWAYS;
    else if (deleteIfExists)
      creationDisposition = CREATE_ALWAYS;
    else
      creationDisposition = CREATE_NEW;

    HANDLE fileHandle =
      CreateFile(fileName.c_str(), GENERIC_WRITE, 0, sa,
                 creationDisposition, FILE_ATTRIBUTE_NORMAL, NULL);
    if (fileHandle == INVALID_HANDLE_VALUE)
      throw jrl::GenIoExceptions::Open();

    if (append) {
      if (SetFilePointer(fileHandle, 0, NULL, FILE_END) == 0xFFFFFFFF) {
	CloseHandle(fileHandle);
        throw jrl::GenIoExceptions::Open();
      }
    }

    return fileHandle;
  }

  // This function finds the full path name of gzip.exe. It searches
  // the normal path for executables.

  string GetGzipPath ()
  {
    char fullPath[MAX_PATH];
    if (SearchPath(NULL, "gzip.exe", NULL, MAX_PATH, fullPath, NULL) == 0) {
      cerr << "Error: Cannot find gzip.exe in path." << endl;
      throw jrl::GenIoExceptions::Open();
    }
    return fullPath;
  }

  // This function launches a gunzip process to uncompress the file
  // with name 'fileName'. It put the handle of that gunzip process in
  // 'childHandle', and returns a handle for reading that process's
  // output.

  HANDLE ReadFromGunzip (string fileName, HANDLE& childHandle)
  {
    // Create some security attributes for use in sharing handles with
    // a created process.

    SECURITY_ATTRIBUTES saAttr;
    saAttr.nLength = sizeof(SECURITY_ATTRIBUTES); 
    saAttr.bInheritHandle = TRUE;
    saAttr.lpSecurityDescriptor = NULL;

    // Create a pipe (two handles, hPipeRead and hPipeWrite) that can be
    // inherited by child processes

    TemporaryHandle hPipeRead, hPipeWrite;
    if (!CreatePipe(&hPipeRead, &hPipeWrite, &saAttr, 0))
      throw jrl::GenIoExceptions::Open();

    // Open the file for reading, so that it's sharable by the child process.

    TemporaryHandle fileHandle = OpenFileForReading(fileName, &saAttr);

    // Save a duplicate of hPipeRead that won't be inherited by the child
    // process in hPipeReadDuplicate.

    TemporaryHandle hPipeReadDuplicate;
    if (!DuplicateHandle(GetCurrentProcess(), hPipeRead, GetCurrentProcess(),
                         &hPipeReadDuplicate, 0, FALSE /* don't inherit */,
                         DUPLICATE_SAME_ACCESS))
      throw jrl::GenIoExceptions::Open();

    // Close hPipeRead so that the child process does not see it.

    if (!hPipeRead.Close())
      throw jrl::GenIoExceptions::Open();

    // Temporarily set stdin to be the file handle, and stdout to be
    // the write end of the pipe.

    TemporaryStdHandleChange inChange, outChange;
    inChange.Change(STD_INPUT_HANDLE, fileHandle);
    outChange.Change(STD_OUTPUT_HANDLE, hPipeWrite);

    // Create the child process to perform the unzipping.

    STARTUPINFO startupInfo;
    GetStartupInfo(&startupInfo);
    PROCESS_INFORMATION processInfo;
    if (!CreateProcess(GetGzipPath().c_str(), "GZIP -d -c", NULL, NULL, TRUE,
                       0, NULL, NULL, &startupInfo, &processInfo))
      throw jrl::GenIoExceptions::Open();

    childHandle = processInfo.hProcess;
    CloseHandle(processInfo.hThread);
    return hPipeReadDuplicate.MakePermanent();
  }

  // This function launches a gzip process to write to the given file.
  // The 'deleteIfExists' parameter indicates whether we should
  // overwrite the file if it exists. This function puts the handle of
  // the gzip child process in 'childHandle' and returns an open
  // handle to use in writing data to the gzip process.

  HANDLE WriteToGzip (string fileName, bool deleteIfExists,
                      HANDLE& childHandle)
  {
    // Create some security attributes for use in sharing handles with
    // a created process.

    SECURITY_ATTRIBUTES saAttr;
    saAttr.nLength = sizeof(SECURITY_ATTRIBUTES); 
    saAttr.bInheritHandle = TRUE;
    saAttr.lpSecurityDescriptor = NULL;

    // Create a pipe (two handles, hPipeRead and hPipeWrite) that can be
    // inherited by child processes

    TemporaryHandle hPipeRead, hPipeWrite;
    if (!CreatePipe(&hPipeRead, &hPipeWrite, &saAttr, 0))
      throw jrl::GenIoExceptions::Open();

    // Open the file for writing, so that it's sharable by the child process.

    TemporaryHandle fileHandle = OpenFileForWriting(fileName, &saAttr,
						    deleteIfExists, false);

    // Save a duplicate of hPipeWrite that won't be inherited by the child
    // process in hPipeWriteDuplicate.

    TemporaryHandle hPipeWriteDuplicate;
    if (!DuplicateHandle(GetCurrentProcess(), hPipeWrite, GetCurrentProcess(),
                         &hPipeWriteDuplicate, 0, FALSE /* don't inherit */,
                         DUPLICATE_SAME_ACCESS))
      throw jrl::GenIoExceptions::Open();

    // Close hPipeWrite so that the child process does not see it.

    if (!hPipeWrite.Close())
      throw jrl::GenIoExceptions::Open();

    // Temporarily set the stdin handle to be the read end of the pipe,
    // and set the stdout handle to be the file handle.

    TemporaryStdHandleChange inChange, outChange;
    inChange.Change(STD_INPUT_HANDLE, hPipeRead);
    outChange.Change(STD_OUTPUT_HANDLE, fileHandle);

    // Create the child process to perform the zipping.

    STARTUPINFO startupInfo;
    GetStartupInfo(&startupInfo);
    PROCESS_INFORMATION processInfo;
    if (!CreateProcess(GetGzipPath().c_str(), "GZIP -c", NULL, NULL, TRUE, 0,
                       NULL, NULL, &startupInfo, &processInfo))
      throw jrl::GenIoExceptions::Open();

    childHandle = processInfo.hProcess;
    CloseHandle(processInfo.hThread);
    return hPipeWriteDuplicate.MakePermanent();
  }

}

//////////////////////////////////////////////////////////////////////////////
// WindowsHandle functions
//////////////////////////////////////////////////////////////////////////////

// This function performs an unbuffered read on a Unix stream. Its
// semantics conform to those of the abstract function declared in
// ReadableObject.

int WindowsHandle::UnbufferedRead (char *buf, int length)
{
  DWORD bytesRead;

  if (!ReadFile(fileHandle, buf, length, &bytesRead, NULL)) {
    if (GetLastError() == ERROR_BROKEN_PIPE)
      throw jrl::GenIoExceptions::Eof();
    else
      throw jrl::GenIoExceptions::Read();
  }
  if (bytesRead > 0 || bytesRead == length)
    return bytesRead;
  else
    throw jrl::GenIoExceptions::Eof();
}

// This function performs an unbuffered write on a Unix stream. Its
// semantics conform to those of the abstract function declared in
// WritableObject.

int WindowsHandle::UnbufferedWrite (const char *buf, int length)
{
  DWORD bytesWritten;

  if (!WriteFile(fileHandle, buf, length, &bytesWritten, NULL))
    throw jrl::GenIoExceptions::Write();
  if (bytesWritten > 0 || bytesWritten == length)
    return bytesWritten;
  else
    throw jrl::GenIoExceptions::Eof();
}

// This function closes the file descriptor represented by this object.

void WindowsHandle::Close ()
{
  // Mark the file descriptor invalid so we don't try to close it
  // again (or do any operation on it, for that matter).

  if (fileHandle != INVALID_HANDLE_VALUE) {
    CloseHandle(fileHandle);
    fileHandle = INVALID_HANDLE_VALUE;
  }
}

//////////////////////////////////////////////////////////////////////////////
// InputFile functions
//////////////////////////////////////////////////////////////////////////////

// This function opens the file of the given name.

void InputFile::Open (string fileName)
{
  Close();
  fileHandle = OpenFileForReading(fileName, NULL);
}

//////////////////////////////////////////////////////////////////////////////
// InputCompressedFile functions
//////////////////////////////////////////////////////////////////////////////

// This function opens a connection to a gunzip process that will
// uncompress the file of the given name for us.

void InputCompressedFile::Open (string fileName)
{
  Close();
  fileHandle = ReadFromGunzip(fileName, childHandle);
}

// This function closes the connection we have open to a gunzip process.

void InputCompressedFile::Close ()
{
  if (fileHandle != INVALID_HANDLE_VALUE) {
    if (childHandle != NULL) {
      TerminateProcess(childHandle, 0);
      CloseHandle(childHandle);
    }
    WindowsHandle::Close();
  }
}

//////////////////////////////////////////////////////////////////////////////
// GeneralInput functions
//////////////////////////////////////////////////////////////////////////////

// This function opens a handle representing the general input source
// described by 'fileName'. This could be a compressed file, if
// fileName ends in .gz, or just a regular file.

void GeneralInput::Open (string fileName)
{
  Close();

  if (fileName.size() > 3 && fileName.substr(fileName.size() - 3) == ".gz")
    fileHandle = ReadFromGunzip(fileName, childHandle);
  else {
    fileHandle = OpenFileForReading(fileName, NULL);
    childHandle = NULL;
  }
}

// This function closes the handle associated with this general input
// source.

void GeneralInput::Close ()
{
  if (fileHandle != INVALID_HANDLE_VALUE) {
    if (childHandle != NULL) {
      TerminateProcess(childHandle, 0);
      CloseHandle(childHandle);
    }
    WindowsHandle::Close();
  }
}

//////////////////////////////////////////////////////////////////////////////
// OutputFile functions
//////////////////////////////////////////////////////////////////////////////

// This function opens the file of name 'fileName' for writing. The
// parameter 'deleteIfExists' indicates whether we should delete a
// file that already exists by that name. The parameter 'append'
// indicates whether we should append to a file that already exists by
// that name. (If they're both true, 'append' wins.)

void OutputFile::Open (string fileName, bool deleteIfExists,
                       bool append)
{
  Close();
  fileHandle = OpenFileForWriting(fileName, NULL, deleteIfExists, append);
}

//////////////////////////////////////////////////////////////////////////////
// OutputCompressedFile functions
//////////////////////////////////////////////////////////////////////////////

// This function opens the file of name 'fileName' for writing as a
// compressed file. The parameter 'deleteIfExists' indicates whether
// we should delete a file that already exists by that name.

void OutputCompressedFile::Open (string fileName,
                                 bool deleteIfExists)
{
  Close();
  fileHandle = WriteToGzip(fileName, deleteIfExists, childHandle);
}

// This function closes any connection we have to a gzip process that
// is compressing our output to create a compressed file.

void OutputCompressedFile::Close ()
{
  if (fileHandle != INVALID_HANDLE_VALUE) {
    WindowsHandle::Close();
    if (childHandle != NULL) {
      while (WaitForSingleObject(childHandle, INFINITE) != WAIT_OBJECT_0)
        ;
      CloseHandle(childHandle);
    }
  }
}

//////////////////////////////////////////////////////////////////////////////
// GeneralOutput functions
//////////////////////////////////////////////////////////////////////////////

// This function opens a handle representing the general output
// destination described by 'fileName'. This could be a compressed
// file, if fileName ends in .gz, or just a regular file.

void GeneralOutput::Open (string fileName, bool deleteIfExists,
                          bool append)
{
  Close();

  if (fileName.size() > 3 && fileName.substr(fileName.size() - 3) == ".gz") {
    if (append)
      throw jrl::GenIoExceptions::Open();
    fileHandle = WriteToGzip(fileName, deleteIfExists, childHandle);
  }
  else {
    fileHandle = OpenFileForWriting(fileName, NULL, deleteIfExists, append);
    childHandle = NULL;
  }
}

// This function closes the file descriptor associated with this
// general output destination.

void GeneralOutput::Close ()
{
  if (fileHandle != INVALID_HANDLE_VALUE) {
    WindowsHandle::Close();
    if (childHandle != NULL) {
      while (WaitForSingleObject(childHandle, INFINITE) != WAIT_OBJECT_0)
        ;
      CloseHandle(childHandle);
    }
  }
}

//////////////////////////////////////////////////////////////////////////////
// RawInput functions
//////////////////////////////////////////////////////////////////////////////

// This function reads exactly 'length' bytes into the given buffer 'buf'.

void RawInput::ReadBytes (char *buf, int length)
{
  int n;
  while (1) {
    n = ReadNoMoreThan(buf, length);
    if (n == length) return;
    buf += n;
    length -= n;
  }
}

// This function skips exactly 'length' bytes of input.

void RawInput::SkipBytes (int length)
{
  char c;
  while (length--)
    ReadNoMoreThan(&c, 1);
}

// This function reads a boolean in raw format and returns it.

bool RawInput::ReadBool ()
{
  char c;
  ReadBytes(&c, 1);
  return c != '\0';
}

// This function reads a character in raw format and returns it.

char RawInput::ReadChar ()
{
  char c;
  ReadBytes(&c, 1);
  return c;
}

// This function reads an unsigned character in raw format and returns
// it.

unsigned char RawInput::ReadUchar ()
{
  char c;
  ReadBytes(&c, 1);
  return static_cast<unsigned char>(c);
}

// This function reads a 2-byte short integer in raw format and
// returns it.

short RawInput::ReadShort ()
{
  char buf[2];
  ReadBytes(buf, 2);
  return ConvertToShort(buf);
}

// This function reads a 2-byte unsigned short integer in raw format
// and returns it.

unsigned short RawInput::ReadUshort ()
{
  char buf[2];
  ReadBytes(buf, 2);
  return ConvertToUshort(buf);
}

// This function reads a 4-byte long integer in raw format and returns
// it.

long RawInput::ReadLong ()
{
  char buf[4];
  ReadBytes(buf, 4);
  return ConvertToLong(buf);
}

// This function reads a 4-byte unsigned long integer in raw format
// and returns it.

unsigned long RawInput::ReadUlong ()
{
  char buf[4];
  ReadBytes(buf, 4);
  return ConvertToUlong(buf);
}

// This function reads a 8-byte long long integer in raw format and
// returns it.

LONGLONG RawInput::ReadLonglong ()
{
  char buf[8];
  ReadBytes(buf, 8);
  return ConvertToLonglong(buf);
}

// This function reads a 8-byte unsigned long long integer in raw
// format and returns it.

ULONGLONG RawInput::ReadUlonglong ()
{
  char buf[8];
  ReadBytes(buf, 8);
  return ConvertToUlonglong(buf);
}

// This function reads a 5-byte unsigned long long integer in raw
// format and returns it as an 8-byte unsigned long long integer.

ULONGLONG RawInput::Read5ByteUlonglong ()
{
  char buf[5];
  ReadBytes(buf, 5);
  return Convert5BytesToUlonglong(buf);
}

// This function reads a 6-byte unsigned long long integer in raw
// format and returns it as an 8-byte unsigned long long integer.

ULONGLONG RawInput::Read6ByteUlonglong ()
{
  char buf[6];
  ReadBytes(buf, 6);
  return Convert6BytesToUlonglong(buf);
}

// This function reads a string in raw format and returns it.

string RawInput::ReadString ()
{
  return ReadBytes(ReadShort());
}

// This function reads a string in raw format and puts it in the
// buffer pointed to by 'buf', terminted by a null character. The
// parameter 'maxLength' indicates the size of the buffer pointed to
// by 'buf'; if this isn't big enough to hold the string read, this
// function throws StringTooLong().

void RawInput::ReadCString (char *buf, int maxLength)
{
  short length = ReadShort();
  if (length >= maxLength)
    throw jrl::GenIoExceptions::StringTooLong();
  ReadBytes(buf, length);
  buf[length] = '\0';
}

// This function reads 'length' bytes and returns them as a string.

string RawInput::ReadBytes (int length)
{
  SafeArray<char> buf(length);
  ReadBytes(buf, length);
  return string(buf, length);
}

//////////////////////////////////////////////////////////////////////////////
// RawBufferedInput functions
//////////////////////////////////////////////////////////////////////////////

// This function conforms to the semantics of the virtual function it
// overrides, but reads its data via the internal buffer. In other
// words, it reads no more than 'length' bytes and puts them at the
// buffer pointed to by 'buf'. It returns the number of bytes read and
// copied in this way. The return value is always either positive or
// equal to length. If it would be otherwise, this function instead
// throws Eof() or Read().

int RawBufferedInput::ReadNoMoreThan (char *buf, int length)
{
  // If all the bytes in the buffer have been read, fill the buffer
  // with more bytes.

  if (validBytesInInternalBuf <= bytesReadFromInternalBuf)
    FillBuffer();

  // Copy at most length bytes, but no more than the number of bytes in
  // the internal buffer.

  int bytesToCopy = min(validBytesInInternalBuf - bytesReadFromInternalBuf,
                        length);

  // If there aren't any bytes in the internal buffer, throw Eof()
  // unless we have been requested to read 0 bytes.
  
  if (bytesToCopy <= 0 && bytesToCopy != length)
     throw jrl::GenIoExceptions::Eof();

  // Copy the bytes from the internal buffer to 'buf', record the fact
  // that we've read those bytes, and return that number of bytes.
  
  memcpy(buf, internalBuf + bytesReadFromInternalBuf, bytesToCopy);
  bytesReadFromInternalBuf += bytesToCopy;
  return bytesToCopy;
}

// This function skips over the given number of bytes in input. It
// takes account of the internal buffer, skipping unread bytes in the
// buffer first.

void RawBufferedInput::SkipBytes (int length)
{
  while (1) {
    bytesReadFromInternalBuf += length;
    if (bytesReadFromInternalBuf <= validBytesInInternalBuf)
      return;
    length = bytesReadFromInternalBuf - validBytesInInternalBuf;
    FillBuffer();
    if (validBytesInInternalBuf < 1)
      throw jrl::GenIoExceptions::Eof();
  }
}

// This function reads a boolean in raw format and returns it. It does
// its reading via the internal buffer.

bool RawBufferedInput::ReadBool ()
{
  if (bytesReadFromInternalBuf < validBytesInInternalBuf)
    return internalBuf[bytesReadFromInternalBuf++] != '\0';
  else
    return RawInput::ReadBool();
}

// This function reads a character in raw format and returns it. It does
// its reading via the internal buffer.

char RawBufferedInput::ReadChar ()
{
  if (bytesReadFromInternalBuf < validBytesInInternalBuf)
    return internalBuf[bytesReadFromInternalBuf++];
  else
    return RawInput::ReadChar();
}

// This function reads an unsigned character in raw format and returns
// it. It does its reading via the internal buffer.

unsigned char RawBufferedInput::ReadUchar ()
{
  if (bytesReadFromInternalBuf < validBytesInInternalBuf)
    return static_cast<unsigned char>(internalBuf[bytesReadFromInternalBuf++]);
  else
    return RawInput::ReadUchar();
}

// This function reads a 2-byte short integer in raw format and
// returns it. It does its reading via the internal buffer.

short RawBufferedInput::ReadShort ()
{
  if (bytesReadFromInternalBuf + 2 <= validBytesInInternalBuf) {
    short n = ConvertToShort(internalBuf + bytesReadFromInternalBuf);
    bytesReadFromInternalBuf += 2;
    return n;
  }
  else
    return RawInput::ReadShort();
}

// This function reads a 2-byte unsigned short integer in raw format
// and returns it. It does its reading via the internal buffer.

unsigned short RawBufferedInput::ReadUshort ()
{
  if (bytesReadFromInternalBuf + 2 <= validBytesInInternalBuf) {
    unsigned short n = ConvertToUshort(internalBuf + bytesReadFromInternalBuf);
    bytesReadFromInternalBuf += 2;
    return n;
  }
  else
    return RawInput::ReadUshort();
}

// This function reads a 4-byte long integer in raw format and returns
// it. It does its reading via the internal buffer.

long RawBufferedInput::ReadLong ()
{
  if (bytesReadFromInternalBuf + 4 <= validBytesInInternalBuf) {
    long n = ConvertToLong(internalBuf + bytesReadFromInternalBuf);
    bytesReadFromInternalBuf += 4;
    return n;
  }
  else
    return RawInput::ReadLong();
}

// This function reads a 4-byte unsigned long integer in raw format
// and returns it. It does its reading via the internal buffer.

unsigned long RawBufferedInput::ReadUlong ()
{
  if (bytesReadFromInternalBuf + 4 <= validBytesInInternalBuf) {
    unsigned long n = ConvertToUlong(internalBuf + bytesReadFromInternalBuf);
    bytesReadFromInternalBuf += 4;
    return n;
  }
  else
    return RawInput::ReadUlong();
}

// This function reads an 8-byte long long integer in raw format and
// returns it. It does its reading via the internal buffer.

LONGLONG RawBufferedInput::ReadLonglong ()
{
  if (bytesReadFromInternalBuf + 8 <= validBytesInInternalBuf) {
    LONGLONG n = ConvertToLonglong(internalBuf + bytesReadFromInternalBuf);
    bytesReadFromInternalBuf += 8;
    return n;
  }
  else
    return RawInput::ReadLonglong();
}

// This function reads an 8-byte unsigned long long integer in raw
// format and returns it. It does its reading via the internal buffer.

ULONGLONG RawBufferedInput::ReadUlonglong ()
{
  if (bytesReadFromInternalBuf + 8 <= validBytesInInternalBuf) {
    ULONGLONG n = ConvertToUlonglong(internalBuf + bytesReadFromInternalBuf);
    bytesReadFromInternalBuf += 8;
    return n;
  }
  else
    return RawInput::ReadUlonglong();
}

// This function reads a 5-byte unsigned long long integer in raw
// format and returns it as an 8-byte unsigned long long integer. It
// does its reading via the internal buffer.

ULONGLONG RawBufferedInput::Read5ByteUlonglong ()
{
  if (bytesReadFromInternalBuf + 5 <= validBytesInInternalBuf) {
    ULONGLONG n = Convert5BytesToUlonglong(internalBuf +
                                           bytesReadFromInternalBuf);
    bytesReadFromInternalBuf += 5;
    return n;
  }
  else
    return RawInput::Read5ByteUlonglong();
}

// This function reads a 6-byte unsigned long long integer in raw
// format and returns it as an 8-byte unsigned long long integer. It
// does its reading via the internal buffer.

ULONGLONG RawBufferedInput::Read6ByteUlonglong ()
{
  if (bytesReadFromInternalBuf + 6 <= validBytesInInternalBuf) {
    ULONGLONG n = Convert6BytesToUlonglong(internalBuf +
                                           bytesReadFromInternalBuf);
    bytesReadFromInternalBuf += 6;
    return n;
  }
  else
    return RawInput::Read6ByteUlonglong();
}

// This function fills the buffer with new data read from the actual
// input source. It assumes that all the data currently in the buffer
// has already been processed.

void RawBufferedInput::FillBuffer ()
{
  bytesReadFromInternalBuf = 0;
  validBytesInInternalBuf = UnbufferedRead(internalBuf, internalBufSize);
  if (validBytesInInternalBuf < 0)
    throw jrl::GenIoExceptions::Read();
}

//////////////////////////////////////////////////////////////////////////////
// RawOutput functions
//////////////////////////////////////////////////////////////////////////////

// This function writes the 'length' bytes starting at 'buf' to the output
// destination.

void RawOutput::WriteBytes (const char *buf, int length)
{
  int n;
  while (1) {
    n = WriteNoMoreThan(buf, length);
    if (n == length) return;
    buf += n;
    length -= n;
  }
}

// This function writes the given boolean in raw format.

void RawOutput::WriteBool (bool b)
{
  char c = b ? '\1' : '\0';
  WriteBytes(&c, 1);
}

// This function writes the given character in raw format.

void RawOutput::WriteChar (char c)
{
  WriteBytes(&c, 1);
}

// This function writes the given unsigned character in raw format.

void RawOutput::WriteUchar (unsigned char c)
{
  char sc = c;
  WriteBytes(&sc, 1);
}

// This function writes the given short integer in raw format.

void RawOutput::WriteShort (short i)
{
  WriteBytes((const char *) &i, 2);
}

// This function writes the given unsigned short integer in raw format.

void RawOutput::WriteUshort (unsigned short i)
{
  WriteBytes((const char *) &i, 2);
}

// This function writes the given long integer in raw format.

void RawOutput::WriteLong (long i)
{
  WriteBytes((const char *) &i, 4);
}

// This function writes the given unsigned long integer in raw format.

void RawOutput::WriteUlong (unsigned long i)
{
  WriteBytes((const char *) &i, 4);
}

// This function writes the given 8-byte long long integer in raw format.

void RawOutput::WriteLonglong (LONGLONG i)
{
  WriteBytes((const char *) &i, 8);
}

// This function writes the given 8-byte unsigned long long integer in
// raw format.

void RawOutput::WriteUlonglong (ULONGLONG i)
{
  WriteBytes((const char *) &i, 8);
}

// This function writes a string in raw format.

void RawOutput::WriteString (const string& s)
{
  short length = s.size();
  WriteShort(length);
  WriteBytes(s.c_str(), length);
}

// This function writes the null-terminated C string located at
// 'buf' in raw format.

void RawOutput::WriteCString (const char *buf)
{
  short length = strlen(buf);
  WriteShort(length);
  WriteBytes(buf, length);
}

// This function writes the bytes of the given string 's'.

void RawOutput::WriteBytes (const string& s)
{
  WriteBytes(s.c_str(), s.size());
}

//////////////////////////////////////////////////////////////////////////////
// RawBufferedOutput functions
//////////////////////////////////////////////////////////////////////////////

// This function writes at most 'length' bytes from the buffer
// 'buf'. It returns the number of bytes it wrote. It always returns
// either a positive value or length. (If it can't, it raises the
// Write() exception.)

int RawBufferedOutput::WriteNoMoreThan (const char *buf, int length)
{
  if (validBytesInInternalBuf + length > internalBufSize) {
    Flush();
    if (length > internalBufSize)
      length = internalBufSize;
  }

  memcpy(&internalBuf[validBytesInInternalBuf], buf, length);
  validBytesInInternalBuf += length;
  return length;
}

// This function writes out to the output destination any bytes that
// are in the internal buffer.

void RawBufferedOutput::Flush ()
{
  char *unflushedBytes = internalBuf;
  int n;
  while (validBytesInInternalBuf) {
    n = UnbufferedWrite(unflushedBytes, validBytesInInternalBuf);
    validBytesInInternalBuf -= n;
    unflushedBytes += n;
  }
}

//////////////////////////////////////////////////////////////////////////////
// RawBufferedOutputFile functions
//////////////////////////////////////////////////////////////////////////////

// This function closes the output file, first flushing any unflushed
// data in its buffer.

void RawBufferedOutputFile::Close ()
{
  if (fileHandle != INVALID_HANDLE_VALUE) {
    Flush();
    OutputFile::Close();
  }
}

//////////////////////////////////////////////////////////////////////////////
// RawBufferedOutputCompressedFile functions
//////////////////////////////////////////////////////////////////////////////

// This function closes the connection to the gzip process writing the
// compressed output file, first flushing any unflushed data in its
// buffer.

void RawBufferedOutputCompressedFile::Close ()
{
  if (fileHandle != INVALID_HANDLE_VALUE) {
    Flush();
    OutputCompressedFile::Close();
  }
}

//////////////////////////////////////////////////////////////////////////////
// RawBufferedGeneralOutput functions
//////////////////////////////////////////////////////////////////////////////

// This function closes the connection to the output destination,
// first flushing any unflushed data in its buffer.

void RawBufferedGeneralOutput::Close ()
{
  if (fileHandle != INVALID_HANDLE_VALUE) {
    Flush();
    GeneralOutput::Close();
  }
}
