// This file contains the main routine for the RSLog log file reader.

#include "stdafx.h"
#include <iostream>
#include <string>

#include "Exceptions.h"
#include "LogFile.h"

// ReadLogFile reads and prints the contents of the log file of the
// given name.

void ReadLogFile (const string& fileName, const Parameters& params)
{
  LogFile logFile(params);

  try {
    logFile.Open(fileName);
    logFile.Print(std::cout);
  }
  catch (Exceptions::InvalidLogFileFormat) {
    cerr << "Error: Invalid log format in file " << fileName << endl;
  }
  catch (Exceptions::VersionNotSupported) {
    cerr << "Error: Version not supported for file " << fileName << endl;
  }
  catch (jrl::GenIoExceptions::Any) {
    cerr << "Error: General I/O exception while reading log file "
         << fileName << endl;
  }
}

// ReadFile reads and prints the contents of the file of the given name.

void ReadFile (const string& fileName, const Parameters& params)
{
  int length = fileName.size();
  if (length < 4) throw Exceptions::BadFileName();

  string extension;
  if (fileName.substr(length-3) == ".gz") {
    if (length < 7) throw Exceptions::BadFileName();
    extension = fileName.substr(length-6, 3);
  }
  else
    extension = fileName.substr(length-3);

  if (extension == "log")
    ReadLogFile(fileName, params);
  else
    throw Exceptions::BadFileName();
}

// This function prints a usage message incorporating the passed program
// name.

void PrintUsage (char *programName)
{
  cerr << "Usage: " << programName << " [arguments] log-file-name..." << endl;
}

// The main program handles all arguments, treating some as parameters and
// some as file names.

int main (int argc, char **argv)
{
  if (argc < 2) {
    PrintUsage(argv[0]);
    return 0;
  }

  Parameters params;

  for (int i = 1; i < argc; i++) {
    try {
      if (argv[i][0] == '-' && argv[i][1] != '.')
        params.HandleArgument(&argv[i][1]);
      else
        ReadFile(argv[i], params);
    }
    catch (Exceptions::BadFileName) {
      cerr << "Bad file name " << argv[i] << endl;
      PrintUsage(argv[0]);
    }
    catch (Exceptions::ParameterNotHandled) {
      cerr << "Parameter " << argv[i] << " not handled." << endl;
      PrintUsage(argv[0]);
    }
  }

  return 0;
}
