#include "stdafx.h"
#include <iostream>
#include <time.h>
#include <sstream>
#include "times.h"
#include "typeconv.h"

using jrl::Time;

// Time::Time initializes the time to be 0.

Time::Time()
{
  sec = 0;
  ns = 0;
}

// Time::Time initializes the time to be the given number of seconds plus
// the given number of microseconds.

Time::Time(unsigned long seconds, unsigned long nanoseconds)
{
  sec = seconds;
  ns = nanoseconds;
}

// Time::Time initializes the time to be the given number of seconds. This
// number of seconds may not necessarily be integral.

Time::Time(double seconds)
{
  seconds += 0.0000000005;
  sec = (unsigned long) seconds;
  ns = (unsigned long) ((seconds - sec) * 1E9);
}

// Time::Time initializes the time to be the same as 'time'.

Time::Time(const Time& time)
{
  sec = time.sec;
  ns = time.ns;
}

// Time::Time initializes the time by converting a given number of cycles
// ('cycles') using a given CPU speed in Hz ('cpuSpeed').

Time::Time(LONGLONG cycles, unsigned long cpuSpeed)
{
  sec = cycles / cpuSpeed;
  cycles = cycles % cpuSpeed;
  ns = ((cycles * 1000000000) + (cpuSpeed / 2)) / cpuSpeed;
  if (ns >= 1000000000) {
    ns -= 1000000000;
    sec++;
  }
}

// Time::Time initializes the time by converting a given number of cycles
// ('cycles') using a given CPU speed in Hz ('cpuSpeed'), then adds a given
// offset.

Time::Time(LONGLONG cycles, unsigned long cpuSpeed, const Time& offset)
{
  sec = cycles / cpuSpeed;
  cycles = cycles % cpuSpeed;
  ns = ((cycles * 1000000000) + (cpuSpeed / 2)) / cpuSpeed;
  *this += offset;
}

// Time::Time initializes the time by converting a given number of cycles
// (high word 'cyclesHigh', low word 'cyclesLow') using a given CPU speed
// in Hz ('cpuSpeed').

Time::Time(unsigned long cyclesHigh, unsigned long cyclesLow,
           unsigned long cpuSpeed)
{
  ULONGLONG cycles;

  cycles = cyclesHigh;
  cycles <<= 32;
  cycles |= cyclesLow;

  sec = cycles / cpuSpeed;
  cycles = cycles % cpuSpeed;
  ns = ((cycles * 1000000000) + (cpuSpeed / 2)) / cpuSpeed;
  if (ns >= 1000000000) {
    ns -= 1000000000;
    sec++;
  }
}

// Time::SetFields sets the time to be the given number of seconds plus
// the given number of microseconds.

void Time::SetFields (unsigned long seconds, unsigned long nanoseconds)
{
  sec = seconds;
  ns = nanoseconds;
}

// Time::GetFields returns the number of seconds and nanoseconds of this
// time in the given parameters.

void Time::GetFields (unsigned long& seconds, unsigned long& nanoseconds) const
{
  seconds = sec;
  nanoseconds = ns;
}

// Time::operator+= adds the amount of time in 'addend' to this time.

Time& Time::operator+= (const Time& addend)
{
  ns += addend.ns;
  sec += addend.sec;
  if (ns >= 1000000000) {
    ns -= 1000000000;
    sec++;
  }
  return *this;
}

// Time::operator+ returns the sum of this time and the time represented by
// 'addend'.

Time Time::operator+ (const Time& addend) const
{
  Time retVal = *this;
  retVal += addend;
  return retVal;
}

// Time::operator-= subtacts the amount of time in 'subtraend' from this time.

Time& Time::operator-= (const Time& subtraend)
{
  if (ns < subtraend.ns) {
    ns = ns + 1000000000 - subtraend.ns;
    sec = sec - subtraend.sec - 1;
  }
  else {
    ns -= subtraend.ns;
    sec -= subtraend.sec;
  }

  return *this;
}

// Time::operator+ returns the difference of this time minus the time
// represented by 'subtraend'.

Time Time::operator- (const Time& subtraend) const
{
  Time retVal = *this;
  retVal -= subtraend;
  return retVal;
}

// Time::operator>= returns the comparison between this time and 'other'.

bool Time::operator>= (const Time& other) const
{
  if (sec > other.sec) return 1;
  if (sec < other.sec) return 0;
  return (ns >= other.ns);
}

// Time::operator> returns the comparison between this time and 'other'.

bool Time::operator> (const Time& other) const
{
  if (sec > other.sec) return 1;
  if (sec < other.sec) return 0;
  return (ns > other.ns);
}

// Time::operator<= returns the comparison between this time and 'other'.

bool Time::operator<= (const Time& other) const
{
  if (sec < other.sec) return 1;
  if (sec > other.sec) return 0;
  return (ns <= other.ns);
}

// Time::operator< returns the comparison between this time and 'other'.

bool Time::operator< (const Time& other) const
{
  if (sec < other.sec) return 1;
  if (sec > other.sec) return 0;
  return (ns < other.ns);
}

// Time::operator== returns the comparison between this time and 'other'.

bool Time::operator== (const Time& other) const
{
  return ((sec == other.sec) && (ns == other.ns));
}

// Time::operator!= returns the comparison between this time and 'other'.

bool Time::operator!= (const Time& other) const
{
  return ((sec != other.sec) || (ns != other.ns));
}

// Time::AsString prints the time to the given string 'str' and returns
// a pointer to that string.  This string buffer should be at least 27 bytes
// long, for safety.

string Time::AsString () const
{
  ostringstream ost;
  ost.fill('0');
  ost << sec / 3600 << ':';
  ost.width(2);
  ost << (sec / 60) % 60 << ':';
  ost.width(2);
  ost << sec % 60 << '.';
  ost.width(9);
  ost << ns;

  return ost.str();
}

// Time::ConvertToDouble returns a double representing the same number of
// seconds this does.

double Time::ConvertToDouble (void) const
{
  return (ns * 0.000000001 + sec);
}

// Time::RoundToMicroseconds returns a number of microseconds representing this
// time rounded to the nearest microsecond.

unsigned long Time::RoundToMicroseconds (void) const
{
  return (sec * 1000000) + ((ns + 500) / 1000);
}

// Time::RoundToMilliseconds returns a number of milliseconds representing this
// time rounded to the nearest millisecond.

unsigned long Time::RoundToMilliseconds (void) const
{
  return (sec * 1000) + ((ns + 500000) / 1000000);
}

// Time::RoundToSeconds returns a number of seconds representing this time
// rounded to the nearest second.

unsigned long Time::RoundToSeconds (void) const
{
  return (ns >= 500000000) ? sec + 1 : sec;
}

// Time::Print prints the time on the given output stream.

void Time::Print (ostream& os) const
{
  os.fill('0');

  if (sec > 0x20000000) {
    time_t unixTime = sec;
    struct tm *calendar = gmtime(&unixTime);
    if (calendar) {
      os.width(2);
      os << calendar->tm_mon + 1 << '/';
      os.width(2);
      os << calendar->tm_mday << '/';
      os.width(2);
      os << (calendar->tm_year % 100) << ' ';
      os.width(2);
      os << calendar->tm_hour << ':';
      os.width(2);
      os << calendar->tm_min << ':';
      os.width(2);
      os << calendar->tm_sec;
    }
    else
      os << "unknown";
  }
  else {
    os << sec / 3600;
    os << ':';
    os.width(2);
    os << (sec / 60) % 60 << ':';
    os.width(2);
    os << sec % 60;
  }

  os << '.';
  os.width(9);
  os << ns;
}

// ConvertMillisecondsToTime returns a Time representing the same number
// of milliseconds as the passed parameter 'milliseconds'.

Time jrl::ConvertMillisecondsToTime (unsigned long milliseconds)
{
  Time time(milliseconds / 1000, (milliseconds % 1000) * 1000000);
  return time;
}

Time jrl::MaxTime ()
{
  return Time(0xFFFFFFFFUL, 999999999);
}

ostream& jrl::operator<< (ostream& os, const Time& t)
{
  t.Print(os);
  return os;
}

ostream& jrl::operator<< (ostream& os, Time *t)
{
  t->Print(os);
  return os;
}

jrl::RawInput& jrl::operator>> (jrl::RawInput& is, Time& t)
{
  t.Read(is);
  return is;
}

jrl::RawOutput& jrl::operator<< (jrl::RawOutput& od, const Time& t)
{
  t.Write(od);
  return od;
}
