// This file contains routines for converting to and from various endian
// representations of numbers. The default when the endianness is not
// specified is little endian.

#include "stdafx.h"
#include <cstring>
#include "typeconv.h"

// This function converts a 5-byte little-endian representation of a
// number to a LONGLONG.

LONGLONG jrl::Convert5BytesToLonglong (const char *buf)
{
  return (((LONGLONG) (unsigned char) buf[4]) << 32) +
         (((LONGLONG) (unsigned char) buf[3]) << 24) +
         (((LONGLONG) (unsigned char) buf[2]) << 16) +
         (((LONGLONG) (unsigned char) buf[1]) << 8) +
         ((LONGLONG) (unsigned char) buf[0]);
}

// This function converts a 5-byte little-endian representation of a
// number to a ULONGLONG.

ULONGLONG jrl::Convert5BytesToUlonglong (const char *buf)
{
  return (((ULONGLONG) (unsigned char) buf[4]) << 32) +
         (((ULONGLONG) (unsigned char) buf[3]) << 24) +
         (((ULONGLONG) (unsigned char) buf[2]) << 16) +
         (((ULONGLONG) (unsigned char) buf[1]) << 8) +
         ((ULONGLONG) (unsigned char) buf[0]);
}

// This function converts a 6-byte little-endian representation of a
// number to a LONGLONG.

LONGLONG jrl::Convert6BytesToLonglong (const char *buf)
{
  return (((LONGLONG) (unsigned char) buf[5]) << 40) +
         (((LONGLONG) (unsigned char) buf[4]) << 32) +
         (((LONGLONG) (unsigned char) buf[3]) << 24) +
         (((LONGLONG) (unsigned char) buf[2]) << 16) +
         (((LONGLONG) (unsigned char) buf[1]) << 8) +
         ((LONGLONG) (unsigned char) buf[0]);
}

// This function converts a 6-byte little-endian representation of a
// number to a ULONGLONG.

ULONGLONG jrl::Convert6BytesToUlonglong (const char *buf)
{
  return (((ULONGLONG) (unsigned char) buf[5]) << 40) +
         (((ULONGLONG) (unsigned char) buf[4]) << 32) +
         (((ULONGLONG) (unsigned char) buf[3]) << 24) +
         (((ULONGLONG) (unsigned char) buf[2]) << 16) +
         (((ULONGLONG) (unsigned char) buf[1]) << 8) +
         ((ULONGLONG) (unsigned char) buf[0]);
}

// This function converts an 8-byte little-endian representation of a
// number to a double.

double jrl::ConvertToDouble (const char *buf)
{
#ifdef BYTE_ORDER_LITTLE_ENDIAN
  return * (double *) buf;
#else
  char reverseBuf[8];
  for (int i = 0; i < 8; i++)
    reverseBuf[i] = buf[7-i];
  return * (double *) reverseBuf;
#endif
}

// This function converts an 8-byte little-endian representation of a
// number to a LONGLONG.

LONGLONG jrl::ConvertToLonglong (const char *buf)
{
#ifdef BYTE_ORDER_LITTLE_ENDIAN
  if (sizeof(LONGLONG) == 8)
    return * (LONGLONG *) buf;
#endif

  LONGLONG n = (((LONGLONG) (unsigned char) (buf[7] & 0x7F)) << 56) +
               (((LONGLONG) (unsigned char) buf[6]) << 48) +
               (((LONGLONG) (unsigned char) buf[5]) << 40) +
               (((LONGLONG) (unsigned char) buf[4]) << 32) +
               (((LONGLONG) (unsigned char) buf[3]) << 24) +
               (((LONGLONG) (unsigned char) buf[2]) << 16) +
               (((LONGLONG) (unsigned char) buf[1]) << 8) +
               ((LONGLONG) (unsigned char) buf[0]);
  if (buf[7] & 0x80)
    return n - MakeUlonglong(0x80000000UL, 0);
  else
    return n;
}

// This function converts an 8-byte little-endian representation of a
// number to a ULONGLONG.

ULONGLONG jrl::ConvertToUlonglong (const char *buf)
{
#ifdef BYTE_ORDER_LITTLE_ENDIAN
  if (sizeof(ULONGLONG) == 8)
    return * (ULONGLONG *) buf;
#endif

  return (((ULONGLONG) (unsigned char) buf[7]) << 56) +
         (((ULONGLONG) (unsigned char) buf[6]) << 48) +
         (((ULONGLONG) (unsigned char) buf[5]) << 40) +
         (((ULONGLONG) (unsigned char) buf[4]) << 32) +
         (((ULONGLONG) (unsigned char) buf[3]) << 24) +
         (((ULONGLONG) (unsigned char) buf[2]) << 16) +
         (((ULONGLONG) (unsigned char) buf[1]) << 8) +
         ((ULONGLONG) (unsigned char) buf[0]);
}

// This function converts a 4-byte little-endian representation of a
// number to a long.

long jrl::ConvertToLong (const char *buf)
{
#ifdef BYTE_ORDER_LITTLE_ENDIAN
  if (sizeof(long) == 4)
    return * (long *) buf;
#endif

  long n = (((long) (unsigned char) (buf[3] & 0x7F)) << 24) +
           (((long) (unsigned char) buf[2]) << 16) +
           (((long) (unsigned char) buf[1]) << 8) +
           ((long) (unsigned char) buf[0]);
  if (buf[3] & 0x80)
    return n - 0x80000000UL;
  else
    return n;
}

// This function converts a 4-byte little-endian representation of a
// number to an unsigned long.

unsigned long jrl::ConvertToUlong (const char *buf)
{
#ifdef BYTE_ORDER_LITTLE_ENDIAN
  if (sizeof(unsigned long) == 4)
    return * (unsigned long *) buf;
#endif

  return (((unsigned long) (unsigned char) buf[3]) << 24) +
         (((unsigned long) (unsigned char) buf[2]) << 16) +
         (((unsigned long) (unsigned char) buf[1]) << 8) +
         ((unsigned long) (unsigned char) buf[0]);
}

// This function converts a 2-byte little-endian representation of a
// number to an unsigned short.

short jrl::ConvertToShort (const char *buf)
{
#ifdef BYTE_ORDER_LITTLE_ENDIAN
  if (sizeof(short) == 2)
    return * (short *) buf;
#endif

  short n = (((short) (unsigned char) (buf[1] & 0x7F)) << 8) +
            ((unsigned short) (unsigned char) buf[0]);
  if (buf[1] & 0x80)
    return n - 0x8000;
  else
    return n;
}

// This function converts a 2-byte little-endian representation of a
// number to an unsigned short.

unsigned short jrl::ConvertToUshort (const char *buf)
{
#ifdef BYTE_ORDER_LITTLE_ENDIAN
  if (sizeof(unsigned short) == 2)
    return * (unsigned short *) buf;
#endif

  return (((unsigned short) (unsigned char) buf[1]) << 8) +
         ((unsigned short) (unsigned char) buf[0]);
}

// This function converts a double into an 8-byte little endian
// representation, placed in 'buf'.

void jrl::ConvertDouble (double number, char *buf)
{
#ifdef BYTE_ORDER_LITTLE_ENDIAN
  memcpy(buf, (char *) &number, 8);
#else
  for (int i = 0; i < 8; i++)
    buf[i] = ((char *) &number)[7-i];
#endif
}

// This function converts a ULONGLONG into an 8-byte little endian
// representation, placed in 'buf'.

void jrl::ConvertLonglong (LONGLONG number, char *buf)
{
#ifdef BYTE_ORDER_LITTLE_ENDIAN
  memcpy(buf, (char *) &number, 8);
#else
  bool negative = (number < 0);

  if (negative)
    number += MakeUlonglong(0x80000000UL, 0);

  buf[0] = (char) (number & 0xFF);
  buf[1] = (char) ((number & 0xFF00) >> 8);
  buf[2] = (char) ((number & 0xFF0000) >> 16);
  buf[3] = (char) ((number & 0xFF000000) >> 24);
  number >>= 32;
  buf[4] = (char) (number & 0xFF);
  buf[5] = (char) ((number & 0xFF00) >> 8);
  buf[6] = (char) ((number & 0xFF0000) >> 16);
  buf[7] = (char) ((number & 0xFF000000) >> 24);

  if (negative)
    buf[7] |= 0x80;
#endif
}

// This function converts a ULONGLONG into an 8-byte little endian
// representation, placed in 'buf'.

void jrl::ConvertUlonglong (ULONGLONG number, char *buf)
{
#ifdef BYTE_ORDER_LITTLE_ENDIAN
  memcpy(buf, (char *) &number, 8);
#else
  buf[0] = (char) (number & 0xFF);
  buf[1] = (char) ((number & 0xFF00) >> 8);
  buf[2] = (char) ((number & 0xFF0000) >> 16);
  buf[3] = (char) ((number & 0xFF000000) >> 24);
  number >>= 32;
  buf[4] = (char) (number & 0xFF);
  buf[5] = (char) ((number & 0xFF00) >> 8);
  buf[6] = (char) ((number & 0xFF0000) >> 16);
  buf[7] = (char) ((number & 0xFF000000) >> 24);
#endif
}

// This function converts a long into a 4-byte little endian
// representation, placed in 'buf'.

void jrl::ConvertLong (long number, char *buf)
{
#ifdef BYTE_ORDER_LITTLE_ENDIAN
  memcpy(buf, (char *) &number, 4);
#else
  bool negative = (number < 0);

  if (negative)
    number += 0x80000000UL;

  buf[0] = (char) (number & 0xFF);
  buf[1] = (char) ((number & 0xFF00) >> 8);
  buf[2] = (char) ((number & 0xFF0000) >> 16);
  buf[3] = (char) ((number & 0xFF000000) >> 24);

  if (negative)
    buf[3] |= 0x80;
#endif
}

// This function converts an unsigned long into a 4-byte little endian
// representation, placed in 'buf'.

void jrl::ConvertUlong (unsigned long number, char *buf)
{
#ifdef BYTE_ORDER_LITTLE_ENDIAN
  memcpy(buf, (char *) &number, 4);
#else
  buf[0] = (char) (number & 0xFF);
  buf[1] = (char) ((number & 0xFF00) >> 8);
  buf[2] = (char) ((number & 0xFF0000) >> 16);
  buf[3] = (char) ((number & 0xFF000000) >> 24);
#endif
}

// This function converts a short into a 2-byte little endian
// representation, placed in 'buf'.

void jrl::ConvertShort (short number, char *buf)
{
#ifdef BYTE_ORDER_LITTLE_ENDIAN
  memcpy(buf, (char *) &number, 2);
#else
  bool negative = (number < 0);

  if (negative)
    number += (short) 0x8000;

  buf[0] = (char) (number & 0xFF);
  buf[1] = (char) ((number & 0xFF00) >> 8);

  if (negative)
    buf[1] |= 0x80;
#endif
}

// This function converts an unsigned short into a 2-byte little
// endian representation, placed in 'buf'.

void jrl::ConvertUshort (unsigned short number, char *buf)
{
#ifdef BYTE_ORDER_LITTLE_ENDIAN
  memcpy(buf, (char *) &number, 2);
#else
  buf[0] = (char) (number & 0xFF);
  buf[1] = (char) ((number & 0xFF00) >> 8);
#endif
}

// This function converts a 4-byte big-endian representation of a
// number to an unsigned long.

unsigned long jrl::ConvertToUlongBigEndian (const char *buf)
{
  return (((unsigned long) (unsigned char) buf[0]) << 24) +
         (((unsigned long) (unsigned char) buf[1]) << 16) +
         (((unsigned long) (unsigned char) buf[2]) << 8) +
         ((unsigned long) (unsigned char) buf[3]);
}

// This function converts a 2-byte big-endian representation of a
// number to an unsigned short.

unsigned short jrl::ConvertToUshortBigEndian (const char *buf)
{
  return (((unsigned short) (unsigned char) buf[0]) << 8) +
         ((unsigned short) (unsigned char) buf[1]);
}

// This function converts an unsigned long into a 4-byte big endian
// representation, placed in 'buf'.

void jrl::ConvertUlongBigEndian (unsigned long number, char *buf)
{
#ifdef BYTE_ORDER_BIG_ENDIAN
  memcpy(buf, (char *) &number, 4);
#else
  buf[3] = (char) (number & 0xFF);
  buf[2] = (char) ((number & 0xFF00) >> 8);
  buf[1] = (char) ((number & 0xFF0000) >> 16);
  buf[0] = (char) ((number & 0xFF000000) >> 24);
#endif
}

// This function converts an unsigned short into a 2-byte big endian
// representation, placed in 'buf'.

void jrl::ConvertUshortBigEndian (unsigned short number, char *buf)
{
#ifdef BYTE_ORDER_BIG_ENDIAN
  memcpy(buf, (char *) &number, 2);
#else
  buf[1] = (char) (number & 0xFF);
  buf[0] = (char) ((number & 0xFF00) >> 8);
#endif
}
