// This file contains functions for checking task type group file
// information.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/Parameters.h"
#include "../Common/Algorithm.h"
#include "../Common/TTG.h"
#include "Check.h"
#include "Convert.h"
#include "Globals.h"
#include "Log.h"
#include "SimpleSched.h"
#include "Update.h"

////////////////////////////////////////////////////////////////////////////
// SAMPLE DATA
////////////////////////////////////////////////////////////////////////////

// This function checks the aged sample data given. It returns a boolean
// indicating if the sample is valid.

BOOLEAN CheckAgedSample (PAGED_SAMPLE sample)
{
  // Make sure the aging factor is valid.

  if (sample->agingFactor < 0 || sample->agingFactor > 1) {
    DebugPrint(("RSTask: ERROR - Aging factor is invalid.\n"));
    BothLogShort(RS_TASK_ERROR_INVALID_AGING_FACTOR,
                 STATUS_UNSUCCESSFUL);
    return FALSE;
  }

  if (sample->n == 0) {
    // If the sample size is 0, the weighted count, sum, and sum of
    // squares should all be 0.

    if (sample->weightedCount || sample->weightedSum ||
        sample->weightedSumSq) {
      DebugPrint(("RSTask: ERROR - Aged sample has sample size 0 but "
                  "summary values are not all 0.\n"));
      BothLogShort(RS_TASK_ERROR_AGED_SAMPLE_EMPTY_BUT_INVALID,
                   STATUS_UNSUCCESSFUL);
      return FALSE;
    }
  }
  else {
    // If the sample size is nonzero, the weighted count should be
    // positive and the variance should be nonnegative.

    if (sample->weightedCount <= 0) {
      DebugPrint(("RSTask: ERROR - Aged sample has a nonzero sample size "
                  "but a nonpositive weighted count.\n"));
      BothLogShort(RS_TASK_ERROR_AGED_SAMPLE_WEIGHTED_COUNT_WRONG,
                   STATUS_UNSUCCESSFUL);
      return FALSE;
    }
    if (sample->weightedSumSq * sample->weightedCount <
        sample->weightedSum * sample->weightedSum) {
      DebugPrint(("RSTask: WARNING - Aged sample amounts are "
                  "impossible since they imply a negative variance.\n"));
      BothLogShort(RS_TASK_ERROR_AGED_SAMPLE_NONEMPTY_AND_INVALID,
                   STATUS_UNSUCCESSFUL);
      return FALSE;
    }
  }

  // If everything has checked out OK, the sample is OK.

  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
// ALGORITHM DATA
////////////////////////////////////////////////////////////////////////////

// This algorithm checks data associated with the aged-gamma basic
// strategy.

BOOLEAN CheckAGBasicData (PTT_INFO ttInfo, PAG_BASIC_DATA agBasicData)
{
  // Check the sample.

  if (!CheckAgedSample(&agBasicData->sample))
    return FALSE;

  // It's just as fast to recompute the deadline in intervals as it
  // is to check it, so just recompute it.

  agBasicData->deadlineI = HnsToIntervals(ttInfo->deadlineHns);

  // Make sure the number of sample values before the first reschedule
  // is at least 2, since we can't compute the variance with fewer than
  // 2 sample values.

  if (agBasicData->sampleValuesBeforeFirstReschedule < 2)
    agBasicData->sampleValuesBeforeFirstReschedule = 2;

  // Make sure the number of sample values between reschedules is at
  // least 1, because it doesn't make sense otherwise.

  if (agBasicData->sampleValuesBetweenReschedules < 1)
    agBasicData->sampleValuesBetweenReschedules = 1;

  // If everything checked out OK, return TRUE.

  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
// TASK TYPE INFO
////////////////////////////////////////////////////////////////////////////

// This function checks the task type info for a certain task type
// in a task type group. This involves checking that the offset to
// this task type information is valid as well as checking that all
// the data for the task type is valid. It returns a boolean indicating
// whether the check was successful.
//
// The parameters are as follows:
//
// ttgInfo --
//   points to the task type group information, i.e., the image of the
//   task type group file in memory
//
// index --
//   the index of the task type to be checked by this function
//
// minOffsetPtr --
//   On entry to this function, '*minOffsetPtr' contains the minimum
//   valid offset for the beginning of this task type's information,
//   computed as the ending offset for the previous feature in the
//   task type group file. On exit from this function, it should
//   contain the ending offset for this task type's information, so
//   that it can be used for the next call to this function for the
//   next task type information.
//
// maxOffset --
//   the maximum offset that this task type information can end at,
//   normally the offset of the end of the file
//
// mustRecomputeSchedule --
//   a boolean which, if true, forces us to recompute the schedule for
//   this task type, normally set because either the CPU information
//   or scheduling interval has changed since we last checked this
//   task type information

BOOLEAN CheckTTInfo (
  PTTG_INFO ttgInfo,
  ULONG index,
  ULONG *minOffsetPtr,
  ULONG maxOffset,
  BOOLEAN mustRecomputeSchedule
  )
{
  PTT_INFO ttInfo;
  ULONG ttInfoOffset, algorithmDataOffset;
  PVOID algorithmDataAddr;

  // Determine the offset for this task type information, and make sure
  // it doesn't violate the minimum-offset restriction, i.e., make sure
  // that it doesn't start before the previous feature ended.

  ttInfoOffset = ttgInfo->ttInfoOffsets[index];
  if (ttInfoOffset < *minOffsetPtr) {
    DebugPrint(("RSTask: ERROR - TT info number %d has too early an offset "
                "(%lu before %lu).\n", index, ttInfoOffset, *minOffsetPtr));
    BothLogOneParameter(RS_TASK_ERROR_TASK_TYPE_INFO_HAS_TOO_EARLY_OFFSET,
                        STATUS_UNSUCCESSFUL,
                        index);
    return FALSE;
  }

  // Set the minimum offset to just beyond the preamble of this task
  // type information. Make sure this does not go beyond the maximum
  // offset. Then, we can safely access parts of the preamble without
  // worrying about memory access violations.

  *minOffsetPtr = ttInfoOffset + sizeof(TT_INFO);
  if (*minOffsetPtr > maxOffset) {
    DebugPrint(("RSTask: ERROR - TT info number %d goes beyond end.\n",
                index));
    BothLogOneParameter(RS_TASK_ERROR_TASK_TYPE_INFO_GOES_PAST_FILE_END,
                        STATUS_UNSUCCESSFUL,
                        index);
    return FALSE;
  }

  // Compute the location in memory of the task type information.

  ttInfo = (PTT_INFO)(ttInfoOffset + (PCHAR) ttgInfo);

  // Make sure that the boolean indicating whether sampling should be
  // performed is set correctly.

  ttInfo->useSample = (AUTO_SCHEDULE_OF_ALGORITHM(ttInfo->algorithmType) !=
                       AUTO_SCHEDULE_TYPE_NONE);

  // Make sure the schedule has at least one phase per available speed
  // and no more than MAX_NUM_SCHEDULE_PHASES phases.

  if (ttInfo->schedule.maxNumPhases < g.registryInfo.numAvailableSpeeds) {
    DebugPrint(("RSTask: ERROR - TT info number %d has too few "
                "schedule phase spots (%d).\n", index,
                ttInfo->schedule.maxNumPhases));
    BothLogOneParameter(RS_TASK_ERROR_TASK_TYPE_INFO_HAS_TOO_FEW_PHASES,
                        STATUS_UNSUCCESSFUL,
                        index);
    return FALSE;
  }
  if (ttInfo->schedule.maxNumPhases > MAX_NUM_SCHEDULE_PHASES) {
    DebugPrint(("RSTask: ERROR - TT info number %d has too many "
                "schedule phase spots (%d).\n", index,
                ttInfo->schedule.maxNumPhases));
    BothLogOneParameter(RS_TASK_ERROR_TASK_TYPE_INFO_HAS_TOO_MANY_PHASES,
                        STATUS_UNSUCCESSFUL,
                        index);
    return FALSE;
  }

  // Now increase the minimum offset to reflect the size of the
  // schedule phases that we now know follow the preamble. There
  // are ttInfo->schedule.maxNumPhases - 1 of them. Make sure this
  // offset does not go beyond the maximum offset.

  *minOffsetPtr += (ttInfo->schedule.maxNumPhases - 1) *
                   sizeof(SPEED_SCHEDULE_PHASE);
  if (*minOffsetPtr > maxOffset) {
    DebugPrint(("RSTask: ERROR - TT info number %d has its schedule go "
                "beyond the end of file.\n", index));
    BothLogOneParameter(RS_TASK_ERROR_TASK_TYPE_INFO_SCHEDULE_PAST_FILE_END,
                        STATUS_UNSUCCESSFUL,
                        index);
    return FALSE;
  }

  // Now, compute the offset of the algorithm data part of this task
  // type information (by adding the offset of the beginning of this
  // task type information to the relative offset of the algorithm
  // data within the task type information). This offset must not be
  // before the minimum offset, which shows where the schedule of this
  // task type ends.

  algorithmDataOffset = ttInfoOffset + ttInfo->algorithmDataOffset;
  if (algorithmDataOffset < *minOffsetPtr) {
    DebugPrint(("RSTask: ERROR - TT info number %d has algorithm data "
                "too early (%lu before %lu).\n", index, algorithmDataOffset,
                *minOffsetPtr));
    BothLogOneParameter(RS_TASK_ERROR_TASK_TYPE_INFO_ALGORITHM_DATA_EARLY,
                        STATUS_UNSUCCESSFUL,
                        index);
    return FALSE;
  }

  // Compute the address of the algorithm data in memory from the
  // algorithm data offset. Make this the new minimum offset
  // reflecting where the algorithm data begins.

  algorithmDataAddr = (PVOID)(algorithmDataOffset + (PCHAR) ttgInfo);
  *minOffsetPtr = algorithmDataOffset;

  // Now, check the algorithm data. This involves incrementing the
  // minimum offset by its size then making sure that this minimum
  // offset, which then reflects the end of the algorithm data
  // feature, does not go beyond the end of the file. It also
  // involves making sure the algorithm data is valid.

  switch (ttInfo->algorithmType) {
  case ALGORITHM_TYPE_FLAT :
    {
      // Increment the minimum offset by the size of this algorithm's data
      // and make sure it doesn't go past the end of the file image.

      *minOffsetPtr += sizeof(ALGORITHM_DATA_FLAT);
      if (*minOffsetPtr > maxOffset) {
        DebugPrint(("RSTask: ERROR - Algorithm data of TT info number %d "
                    "(type %d) goes past end of file.\n", index,
                    ttInfo->algorithmType));
        BothLogTwoParameters(RS_TASK_ERROR_ALGORITHM_DATA_GOES_PAST_EOF,
                             STATUS_UNSUCCESSFUL,
                             ttInfo->algorithmType,
                             index);
        return FALSE;
      }
    }
    break;

  case ALGORITHM_TYPE_FLAT_AG_BASIC :
    {
      PALGORITHM_DATA_FLAT_AG_BASIC algorithmData = algorithmDataAddr;

      // Increment the minimum offset by the size of this algorithm's
      // data and make sure it doesn't go past the end of the file
      // image. Then we'll be able to access fields in the data
      // without worrying about memory access violations.

      *minOffsetPtr += sizeof(ALGORITHM_DATA_FLAT_AG_BASIC);
      if (*minOffsetPtr > maxOffset) {
        DebugPrint(("RSTask: ERROR - Algorithm data of TT info number %d "
                    "(type %d) goes past end of file.\n", index,
                    ttInfo->algorithmType));
        BothLogTwoParameters(RS_TASK_ERROR_ALGORITHM_DATA_GOES_PAST_EOF,
                             STATUS_UNSUCCESSFUL,
                             ttInfo->algorithmType,
                             index);
        return FALSE;
      }

      // Check the data related to the aged-gamma-basic automatic
      // scheduler.

      if (!CheckAGBasicData(ttInfo, &algorithmData->agBasicData))
        return FALSE;
    }
    break;

  case ALGORITHM_TYPE_LONGSHORT :
    {
      // Increment the minimum offset by the size of this algorithm's data
      // and make sure it doesn't go past the end of the file image.

      *minOffsetPtr += sizeof(ALGORITHM_DATA_LONGSHORT);
      if (*minOffsetPtr > maxOffset) {
        DebugPrint(("RSTask: ERROR - Algorithm data of TT info number %d "
                    "(type %d) goes past end of file.\n", index,
                    ttInfo->algorithmType));
        BothLogTwoParameters(RS_TASK_ERROR_ALGORITHM_DATA_GOES_PAST_EOF,
                             STATUS_UNSUCCESSFUL,
                             ttInfo->algorithmType,
                             index);
        return FALSE;
      }
    }
    break;

  case ALGORITHM_TYPE_LONGSHORT_AG_BASIC :
    {
      PALGORITHM_DATA_LONGSHORT_AG_BASIC algorithmData = algorithmDataAddr;

      // Increment the minimum offset by the size of this algorithm's
      // data and make sure it doesn't go past the end of the file
      // image. Then we'll be able to access fields in the data
      // without worrying about memory access violations.

      *minOffsetPtr += sizeof(ALGORITHM_DATA_LONGSHORT_AG_BASIC);
      if (*minOffsetPtr > maxOffset) {
        DebugPrint(("RSTask: ERROR - Algorithm data of TT info number %d "
                    "(type %d) goes past end of file.\n", index,
                    ttInfo->algorithmType));
        BothLogTwoParameters(RS_TASK_ERROR_ALGORITHM_DATA_GOES_PAST_EOF,
                             STATUS_UNSUCCESSFUL,
                             ttInfo->algorithmType,
                             index);
        return FALSE;
      }

      // Check the data related to the aged-gamma-basic automatic
      // scheduler.

      if (!CheckAGBasicData(ttInfo, &algorithmData->agBasicData))
        return FALSE;
    }
    break;

  case ALGORITHM_TYPE_PASTPEG :
    {
      // Increment the minimum offset by the size of this algorithm's data
      // and make sure it doesn't go past the end of the file image.

      *minOffsetPtr += sizeof(ALGORITHM_DATA_PASTPEG);
      if (*minOffsetPtr > maxOffset) {
        DebugPrint(("RSTask: ERROR - Algorithm data of TT info number %d "
                    "(type %d) goes past end of file.\n", index,
                    ttInfo->algorithmType));
        BothLogTwoParameters(RS_TASK_ERROR_ALGORITHM_DATA_GOES_PAST_EOF,
                             STATUS_UNSUCCESSFUL,
                             ttInfo->algorithmType,
                             index);
        return FALSE;
      }
    }
    break;

  case ALGORITHM_TYPE_PASTPEG_AG_BASIC :
    {
      PALGORITHM_DATA_PASTPEG_AG_BASIC algorithmData = algorithmDataAddr;

      // Increment the minimum offset by the size of this algorithm's
      // data and make sure it doesn't go past the end of the file
      // image. Then we'll be able to access fields in the data
      // without worrying about memory access violations.

      *minOffsetPtr += sizeof(ALGORITHM_DATA_PASTPEG_AG_BASIC);
      if (*minOffsetPtr > maxOffset) {
        DebugPrint(("RSTask: ERROR - Algorithm data of TT info number %d "
                    "(type %d) goes past end of file.\n", index,
                    ttInfo->algorithmType));
        BothLogTwoParameters(RS_TASK_ERROR_ALGORITHM_DATA_GOES_PAST_EOF,
                             STATUS_UNSUCCESSFUL,
                             ttInfo->algorithmType,
                             index);
        return FALSE;
      }

      // Check the data related to the aged-gamma-basic automatic
      // scheduler.

      if (!CheckAGBasicData(ttInfo, &algorithmData->agBasicData))
        return FALSE;
    }
    break;

  case ALGORITHM_TYPE_LONGRUN :
    {
      // There is no algorithm data for this algorithm, so just make
      // sure the end of this no data does not go beyond the end of
      // the file image.

      if (*minOffsetPtr > maxOffset) {
        DebugPrint(("RSTask: ERROR - Algorithm data of TT info number %d "
                    "(type %d) goes past end of file.\n", index,
                    ttInfo->algorithmType));
        BothLogTwoParameters(RS_TASK_ERROR_ALGORITHM_DATA_GOES_PAST_EOF,
                             STATUS_UNSUCCESSFUL,
                             ttInfo->algorithmType,
                             index);
        return FALSE;
      }
    }
    break;

  case ALGORITHM_TYPE_LONGRUN_AG_BASIC :
    {
      PALGORITHM_DATA_LONGRUN_AG_BASIC algorithmData = algorithmDataAddr;

      // Increment the minimum offset by the size of this algorithm's
      // data and make sure it doesn't go past the end of the file
      // image. Then we'll be able to access fields in the data
      // without worrying about memory access violations.

      *minOffsetPtr += sizeof(ALGORITHM_DATA_LONGRUN_AG_BASIC);
      if (*minOffsetPtr > maxOffset) {
        DebugPrint(("RSTask: ERROR - Algorithm data of TT info number %d "
                    "(type %d) goes past end of file.\n", index,
                    ttInfo->algorithmType));
        BothLogTwoParameters(RS_TASK_ERROR_ALGORITHM_DATA_GOES_PAST_EOF,
                             STATUS_UNSUCCESSFUL,
                             ttInfo->algorithmType,
                             index);
        return FALSE;
      }

      // Check the data related to the aged-gamma-basic automatic
      // scheduler.

      if (!CheckAGBasicData(ttInfo, &algorithmData->agBasicData))
        return FALSE;
    }
    break;

  default :
    DebugPrint(("RSTask: ERROR - Invalid algorithm type %d in task type %d.\n",
                ttInfo->algorithmType, index));
    BothLogTwoParameters(RS_TASK_ERROR_ALGORITHM_CANT_CHECK_ALGORITHM_TYPE,
                         STATUS_UNSUCCESSFUL,
                         ttInfo->algorithmType,
                         index);
    return FALSE;
    break;
  }

  // If we are forced to recompute the schedule for this schedule, or
  // if the current schedule is not valid, compute a new initial
  // schedule.

  if (mustRecomputeSchedule || !IsSpeedScheduleValid(&ttInfo->schedule))
    ComputeInitialSchedule(ttInfo);

  // Now that all the checks were successful, return TRUE.

  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
// TASK TYPE GROUP INFO
////////////////////////////////////////////////////////////////////////////

// This function checks the task type group information passed to make
// sure it's valid and doesn't reference anything beyond its place in
// memory. The other parameter is the size of this information in
// memory.

BOOLEAN CheckTTGInfo (
  PTTG_INFO ttgInfo,
  ULONG infoSize
  )
{
  KFLOATING_SAVE floatingPointStateSave;
  BOOLEAN success = TRUE, mustRecomputeSchedules = FALSE;
  PTT_INFO ttInfo;
  ULONG i, minOffset;

  DebugPrint(("RSTask: Checking TTG info.\n"));

  // First, make sure the size in memory is at least the size of the
  // preamble. Then we will be able to reference fields in that preamble
  // without worrying about violating memory restrictions.

  if (infoSize < TTG_INFO_PREAMBLE_SIZE) {
    DebugPrint(("RSTask: ERROR - TTG info is too small (%lu).\n", infoSize));
    BothLogShort(RS_TASK_ERROR_TTG_INFO_TOO_SMALL,
                 STATUS_UNSUCCESSFUL);
    return FALSE;
  }

  // Check the signature to make sure it's a valid task type group file.

  if (ttgInfo->signature != TTG_INFO_SIGNATURE) {
    DebugPrint(("RSTask: ERROR - TTG info signature is wrong.\n"));
    BothLogShort(RS_TASK_ERROR_TTG_SIGNATURE_WRONG,
                 STATUS_UNSUCCESSFUL);
    return FALSE;
  }

  // Check the version to make sure we know how to interpret this
  // task type group file.

  if (ttgInfo->version != TTG_INFO_VERSION) {
    DebugPrint(("RSTask: ERROR - TTG info version is wrong (%x).\n",
                ttgInfo->version));
    BothLogShort(RS_TASK_ERROR_INVALID_TTG_VERSION,
                 STATUS_UNSUCCESSFUL);
    return FALSE;
  }

  // Check the preamble reported size to make sure it's the size of
  // the file image in memory.

  if (ttgInfo->size != infoSize) {
    DebugPrint(("RSTask: ERROR - TTG info size is wrong (%lu instead of %lu)"
                ".\n", ttgInfo->size, infoSize));
    BothLogShort(RS_TASK_ERROR_INVALID_TTG_SIZE_FIELD,
                 STATUS_UNSUCCESSFUL);
    return FALSE;
  }

  // Make sure the number of task types is not excessively large.

  if (ttgInfo->numTaskTypes > MAX_NUM_TASK_TYPES) {
    DebugPrint(("RSTask: ERROR - TTG info has too many task types (%ld).\n",
                ttgInfo->numTaskTypes));
    BothLogShort(RS_TASK_ERROR_TTG_INFO_TOO_MANY_TASK_TYPES,
                 STATUS_UNSUCCESSFUL);
    return FALSE;
  }

  // Now we begin checking of the rest of the file image. The way we
  // do this is we set 'minOffset' to be the minimum offset the next
  // feature should have in the file (just beyond the previous
  // feature), and make sure this condition isn't violated. Then we
  // check that the feature doesn't go beyond the end of the file
  // image. Finally, we set 'minOffset' to be the offset just beyond
  // that feature. Going like this, we eventually check the entire
  // file image.

  // Initialize 'minOffset' to be the position just beyond the
  // array of task type information offsets. That's the size of the
  // preamble plus sizeof(ULONG) for each task type. Make sure this
  // offset is not beyond the end of the file.

  minOffset = TTG_INFO_PREAMBLE_SIZE + ttgInfo->numTaskTypes * sizeof(ULONG);
  if (minOffset > infoSize) {
    DebugPrint(("RSTask: ERROR - TTG info preamble is too big.\n"));
    BothLogShort(RS_TASK_ERROR_TTG_INFO_PREAMBLE_GOES_PAST_FILE_END,
                 STATUS_UNSUCCESSFUL);
    return FALSE;
  }

  // Before checking all the task types, determine if we will need
  // to recompute all their schedules. This is the case if the file
  // hasn't been checked since the last time the CPU information in
  // the registry was changed, or if the scheduling interval in
  // force when this file was last checked was different from what
  // it is now.

  if (ttgInfo->timeLastChecked.QuadPart <
        g.registryInfo.cpuInfoModificationTime.QuadPart ||
      ttgInfo->schedulingIntervalHnsWhenLastChecked !=
        g.timerResolutionInfo.schedulingIntervalHns)
    mustRecomputeSchedules = TRUE;

  // Loop, checking all the task types. These checks require floating-
  // point operations, so we save and restore the floating-point state
  // before and after this loop.
  // Note: Checking a task type involves making sure the offset to it
  // is valid and also checking all the data for that task type.

  KeSaveFloatingPointState(&floatingPointStateSave);
  for (i = 0; i < ttgInfo->numTaskTypes; ++i) {
    if (!CheckTTInfo(ttgInfo,
                     i,
                     &minOffset,
                     infoSize,
                     mustRecomputeSchedules)) {
      success = FALSE;
      break;
    }
  }
  KeRestoreFloatingPointState(&floatingPointStateSave);

  // If the check was successful, then update the file's
  // time-last-checked and scheduling-interval-when-last-checked.

  if (success) {
    KeQuerySystemTime(&ttgInfo->timeLastChecked);
    ttgInfo->schedulingIntervalHnsWhenLastChecked =
      g.timerResolutionInfo.schedulingIntervalHns;
  }

  // Return the boolean indicating if the check was successful.

  return success;
}
