// This file contains routines for finding function pointers in the
// NTDLL.DLL image loaded into memory.
//
// These routines borrow much from Gary Nebbett's book, Windows
// NT/2000 Native API Reference, pages 433-434.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "Log.h"
#include "FindFunc.h"
#include "Globals.h"

///////////////////////////////////////////////////////////////////////////
// SYSTEM DEFINITIONS
///////////////////////////////////////////////////////////////////////////

typedef enum _SYSTEM_INFORMATION_CLASS {
  SystemBasicInformation,             // 0x002C
  SystemProcessorInformation,         // 0x000C
  SystemPerformanceInformation,       // 0x0138
  SystemTimeInformation,              // 0x0020
  SystemNotImplemented1,
  SystemProcessInformation,           // 0x00C8+ per process
  SystemCallInformation,              // 0x0018 + (n * 0x0004)
  SystemConfigurationInformation,     // 0x0018
  SystemProcessorCounters,            // 0x0030 per cpu
  SystemGlobalFlag,                   // 0x0004
  SystemNotImplemented2,
  SystemModuleInformation,            // 0x0004 + (n * 0x011C)
  SystemLockInformation,              // 0x0004 + (n * 0x0024)
  SystemNotImplemented3,
  SystemPagedPoolInformation,         // checked build only
  SystemNonPagedPoolInformation,      // checked build only
  SystemHandleInformation,            // 0x0004  + (n * 0x0010)
  SystemObjectInformation,            // 0x0038+ + (n * 0x0030+)
  SystemPageFileInformation,          // 0x0018+ per page file
  SystemInstemulInformation,          // 0x0088
  SystemInvalidInfoClass1,
  SystemCacheInformation,             // 0x0024
  SystemPoolTagInformation,           // 0x0004 + (n * 0x001C)
  SystemProcessorStatistics,          // 0x0000, or 0x0018 per cpu
  SystemDpcInformation,               // 0x0014
  SystemInfo25,                       // checked build only
  SystemLoadDriver,                   // 0x0018, set mode only
  SystemUnloadDriver,                 // 0x0004, set mode only
  SystemTimeAdjustmentInformation,    // 0x000C, 0x0008 writeable
  SystemInfo29,                       // checked build only
  SystemInfo30,                       // checked build only
  SystemInfo31,                       // checked build only
  SystemCrashDumpInformation,         // 0x0004
  SystemInfo33,                       // 0x0010
  SystemCrashDumpStateInformation,    // 0x0004
  SystemDebuggerInformation,          // 0x0002
  SystemThreadSwitchInformation,      // 0x0030
  SystemRegistryQuotaInformation,     // 0x000C
  SystemAddDriver,                    // 0x0008, set mode only
  SystemPrioritySeparationInformation,// 0x0004, set mode only
  SystemInfo40,                       // not implemented
  SystemInfo41,                       // not implemented
  SystemInfo42,                       // invalid info class
  SystemInfo43,                       // invalid info class
  SystemTimeZoneInformation,          // 0x00AC
  SystemLookasideInformation,         // n * 0x0020
  MaxSystemInfoClass
} SYSTEM_INFORMATION_CLASS;

typedef struct _SYSTEM_MODULE_INFORMATION_PART {
  ULONG Reserved[2];
  PVOID Base;
  ULONG Size;
  ULONG Flags;
  USHORT Index;
  USHORT Unknown;
  USHORT LoadCount;
  USHORT ModuleNameOffset;
  CHAR ImageName[256];
} SYSTEM_MODULE_INFORMATION_PART, *PSYSTEM_MODULE_INFORMATION_PART;

typedef struct _SYSTEM_MODULE_INFORMATION {
  ULONG numModules;
  SYSTEM_MODULE_INFORMATION_PART module[1];
} SYSTEM_MODULE_INFORMATION, *PSYSTEM_MODULE_INFORMATION;

// Definitions from WINNT.H (and other sources) describing the length
// of PE file format structures

#define SIZE_OF_NT_SIGNATURE              4
#define IMAGE_SIZEOF_FILE_HEADER          20
#define IMAGE_SIZEOF_NT_OPTIONAL_HEADER   224
#define IMAGE_SIZEOF_SHORT_NAME           8
#define IMAGE_NUMBEROF_DIRECTORY_ENTRIES    16

// Directory Entries

#define IMAGE_DIRECTORY_ENTRY_EXPORT          0   // Export Directory
#define IMAGE_DIRECTORY_ENTRY_IMPORT          1   // Import Directory
#define IMAGE_DIRECTORY_ENTRY_RESOURCE        2   // Resource Directory
#define IMAGE_DIRECTORY_ENTRY_EXCEPTION       3   // Exception Directory
#define IMAGE_DIRECTORY_ENTRY_SECURITY        4   // Security Directory
#define IMAGE_DIRECTORY_ENTRY_BASERELOC       5   // Base Relocation Table
#define IMAGE_DIRECTORY_ENTRY_DEBUG           6   // Debug Directory
//      IMAGE_DIRECTORY_ENTRY_COPYRIGHT       7   // (X86 usage)
#define IMAGE_DIRECTORY_ENTRY_ARCHITECTURE    7   // Architecture Specific Data
#define IMAGE_DIRECTORY_ENTRY_GLOBALPTR       8   // RVA of GP
#define IMAGE_DIRECTORY_ENTRY_TLS             9   // TLS Directory
#define IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG    10   // Load Configuration Dir
#define IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT   11   // Bound Import Dir in hdrs
#define IMAGE_DIRECTORY_ENTRY_IAT            12   // Import Address Table
#define IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT   13   // Delay Load Import Desc
#define IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR 14   // COM Runtime descriptor

typedef struct _IMAGE_DOS_HEADER {    // DOS .EXE header
  SHORT  e_magic;                     // Magic number
  SHORT  e_cblp;                      // Bytes on last page of file
  SHORT  e_cp;                        // Pages in file
  SHORT  e_crlc;                      // Relocations
  SHORT  e_cparhdr;                   // Size of header in paragraphs
  SHORT  e_minalloc;                  // Minimum extra paragraphs needed
  SHORT  e_maxalloc;                  // Maximum extra paragraphs needed
  SHORT  e_ss;                        // Initial (relative) SS value
  SHORT  e_sp;                        // Initial SP value
  SHORT  e_csum;                      // Checksum
  SHORT  e_ip;                        // Initial IP value
  SHORT  e_cs;                        // Initial (relative) CS value
  SHORT  e_lfarlc;                    // File address of relocation table
  SHORT  e_ovno;                      // Overlay number
  SHORT  e_res[4];                    // Reserved words
  SHORT  e_oemid;                     // OEM identifier (for e_oeminfo)
  SHORT  e_oeminfo;                   // OEM information; e_oemid specific
  SHORT  e_res2[10];                  // Reserved words
  LONG   e_lfanew;                    // File address of new exe header
} IMAGE_DOS_HEADER, *PIMAGE_DOS_HEADER;

typedef struct _IMAGE_FILE_HEADER {
  SHORT   Machine;
  SHORT   NumberOfSections;
  ULONG   TimeDateStamp;
  ULONG   PointerToSymbolTable;
  ULONG   NumberOfSymbols;
  SHORT   SizeOfOptionalHeader;
  SHORT   Characteristics;
} IMAGE_FILE_HEADER, *PIMAGE_FILE_HEADER;

typedef struct _IMAGE_DATA_DIRECTORY {
  ULONG   VirtualAddress;
  ULONG   Size;
} IMAGE_DATA_DIRECTORY, *PIMAGE_DATA_DIRECTORY;

typedef struct _IMAGE_OPTIONAL_HEADER {
  //
  // Standard fields.
  //

  SHORT   Magic;
  CHAR    MajorLinkerVersion;
  CHAR    MinorLinkerVersion;
  ULONG   SizeOfCode;
  ULONG   SizeOfInitializedData;
  ULONG   SizeOfUninitializedData;
  ULONG   AddressOfEntryPoint;
  ULONG   BaseOfCode;
  ULONG   BaseOfData;

  //
  // NT additional fields.
  //

  ULONG   ImageBase;
  ULONG   SectionAlignment;
  ULONG   FileAlignment;
  SHORT   MajorOperatingSystemVersion;
  SHORT   MinorOperatingSystemVersion;
  SHORT   MajorImageVersion;
  SHORT   MinorImageVersion;
  SHORT   MajorSubsystemVersion;
  SHORT   MinorSubsystemVersion;
  ULONG   Win32VersionValue;
  ULONG   SizeOfImage;
  ULONG   SizeOfHeaders;
  ULONG   CheckSum;
  SHORT   Subsystem;
  SHORT   DllCharacteristics;
  ULONG   SizeOfStackReserve;
  ULONG   SizeOfStackCommit;
  ULONG   SizeOfHeapReserve;
  ULONG   SizeOfHeapCommit;
  ULONG   LoaderFlags;
  ULONG   NumberOfRvaAndSizes;
  IMAGE_DATA_DIRECTORY DataDirectory[IMAGE_NUMBEROF_DIRECTORY_ENTRIES];
} IMAGE_OPTIONAL_HEADER32, *PIMAGE_OPTIONAL_HEADER32;

typedef struct _IMAGE_NT_HEADERS {
  ULONG Signature;
  IMAGE_FILE_HEADER FileHeader;
  IMAGE_OPTIONAL_HEADER32 OptionalHeader;
} IMAGE_NT_HEADERS32, *PIMAGE_NT_HEADERS32;

#ifdef _WIN64
typedef IMAGE_NT_HEADERS64                  IMAGE_NT_HEADERS;
typedef PIMAGE_NT_HEADERS64                 PIMAGE_NT_HEADERS;
#else
typedef IMAGE_NT_HEADERS32                  IMAGE_NT_HEADERS;
typedef PIMAGE_NT_HEADERS32                 PIMAGE_NT_HEADERS;
#endif

// IMAGE_EXPORT_DIRECTORY is lifted from WINNT.H.  It describes the
// data directory describing the exports section of a file in PE format.

typedef struct _IMAGE_EXPORT_DIRECTORY {
  ULONG   Characteristics;
  ULONG   TimeDateStamp;
  USHORT  MajorVersion;
  USHORT  MinorVersion;
  ULONG   Name;
  ULONG   Base;
  ULONG   NumberOfFunctions;
  ULONG   NumberOfNames;
  PULONG  *AddressOfFunctions;
  PULONG  *AddressOfNames;
  PUSHORT *AddressOfNameOrdinals;
} IMAGE_EXPORT_DIRECTORY, *PIMAGE_EXPORT_DIRECTORY;

// IMAGE_SECTION_HEADER is lifted from WINNT.H.  It describes the
// section header for each section in a file in PE format.

typedef struct _IMAGE_SECTION_HEADER {
  char    Name[IMAGE_SIZEOF_SHORT_NAME];
  union {
    ULONG  PhysicalAddress;
    ULONG  VirtualSize;
  } Misc;
  ULONG   VirtualAddress;
  ULONG   SizeOfRawData;
  ULONG   PointerToRawData;
  ULONG   PointerToRelocations;
  ULONG   PointerToLinenumbers;
  USHORT  NumberOfRelocations;
  USHORT  NumberOfLinenumbers;
  ULONG   Characteristics;
} IMAGE_SECTION_HEADER, *PIMAGE_SECTION_HEADER;

NTSYSAPI
NTSTATUS
NTAPI
ZwQuerySystemInformation(
  IN SYSTEM_INFORMATION_CLASS SystemInformationClass,
  IN OUT PVOID SystemInformation,
  IN ULONG SystemInformationLength,
  OUT PULONG ReturnLength OPTIONAL
  );

///////////////////////////////////////////////////////////////////////////
// EXPORTED FUNCTIONS
///////////////////////////////////////////////////////////////////////////

// This function finds the base address of NTDLL.DLL in memory and
// returns it in the pointer location given. It returns a status
// value.

NTSTATUS FindNtdllBase (PVOID *ntdllBasePtr)
{
  ULONG requiredBufferSize, returnedBytes, i;
  PSYSTEM_MODULE_INFORMATION systemModuleInfo;
  PSYSTEM_MODULE_INFORMATION_PART currentModule;
  NTSTATUS status;

  // Query system module information with an empty buffer just to
  // figure out how many bytes we need.

  status = ZwQuerySystemInformation(SystemModuleInformation,
                                    &i,  // a fake buffer pointer
                                    0,   // buffer length: none
                                    &requiredBufferSize);
  if (!NT_SUCCESS(status) && status != STATUS_INFO_LENGTH_MISMATCH) {
    DebugPrint(("RSTask: ERROR - Could not query system module "
                "information size (%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_QUERY_SYSTEM_MODULE_INFO_SIZE,
                 status);
    return status;
  }

  // Allocate space to receive system module information.

  systemModuleInfo = ExAllocatePool(NonPagedPool, requiredBufferSize);
  if (!systemModuleInfo) {
    DebugPrint(("RSTask: ERROR - Could not allocate space for system "
                "module information of size %lu.\n", requiredBufferSize));
    BothLogOneParameter(RS_TASK_ERROR_CANT_ALLOCATE_SYSTEM_MODULE_INFO_BUF,
                        STATUS_INSUFFICIENT_RESOURCES,
                        requiredBufferSize);
    return STATUS_INSUFFICIENT_RESOURCES;
  }

  // Actually query the system module information.
  
  status = ZwQuerySystemInformation(SystemModuleInformation,
                                    systemModuleInfo,
                                    requiredBufferSize,
                                    &returnedBytes);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not query system module "
                "information (%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_QUERY_SYSTEM_MODULE_INFO,
                 status);
    return status;
  }

  // Make sure the returned number of bytes is consistent with the
  // information returned. First, make sure enough bytes were returned
  // so we can check how many modules there are. Then, make sure enough
  // bytes were returned to describe all those modules.

  if (returnedBytes < sizeof(ULONG)) {
    DebugPrint(("RSTask: System module information has an invalid size "
                "(%lu bytes).\n", returnedBytes));
    BothLogOneParameter(RS_TASK_ERROR_INVALID_SYSTEM_MODULE_INFO_SIZE,
                        STATUS_UNSUCCESSFUL,
                        returnedBytes);
    return STATUS_UNSUCCESSFUL;
  }
  if (returnedBytes != sizeof(ULONG) +
      systemModuleInfo->numModules * sizeof(SYSTEM_MODULE_INFORMATION_PART)) {
    DebugPrint(("RSTask: System module information has an invalid size "
                "(%lu for %lu modules).\n",
                returnedBytes,
                systemModuleInfo->numModules));
    BothLogOneParameter(RS_TASK_ERROR_INVALID_SYSTEM_MODULE_INFO_SIZE,
                        STATUS_UNSUCCESSFUL,
                        returnedBytes);
    return STATUS_UNSUCCESSFUL;
  }

  // Search through all the modules to find NTDLL.DLL.

  for (i = 0; i < systemModuleInfo->numModules; ++i) {
    currentModule = &systemModuleInfo->module[i];
    if (!_stricmp(currentModule->ImageName + currentModule->ModuleNameOffset,
                  "ntdll.dll")) {
      *ntdllBasePtr = currentModule->Base;
      ExFreePool(systemModuleInfo);
      return STATUS_SUCCESS;
    }
  }

  // If we didn't find it, return failure.

  ExFreePool(systemModuleInfo);
  BothLogShort(RS_TASK_ERROR_CANT_FIND_NTDLL_MODULE,
               STATUS_UNSUCCESSFUL);
  return STATUS_UNSUCCESSFUL;
}

// This function searches NTDLL.DLL in memory for the function of the
// given name and returns its address in 'addrPtr'. The 'ntdllBase'
// parameter is assumed to be the base address of NTDLL.DLL in memory.
// This function returns a status value.

NTSTATUS FindNtdllFunction (PCSTR name, PVOID ntdllBase, PVOID *addrPtr)
{
  PIMAGE_DOS_HEADER dos;
  PIMAGE_NT_HEADERS nt;
  PIMAGE_DATA_DIRECTORY expdir;
  PIMAGE_EXPORT_DIRECTORY exports;
  ULONG size, addr, i, ord;
  PULONG functions, names;
  PSHORT ordinals;

  // Make sure the base address passed is not NULL.

  if (ntdllBase == NULL) {
    DebugPrint(("RSTask: ERROR - Passed NULL ntdll base to function-"
                "finding routine.\n"));
    BothLogShort(RS_TASK_ERROR_CANT_FIND_FUNCTION_WITH_NULL_NTDLL,
                 STATUS_UNSUCCESSFUL);
    return STATUS_UNSUCCESSFUL;
  }

  // Interpret the base address as the beginning of a DOS header. Use
  // this to find the NT headers. Then, use that to find the information
  // about the export directory.

  dos = (PIMAGE_DOS_HEADER)(ntdllBase);
  nt = (PIMAGE_NT_HEADERS)((PCHAR)(ntdllBase) + dos->e_lfanew);
  expdir = nt->OptionalHeader.DataDirectory + IMAGE_DIRECTORY_ENTRY_EXPORT;

  // Find the relative address and size of the export directory.
  // Convert the relative address into a real address, 'exports'.

  addr = expdir->VirtualAddress;
  size = expdir->Size;
  exports = (PIMAGE_EXPORT_DIRECTORY)((PCHAR)(ntdllBase) + addr);

  // Find the function, ordinal, and name arrays in the export
  // directory.

  functions = (PULONG)((PCHAR)(ntdllBase) +
                       (ULONG)(exports->AddressOfFunctions));
  ordinals = (PSHORT)((PCHAR)(ntdllBase) +
                      (ULONG)(exports->AddressOfNameOrdinals));
  names = (PULONG)((PCHAR)(ntdllBase) +
                   (ULONG)(exports->AddressOfNames));

  // Loop through all the names, searching for one that matches the
  // requested one.

  for (i = 0; i < exports->NumberOfNames; ++i) {
    ord = ordinals[i];

    // For some reason, Gary Nebbett checks not only that the name
    // matches, but also that the function pointer doesn't lie in the
    // export directory itself.

    if (functions[ord] < addr || functions[ord] >= addr + size) {
      if (!strcmp((PSTR)((PCHAR)(ntdllBase) + names[i]), name)) {
        *addrPtr = (PCHAR)(ntdllBase) + functions[ord];
        return STATUS_SUCCESS;
      }
    }
  }

  // If we didn't find a matching one, return failure.

  DebugPrint(("RSTask: ERROR - Could not find function named \"%s\" in "
              "ntdll.dll.\n", name));
  BothLogShort(RS_TASK_ERROR_CANT_FIND_FUNCTION_IN_NTDLL,
               STATUS_UNSUCCESSFUL);
  return STATUS_UNSUCCESSFUL;
}
