// This file contains routines for flushing TTG file information to disk.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "Globals.h"
#include "Log.h"
#include "Flush.h"
#include "TTGFile.h"

/////////////////////////////////////////////////////////////////////////////
// MACROS
/////////////////////////////////////////////////////////////////////////////

// This macro is the number of hundreds of nanoseconds per second:
#define HNS_PER_SEC                10000000

/////////////////////////////////////////////////////////////////////////////
// WORK ITEM
/////////////////////////////////////////////////////////////////////////////

// This routine performs a flush of all TTG files, but must be called
// at passive level. It is called from a work item, the address of
// which is passed as the context parameter. This routine frees that
// work item when it is complete.

void PeriodicFlushPassiveRoutine (
  IN PDEVICE_OBJECT DeviceObject,
  IN PVOID Context
  )
{
  DebugPrint(("RSTask: Inside periodic-flush work item.\n"));

  FlushAllTTGFiles();

  DebugPrint(("RSTask: Periodic-flush work item complete.\n"));

  IoFreeWorkItem((PIO_WORKITEM) Context);
}

// This routine schedules a flush of all TTG files to occur when the
// system is at passive level.

void PeriodicFlushDpcRoutine (
  IN PKDPC Dpc,
  IN PVOID DeferredContext,
  IN PVOID SystemArgument1,
  IN PVOID SystemArgument2
  )
{
  PIO_WORKITEM workItem;

  // Allocate a new work item for this operation.

  workItem = IoAllocateWorkItem(g.interfaceDeviceObject);
  if (!workItem) {
    DebugPrint(("RSTask: ERROR - IoAllocateWorkItem failed.\n"));
    BothLogShort(RS_TASK_ERROR_CANT_ALLOCATE_PERIODIC_FLUSH_WORK_ITEM,
                 STATUS_INSUFFICIENT_RESOURCES);
    return;
  }

  // Queue the work item to occur eventually when the system is at
  // passive level. We put it on the 'delayed' queue since it's not
  // a critical operation.

  DebugPrint(("RSTask: Queueing periodic-flush work item.\n"));
  IoQueueWorkItem(workItem,
                  PeriodicFlushPassiveRoutine,
                  DelayedWorkQueue,
                  workItem); // passing work item as context
}

/////////////////////////////////////////////////////////////////////////////
// HIGH-LEVEL PERIODIC FLUSH ROUTINES
/////////////////////////////////////////////////////////////////////////////

// This routine initializes objects needed to perform a flush of all
// TTG files periodically.

void InitializePeriodicFlush (PPERIODIC_FLUSH_INFO info)
{
  KeInitializeDpc(&info->dpc,
                  PeriodicFlushDpcRoutine,
                  NULL);  // context
  KeInitializeTimerEx(&info->timer, SynchronizationTimer);
}

// This routine begins the periodic flush.

void StartPeriodicFlush (PPERIODIC_FLUSH_INFO info)
{
  LARGE_INTEGER relativeTimeHns;

  // Compute the time to first flush as a relative time in hundreds of
  // nanoseconds. This involves negating it (to indicate relative
  // rather than absolute time) and multiplying by a conversion from
  // seconds to hundreds of nanoseconds;

  relativeTimeHns.QuadPart = g.registryInfo.ttgFlushPeriodSec;
  relativeTimeHns.QuadPart *= -HNS_PER_SEC;

  // Set the periodic timer.

  KeSetTimerEx(&info->timer,
               relativeTimeHns,                           // time in hns
               g.registryInfo.ttgFlushPeriodSec * 1000,   // period in ms
               &info->dpc);
}

// This routine stops the periodic flush.

void StopPeriodicFlush (PPERIODIC_FLUSH_INFO info)
{
  KeCancelTimer(&info->timer);
}
