// This file contains functions for satisfying I/O control requests.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/ErrCodes.h"
#include "../Common/FuncPtrs.h"
#include "Ioctls.h"
#include "Policy.h"
#include "TTGFile.h"
#include "Globals.h"
#include "Log.h"
#include "Update.h"
#include "Sample.h"
#include "SimpleSched.h"
#include "Gamma.h"

/////////////////////////////////////////////////////////////////////////////
// I/O CONTROL FUNCTIONS
/////////////////////////////////////////////////////////////////////////////

// This function satisfies an I/O control request to end one task and
// begin another. The input buffer should contain two 4-byte values,
// the first of which is the task ID of the task to end and the second
// of which is the task type of the task to begin. The task to begin
// is assumed to belong to the task type group file passed as
// 'ttgFile'. This function returns the task ID of the newly begun
// task.

NTSTATUS IoctlEndTaskAndBeginAnother (
  PTTG_FILE ttgFile,
  PCHAR inBuf,
  ULONG inBufLength,
  PCHAR outBuf,
  ULONG outBufLength,
  PULONG returnedBytesPtr)
{
  NTSTATUS status;
  ULONG taskType;
  long taskId;

  // Make sure there's enough data in the input buffer.

  if (inBufLength < sizeof(ULONG) * 2) {
    DebugPrint(("RSTask: ERROR - Received too few bytes to "
                "end-and-begin-task I/O control request.\n"));
    return STATUS_INVALID_PARAMETER;
  }

  // Make sure this task type group file is not the NULL file.

  if (!ttgFile) {
    DebugPrint(("RSTask: ERROR - Received end-and-begin-task request to "
                "a handle not corresponding to a TTG file.\n"));
    return STATUS_INVALID_HANDLE;
  }

  // Extract the task ID and task type from the input buffer, and make
  // sure this task type is a valid one for this task type group file.

  taskId = * (ULONG *) inBuf;
  taskType = * (ULONG *) &inBuf[4];
  if (taskType >= ttgFile->contents->numTaskTypes) {
    DebugPrint(("RSTask: ERROR - Received an end-and-begin-task request "
                "for an invalid task type (%lu when there are only %lu).\n",
                taskType, ttgFile->contents->numTaskTypes));
    return RIGHTSPEED_ERROR_INVALID_TASK_TYPE;
  }

  // End the task of the given ID.

  EndTask(taskId);

  // Begin the new task.

  status = BeginTask(ttgFile,
                     GET_TT_INFO(ttgFile->contents, taskType),
                     &taskId);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - BeginTask failed (%x).\n", status));
    return status;
  }

  // If the begin-task operation was successful, and there's room in
  // the output buffer for that task's ID, put it there and indicate
  // that we're returning that number of bytes.

  if (outBufLength >= sizeof(long)) {
    * (long *) outBuf = taskId;
    *returnedBytesPtr = sizeof(long);
  }

  // Return success.

  return STATUS_SUCCESS;
}

// This function satisfies an I/O control request to begin a task. The
// input buffer should contain the 4-byte task type, assumed to belong
// to the task type group file opened by 'ttgFile'. This function
// returns the ID of the newly begun task in the output buffer.

NTSTATUS IoctlBeginTask (
  PTTG_FILE ttgFile,
  PCHAR inBuf,
  ULONG inBufLength,
  PCHAR outBuf,
  ULONG outBufLength,
  PULONG returnedBytesPtr
  )
{
  NTSTATUS status;
  ULONG taskType;
  long taskId;

  // Make sure there's enough data in the input buffer.

  if (inBufLength < sizeof(ULONG)) {
    DebugPrint(("RSTask: ERROR - Received too few bytes to begin-task "
                "I/O control request.\n"));
    return STATUS_INVALID_PARAMETER;
  }

  // Make sure this task type group file is not the NULL file.

  if (!ttgFile) {
    DebugPrint(("RSTask: ERROR - Received begin-task request to "
                "a handle not corresponding to a TTG file.\n"));
    return STATUS_INVALID_HANDLE;
  }

  // Extract the task type from the input buffer, and make sure this
  // task type is a valid one for this task type group file.

  taskType = * (ULONG *) inBuf;
  if (taskType >= ttgFile->contents->numTaskTypes) {
    DebugPrint(("RSTask: ERROR - Received a begin-task request "
                "for an invalid task type (%lu when there are only %lu).\n",
                taskType, ttgFile->contents->numTaskTypes));
    return RIGHTSPEED_ERROR_INVALID_TASK_TYPE;
  }

  // Begin the new task.

  status = BeginTask(ttgFile,
                     GET_TT_INFO(ttgFile->contents, taskType),
                     &taskId);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - BeginTask failed (%x).\n", status));
    return status;
  }

  // If the begin-task operation was successful, and there's room in
  // the output buffer for that task's ID, put it there and indicate
  // that we're returning that number of bytes.

  if (outBufLength >= sizeof(long)) {
    * (long *) outBuf = taskId;
    *returnedBytesPtr = sizeof(long);
  }

  // Return success.

  return STATUS_SUCCESS;
}

// This function satisfies an I/O control request to end a task.  The
// input buffer should contain the 4-byte task ID of the task to end.

NTSTATUS IoctlEndTask (
  PCHAR inBuf,
  ULONG inBufLength
  )
{
  long taskId;

  // Make sure there's enough data in the input buffer.

  if (inBufLength < sizeof(long)) {
    DebugPrint(("RSTask: ERROR - Received too few bytes to "
                "end-task I/O control request.\n"));
    return STATUS_INVALID_PARAMETER;
  }

  // Extract the task ID from the input buffer.

  taskId = * (long *) inBuf;

  // End the task of the given ID and return the status value from
  // that operation.

  return EndTask(taskId);
}

// This function satisfies an I/O control request to set policy
// parameters. The first ULONG in the input buffer is the policy type
// and the second ULONG in the input buffer is the policy value.

NTSTATUS IoctlSetGlobalPolicy (
  PCHAR inBuf,
  ULONG inBufLength
  )
{
  // Make sure there are at least two ULONGs' worth of data in the
  // input buffer.

  if (inBufLength < 8) {
    DebugPrint(("RSTask: ERROR - Not enough bytes to set-policy "
                "request.\n"));
    return STATUS_INVALID_PARAMETER;
  }

  // Set the policy using the first ULONG in the input buffer as the
  // policy type and the second ULONG as the policy value.

  return SetPolicy(* (ULONG *) inBuf, * (ULONG *) &inBuf[4]);
}

// This function satisfies an I/O control request to determine the
// deadline for a certain task type of a given task type group file.
// The input buffer should contain the task type number, and the
// output buffer should have enough room to contain the deadline
// in hundreds of nanoseconds.

NTSTATUS IoctlQueryDeadline (
  PTTG_FILE ttgFile,
  PCHAR inBuf,
  ULONG inBufLength,
  PCHAR outBuf,
  ULONG outBufLength,
  PULONG returnedBytesPtr
  )
{
  ULONG taskType;

  // Make sure there's data in the input buffer specifying the task type
  // number.

  if (inBufLength < sizeof(ULONG)) {
    DebugPrint(("RSTask: ERROR - Received too few bytes to query-deadline "
                "I/O control request.\n"));
    return STATUS_INVALID_PARAMETER;
  }

  // Make sure there's room in the output buffer to receive the
  // deadline.

  if (outBufLength < sizeof(unsigned long)) {
    DebugPrint(("RSTask: ERROR - Too little output buffer space in "
                "query-deadline request.\n"));
    return STATUS_BUFFER_TOO_SMALL;
  }

  // Make sure this task type group file is not the NULL file.

  if (!ttgFile) {
    DebugPrint(("RSTask: ERROR - Received query-deadline request to a "
                "handle not corresponding to a TTG file.\n"));
    return STATUS_INVALID_HANDLE;
  }

  // Extract the task type number from the input buffer, and make sure
  // it's a valid one for this task type group file.

  taskType = * (ULONG *) inBuf;
  if (taskType >= ttgFile->contents->numTaskTypes) {
    DebugPrint(("RSTask: ERROR - Received a query-deadline request for an "
                "invalid task type (%lu when there are only %lu).\n",
                taskType, ttgFile->contents->numTaskTypes));
    return RIGHTSPEED_ERROR_INVALID_TASK_TYPE;
  }

  // Put the deadline in the output buffer and indicate the number of
  // bytes returned in the output buffer in `*returnedBytesPtr'.

  * (unsigned long *) outBuf =
      GET_TT_INFO(ttgFile->contents, taskType)->deadlineHns;
  *returnedBytesPtr = sizeof(unsigned long);

  // Return success.

  return STATUS_SUCCESS;
}

// This function satisfies an I/O control request to change the deadline
// of a certain task type in the given task type group file. The input
// buffer should contain the task type number followed by the new deadline
// in hundreds of nanoseconds..

NTSTATUS IoctlChangeDeadline (
  PTTG_FILE ttgFile,
  PCHAR inBuf,
  ULONG inBufLength
  )
{
  KIRQL oldirql;
  ULONG taskType, newDeadlineHns;

  // Make sure there's enough information in the input buffer to specify
  // what we have to do.

  if (inBufLength < 8) {
    DebugPrint(("RSTask: ERROR - Received too few bytes to change-deadline "
                "I/O control request.\n"));
    return STATUS_INVALID_PARAMETER;
  }

  // Make sure this task type group file is not the NULL one.

  if (!ttgFile) {
    DebugPrint(("RSTask: ERROR - Received change-deadline request to a "
                "handle not corresponding to a TTG file.\n"));
    return STATUS_INVALID_HANDLE;
  }

  // Extract the task type from the input buffer and make sure it's a
  // valid one for this task type group file.

  taskType = * (ULONG *) inBuf;
  if (taskType >= ttgFile->contents->numTaskTypes) {
    DebugPrint(("RSTask: ERROR - Received a change-deadline request for an "
                "invalid task type (%lu when there are only %lu).\n",
                taskType, ttgFile->contents->numTaskTypes));
    return RIGHTSPEED_ERROR_INVALID_TASK_TYPE;
  }

  // Extract the new deadline from the input buffer.

  newDeadlineHns = * (ULONG *) &inBuf[4];

  // Perform the deadline-change while holding the spin lock for the
  // task type group file.

  KeAcquireSpinLock(&ttgFile->spinLock, &oldirql);
  ChangeDeadline(GET_TT_INFO(ttgFile->contents, taskType),
                 newDeadlineHns);
  KeReleaseSpinLock(&ttgFile->spinLock, oldirql);

  // Return success.

  return STATUS_SUCCESS;
}

// This function satisfies an I/O control request to reset the sample
// for a certain task type of the given task type group file. The input
// buffer should contain the task type number.

NTSTATUS IoctlResetSample (
  PTTG_FILE ttgFile,
  PCHAR inBuf,
  ULONG inBufLength
  )
{
  ULONG taskType;
  KIRQL oldirql;

  // Make sure there's enough data in the input buffer to specify what
  // we have to do.

  if (inBufLength < sizeof(ULONG)) {
    DebugPrint(("RSTask: ERROR - Received too few bytes to reset-sample "
                "I/O control request.\n"));
    return STATUS_INVALID_PARAMETER;
  }

  // Make sure this request is on a non-NULL task type group file.

  if (!ttgFile) {
    DebugPrint(("RSTask: ERROR - Received reset-sample request to a "
                "handle not corresponding to a TTG file.\n"));
    return STATUS_INVALID_HANDLE;
  }

  // Extract the task type number from the input buffer and make sure
  // it is valid for this task type group file.

  taskType = * (ULONG *) inBuf;
  if (taskType >= ttgFile->contents->numTaskTypes) {
    DebugPrint(("RSTask: ERROR - Received a reset-sample request for an "
                "invalid task type (%lu when there are only %lu).\n",
                taskType, ttgFile->contents->numTaskTypes));
    return RIGHTSPEED_ERROR_INVALID_TASK_TYPE;
  }

  // Perform the reset-sample operation while holding the spin lock
  // for the task type group file.

  KeAcquireSpinLock(&ttgFile->spinLock, &oldirql);
  ResetSample(GET_TT_INFO(ttgFile->contents, taskType));
  KeReleaseSpinLock(&ttgFile->spinLock, oldirql);

  // Return success.

  return STATUS_SUCCESS;
}

// This function satisfies an I/O control request to reset all the
// samples for all the task types of the given task type group
// file.

NTSTATUS IoctlResetAllSamples (
  PTTG_FILE ttgFile
  )
{
  KIRQL oldirql;
  ULONG taskType;

  // Make sure this request is on a non-NULL task type group file.

  if (!ttgFile) {
    DebugPrint(("RSTask: ERROR - Received reset-sample request to a "
                "handle not corresponding to a TTG file.\n"));
    return STATUS_INVALID_HANDLE;
  }

  // Perform the reset-sample operation on all task types while
  // holding the spin lock for the task type group file. We release
  // the spin lock between task types to give other threads a chance
  // to do something, since there may be several task types.

  for (taskType = 0; taskType < ttgFile->contents->numTaskTypes; ++taskType) {
    KeAcquireSpinLock(&ttgFile->spinLock, &oldirql);
    ResetSample(GET_TT_INFO(ttgFile->contents, taskType));
    KeReleaseSpinLock(&ttgFile->spinLock, oldirql);
  }

  // Return success.

  return STATUS_SUCCESS;
}

// This function satisfies an I/O control request to allow the given
// task type group file's corresponding on-disk file to be modified.
// In other words, it disassociates the given file from that on-disk
// file.

NTSTATUS IoctlAllowFileModification (
  PTTG_FILE ttgFile
  )
{
  // Make sure the file is not the NULL file.

  if (!ttgFile) {
    DebugPrint(("RSTask: ERROR - Received flush request to a handle "
                "not corresponding to a TTG file.\n"));
    return STATUS_INVALID_HANDLE;
  }

  return AllowFileModificationTTGFile(ttgFile);
}

// This function satisfies an I/O control request to flush the given
// task type group file.

NTSTATUS IoctlFlush (
  PTTG_FILE ttgFile
  )
{
  PVOID buf;
  NTSTATUS status;

  // Make sure the file is not the NULL file.

  if (!ttgFile) {
    DebugPrint(("RSTask: ERROR - Received flush request to a handle "
                "not corresponding to a TTG file.\n"));
    return STATUS_INVALID_HANDLE;
  }

  // Allocate a buffer big enough to hold the file contents in memory.

  buf = ExAllocatePool(NonPagedPool, ttgFile->contents->size);
  if (!buf) {
    DebugPrint(("RSTask: ERROR - Could not allocate space for TTG info "
                "copy for flush.\n"));
    SysLog(g.interfaceDeviceObject,
           RS_TASK_ERROR_CANT_ALLOCATE_SPACE_TO_FLUSH_TTG_INFO,
           STATUS_INSUFFICIENT_RESOURCES,
           0,  // unique error value
           IRP_MJ_DEVICE_CONTROL,
           RS_TASK_IOCTL_FLUSH,
           0,  // dump data size
           0); // dump data
    if (LOGGING_ERRORS)
      LogError(RS_TASK_ERROR_CANT_ALLOCATE_SPACE_TO_FLUSH_TTG_INFO);
    return STATUS_INSUFFICIENT_RESOURCES;
  }

  // Do the flush using the allocated buffer.

  status = FlushTTGFile(ttgFile, buf);

  // Free the allocated buffer.

  ExFreePool(buf);

  // Return the status value obtained when doing the flush.
  
  return status;
}

// This function satisfies an I/O control request to flush all open
// task type group files.

NTSTATUS IoctlFlushAllTTGFiles (void)
{
  FlushAllTTGFiles();
  return STATUS_SUCCESS;
}

// This function satisfies an I/O control request to obtain a copy of the
// preamble of the given open task type group file.

NTSTATUS IoctlQueryTTGFilePreamble (
  PTTG_FILE ttgFile,
  PCHAR outBuf,
  ULONG outBufLength,
  PULONG returnedBytesPtr
  )
{
  // Make sure this isn't referring to the NULL file.

  if (!ttgFile) {
    DebugPrint(("RSTask: ERROR - Received query-TTG-info-preamble "
                "request to a handle not corresponding to a TTG file.\n"));
    return STATUS_INVALID_HANDLE;
  }

  // We'll copy as much data as there is room for in the output
  // buffer, but no more than the size of the preamble. So, if the
  // output buffer length is longer than the size of the preamble,
  // just set it to the size of the preamble. The preamble consists
  // of everything in the task type group information structure other
  // than the array of task type information offsets.

  if (outBufLength > TTG_INFO_PREAMBLE_SIZE)
    outBufLength = TTG_INFO_PREAMBLE_SIZE;

  // Copy the data into the output buffer and record in 'returnedBytesPtr'
  // the number of bytes we copied.

  RtlCopyMemory(outBuf, ttgFile->contents, outBufLength);
  *returnedBytesPtr = outBufLength;

  // Return success.

  return STATUS_SUCCESS;
}

// This function satisfies an I/O control request to obtain the number
// of speeds available on this CPU.

NTSTATUS IoctlQueryNumSpeeds (
  PCHAR outBuf,
  ULONG outBufLength,
  PULONG returnedBytesPtr
  )
{
  // Make sure there's enough room in the output buffer to put the
  // number of speeds.

  if (outBufLength < sizeof(ULONG)) {
    DebugPrint(("RSTask: ERROR - Received a query-num-speeds request "
                "with too few bytes.\n"));
    return STATUS_BUFFER_TOO_SMALL;
  }

  // Put the number of speeds in the output buffer and change the
  // 'returned bytes' variable to reflect the bytes being returned.

  * (ULONG *) outBuf = g.registryInfo.numAvailableSpeeds;
  *returnedBytesPtr = sizeof(ULONG);

  // Return success.
  
  return STATUS_SUCCESS;
}

// This function satisfies an I/O control request to obtain the name of
// the task type group file directory.

NTSTATUS IoctlQueryTTGDirectory (
  PCHAR outBuf,
  ULONG outBufLength,
  PULONG returnedBytesPtr
  )
{
  ANSI_STRING ansiString;
  NTSTATUS status;
  ULONG length = g.registryInfo.ttgDirectory.Length / 2; // sizeof(WCHAR) == 2

  // Make sure there's enough room in the output buffer for all the
  // characters in the directory name plus one for the terminating
  // NULL.

  if (outBufLength < length + 1) {
    DebugPrint(("RSTask: ERROR - Received a query-TTG-directory request "
                "with too few bytes.\n"));
    return STATUS_BUFFER_TOO_SMALL;
  }

  // Initialize the ANSI string 'ansiString' to be associated with the
  // output buffer but to have initial length 0.

  ansiString.Length = 0;
  ansiString.MaximumLength = (USHORT) outBufLength;
  ansiString.Buffer = outBuf;

  // Convert the Unicode task type group directory string to ANSI,
  // effectively putting the task type group directory in the output
  // buffer.

  status = RtlUnicodeStringToAnsiString(&ansiString,
                                        &g.registryInfo.ttgDirectory,
                                        FALSE); // don't allocate destination
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Cannot copy TTG directory to ANSI string "
                "(%x).\n", status));
    SysLog(g.interfaceDeviceObject,
           RS_TASK_ERROR_CANT_CONVERT_TTG_DIRECTORY_TO_ANSI,
           status,
           0,  // unique error value
           IRP_MJ_DEVICE_CONTROL,
           RS_TASK_IOCTL_QUERY_TTG_DIRECTORY,
           0,  // dump data size
           0); // dump data
    if (LOGGING_ERRORS)
      LogError(RS_TASK_ERROR_CANT_CONVERT_TTG_DIRECTORY_TO_ANSI);
    return status;
  }

  // Make sure a NULL character terminates the ANSI string returned in
  // the output buffer.

  outBuf[length] = '\0';

  // Indicate in 'returnedBytesPtr' how many bytes are being returned:
  // one for each character in the string plus one for the terminating
  // NULL.
  
  *returnedBytesPtr = length + 1;

  // Return success.

  return STATUS_SUCCESS;
}

// This function satisfies an I/O control request to prepare RSTask
// for the uninstallation of RightSpeed.

NTSTATUS IoctlPrepareForUninstall (void)
{
  return PrepareTTGFilesInfoForUninstall(&g.openTTGFilesInfo);
}

// This function satisfies an I/O control request to obtain the
// function pointer (address) of a function. The function is specified
// by a 4-byte value in the input buffer. The function pointer should
// be placed in the output buffer.

NTSTATUS IoctlGetFunctionPointer (
  PCHAR inBuf,
  ULONG inBufLength,
  PCHAR outBuf,
  ULONG outBufLength,
  PULONG returnedBytesPtr
  )
{
  ULONG whichFnPtr;
  PVOID fnPtr;

  // Make sure the buffer sizes are sufficient.

  if (inBufLength < sizeof(ULONG)) {
    DebugPrint(("RSTask: ERROR - Received too few bytes to get-fn-pointer "
                "request.\n"));
    return STATUS_INVALID_PARAMETER;
  }

  if (outBufLength < sizeof(PVOID)) {
    DebugPrint(("RSTask: ERROR - Output buffer length too small for "
                "get-fn-pointer request.\n"));
    return STATUS_BUFFER_TOO_SMALL;
  }

  // Extract the value indicating which function pointer is being
  // requested.

  whichFnPtr = * (ULONG *) inBuf;

  // Determine the corresponding function pointer.

  switch (whichFnPtr) {
  case FN_PTR_COMPUTE_MULTISPEED_GAMMA_SCHEDULE :
    fnPtr = (PVOID) &ComputeMultispeedGammaSchedule;
    break;
    
  case FN_PTR_ADD_AGED_SAMPLE_VALUE :
    fnPtr = (PVOID) &AddAgedSampleValue;
    break;
    
  case FN_PTR_GET_AGED_SAMPLE_MEAN_AND_VARIANCE :
    fnPtr = (PVOID) &GetAgedSampleMeanAndVariance;
    break;
    
  case FN_PTR_COMPUTE_FLAT_SCHEDULE :
    fnPtr = (PVOID) &ComputeFlatSchedule;
    break;

  case FN_PTR_COMPUTE_GAMMA_SCHEDULE :
    fnPtr = (PVOID) &ComputeGammaSchedule;
    break;

  case FN_PTR_ADD_SAMPLE_VALUE :
    fnPtr = (PVOID) &AddSampleValue;
    break;

  case FN_PTR_ADD_SAMPLE_VALUE_AND_RECOMPUTE_SCHEDULE :
    fnPtr = (PVOID) &AddSampleValueAndRecomputeSchedule;
    break;

  case FN_PTR_PROCESS_SAMPLES_IN_QUEUE :
    fnPtr = (PVOID) &ProcessSamplesInQueue;
    break;

  default :
    DebugPrint(("RSTask: ERROR - Invalid function pointer requested.\n"));
    return STATUS_INVALID_PARAMETER;
  }

  // Put the function pointer in the output buffer, indicate how many
  // bytes are being returned, and return success.

  * (PVOID *) outBuf = fnPtr;
  *returnedBytesPtr = sizeof(PVOID);
  return STATUS_SUCCESS;
}

// This function satisfies an I/O control request to obtain the speeds,
// in KHz, of the speeds available.

NTSTATUS IoctlQueryAvailableSpeedsKHz (
  PCHAR outBuf,
  ULONG outBufLength,
  PULONG returnedBytesPtr
  )
{
  PULONG outBufAsUlongPtr;
  int i;

  // Make sure there's enough room in the output buffer to put the
  // speeds.

  if (outBufLength < sizeof(ULONG) * g.registryInfo.numAvailableSpeeds) {
    DebugPrint(("RSTask: ERROR - Received a query-num-speeds request "
                "with too few bytes.\n"));
    return STATUS_BUFFER_TOO_SMALL;
  }

  // Put the speeds in the output buffer and change the 'returned
  // bytes' variable to reflect the bytes being returned.

  outBufAsUlongPtr = (ULONG *) outBuf;
  for (i = 0; i < g.registryInfo.numAvailableSpeeds; ++i)
    outBufAsUlongPtr[i] = g.registryInfo.availableSpeedsInfo[i].cycPerMs;
  *returnedBytesPtr = sizeof(ULONG) * g.registryInfo.numAvailableSpeeds;

  // Return success.
  
  return STATUS_SUCCESS;
}

// This function satisfies an I/O request to obtain the number of
// hundreds of nanoseconds the idle thread used.

NTSTATUS IoctlQueryIdleThreadTime (
  PCHAR outBuf,
  ULONG outBufLength,
  PULONG returnedBytesPtr
  )
{
  NTSTATUS status;

  // Make sure the output buffer is big enough to hold a LARGE_INTEGER.

  if (outBufLength < sizeof(LARGE_INTEGER)) {
    DebugPrint(("RSTask: ERROR - Received a query-idle-thread-time request "
                "with too few bytes.\n"));
    return STATUS_BUFFER_TOO_SMALL;
  }

  // Get the time and put it in outBuf.  If this succeeds, set the number
  // of returned bytes.

  status = GetIdleThreadTime((PLARGE_INTEGER) outBuf);
  if (NT_SUCCESS(status))
    *returnedBytesPtr = sizeof(LARGE_INTEGER);

  // Return the status value of GetIdleThreadTime.

  return status;
}
