// This file contains routines for performing floating-point operations
// that are otherwise unsupported by the DDK.

#include <ntddk.h>
#include "MyFloat.h"
#include "Log.h"

/////////////////////////////////////////////////////////////////////////////
// MACROS
/////////////////////////////////////////////////////////////////////////////

// This is the bit mask describing the set of bits in the floating-point
// state that specify the rounding mode:

#define ROUNDING_MODE_BIT_MASK            0xc

// This is the inverse of the above bit mask:

#define ROUNDING_MODE_INVERSE_BIT_MASK    0xf3

// This is two to the thirtieth power:

#define TWO_TO_THIRTIETH_POWER            1073741824

// This is to to the negative thirtieth power:

#define TWO_TO_NEGATIVE_THIRTIETH_POWER   9.31322574615478515625e-10

/////////////////////////////////////////////////////////////////////////////
// GLOBAL VARIABLES
/////////////////////////////////////////////////////////////////////////////

// Defining _fltused allows floating-point operations to occur:

int _fltused;

/////////////////////////////////////////////////////////////////////////////
// ROUNDING MODE
/////////////////////////////////////////////////////////////////////////////

// This function returns the active rounding mode.

char GetDefaultRoundingMode (void)
{
  short controlWord;

  __asm {
    fstcw   controlWord    ; get the control word
  }

  return (controlWord >> 8) & ROUNDING_MODE_BIT_MASK;
}

// This function sets the active rounding mode to the given one.

void SetDefaultRoundingMode (char mode)
{
  short newControlWord;

  __asm {
    fstsw   ax                   ; get the control word in ax
    and     ah, ROUNDING_MODE_INVERSE_BIT_MASK ; remove rounding mode bits
    or      ah, mode             ; add desired rounding mode bits
    mov     newControlWord, ax   ; move new control word to memory
    fldcw   newControlWord       ; set the new control word from memory
  }
}

/////////////////////////////////////////////////////////////////////////////
// ROUNDING
/////////////////////////////////////////////////////////////////////////////

// This function rounds the given floating-point number toward zero.

long RoundTowardZero (double x)
{
  long n;
  short savedControlWord, desiredControlWord;

  __asm {
     fstcw   savedControlWord         ; save control word
     mov     ax, savedControlWord     ; move control word to ax
     or      ah, ROUNDING_MODE_BIT_MASK  ; set bits 10 and 11 to 0x11
     mov     desiredControlWord, ax   ; move new control word to memory
     fldcw   desiredControlWord       ; set control word from memory
                                          
     fld     x                        ; push x on the floating-point stack
     fistp   dword ptr n              ; pop x as a long into n
                                          
     fldcw   savedControlWord         ; restore saved control word
  }

  return n;
}

// This function rounds the given floating-point number to the nearest
// integer.

long RoundToNearest (double x)
{
  long n;
  short savedControlWord, desiredControlWord;

  __asm {
     fstcw   savedControlWord         ; save control word
     mov     ax, savedControlWord     ; move control word to ax
     and     ah, ROUNDING_MODE_INVERSE_BIT_MASK  ; set bits 10 and 11 to 0x00
     mov     desiredControlWord, ax   ; move new control word to memory
     fldcw   desiredControlWord       ; set control word from memory

     fld     x                        ; push x on the floating-point stack
     fistp   dword ptr n              ; pop x as a long into n

     fldcw   savedControlWord         ; restore saved control word
  }

  return n;
}

// This function rounds the given floating-point number down.

long RoundDown (double x)
{
  long n;
  short savedControlWord, desiredControlWord;

  __asm {
     fstcw   savedControlWord         ; save control word
     mov     ax, savedControlWord     ; move control word to ax
     and     ah, ROUNDING_MODE_INVERSE_BIT_MASK  ; remove rounding mode bits
     or      ah, ROUNDING_MODE_DOWN   ; add desired rounding mode bits
     mov     desiredControlWord, ax   ; move new control word to memory
     fldcw   desiredControlWord       ; set control word from memory
                                          
     fld     x                        ; push x on the floating-point stack
     fistp   dword ptr n              ; pop x as a long into n
                                          
     fldcw   savedControlWord         ; restore saved control word
  }

  return n;
}

// This function rounds the given floating-point number up.

long RoundUp (double x)
{
  long n;
  short savedControlWord, desiredControlWord;

  __asm {
     fstcw   savedControlWord         ; save control word
     mov     ax, savedControlWord     ; move control word to ax
     and     ah, ROUNDING_MODE_INVERSE_BIT_MASK  ; remove rounding mode bits
     or      ah, ROUNDING_MODE_UP     ; add desired rounding mode bits
     mov     desiredControlWord, ax   ; move new control word to memory
     fldcw   desiredControlWord       ; set control word from memory
                                          
     fld     x                        ; push x on the floating-point stack
     fistp   dword ptr n              ; pop x as a long into n
                                          
     fldcw   savedControlWord         ; restore saved control word
  }

  return n;
}

// This routine rounds in an unspecified manner, depending on the
// current state of the FPU control word.

long RoundDefault (double x)
{
  long n;

  __asm {
     fld     x                 ; push x on the floating-point stack
     fistp   dword ptr n       ; pop x as a long into n
  }

  return n;
}

// This function computes and returns the square root of x.

double SquareRoot (double x)
{
  __asm {
     fld     x                 ; push x on the floating-point stack
     fsqrt                     ; perform square-root operation
     fstp    x                 ; pop square root and put in x
  }

  return x;
}

// This function computes two to the power of the given number.

double TwoToThePowerOf (double x)
{
  long n;
  double excess, result;

  // Round x to an integer. It doesn't matter what the rounding mode
  // is, since all we care about is finding an integer no more than
  // 1.0 away in either direction, and any of the rounding modes will
  // do that. Then, compute the excess as x - n.

  n = RoundDefault(x);
  excess = x - n;

  // Now, x = excess + n, and excess is in the range -1.0 to 1.0.
  // This makes excess a viable candidate for the x86 instruction
  // f2xm1, allowing us to compute result = 2^(excess).

  __asm {
    fld      excess            ; push the excess on the floating-point stack
    f2xm1                      ; compute 2^(excess) - 1
    fld1                       ; push 1 on the stack
    fadd                       ; add 2^(excess) - 1 plus 1
    fstp     result            ; pop the result as 2^(excess)
  }

  // Now, we begin computing 2^x = 2^(excess + n) = 2^excess * 2^n =
  // result * 2^n.

  // If n is negative, keep multiplying the result by 2^(-30) and
  // adding 30 to n until n is nonnegative. This way, we don't have
  // to do any divisions.

  while (n < 0) {
    result *= TWO_TO_NEGATIVE_THIRTIETH_POWER;
    n += 30;
  }

  // If n is bigger than 30, keep multiplying the result by 2^30 and
  // subtracting 30 from n until n is no bigger than 30. That way we
  // can compute 1 << n without overflowing or worrying about sign
  // change.
  
  while (n > 30) {
    result *= TWO_TO_THIRTIETH_POWER;
    n -= 30;
  }

  // Now that we know 0 <= n <= 30, we can compute result * 2^n as
  // follows:
  
  return result * (1 << n);
}

// This function computes and returns 'base' to the 'exponent' power.

double Exponentiate (double base, double exponent)
{
  __asm {
    fld      exponent          ; push exponent on the floating-point stack
    fld      base              ; push base on the floating-point stack
    fyl2x                      ; compute exponent * log_2(base)
    fstp     base              ; pop the result into base
  }

  return TwoToThePowerOf(base);
}
