// This file contains routines for dealing with samples.

#include <ntddk.h>
#include "Sample.h"

/////////////////////////////////////////////////////////////////////////////
// AGED SAMPLES
/////////////////////////////////////////////////////////////////////////////

// This function empties the sample of all sample values.

void ResetAgedSample (PAGED_SAMPLE sample)
{
  sample->n = 0;
  sample->weightedCount = 0;
  sample->weightedSum = 0;
  sample->weightedSumSq = 0;
}

// This function adds a new value to the given aged sample. It returns
// the new number of sample values in the sample.

unsigned long AddAgedSampleValue (
  PAGED_SAMPLE sample,
  LONGLONG cycles
  )
{
  double value = (double) cycles;

  sample->weightedCount = sample->weightedCount * sample->agingFactor + 1;
  sample->weightedSum = sample->weightedSum * sample->agingFactor + value;
  sample->weightedSumSq = sample->weightedSumSq * sample->agingFactor +
                          value * value;
  return ++sample->n;
}

// This function computes the mean and variance of the given aged
// sample. It puts these computed values in the pointers given.

void GetAgedSampleMeanAndVariance (
  PAGED_SAMPLE sample,
  double *meanPtr,
  double *variancePtr
  )
{
  double inverseCount;

  // Because we will have to divide by the weighted count twice, it's
  // more efficient to first compute the inverse and multiply by it
  // whenever necessary.

  inverseCount = 1 / sample->weightedCount;

  // The mean is the weighted sum divided by the weighted count.
  
  *meanPtr = sample->weightedSum * inverseCount;

  // The variance is the average sum of squares minus the square of
  // the mean. (Var = E[X^2] - (E[X])^2.)
  
  *variancePtr = sample->weightedSumSq * inverseCount - *meanPtr * *meanPtr;

  // To compute the sample variance, we must multiply by n/(n-1),
  // where n is the count of sample values. Instead of doing a
  // multiply and a divide, we rearrange the formula so that we only
  // need an add and a divide. So, we use v += v/(n-1) instead of v =
  // vn/(n-1).
  
  *variancePtr += *variancePtr / (sample->n - 1);
}
