// This file contains functions for controlling the CPU speed.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/MSR.h"
#include "../Common/Schedule.h"
#include "SpeedControl.h"
#include "Globals.h"
#include "Log.h"
#include "TSC.h"
#include "Convert.h"

/////////////////////////////////////////////////////////////////////////////
// MACROS
/////////////////////////////////////////////////////////////////////////////

// This is the time to stay at the minimum speed when in the idle state
// before restoring the default CPU speed-setting policy, in hundreds of
// nanoseconds. It's equivalent to 50 ms:
#define IDLE_HNS_BEFORE_RESTORE_DEFAULT_POLICY          500000

// This is the number of microcycles per cycle:
#define UCYC_PER_CYC                                    1000000

/////////////////////////////////////////////////////////////////////////////
// FORWARD DEFINITIONS
/////////////////////////////////////////////////////////////////////////////

void ReduceSchedule (
  PMAXIMAL_SPEED_SCHEDULE schedule,
  LONG phase0Hns,
  PLONG timeToReduceHnsPtr
  );

void AddPhaseToBackwardSchedule (
  PMAXIMAL_SPEED_SCHEDULE schedule,
  ULONG speedIndex,
  LONG intervals
  );

/////////////////////////////////////////////////////////////////////////////
// SETTING SPEED
/////////////////////////////////////////////////////////////////////////////

// This function sets the speed to the one corresponding to the given
// index.

void SetSpeed (ULONG speedIndex)
{
  ULONG speedCode;

  speedCode = g.registryInfo.availableSpeedsInfo[speedIndex].speedCode;

  *** ELIDED DUE TO CONFIDENTIALITY AGREEMENTS ***

  if (LOGGING_SPEEDS)
    LogAnyEventTypeAndLong(EVENT_TYPE_SPEED_CHANGED, speedIndex);
}

/////////////////////////////////////////////////////////////////////////////
// DPC ROUTINES
/////////////////////////////////////////////////////////////////////////////

// This DPC routine restores the default CPU speed-setting policy.

VOID RestoreDefaultPolicyDpcRoutine (
  IN PKDPC Dpc,
  IN PVOID DeferredContext,
  IN PVOID SystemArgument1,
  IN PVOID SystemArgument2
  )
{
  // Restore the default speed policy.

  *** ELIDED DUE TO CONFIDENTIALITY AGREEMENTS ***

  if (LOGGING_SPEEDS)
    LogAnyEventTypeAndLong(EVENT_TYPE_SPEED_CHANGED,
                           g.registryInfo.numAvailableSpeeds);
}

// This DPC routine advances the current schedule to the next phase.

VOID AdvanceToNextPhaseDpcRoutine (
  IN PKDPC Dpc,
  IN PVOID DeferredContext,
  IN PVOID SystemArgument1,
  IN PVOID SystemArgument2
  )
{
  KIRQL oldirql;
  LARGE_INTEGER duration;
  LONGLONG previousSpeedChangeTSC;

  // Acquire the spin lock so we can access and modify the speed control
  // information.

  KeAcquireSpinLock(&g.speedControllerInfo.spinLock, &oldirql);

  // Make sure we're still in the 'schedule' state. If not, we
  // shouldn't have been called and we have nothing to do.

  if (g.speedControllerInfo.state != SPEED_CONTROLLER_STATE_SCHEDULE) {
    KeReleaseSpinLock(&g.speedControllerInfo.spinLock, oldirql);
    return;
  }

  // Increment the ucycle counter appropriately before we change
  // speed.

  previousSpeedChangeTSC = g.speedControllerInfo.lastSpeedChangeTSC;
  GetTSC(&g.speedControllerInfo.lastSpeedChangeTSC);
  g.speedControllerInfo.ucycFromScheduleStartToLastSpeedChange +=
    (g.speedControllerInfo.lastSpeedChangeTSC - previousSpeedChangeTSC) *
    g.registryInfo.availableSpeedsInfo[g.speedControllerInfo.schedule.phase->
                                       speedIndex].ucycPerTSC;

  // Set the speed to the next speed in the schedule.

  ++g.speedControllerInfo.schedule.phase;
  --g.speedControllerInfo.schedule.numPhases;
  SetSpeed(g.speedControllerInfo.schedule.phase->speedIndex);

  // If this schedule has any further phases, set the timer to go off
  // after the appropriate number of hundreds of nanoseconds relative
  // to now. (A negative number means "relative to now".) Set the DPC
  // that timer triggers to be this advance-schedule DPC.

  if (g.speedControllerInfo.schedule.phase->intervals) {
    // The number of hundreds of nanoseconds to wait is the scheduling
    // interval in hundreds of nanoseconds times the number of
    // intervals.  However, setting the timer to exactly this amount
    // poses a problem: the timer goes off at fixed intervals, and if
    // you schedule the timer to exactly when you want it to go off,
    // it might wind up triggering just after one of the fixed points
    // that it can go off. We avoid this by subtracting half the
    // scheduling interval from the timer duration.

    duration.QuadPart = g.timerResolutionInfo.negSchedulingIntervalHns;
    duration.QuadPart *= g.speedControllerInfo.schedule.phase->intervals;
    duration.QuadPart += g.timerResolutionInfo.halfSchedulingIntervalHns;
    KeSetTimer(&g.speedControllerInfo.timer,
               duration,
               &g.speedControllerInfo.phaseAdvanceDpc);
  }

  // Release the spin lock since we're done accessing and modifying
  // the speed controller information.

  KeReleaseSpinLock(&g.speedControllerInfo.spinLock, oldirql);
}

/////////////////////////////////////////////////////////////////////////////
// INITIALIZATION / TERMINATION
/////////////////////////////////////////////////////////////////////////////

// This function initializes the speed controller.

void InitializeSpeedController (PSPEED_CONTROLLER_INFO info)
{
  // Initialize the spin lock, timer, and DPC's.

  KeInitializeSpinLock(&info->spinLock);
  KeInitializeDpc(&info->phaseAdvanceDpc,
                  AdvanceToNextPhaseDpcRoutine,
                  NULL); // DeferredContext parameter passed to routine
  KeInitializeDpc(&info->restorePolicyDpc,
                  RestoreDefaultPolicyDpcRoutine,
                  NULL); // DeferredContext parameter passed to routine
  KeInitializeTimerEx(&info->timer,
                      SynchronizationTimer);

  // Initialize the current speed codes to what they are by default.

  info->currentSpeedCode1 = g.registryInfo.minSpeedCode;
  info->currentSpeedCode2 = g.registryInfo.maxSpeedCode;

  // Initialize the current schedule ID.

  info->currentScheduleId = 0;

  // Set the state to "idle". All other fields are irrelevant in this
  // state.

  info->state = SPEED_CONTROLLER_STATE_IDLE;
}

// This function closes the speed controller. This involves cancelling
// any pending operations and timers, and restoring the default speed
// policy for the CPU.

void CloseSpeedController (PSPEED_CONTROLLER_INFO info)
{
  KIRQL oldirql;

  // Acquire the spin lock controlling access to the state. Within
  // that spin lock, cancel the timer and change the state to "idle".

  KeAcquireSpinLock(&info->spinLock, &oldirql);
  KeCancelTimer(&info->timer);
  info->state = SPEED_CONTROLLER_STATE_IDLE;
  KeReleaseSpinLock(&info->spinLock, oldirql);

  // Restore the default speed policy.

  *** ELIDED DUE TO CONFIDENTIALITY AGREEMENTS ***

  if (LOGGING_SPEEDS)
    LogAnyEventTypeAndLong(EVENT_TYPE_SPEED_CHANGED,
                           g.registryInfo.numAvailableSpeeds);
}

/////////////////////////////////////////////////////////////////////////////
// SIMPLE STATES
/////////////////////////////////////////////////////////////////////////////

// This function changes the current state to 'maximum', meaning that
// the maximum speed is in force.

void UseMaximumSpeed (void)
{
  KIRQL oldirql;

  // Acquire the spin lock to change the state.

  KeAcquireSpinLock(&g.speedControllerInfo.spinLock, &oldirql);

  // Change the state. (This involves canceling the timer to avoid any
  // operations pending from another state.)

  KeCancelTimer(&g.speedControllerInfo.timer);
  g.speedControllerInfo.state = SPEED_CONTROLLER_STATE_MAXIMUM;

  // Set the speed to maximum.

  *** ELIDED DUE TO CONFIDENTIALITY AGREEMENTS ***

  if (LOGGING_SPEEDS)
    LogAnyEventTypeAndLong(EVENT_TYPE_SPEED_CHANGED,
                           g.registryInfo.maxAvailableSpeedIndex);

  // Release the spin lock since we're done changing the state.

  KeReleaseSpinLock(&g.speedControllerInfo.spinLock, oldirql);
}

// This function switches to the idle state. This means that the minimum
// speed is in force for a short period, followed by restoration of the
// default CPU speed policy.

void UseIdleSpeed (void)
{
  KIRQL oldirql;
  LARGE_INTEGER duration;

  // Acquire the spin lock to change the state.

  KeAcquireSpinLock(&g.speedControllerInfo.spinLock, &oldirql);

  // Change the state. (This involves canceling the timer to avoid any
  // operations pending from another state.)

  KeCancelTimer(&g.speedControllerInfo.timer);
  g.speedControllerInfo.state = SPEED_CONTROLLER_STATE_IDLE;

  // Set the speed to its minimum.

  *** ELIDED DUE TO CONFIDENTIALITY AGREEMENTS ***

  if (LOGGING_SPEEDS)
    LogAnyEventTypeAndLong(EVENT_TYPE_SPEED_CHANGED, 0);

  // Set the timer to go off after the appropriate number of hundreds
  // of nanoseconds relative to now. (A negative number means
  // "relative to now".) Set the DPC that timer triggers to be the
  // idle DPC.

  duration.QuadPart = -IDLE_HNS_BEFORE_RESTORE_DEFAULT_POLICY;
  KeSetTimer(&g.speedControllerInfo.timer,
             duration,
             &g.speedControllerInfo.restorePolicyDpc);

  // Release the spin lock since we're done changing the state.

  KeReleaseSpinLock(&g.speedControllerInfo.spinLock, oldirql);
}

/////////////////////////////////////////////////////////////////////////////
// SCHEDULING
/////////////////////////////////////////////////////////////////////////////

// This function causes the speed controller to enter its 'schedule'
// state, in which it follows a given speed schedule. The spinLock
// parameter to this function gives a spin lock that controls access
// to the schedule passed as a parameter.

void StartSpeedSchedule (PSPEED_SCHEDULE schedule, PKSPIN_LOCK spinLock)
{
  KIRQL oldirql, oldirql2;
  LARGE_INTEGER duration;

  if (LOGGING_SCHEDULES)
    LogSchedule(schedule);

  // Acquire the spin lock to change the state.

  KeAcquireSpinLock(&g.speedControllerInfo.spinLock, &oldirql);

  // Change the state. (This involves canceling the timer to avoid any
  // operations pending from another state.)

  KeCancelTimer(&g.speedControllerInfo.timer);
  g.speedControllerInfo.state = SPEED_CONTROLLER_STATE_SCHEDULE;

  // Copy the schedule to the speed controller's schedule buffer. When
  // accessing any part of the passed schedule, hold the passed spin
  // lock.

  KeAcquireSpinLock(spinLock, &oldirql2);
  g.speedControllerInfo.schedule.numPhases = schedule->numPhases;
  RtlCopyMemory(g.speedControllerInfo.phaseBuffer,
                schedule->phase,
                schedule->numPhases * sizeof(SPEED_SCHEDULE_PHASE));
  KeReleaseSpinLock(spinLock, oldirql2);

  // Set other fields related to the new schedule.

  g.speedControllerInfo.schedule.phase = g.speedControllerInfo.phaseBuffer;
  ++g.speedControllerInfo.currentScheduleId;
  g.speedControllerInfo.accruedTimeToReduceHns = 0;

  // Initialize the ucycle counter information.

  GetTSC(&g.speedControllerInfo.lastSpeedChangeTSC);
  g.speedControllerInfo.ucycFromScheduleStartToLastSpeedChange = 0;

  // Set the speed to the first speed in the schedule.

  SetSpeed(g.speedControllerInfo.schedule.phase->speedIndex);

  // If this schedule has any further phases, set the timer to go off
  // after the appropriate number of hundreds of nanoseconds relative
  // to now. (A negative number means "relative to now".) Set the DPC
  // that timer triggers to be the advance-schedule DPC.

  if (g.speedControllerInfo.schedule.phase->intervals) {
    // The number of hundreds of nanoseconds to wait is the scheduling
    // interval in hundreds of nanoseconds times the number of
    // intervals.  However, setting the timer to exactly this amount
    // poses a problem: the timer goes off at fixed intervals, and if
    // you schedule the timer to exactly when you want it to go off,
    // it might wind up triggering just after one of the fixed points
    // that it can go off. We avoid this by subtracting half the
    // scheduling interval from the timer duration.

    duration.QuadPart = g.timerResolutionInfo.negSchedulingIntervalHns;
    duration.QuadPart *= g.speedControllerInfo.schedule.phase->intervals;
    duration.QuadPart += g.timerResolutionInfo.halfSchedulingIntervalHns;
    KeSetTimer(&g.speedControllerInfo.timer,
               duration,
               &g.speedControllerInfo.phaseAdvanceDpc);
  }

  // Release the spin lock since we're done changing the state.

  KeReleaseSpinLock(&g.speedControllerInfo.spinLock, oldirql);
}

// This function returns the number of CPU cycles that have passed
// since the schedule started, not counting cycles spent with the
// schedule paused.

LONGLONG GetCyclesSinceScheduleStart (void)
{
  LONGLONG nowTSC, ucycSinceScheduleStart;
  KIRQL oldirql;
  ULONG currentUcycPerTSC;

  // Read all the data while holding the spin lock so that it's in a
  // consistent state.

  KeAcquireSpinLock(&g.speedControllerInfo.spinLock, &oldirql);

  // Determine the microcycles per timestamp counter at the current
  // speed.

  currentUcycPerTSC =
    g.registryInfo.availableSpeedsInfo
    [g.speedControllerInfo.schedule.phase->speedIndex].ucycPerTSC;

  switch (g.speedControllerInfo.state) {
  case SPEED_CONTROLLER_STATE_SCHEDULE :
    // Compute the microcycles since the schedule started by adding
    // the ucycles between the schedule start and the last speed change
    // and the ucycles between the last speed change and 'now'.

    GetTSC(&nowTSC);
    ucycSinceScheduleStart =
      g.speedControllerInfo.ucycFromScheduleStartToLastSpeedChange +
      (nowTSC - g.speedControllerInfo.lastSpeedChangeTSC) *
      currentUcycPerTSC;
    break;

  case SPEED_CONTROLLER_STATE_SCHEDULE_PAUSED :
  case SPEED_CONTROLLER_STATE_SCHEDULE_RESUMING :
    // If we're in the 'schedule-paused' or 'schedule-resuming' state,
    // add the ucycles between the schedule start and last speed change
    // and the ucycles between the last speed change and the time paused.

    ucycSinceScheduleStart =
      g.speedControllerInfo.ucycFromScheduleStartToLastSpeedChange +
      (g.speedControllerInfo.pauseTSC -
       g.speedControllerInfo.lastSpeedChangeTSC) *
      currentUcycPerTSC;
    break;

  default :
    // If we're in any other state, this is not a valid call, so just
    // use 0 as the answer.

    ucycSinceScheduleStart = 0;
    break;
  }

  // We're done with the speed control information, so release the
  // spin lock.

  KeReleaseSpinLock(&g.speedControllerInfo.spinLock, oldirql);

  // Convert microcycles into cycles to return the cycle count.

  return ucycSinceScheduleStart / UCYC_PER_CYC;
}

/////////////////////////////////////////////////////////////////////////////
// PAUSING / RESUMING
/////////////////////////////////////////////////////////////////////////////

// This function pauses the current schedule and goes into the 'schedule
// paused' state.

void PauseSchedule (void)
{
  KIRQL oldirql;
  ULONG speedIndex;

  // Start with a quick check to make sure this call is worthwhile.

  if (g.speedControllerInfo.state != SPEED_CONTROLLER_STATE_SCHEDULE)
    return;

  // If we're logging I/O, log that a wait for I/O is starting
  // now.

  if (LOGGING_IO)
    LogAnyEventType(EVENT_TYPE_IO_WAIT_STARTS);

  // Acquire the speed controller spin lock to change the state.

  KeAcquireSpinLock(&g.speedControllerInfo.spinLock, &oldirql);

  // If we're not in the 'schedule' state, we can't pause.

  if (g.speedControllerInfo.state != SPEED_CONTROLLER_STATE_SCHEDULE) {
    KeReleaseSpinLock(&g.speedControllerInfo.spinLock, oldirql);
    return;
  }

  // Change the state. (This involves canceling the timer to avoid any
  // schedule-advance operations pending.)

  KeCancelTimer(&g.speedControllerInfo.timer);
  g.speedControllerInfo.state = SPEED_CONTROLLER_STATE_SCHEDULE_PAUSED;

  // Record the time we paused.

  GetTSC(&g.speedControllerInfo.pauseTSC);

  // Change to the minimum speed.

  *** ELIDED DUE TO CONFIDENTIALITY AGREEMENTS ***

  if (LOGGING_SPEEDS)
    LogAnyEventTypeAndLong(EVENT_TYPE_SPEED_CHANGED, 0);

  // Release the spin lock now that we've updated the state.

  KeReleaseSpinLock(&g.speedControllerInfo.spinLock, oldirql);
}

// This function resumes a paused schedule.

void ResumeSchedule (void)
{
  KIRQL oldirql;
  LONGLONG nowTSC, pauseDurationTSC, completedPhase0TSC;
  MAXIMAL_SPEED_SCHEDULE scheduleCopy;
  LONG currentScheduleId, remainingPhase0Hns, timeToReduceHns;
  ULONG interruptedSpeedIndex;
  LARGE_INTEGER duration;

  // Start with a quick check to make sure this call is worthwhile.

  if (g.speedControllerInfo.state != SPEED_CONTROLLER_STATE_SCHEDULE_PAUSED)
    return;

  // If we're logging I/O, log that a wait for I/O is stopping now.

  if (LOGGING_IO)
    LogAnyEventType(EVENT_TYPE_IO_WAIT_STOPS);

  ////////////////////////////////////////////////////////////////////
  // Copy enough information to allow us to compute a new schedule. //
  ////////////////////////////////////////////////////////////////////

  // Acquire the speed controller spin lock to change the state.

  KeAcquireSpinLock(&g.speedControllerInfo.spinLock, &oldirql);

  // If we're not in the 'schedule paused' state, we can't resume.

  if (g.speedControllerInfo.state != SPEED_CONTROLLER_STATE_SCHEDULE_PAUSED) {
    KeReleaseSpinLock(&g.speedControllerInfo.spinLock, oldirql);
    return;
  }

  // Determine how long we've been paused and what speed we were
  // running at before we paused.

  GetTSC(&nowTSC);
  pauseDurationTSC = nowTSC - g.speedControllerInfo.pauseTSC;
  interruptedSpeedIndex = g.speedControllerInfo.schedule.phase->speedIndex;

  // If there's only one phase left in the schedule, all we have to do
  // is restore the speed for that phase and we'll be back in the
  // 'schedule' state.

  if (g.speedControllerInfo.schedule.phase->intervals == 0) {
    g.speedControllerInfo.state = SPEED_CONTROLLER_STATE_SCHEDULE;

    // To reflect the fact that we were paused for a timestamp count
    // of 'pauseDurationTSC', we increment the last speed-change time
    // by that amount. This is a lie, but it yields the proper results
    // and it's efficient.
    
    g.speedControllerInfo.lastSpeedChangeTSC += pauseDurationTSC;
    SetSpeed(interruptedSpeedIndex);
    KeReleaseSpinLock(&g.speedControllerInfo.spinLock, oldirql);
    return;
  }

  // Change the state to 'schedule resuming'.

  g.speedControllerInfo.state = SPEED_CONTROLLER_STATE_SCHEDULE_RESUMING;

  // Copy locally all the information we will need to compute and
  // apply a new schedule.

  scheduleCopy.numPhases = g.speedControllerInfo.schedule.numPhases;
  RtlCopyMemory(scheduleCopy.phase,
                g.speedControllerInfo.schedule.phase,
                scheduleCopy.numPhases * sizeof(SPEED_SCHEDULE_PHASE));
  currentScheduleId = g.speedControllerInfo.currentScheduleId;
  completedPhase0TSC = g.speedControllerInfo.pauseTSC -
                       g.speedControllerInfo.lastSpeedChangeTSC;
  timeToReduceHns = g.speedControllerInfo.accruedTimeToReduceHns;

  // Release the spin lock now that we've updated the state.

  KeReleaseSpinLock(&g.speedControllerInfo.spinLock, oldirql);

  /////////////////////////////
  // Compute a new schedule. //
  /////////////////////////////

  // Determine how many hundreds of nanoseconds to reduce by. This is
  // the accrued amount of time we have left over from the last time
  // we resumed this schedule, plus the amount of time we spent paused.

  timeToReduceHns += (long) TSCToHns(pauseDurationTSC);

  // Compute how many hundreds of nanoseconds remain of the first
  // phase of the old schedule. This is the length of that phase minus
  // the amount of time spent running before we paused. If this amount
  // is negative, treat it as zero.

  remainingPhase0Hns = scheduleCopy.phase[0].intervals *
                       g.timerResolutionInfo.schedulingIntervalHns -
                       (long) TSCToHns(completedPhase0TSC);
  if (remainingPhase0Hns < 0)
    remainingPhase0Hns = 0;

  // Reduce the schedule by 'timeToReduceHns'. Set 'timeToReduceHns' to
  // the amount of hundreds of nanoseconds we still have to reduce by
  // (the amount by which we fell short of our goal); note that this
  // amount may be negative if we go over our goal.

  if (LOGGING_SCHEDULES)
    LogSchedule((PSPEED_SCHEDULE) &scheduleCopy);

  ReduceSchedule(&scheduleCopy,
                 remainingPhase0Hns,
                 &timeToReduceHns);

  if (LOGGING_SCHEDULES)
    LogSchedule((PSPEED_SCHEDULE) &scheduleCopy);

  ///////////////////////////////////////
  // Put the new schedule into effect. //
  ///////////////////////////////////////

  // Acquire the spin lock to change the state.

  KeAcquireSpinLock(&g.speedControllerInfo.spinLock, &oldirql);

  // Make sure we're still working on the same schedule now. If we
  // aren't, just return.

  if (g.speedControllerInfo.state !=SPEED_CONTROLLER_STATE_SCHEDULE_RESUMING ||
      g.speedControllerInfo.currentScheduleId != currentScheduleId) {
    KeReleaseSpinLock(&g.speedControllerInfo.spinLock, oldirql);
    return;
  }

  // Change the state to just 'schedule'.

  g.speedControllerInfo.state = SPEED_CONTROLLER_STATE_SCHEDULE;

  // Copy the schedule to the speed controller's schedule buffer.

  g.speedControllerInfo.schedule.numPhases = scheduleCopy.numPhases;
  g.speedControllerInfo.schedule.phase = g.speedControllerInfo.phaseBuffer;
  RtlCopyMemory(g.speedControllerInfo.phaseBuffer,
                scheduleCopy.phase,
                scheduleCopy.numPhases * sizeof(SPEED_SCHEDULE_PHASE));

  // Update the time still left to reduce in the schedule to the
  // new left-over time.

  g.speedControllerInfo.accruedTimeToReduceHns = timeToReduceHns;

  // Update the microcycle counter to reflect the time spent running
  // before we paused.

  g.speedControllerInfo.ucycFromScheduleStartToLastSpeedChange +=
    completedPhase0TSC *
    g.registryInfo.availableSpeedsInfo[interruptedSpeedIndex].ucycPerTSC;
  GetTSC(&g.speedControllerInfo.lastSpeedChangeTSC);

  // Set the speed to the first speed in the schedule.

  SetSpeed(scheduleCopy.phase[0].speedIndex);

  // If this schedule has any further phases, set the timer to go off
  // after the appropriate number of hundreds of nanoseconds relative
  // to now. (A negative number means "relative to now".) Set the DPC
  // that timer triggers to be the advance-schedule DPC.

  if (scheduleCopy.phase[0].intervals) {
    // The number of hundreds of nanoseconds to wait is the scheduling
    // interval in hundreds of nanoseconds times the number of
    // intervals.  However, setting the timer to exactly this amount
    // poses a problem: the timer goes off at fixed intervals, and if
    // you schedule the timer to exactly when you want it to go off,
    // it might wind up triggering just after one of the fixed points
    // that it can go off. We avoid this by subtracting half the
    // scheduling interval from the timer duration.

    duration.QuadPart = g.timerResolutionInfo.negSchedulingIntervalHns;
    duration.QuadPart *= scheduleCopy.phase[0].intervals;
    duration.QuadPart += g.timerResolutionInfo.halfSchedulingIntervalHns;
    KeSetTimer(&g.speedControllerInfo.timer,
               duration,
               &g.speedControllerInfo.phaseAdvanceDpc);
  }

  // Release the spin lock since we're done changing the state.

  KeReleaseSpinLock(&g.speedControllerInfo.spinLock, oldirql);
}

// This function reduces the given schedule by a certain amount of
// time, pointed to by 'timeToReduceHnsPtr'. It assumes the given
// schedule has at least two phases. It ignores the 'intervals' field
// of its first phase, and instead treats that phase as having length
// 'phase0Hns' hundreds of nanoseconds. When finished, it sets the
// contents of 'timeToReduceHnsPtr' by the amount still needing to be
// reduced by; this amount may be negative if this function "overdoes"
// the reduction.

void ReduceSchedule (
  PMAXIMAL_SPEED_SCHEDULE schedule,
  LONG phase0Hns,
  PLONG timeToReduceHnsPtr
  )
{
  MAXIMAL_SPEED_SCHEDULE newSchedule;
  PSPEED_SCHEDULE_PHASE phase, sourcePhase, destPhase;
  LONG reducibleHns, oldHns, denominator;
  LONG totalIntervals, newIntervals, rightIntervals, leftIntervals;
  LONGLONG newHns, oldHucyc, x;
  short leftSetting, rightSetting, i;
  ULONG oldCycPerMs, newCycPerMs;

  // Determine how much reducible time is in the schedule. Do this by
  // adding up the intervals in all but the first and last phases,
  // then adding the corresponding amount of hundreds of nanoseconds
  // to the hundreds of nanoseconds in the first phase. We don't use
  // the last phase, which is of indefinite length and is not to be
  // reduced.

  totalIntervals = 0;
  for (phase = &schedule->phase[1]; phase->intervals; ++phase)
    totalIntervals += phase->intervals;
  reducibleHns =
    phase0Hns + totalIntervals * g.timerResolutionInfo.schedulingIntervalHns;

  // Initialize the new schedule, which we'll fill in backward, with
  // the last phase of the schedule.

  newSchedule.numPhases = 1;
  newSchedule.phase[0].speedIndex =
    schedule->phase[schedule->numPhases - 1].speedIndex;
  newSchedule.phase[0].intervals = 0;

  // Loop backward through the phases of the schedule, starting with
  // the next-to-last one (since the last one is of indefinite length
  // and we won't be changing it). For each phase, determine how many
  // intervals it should now be, and use that information to convert
  // it into a new reduced phase. It may be that it becomes two new
  // phases, since there may be an ideal speed for it that falls between
  // two consecutive available speeds.

  for (i = schedule->numPhases - 2; i >= 0; --i) {
    phase = &schedule->phase[i];

    // Determine how many hundreds of nanoseconds the old phase has.
    // If it doesn't have any time, skip this phase.

    oldHns = (i > 0) ?
      phase->intervals * g.timerResolutionInfo.schedulingIntervalHns :
      phase0Hns;
    if (oldHns < 1)
      continue;

    // Determine the speed the old phase used in cycles per
    // millisecond.

    oldCycPerMs = CYC_PER_MS_OF_INDEX(phase->speedIndex);

    // Determine the number of hundreds of microcycles the old phase
    // performed, by multiplying the speed in cycles per millisecond by
    // the length in hundreds of nanoseconds.

    oldHucyc = oldCycPerMs;
    oldHucyc *= oldHns;

    // Determine how many hundreds of nanoseconds we want this phase
    // to be reduced to. We have 'reducibleHns' total hundreds of
    // nanoseconds in the schedule, which we want reduced by
    // '*timeToReduceHnsPtr', so this should be:
    //
    // oldHns * (reducibleHns - *timeToReduceHnsPtr) / reducibleHns.

    newHns = oldHns;
    newHns *= (reducibleHns - *timeToReduceHnsPtr);
    newHns /= reducibleHns;

    // The phase must be an integral number of intervals, so round to
    // the nearest number of intervals (but don't use 0 intervals).

    newIntervals = (long)
      ((newHns + g.timerResolutionInfo.halfSchedulingIntervalHns) /
       g.timerResolutionInfo.schedulingIntervalHns);
    if (newIntervals < 1)
      newIntervals = 1;
    newHns = newIntervals * g.timerResolutionInfo.schedulingIntervalHns;

    // Now that we know the length of the new phase in hundreds of
    // nanoseconds and the number of hundreds of microcycles it should
    // perform, we can compute the speed in cycles per millisecond.

    newCycPerMs = (long) (oldHucyc / newHns);

    if (newCycPerMs >= g.registryInfo.maxWorthwhileCycPerMs) {
      // If the new speed is higher than the maximum worthwhile speed,
      // just use the maximum worthwhile speed for this phase.

      // Compute the new length as the old work divided by the new
      // speed. Round the new length to the nearest multiple of the
      // interval length.

      newHns = oldHucyc / g.registryInfo.maxWorthwhileCycPerMs;
      newIntervals = (long)
        ((newHns + g.timerResolutionInfo.halfSchedulingIntervalHns) /
         g.timerResolutionInfo.schedulingIntervalHns);
      newHns = newIntervals * g.timerResolutionInfo.schedulingIntervalHns;

      // Push the new phase onto the new schedule.

      AddPhaseToBackwardSchedule(&newSchedule,
                                 g.registryInfo.maxWorthwhileSpeedIndex,
                                 newIntervals);
    }
    else {
      // If the new speed is below the maximum worthwhile speed,
      // search for the pair of worthwhile speeds that flank it and
      // use those speeds in new phases.
      
      // In some cases, rounding can cause the new speed to be less than
      // the old speed. Obviously, pausing can't make you want to run
      // more slowly, so just bump up the speed to the old speed.

      if (newCycPerMs < oldCycPerMs)
        newCycPerMs = oldCycPerMs;

      // If for some reason the desired speed is less than the minimum
      // worthwhile speed, bump it up to the minimum worthwhile speed.

      if (newCycPerMs < g.registryInfo.minWorthwhileCycPerMs)
        newCycPerMs = g.registryInfo.minWorthwhileCycPerMs;

      // Search for the 'left' and 'right' worthwhile settings that
      // flank the new speed desired, 'newCycPerMs'.

      for (leftSetting = 0;
           CYC_PER_MS_OF_SETTING(leftSetting + 1) <= newCycPerMs;
           ++leftSetting)
        ;
      rightSetting = leftSetting + 1;

      // The fraction of the intervals spent at the 'left' setting
      // should be approximately the fraction:
      //
      //             cycPerMsAtRightSetting - newCycPerMs
      //         ----------------------------------------------
      //         cycPerMsAtRightSetting - cycPerMsAtLeftSetting
      //
      // of the new intervals.

      denominator = CYC_PER_MS_OF_SETTING(rightSetting) -
                    CYC_PER_MS_OF_SETTING(leftSetting);
      x = newIntervals;
      x *= (CYC_PER_MS_OF_SETTING(rightSetting) - newCycPerMs);
      x += (denominator >> 1);
      x /= denominator;
      leftIntervals = (long) x;
      rightIntervals = newIntervals - leftIntervals;

      // Now push the two new phases onto the new schedule.

      AddPhaseToBackwardSchedule(&newSchedule,
                                 INDEX_OF_SETTING(rightSetting),
                                 rightIntervals);
      AddPhaseToBackwardSchedule(&newSchedule,
                                 INDEX_OF_SETTING(leftSetting),
                                 leftIntervals);
    }

    // At this point, we've reduced the length of this phase from
    // 'oldHns' to 'newHns'. So we have oldHns - newHns less time to
    // reduce by, but we also have oldHns less reducible time.

    *timeToReduceHnsPtr -= (long)(oldHns - newHns);
    reducibleHns -= oldHns;
  }

  // Copy the backwards schedule 'newSchedule' to 'schedule'.

  schedule->numPhases = newSchedule.numPhases;
  for (i = 0, sourcePhase = &newSchedule.phase[newSchedule.numPhases - 1],
         destPhase = schedule->phase;
       i < newSchedule.numPhases;
       ++i, ++destPhase, --sourcePhase)
    RtlCopyMemory(destPhase, sourcePhase, sizeof(SPEED_SCHEDULE_PHASE));
}

/////////////////////////////////////////////////////////////////////////////
// UTILITY FUNCTIONS
/////////////////////////////////////////////////////////////////////////////

// This function adds a phase to the beginning of the given backward
// schedule (which will then be the highest-numbered phase of the
// schedule, since the schedule is backward). The new phase is given
// as a speed index and a number of intervals. It assumes that the
// backward schedule passed has at least one phase in it already.
// It assumes that speeds are being added in decreasing order.

void AddPhaseToBackwardSchedule (
  PMAXIMAL_SPEED_SCHEDULE schedule,
  ULONG speedIndex,
  LONG intervals
  )
{
  PSPEED_SCHEDULE_PHASE bottomPhase;

  // If there are no intervals to add, we're done.

  if (intervals < 1)
    return;

  bottomPhase = &schedule->phase[schedule->numPhases - 1];
  if (speedIndex == bottomPhase->speedIndex) {
    // If the bottom phase uses the same speed index as this new phase,
    // just add the intervals of the new phase to that phase. If it's
    // an 'indefinite' phase, don't add any intervals to it.

    if (bottomPhase->intervals)
      bottomPhase->intervals += intervals;
  }
  else if (speedIndex < bottomPhase->speedIndex) {
    // If the new speed is slower than the bottom speed, then make a
    // new phase below the bottom speed for the new phase.

    ++bottomPhase;
    bottomPhase->speedIndex = speedIndex;
    bottomPhase->intervals = intervals;
    ++schedule->numPhases;
  }
  else {
    // If the new speed is higher than the bottom speed, search for the
    // last phase that's at least as high as it, and pretend it's the
    // same speed as that phase.

    while (bottomPhase > schedule->phase &&
           bottomPhase->speedIndex < speedIndex)
      --bottomPhase;
    if (bottomPhase->intervals)
      bottomPhase->intervals += intervals;
  }
}
