// This file contains routines for dealing with task type group files.

#include <ntddk.h>
#include "../Common/Drivers.h"
#include "../Common/ErrCodes.h"
#include "../Common/Parameters.h"
#include "Check.h"
#include "Globals.h"
#include "Log.h"
#include "Sleep.h"
#include "TTGFile.h"

/////////////////////////////////////////////////////////////////////////////
// MACROS
/////////////////////////////////////////////////////////////////////////////

// This is the maximum number of times in a row we should sleep while
// waiting for a task type group file we want to open to be opened or
// closed by some other thread:
#define MAX_TIMES_TO_WAIT_FOR_OPEN_OR_CLOSE         10

// This is the time to wait for another thread that's loading or
// unloading a task type group file we want to open to finish its
// load or unload operation:
#define MILLISECONDS_TO_WAIT_FOR_OPEN_OR_CLOSE      500

/////////////////////////////////////////////////////////////////////////////
// FORWARD DEFINITIONS
/////////////////////////////////////////////////////////////////////////////

int GetTTGFileHashBucketNum (PUNICODE_STRING name);
PTTG_FILE FindInTTGFileHash (PUNICODE_STRING name, int bucketNum);
void InsertInTTGFileHash (
  PTTG_FILE ttgFile,
  PUNICODE_STRING name,
  int bucketNum
  );
void RemoveFromTTGFileHash (PTTG_FILE ttgFile);

PVOID AllocateTTGFile (
  IN POOL_TYPE PoolType,
  IN ULONG NumberOfBytes,
  IN ULONG Tag
  );
NTSTATUS MyWriteBackTTGFileContents (
  HANDLE fileHandle,
  PVOID buf,
  ULONG bufLength
  );
BOOLEAN ReallocateBuffer (
  PVOID *bufPtr,
  int *currentSizePtr,
  int desiredSize
  );

/////////////////////////////////////////////////////////////////////////////
// OPEN FILE INFO
/////////////////////////////////////////////////////////////////////////////

// This function initializes the structures needed to keep track of open
// task type group files. This includes the lookaside list that is used
// for allocating and freeing such files, as well as the hash table that
// is used for searching for an open file of a given name. It returns
// a status value indicating if it was successful.

NTSTATUS InitializeOpenTTGFilesInfo (POPEN_TTG_FILES_INFO info)
{
  NTSTATUS status;
  OBJECT_ATTRIBUTES objectAttributes;
  IO_STATUS_BLOCK ioStatusBlock;
  int i;

  // Open the task type group directory handle.

  InitializeObjectAttributes(&objectAttributes,
                             &g.registryInfo.ttgDirectory,
                             OBJ_CASE_INSENSITIVE | OBJ_KERNEL_HANDLE,
                             NULL,  // root directory
                             NULL); // security descriptor
  status = ZwOpenFile(&info->ttgDirectoryHandle,
                      GENERIC_READ | GENERIC_WRITE | FILE_TRAVERSE,
                      &objectAttributes,
                      &ioStatusBlock,
                      FILE_SHARE_READ | FILE_SHARE_WRITE,
                      FILE_DIRECTORY_FILE);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Can't open task type group directory "
                "handle (%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_OPEN_TTG_DIRECTORY,
                 status);
    return status;
  }

  // Initialize the remaining fields of the structure.

  KeInitializeSpinLock(&info->spinLock);
  ExInitializeNPagedLookasideList(&info->lookasideList,
                                  AllocateTTGFile,
                                  NULL,   // use generic free function
                                  0,      // no flags (must be zero anyway)
                                  sizeof(TTG_FILE),
                                  'ttgf', // pool tag
                                  0);     // depth must be zero
  info->numOpenFiles = 0;
  info->uninstalling = FALSE;

  // Initialize the hash bucket by making all bucket heads NULL.

  for (i = 0; i < NUM_TTG_FILE_HASH_BUCKETS; ++i)
    info->hashBucketHead[i] = NULL;

  // Return success.

  return STATUS_SUCCESS;
}

// This function frees any resources obtained for the task type group
// file lists.

void CloseOpenTTGFilesInfo (POPEN_TTG_FILES_INFO info)
{
  // Check for any still-open files.

  if (info->numOpenFiles > 0) {
    DebugPrint(("RSTask: ERROR - Closing TTG files when %d "
                "are still open.\n", info->numOpenFiles));
    BothLogShort(RS_TASK_ERROR_CLOSING_TTG_FILE_LISTS_WHEN_SOME_ARE_OPEN,
                 STATUS_UNSUCCESSFUL);
  }

  // Free resources associated with the lookaside list.

  ExDeleteNPagedLookasideList(&info->lookasideList);

  // Close the task type group directory handle if it's open.

  if (info->ttgDirectoryHandle)
    ZwClose(info->ttgDirectoryHandle);
}

// This function prepares for the uninstallation of RightSpeed. It
// disassociates all task type group files from their on-disk files,
// and closes the task type group directory handle. It returns a
// status value.

NTSTATUS PrepareTTGFilesInfoForUninstall (POPEN_TTG_FILES_INFO info)
{
  PTTG_FILE ttgFile;
  KIRQL oldirql;
  NTSTATUS status;
  HANDLE ttgDirectoryHandle;
  int bucketNum, timesWaited = 0;

  // Indicate that we're uninstalling and close the task type group
  // directory handle.
  
  KeAcquireSpinLock(&g.openTTGFilesInfo.spinLock, &oldirql);
  g.openTTGFilesInfo.uninstalling = TRUE;
  ttgDirectoryHandle = info->ttgDirectoryHandle;
  info->ttgDirectoryHandle = NULL;
  KeReleaseSpinLock(&g.openTTGFilesInfo.spinLock, oldirql);

  if (ttgDirectoryHandle)
    ZwClose(ttgDirectoryHandle);

  // Loop through all the buckets, disassociating each file in each
  // bucket from its on-disk file.

  KeAcquireSpinLock(&g.openTTGFilesInfo.spinLock, &oldirql);
  for (bucketNum = 0; bucketNum < NUM_TTG_FILE_HASH_BUCKETS; ++bucketNum) {
    // Disassociate the first file in the bucket repeatedly until
    // there is no file in the bucket. (Disassociating a file removes
    // it from its bucket.)

    while ((ttgFile = g.openTTGFilesInfo.hashBucketHead[bucketNum]) != NULL) {
      // If the reference count on this file is 0, it's in the process
      // of being opened or closed. In either case, we must wait for
      // it to no longer be so before proceeding.

      if (ttgFile->referenceCount == 0) {
        KeReleaseSpinLock(&g.openTTGFilesInfo.spinLock, oldirql);
        ++timesWaited;
        if (timesWaited < MAX_TIMES_TO_WAIT_FOR_OPEN_OR_CLOSE) {
          DebugPrint(("RSTask: Sleeping to wait for open or close before "
                      "disassociating.\n"));
          Sleep(MILLISECONDS_TO_WAIT_FOR_OPEN_OR_CLOSE);
        }
        else {
          DebugPrint(("RSTask: ERROR - Waited repeatedly for file to "
                      "open or close without it happening.\n"));
          BothLogShort(RS_TASK_ERROR_UNINSTALL_DURING_OPEN_OR_CLOSE,
                       RIGHTSPEED_ERROR_TTG_OPENING_OR_CLOSING);
          return RIGHTSPEED_ERROR_TTG_OPENING_OR_CLOSING;
        }
      }
      else {
        timesWaited = 0;

        // In order to do the actual disassociation, which might take a
        // while, we have to release the spin lock controlling access to
        // the information about open TTG files. To ensure that the TTG
        // file pointer remains valid throughout the flush, we reference
        // it before releasing the lock.

        ReferenceTTGFile(ttgFile);
        KeReleaseSpinLock(&g.openTTGFilesInfo.spinLock, oldirql);
        status = AllowFileModificationTTGFile(ttgFile);
        DereferenceTTGFile(ttgFile);

        // If the disassociation failed, e.g., because a flush was in
        // progress, fail this entire call.

        if (!NT_SUCCESS(status))
          return status;
      }

      // Before proceeding to the next loop iteration, reacquire
      // the spin lock.

      KeAcquireSpinLock(&g.openTTGFilesInfo.spinLock, &oldirql);
    }
  }

  // Release the spin lock obtained at the beginning of the loop and
  // return success.
  
  KeReleaseSpinLock(&g.openTTGFilesInfo.spinLock, oldirql);
  return STATUS_SUCCESS;
}

/////////////////////////////////////////////////////////////////////////////
// OPENING
/////////////////////////////////////////////////////////////////////////////

// This function opens a task type group file given its name relative
// to the task type group directory and puts a pointer to the open
// task type group file information in the given location. It returns
// a status value. It assumes that the file name is in nonpaged memory
// and is null-terminated. It only changes the contents of
// 'ttgFilePtr' on successful completion.

NTSTATUS OpenTTGFileByName (PUNICODE_STRING fileName, PTTG_FILE *ttgFilePtr)
{
  KIRQL oldirql;
  NTSTATUS status;
  TTG_FILE *ttgFile;
  UNICODE_STRING uppercaseFileName;
  int bucketNum, timesWaited = 0;
  BOOLEAN alreadyOpen = FALSE, openingOrClosing = FALSE, uninstalling = FALSE;
  IO_STATUS_BLOCK ioStatusBlock;
  OBJECT_ATTRIBUTES objectAttributes;
  FILE_STANDARD_INFORMATION fileInfo;
  ULONG fileSize;

  DebugPrint(("RSTask: Opening TTG file named \"%ws\".\n", fileName->Buffer));

  // Create an uppercase version of the file name from the file name.
  // We use this for hashing. This way, if the file is opened twice
  // with the same name but different case, we still note them as
  // matching. Note that RtlUpcaseUnicodeString allocates space for
  // the new string, so we must eventually free it.

  status = RtlUpcaseUnicodeString(&uppercaseFileName,
                                  fileName,
                                  TRUE); // allocate space for this string
  if (!NT_SUCCESS(status) ||
      uppercaseFileName.Buffer == NULL) {  // paranoid check for null buffer
    DebugPrint(("RSTask: ERROR - Could not convert file name to uppercase "
                "(%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_CONVERT_FILE_NAME_TO_UPPERCASE,
                 status);
    return status;
  }

  // Determine which hash bucket this file name appears in. We do
  // this now because it could take a while and we aren't holding
  // a spin lock now.

  bucketNum = GetTTGFileHashBucketNum(&uppercaseFileName);
  DebugPrint(("RSTask: Hash number is %d.\n", bucketNum));

  // Loop until we have either:
  // (a) found that the file is not open and put an entry for the
  //     new file in the hash table;
  // (b) found that the file is already open and not in a state of
  //     being opened or closed by some other thread; or
  // (c) been through the loop several times unsuccessfully, meaning
  //     that the file is open and in a perpetual state of being
  //     opened or closed by some other thread.

  while (1) {
    // Access the information about open task type group file only
    // while holding the spin lock.

    KeAcquireSpinLock(&g.openTTGFilesInfo.spinLock, &oldirql);

    // Search the hash bucket for any open file that matches this
    // uppercase file name.

    ttgFile = FindInTTGFileHash(&uppercaseFileName, bucketNum);

    if (ttgFile) {
      // If we found a matching file, check to see if it's being
      // opened or closed by some other thread. If it isn't, then
      // reference the file to indicate that our caller is now also
      // referencing it. If it is being opened or closed by some other
      // thread, then set 'openingOrClosing' to TRUE to indicate we
      // should wait for the open/close to be finished and then return
      // to the beginning of the loop.
      
      if (ttgFile->referenceCount > 0) {
        ReferenceTTGFile(ttgFile);
        alreadyOpen = TRUE;
      }
      else
        openingOrClosing = TRUE;
    }
    else {
      // If we found no matching file, allocate a new one and put
      // it in the hash table. However, if we're uninstalling
      // RightSpeed, don't create a new task type group file.

      if (g.openTTGFilesInfo.uninstalling)
        uninstalling = TRUE;
      else {
        ttgFile =
          ExAllocateFromNPagedLookasideList(&g.openTTGFilesInfo.lookasideList);
        if (ttgFile) {
          InsertInTTGFileHash(ttgFile, &uppercaseFileName, bucketNum);
          ++g.openTTGFilesInfo.numOpenFiles;
        }
      }
    }

    // We're now done accessing the open task type group file lists,
    // so we can release the spin lock.
    
    KeReleaseSpinLock(&g.openTTGFilesInfo.spinLock, oldirql);

    // Unless we found that the file is being opened or closed by some
    // other thread, we're done.

    if (!openingOrClosing)
      break;

    // If we found a matching file, but it's being opened or closed
    // by some other thread, then wait for that thread to finish and
    // then loop again. If we've already waited many times, though,
    // just fail.

    if (timesWaited < MAX_TIMES_TO_WAIT_FOR_OPEN_OR_CLOSE) {
      DebugPrint(("RSTask: Sleeping to wait for open or close.\n"));
      Sleep(MILLISECONDS_TO_WAIT_FOR_OPEN_OR_CLOSE);
      openingOrClosing = FALSE; // reset this for the next loop iteration
      ++timesWaited;            // increment the count of times we've waited
    }
    else {
      DebugPrint(("RSTask: WARNING - Attempt to open a file that's being "
                  "opened or closed by another call.\n"));
      BothLogShort(RS_TASK_WARNING_ATTEMPT_TO_OPEN_FILE_DURING_OPEN_OR_CLOSE,
                   RIGHTSPEED_ERROR_TTG_OPENING_OR_CLOSING);
      RtlFreeUnicodeString(&uppercaseFileName);
      return RIGHTSPEED_ERROR_TTG_OPENING_OR_CLOSING;
    }
  }

  // If the file was already open, we're done. Just record the task
  // type group file pointer in the appropriate location and free the
  // now-unneeded allocated uppercase file name.

  if (alreadyOpen) {
    DebugPrint(("RSTask: Found already-loaded file at %x.\n", ttgFile));
    RtlFreeUnicodeString(&uppercaseFileName);
    *ttgFilePtr = ttgFile;
    return STATUS_SUCCESS;
  }

  // If we were uninstalling RightSpeed, return failure.

  if (uninstalling) {
    DebugPrint(("RSTask: ERROR - Can't open task type group file because "
                "RightSpeed is being uninstalled.\n"));
    return RIGHTSPEED_ERROR_UNINSTALLING;
  }

  // If we weren't able to allocate a new task type group file structure,
  // return failure.

  if (!ttgFile) {
    DebugPrint(("RSTask: ERROR - Could not allocate a new TTG file "
                "from lookaside list.\n"));
    BothLogShort(RS_TASK_ERROR_CANT_ALLOCATE_TTG_FILE,
                 STATUS_INSUFFICIENT_RESOURCES);
    RtlFreeUnicodeString(&uppercaseFileName);
    return STATUS_INSUFFICIENT_RESOURCES;
  }

  // At this point, we know we allocated a new task type group file
  // structure and inserted it in the hash. The reference count is
  // zero, meaning we have exclusive access to the structure and other
  // threads will know that it's in a state of being opened or closed
  // by us. Thus, if we fail at this point, we must close the task
  // type group file.
  //
  // Note: The uppercase file name has by now been copied to
  // ttgFile->name, so we don't have to worry about freeing it;
  // closing the task type group file structure will do this.

  // First, open the file on disk. Use a kernel handle so that the file
  // handle can be used from any process context. Share it for reading
  // but *not* for writing so that no one can change the file while we
  // have it open. Note that closing this TTG file will automatically
  // close this handle since we're storing it in ttgFile->fileHandle.

  DebugPrint(("RSTask: Opening TTG file at %x with ZwOpenFile.\n",
              ttgFile));
  InitializeObjectAttributes(&objectAttributes,
                             fileName,
                             OBJ_CASE_INSENSITIVE | OBJ_KERNEL_HANDLE,
                             g.openTTGFilesInfo.ttgDirectoryHandle, // root dir
                             NULL); // security descriptor
  status = ZwOpenFile(&ttgFile->fileHandle,
                      GENERIC_READ | GENERIC_WRITE | SYNCHRONIZE,
                      &objectAttributes,
                      &ioStatusBlock,
                      FILE_SHARE_READ, // don't allow others to write
                      FILE_NON_DIRECTORY_FILE | FILE_SEQUENTIAL_ONLY |
                        FILE_WRITE_THROUGH | FILE_SYNCHRONOUS_IO_ALERT);
  if (!NT_SUCCESS(status)) {
    // If we fail for some reason other than simply that the file
    // doesn't exist, that error is worth noting and logging.

    if (status != STATUS_OBJECT_NAME_NOT_FOUND) {
      DebugPrint(("RSTask: ERROR - Cannot open TTG file (%x).\n", status));
      BothLogShort(RS_TASK_ERROR_CANT_OPEN_TTG_FILE,
                   status);
    }

    ttgFile->fileHandle = NULL;  // paranoia - ensure ZwOpenFile didn't set it
    CloseTTGFile(ttgFile);
    return status;
  }

  // Now, determine the size of the file so that we can allocate memory
  // to store its contents. To do this, we get the 'standard information'
  // about the file from ZwQueryInformationFile.

  status = ZwQueryInformationFile(ttgFile->fileHandle,
                                  &ioStatusBlock,
                                  &fileInfo,
                                  sizeof(fileInfo),
                                  FileStandardInformation);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Cannot get size of file (%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_GET_TTG_FILE_SIZE,
                 status);
    CloseTTGFile(ttgFile);
    return status;
  }

  // If the file is bigger than the maximum valid size, fail.

  if (fileInfo.EndOfFile.QuadPart > MAX_TTG_INFO_SIZE) {
    DebugPrint(("RSTask: ERROR - File is too big (%ld).\n",
                fileInfo.EndOfFile.LowPart));
    BothLogOneParameter(RS_TASK_ERROR_TTG_FILE_SIZE_TOO_BIG,
                        status,
                        fileInfo.EndOfFile.LowPart);
    CloseTTGFile(ttgFile);
    return status;
  }

  // Record the size of the file in 'fileSize'. (Since it's no bigger
  // than MAX_TTG_INFO_SIZE, we know it fits in 4 bytes, and we can
  // just copy the low part of the file size into 'fileSize'.)
  
  fileSize = fileInfo.EndOfFile.LowPart;

  // Allocate memory for the file contents now that we know how big
  // they are. Note that closing the file will automatically free this
  // memory since we're storing the pointer to it in ttgFile->contents.

  ttgFile->contents = ExAllocatePool(NonPagedPool, fileSize);
  if (!ttgFile->contents) {
    DebugPrint(("RSTask: ERROR - Could not allocate space for TTG info.\n"));
    BothLogShort(RS_TASK_ERROR_CANT_ALLOCATE_TTG_INFO_SPACE,
                 status);
    CloseTTGFile(ttgFile);
    return STATUS_INSUFFICIENT_RESOURCES;
  }

  // Read the whole file into the allocated memory space.

  status = ZwReadFile(ttgFile->fileHandle,
                      NULL,  // no event
                      NULL,  // no APC routine
                      NULL,  // no APC context
                      &ioStatusBlock,
                      ttgFile->contents,
                      fileSize,
                      NULL,  // byte offset: "use current file pointer"
                      NULL); // no key
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not read data from file (%x).\n",
                status));
    BothLogShort(RS_TASK_ERROR_CANT_READ_TTG_FILE,
                 status);
    CloseTTGFile(ttgFile);
    return status;
  }
  if (ioStatusBlock.Information != fileSize) {
    DebugPrint(("RSTask: ERROR - Read %d bytes instead of %d.\n",
                ioStatusBlock.Information, fileSize));
    BothLogTwoParameters(RS_TASK_ERROR_READ_WRONG_NUMBER_OF_TTG_FILE_BYTES,
                         STATUS_UNSUCCESSFUL,
                         ioStatusBlock.Information,
                         fileSize);
    CloseTTGFile(ttgFile);
    return STATUS_UNSUCCESSFUL;
  }

  // Perform checks on the file contents to make sure they're valid
  // and to update the task type group info to reflect the current
  // machine parameters. If these checks fail, we fail.

  if (!CheckTTGInfo(ttgFile->contents, fileSize)) {
    DebugPrint(("RSTask: ERROR - TTG info was invalid.\n"));
    CloseTTGFile(ttgFile);
    return RIGHTSPEED_ERROR_BAD_TTG_FILE_FORMAT;
  }

  // Now that we know the task type group info is OK, allow writeback
  // of its contents to disk when we eventually close it.

  ttgFile->writeBack = TRUE;

  // Release our exclusive access to the task type group file structure
  // and indicate that our caller is referencing it, all by the simple
  // act of setting the reference count to 1.

  ttgFile->referenceCount = 1;

  // Return the task type group file pointer successfully.
  
  DebugPrint(("RSTask: Opened file successfully at %x.\n", ttgFile));
  *ttgFilePtr = ttgFile;
  return STATUS_SUCCESS;
}

// This function opens the task type group file corresponding to the
// given RSTask file object. It derives the task type group file name
// from the name of the RSTask file object. For instance, if someone
// opens \\.\RSTask\Foo\Bar.ttg, then the file object will have the
// name \Foo\Bar.ttg. We then open the task type group file of this
// name, which is on disk in DIR\Foo\Bar.ttg where DIR is the task
// type group file directory. As a special case, if someone opens
// just plain \\.\RSTask, the file name we see will be empty; we then
// use the special task type group file 'NULL'.

NTSTATUS OpenTTGFile (PFILE_OBJECT fileObject)
{
  NTSTATUS status;
  UNICODE_STRING fileName;

  DebugPrint(("RSTask: Opening TTG file for new file object at %x.\n",
              fileObject));

  // Make sure we know the file object so that we can record the
  // reference to the TTG file there.

  if (fileObject == NULL) {
    DebugPrint(("RSTask: ERROR - Could not get file object pointer "
                "to fill in reference.\n"));
    BothLogShort(RS_TASK_ERROR_CANT_GET_FILE_OBJECT_POINTER,
                 STATUS_UNSUCCESSFUL);
    return STATUS_UNSUCCESSFUL;
  }

  // Use NULL as the TTG file pointer by default.

  fileObject->FsContext = NULL;

  // If there is no file name, or there's nothing more to the file
  // name than the leading backslash, we're done; we just use NULL as
  // the TTG file pointer.

  if (fileObject->FileName.Length < 2) {
    DebugPrint(("RSTask: Received request for the null TTG file.\n"));
    return STATUS_SUCCESS;
  }

  // Try to open the new TTG file using the given file name without
  // the leading backslash. Return the success value.

  fileName.Length = fileObject->FileName.Length - 2; // sizeof(WCHAR) == 2
  fileName.MaximumLength = fileObject->FileName.MaximumLength - 2;
  fileName.Buffer = fileObject->FileName.Buffer + 1; // advance past backslash
  return OpenTTGFileByName(&fileName,
                           (PPTTG_FILE) &fileObject->FsContext);
}

/////////////////////////////////////////////////////////////////////////////
// FLUSHING
/////////////////////////////////////////////////////////////////////////////

// This function flushes the contents of the given file 'ttgFile' to
// disk. The other parameter to this function is 'buf', a buffer in
// nonpaged pool that contains at least enough room to hold the
// in-memory contents of the file. This function does not assume that
// it has exclusive access to the file or that the file has writeback
// enabled; it checks for these conditions before proceeding. This
// function assumes that the caller has referenced the file, so it
// won't be closed while this function is running.

NTSTATUS FlushTTGFile (PTTG_FILE ttgFile, PVOID buf)
{
  KIRQL oldirql;
  NTSTATUS status = STATUS_SUCCESS;

  DebugPrint(("RSTask: Flushing TTG file %x using copy buffer %x.\n",
              ttgFile, buf));

  // Acquire the spin lock for the file to make sure that we have
  // exclusive access to the file for writing and that writeback is
  // enabled.

  KeAcquireSpinLock(&ttgFile->spinLock, &oldirql);
  if (!ttgFile->writeBack || !ttgFile->fileHandle)
    status = RIGHTSPEED_ERROR_WRITEBACK_DISABLED;
  else if (ttgFile->flushInProgress)
    status = RIGHTSPEED_ERROR_FLUSH_IN_PROGRESS;
  else {
    // Everything is OK. While we have the spin lock, copy the task
    // type group information from memory into 'buf' so that we have a
    // consistent snapshot of the information. Also, acquire exclusive
    // access to the on-disk file by setting ttgFile->flushInProgress
    // to TRUE.

    RtlCopyMemory(buf, ttgFile->contents, ttgFile->contents->size);
    ttgFile->flushInProgress = TRUE;
  }
  KeReleaseSpinLock(&ttgFile->spinLock, oldirql);

  // If we can't do the flush for some reason, return the status value.

  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Cannot flush TTG file %x (%x).\n",
                ttgFile, status));
    return status;
  }

  // Call MyWriteBackTTGFileContents to do the actual writing to the
  // file, now that we've obtained exclusive access to write the file.

  status = MyWriteBackTTGFileContents(ttgFile->fileHandle,
                                      buf,
                                      ttgFile->contents->size);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not write back TTG file %x to "
                "flush it (%x).\n", ttgFile, status));
    // Don't return; we haven't released exclusive access to the
    // on-disk file yet!
  }

  // Release exclusive access to the on-disk file.

  ttgFile->flushInProgress = FALSE;

  // Return the status value corresponding to how the writeback went.

  return status;
}

// This function flushes all open task type group files to disk.

void FlushAllTTGFiles (void)
{
  PVOID buf = NULL;  // buffer starts out NULL, meaning unallocated
  long bufSize = 0;  // current size of allocated buffer
  int bucketNum;
  KIRQL oldirql;
  PTTG_FILE ttgFile, next;

  DebugPrint(("RSTask: Flushing all TTG files.\n"));

  // Loop through all the buckets, flushing all files in each bucket.
  // While accessing the hash table, we must hold the spin lock
  // controlling access to the information about open TTG files.

  KeAcquireSpinLock(&g.openTTGFilesInfo.spinLock, &oldirql);
  for (bucketNum = 0; bucketNum < NUM_TTG_FILE_HASH_BUCKETS; ++bucketNum) {

    // Loop through all the files in the bucket.

    for (ttgFile = g.openTTGFilesInfo.hashBucketHead[bucketNum];
         ttgFile != NULL;
         ttgFile = next) {

      // If the reference count on this file is 0, it's in the process
      // of being opened or closed. In either case, we shouldn't flush
      // it, so we just go on to the next entry in the hash table.

      if (ttgFile->referenceCount == 0) {
        next = ttgFile->hashTableInfo.next;
        continue;
      }

      // In order to do the actual flush, which might take a while, we
      // have to release the spin lock controlling access to the
      // information about open TTG files. To ensure that the TTG file
      // pointer remains valid throughout the flush, we reference it
      // before releasing the lock.

      ReferenceTTGFile(ttgFile);
      KeReleaseSpinLock(&g.openTTGFilesInfo.spinLock, oldirql);

      // To flush, first make sure the buffer is big enough to hold
      // the file contents, then do the flush if the allocation was
      // successful.

      if (ReallocateBuffer(&buf, &bufSize, ttgFile->contents->size))
        FlushTTGFile(ttgFile, buf);

      // Now that the flush is complete, reacquire the spin lock
      // controlling access to information about open files.

      KeAcquireSpinLock(&g.openTTGFilesInfo.spinLock, &oldirql);

      // Now that we're back holding the spin lock controlling access
      // to the lists of open TTG files, we can determine the next TTG
      // file in the hash in this bucket so we know what to flush
      // next. Once we've done that, we'll be done with the TTG file
      // and can dereference it. Note that we can't rely on the TTG
      // file still being in the hash table; it might have been
      // removed while we didn't hold the spin lock. (This is rather
      // unlikely, as it would require someone to have disassociated
      // the file in a very narrow window of time, but it could
      // happen.) So, we should check to make sure the hash table
      // information is valid before following the next-in-hash
      // pointer blindly.

      next = ttgFile->hashTableInfo.valid ? ttgFile->hashTableInfo.next : NULL;
      DereferenceTTGFile(ttgFile);
    }
  }
  KeReleaseSpinLock(&g.openTTGFilesInfo.spinLock, oldirql);

  // If the buffer is non-NULL, it must be freed.

  if (buf)
    ExFreePool(buf);
}

/////////////////////////////////////////////////////////////////////////////
// DISASSOCIATION
/////////////////////////////////////////////////////////////////////////////

// This routine disassociates a given open task type group file from
// the on-disk file so that the on-disk file can be modified. This
// involves doing three things: (a) disabling writeback on the file,
// (b) removing the open file from the hash table of open files so
// that further opens to the same file name will not match this file,
// and (c) closing the file handle so that the on-disk file can be
// opened by another handle for writing. This function assumes that
// the caller has referenced the file, so it won't be closed while
// this function is running.

NTSTATUS AllowFileModificationTTGFile (PTTG_FILE ttgFile)
{
  KIRQL oldirql;
  HANDLE fileHandle;

  DebugPrint(("RSTask: Allowing file modification of TTG file %x.\n",
              ttgFile));

  // Acquire exclusive access to the file structure so that we can
  // read and modify the writeback state and the file handle.

  KeAcquireSpinLock(&ttgFile->spinLock, &oldirql);

  // If someone is in the process of flushing the file to disk, we
  // cannot close the file handle and thus cannot proceed.

  if (ttgFile->flushInProgress) {
    KeReleaseSpinLock(&ttgFile->spinLock, oldirql);
    DebugPrint(("RSTask: ERROR - Cannot allow overwrite of TTG file %x "
                "because a flush is in progress.\n", ttgFile));
    return RIGHTSPEED_ERROR_FLUSH_IN_PROGRESS;
  }

  // Disable writeback on the file.

  ttgFile->writeBack = FALSE;

  // Set the file handle to NULL since we will be closing it. (But,
  // store the file handle in 'fileHandle' before we set it to NULL so
  // we know what to close!)

  fileHandle = ttgFile->fileHandle;
  ttgFile->fileHandle = NULL;

  // We're done reading and modifying the file structure, so we can
  // release its spin lock.
  
  KeReleaseSpinLock(&ttgFile->spinLock, oldirql);

  // Remove the file from the hash list of open files so that further
  // opens to the same file name will not match this file.

  KeAcquireSpinLock(&g.openTTGFilesInfo.spinLock, &oldirql);
  RemoveFromTTGFileHash(ttgFile);
  KeReleaseSpinLock(&g.openTTGFilesInfo.spinLock, oldirql);

  // Close the file handle if there was one open for the file.

  if (fileHandle)
    ZwClose(fileHandle);

  // Return success.

  return STATUS_SUCCESS;
}

//////////////////////////////////////////////////////////////////////////////
// CLOSING
//////////////////////////////////////////////////////////////////////////////

// This function closes the given task type group file. It frees all
// resources obtained for this file and removes the file from all the
// open-file lists. This function assumes that there are no
// outstanding references to the file, so it won't be accessed by
// anyone else while this function is running. (Others may access the
// reference count and hash table info of this file, though. They'll
// notice the file is in the process of being closed from the fact
// that the reference count is zero.) Eventually, this function frees
// the file structure to the lookaside list, so it makes sure before
// doing so that the structure is in an initialized state.

void CloseTTGFile (PTTG_FILE ttgFile)
{
  KIRQL oldirql;
  NTSTATUS status;

  DebugPrint(("RSTask: Closing file at %x.\n", ttgFile));

  // Free the buffer used to hold the file contents, if one has been
  // so allocated. However, before doing this, first write back any
  // changes that have been made to the file contents while in memory.

  if (ttgFile->contents) {
    if (ttgFile->writeBack && ttgFile->fileHandle)
      MyWriteBackTTGFileContents(ttgFile->fileHandle,
                                 ttgFile->contents,
                                 ttgFile->contents->size);
    ExFreePool(ttgFile->contents);
    ttgFile->contents = NULL;
  }

  // Close the file handle associated with this file if we have it
  // open, to allow others to eventually open this file for writing.

  if (ttgFile->fileHandle) {
    status = ZwClose(ttgFile->fileHandle);
    if (!NT_SUCCESS(status)) {
      DebugPrint(("RSTask: ERROR - Could not close file (%x).\n", status));
      BothLogShort(RS_TASK_ERROR_CANT_CLOSE_TTG_FILE,
                   status);
    }
    ttgFile->fileHandle = NULL;
  }

  // Remove the file from the hash table and decrement the count of
  // open files, both while holding the spin lock controlling access
  // to information about open TTG files.

  KeAcquireSpinLock(&g.openTTGFilesInfo.spinLock, &oldirql);
  RemoveFromTTGFileHash(ttgFile);
  --g.openTTGFilesInfo.numOpenFiles;
  KeReleaseSpinLock(&g.openTTGFilesInfo.spinLock, oldirql);

  // Clean up any remaining fields of the file structure and free it
  // to the lookaside list for later allocation.

  ttgFile->writeBack = FALSE;
  ExFreeToNPagedLookasideList(&g.openTTGFilesInfo.lookasideList, ttgFile);

  DebugPrint(("RSTask: Successfully closed file.\n"));
}

/////////////////////////////////////////////////////////////////////////////
// REFERENCING / DEREFERENCING
/////////////////////////////////////////////////////////////////////////////

// This function references the given task type group file.

void ReferenceTTGFile (PTTG_FILE ttgFile)
{
  InterlockedIncrement(&ttgFile->referenceCount);
}

// This function dereferences the given task type group file. If this
// causes its reference count to drop to 0, it also closes that file.

void DereferenceTTGFile (PTTG_FILE ttgFile)
{
  KIRQL oldirql;
  LONG referenceCount;

  if (InterlockedDecrement(&ttgFile->referenceCount) == 0) {

    // Make sure the reference count is still 0 while holding the spin
    // lock controlling it. (Otherwise, we might notice it's 0 just
    // before someone else increments it while holding the spin lock.)

    KeAcquireSpinLock(&g.openTTGFilesInfo.spinLock, &oldirql);
    referenceCount = ttgFile->referenceCount;
    KeReleaseSpinLock(&g.openTTGFilesInfo.spinLock, oldirql);

    // If the reference count is indeed still 0, close the file.

    if (referenceCount == 0)
      CloseTTGFile(ttgFile);
  }
}

/////////////////////////////////////////////////////////////////////////////
// HASH TABLE
/////////////////////////////////////////////////////////////////////////////

// This function returns a bucket number based deterministically on
// only the given file name. It returns a nonnegative integer less
// than NUM_TTG_FILE_HASH_BUCKETS.

int GetTTGFileHashBucketNum (PUNICODE_STRING name)
{
  int sum = 0, i, length = name->Length / 2;  // sizeof(WCHAR) == 2
  WCHAR *p = name->Buffer;

  // Add up the values of all the characters.

  for (i = 0; i < length; ++i)
    sum += (int)(*p++);

  // Then, return the sum modulo NUM_TTG_FILE_HASH_BUCKETS.

  return sum % NUM_TTG_FILE_HASH_BUCKETS;
}

// This function searches the hash table for a file that has the given
// name. The 'bucketNum' parameter indicates the bucket the file will
// be in if it is in any bucket. It returns a pointer to the file it
// finds, or NULL if it finds no matching file.

PTTG_FILE FindInTTGFileHash (PUNICODE_STRING name, int bucketNum)
{
  PTTG_FILE ttgFile = g.openTTGFilesInfo.hashBucketHead[bucketNum];

  // Loop through all the files in this bucket searching for one whose
  // name matches. When we find one that so matches, return it.

  while (ttgFile) {
    if (name->Length == ttgFile->hashTableInfo.name.Length &&
        RtlEqualMemory(name->Buffer, ttgFile->hashTableInfo.name.Buffer,
                       name->Length))
      return ttgFile;
    else
      ttgFile = ttgFile->hashTableInfo.next;
  }

  // We haven't found one, so return NULL.

  return NULL;
}

// This function inserts the given file into the hash table using the
// given name. The 'bucketNum' parameter indicates which bucket should
// be used. This function assumes the bucket number was computed
// properly. It also assumes no file by this name is already in the
// hash table. It assumes that the name has been allocated by
// RtlUpcaseUnicodeString, and thus refers to a buffer that must later
// be freed by RtlFreeUnicodeString when this file is removed from the
// hash.

void InsertInTTGFileHash (
  PTTG_FILE ttgFile,
  PUNICODE_STRING name,
  int bucketNum
  )
{
  PPTTG_FILE bucketHeadPtr = &g.openTTGFilesInfo.hashBucketHead[bucketNum];
  PTTG_FILE formerFirstInBucket = *bucketHeadPtr;
  PTTG_FILE_HASH_INFO hashTableInfo = &ttgFile->hashTableInfo;

  // Have the former first file in the bucket point backward to this
  // file, and have this file point forward to that former first file.
  // Note that the former first file may be NULL.

  if (formerFirstInBucket)
    formerFirstInBucket->hashTableInfo.prev = &(hashTableInfo->next);
  hashTableInfo->next = formerFirstInBucket;

  // Have the bucket head point forward to this file, and have this
  // file point backward to the bucket head.

  hashTableInfo->prev = bucketHeadPtr;
  *bucketHeadPtr = ttgFile;

  // Copy the name string to the hash table info. Then it becomes the
  // hash table's responsibility to ultimately free the name buffer.

  RtlCopyMemory(&hashTableInfo->name, name, sizeof(UNICODE_STRING));

  // Indicate that the file is now in the hash table and that all the
  // hash table information fields are valid.

  hashTableInfo->valid = TRUE;
}

// This function removes the given file from the hash table. Note that
// if the file is actually in the hash table, removing it involves
// freeing the name buffer associated with the hash table entry.

void RemoveFromTTGFileHash (PTTG_FILE ttgFile)
{
  PTTG_FILE_HASH_INFO hashTableInfo = &ttgFile->hashTableInfo;

  // If it's not in the hash table, we're done.

  if (!hashTableInfo->valid)
    return;

  // Have its predecessor point forward to its successor, and have its
  // successor point backward to its predecessor. Note that it may not
  // have a successor.

  *(hashTableInfo->prev) = hashTableInfo->next;
  if (hashTableInfo->next)
    hashTableInfo->next->hashTableInfo.prev = hashTableInfo->prev;

  // Free the name buffer associated with this hash table entry.

  RtlFreeUnicodeString(&hashTableInfo->name);

  // Indicate that the file is no longer in a hash table and that the
  // hash table information is invalid.

  hashTableInfo->valid = FALSE;
}

/////////////////////////////////////////////////////////////////////////////
// UTILITY FUNCTIONS
/////////////////////////////////////////////////////////////////////////////

// This routine allocates a new task type group file description
// structure. The parameter format is so that it can be used with
// NT lookaside lists.

PVOID AllocateTTGFile (
  IN POOL_TYPE PoolType,
  IN ULONG NumberOfBytes,
  IN ULONG Tag
  )
{
  PTTG_FILE ttgFile;

  // Allocate the space.

  ttgFile = (PTTG_FILE) ExAllocatePoolWithTag(PoolType, NumberOfBytes, Tag);
  if (ttgFile) {
    // Initialize the structure.

    ttgFile->hashTableInfo.valid = FALSE;
    ttgFile->referenceCount = 0;
    KeInitializeSpinLock(&ttgFile->spinLock);
    ttgFile->contents = NULL;
    ttgFile->fileHandle = NULL;
    ttgFile->writeBack = FALSE;
    ttgFile->flushInProgress = FALSE;
  }

  // Return the allocated pointer (possibly NULL if the allocation failed).

  return ttgFile;
}

// This function writes the data pointed to by 'buf' of length
// 'bufLength' to the beginning of the open file referred to by
// 'fileHandle'.

NTSTATUS MyWriteBackTTGFileContents (
  HANDLE fileHandle,
  PVOID buf,
  ULONG bufLength
  )
{
  FILE_POSITION_INFORMATION filePositionInfo;
  IO_STATUS_BLOCK ioStatusBlock;
  NTSTATUS status;

  DebugPrint(("RSTask: Writing back TTG information.\n"));

  // First, set the file pointer to the beginning of the file.

  filePositionInfo.CurrentByteOffset.QuadPart = 0;
  status = ZwSetInformationFile(fileHandle,
                                &ioStatusBlock,
                                &filePositionInfo,
                                sizeof(filePositionInfo),
                                FilePositionInformation);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Could not set file position to start of "
                "file in order to write (%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_SET_TTG_FILE_POSITION_FOR_WRITING,
                 status);
    return status;
  }

  // Next, write the bytes to the file.

  status = ZwWriteFile(fileHandle,
                       NULL,  // no event
                       NULL,  // no APC routine
                       NULL,  // no APC context
                       &ioStatusBlock,
                       buf,
                       bufLength,
                       NULL,  // byte offset: "use current file position"
                       NULL); // no key
  if (!NT_SUCCESS(status)) {
    DebugPrint(("RSTask: ERROR - Failed to write TTG file (%x).\n", status));
    BothLogShort(RS_TASK_ERROR_CANT_WRITE_TTG_FILE,
                 status);
    return status;
  }
  if (ioStatusBlock.Information != bufLength) {
    DebugPrint(("RSTask: ERROR - Wrote wrong number of bytes to TTG "
                "file (%lu instead of %lu).\n",
                ioStatusBlock.Information,
                bufLength));
    BothLogTwoParameters(RS_TASK_ERROR_WROTE_WRONG_NUMBER_OF_TTG_FILE_BYTES,
                         STATUS_UNSUCCESSFUL,
                         ioStatusBlock.Information,
                         bufLength);
    return STATUS_UNSUCCESSFUL;
  }

  // Return success.

  return STATUS_SUCCESS;
}

// This function helps ensure that an allocated buffer is at least a
// given size. It maintains the invariant that *bufPtr is an allocated
// buffer of size '*currentSizePtr' or else is NULL. It returns a
// boolean indicating whether it is able to make '*bufPtr' non-NULL
// and '*currentSizePtr' at least 'desiredSize'.

BOOLEAN ReallocateBuffer (
  PVOID *bufPtr,
  int *currentSizePtr,
  int desiredSize
  )
{
  // If a buffer is already allocated and it's big enough, just
  // return success.

  if (*bufPtr != NULL && *currentSizePtr >= desiredSize)
    return TRUE;

  // If a buffer is already allocated, we know it isn't big enough
  // so we free it.

  if (*bufPtr != NULL)
    ExFreePool(*bufPtr);

  // Now, allocate a buffer at least size 'desiredSize'. So that we
  // don't have to go through this allocation/deallocation too many
  // times, we actually allocate twice as many bytes as we need.
    
  *bufPtr = ExAllocatePool(NonPagedPool, desiredSize * 2);
  if (*bufPtr == NULL) {
    DebugPrint(("RSTask: ERROR - Could not reallocate buffer to be "
                "of size %d.\n", desiredSize * 2));
    BothLogOneParameter(RS_TASK_ERROR_CANT_REALLOCATE_BUFFER,
                        STATUS_INSUFFICIENT_RESOURCES,
                        desiredSize * 2);
    *currentSizePtr = 0;
    return FALSE;
  }

  // At this point, we've successfully allocated a buffer of size
  // desiredSize * 2. Set '*currentSizePtr' to reflect this and
  // return success.
  
  *currentSizePtr = desiredSize * 2;
  return TRUE;
}
