// This file contains functions exported by this library.

#include <windows.h>
#include <string>
#include "../Common/ErrCodes.h"
#include "../Common/Policy.h"
#include "Globals.h"
#include "Log.h"
#include "TTGFile.h"
#include "RSTaskHandle.h"

using namespace std;

#define DLLEXPORT __declspec( dllexport )

//////////////////////////////////////////////////////////////////////////////
// UTILITY FUNCTIONS
//////////////////////////////////////////////////////////////////////////////

namespace {

  // This function copies the string 'source' to the buffer 'buf' of size
  // 'bufSize'. It copies as much as possible while leaving room for a
  // terminating null, then terminates the buffer with a null.

  DWORD CopyIntoBuffer (char *buf, DWORD bufSize, char *source)
  {
    DWORD bytesCopied = 0;

    while (bufSize > 1 && *source) {
      *buf++ = *source++;
      --bufSize;
      ++bytesCopied;
    }
    *buf = '\0';

    return bytesCopied;
  }

  // This function checks a task type group file name passed to an
  // exported function to make sure it's a reasonable name. It returns
  // a boolean indicating if it is. A file name is reasonable if it's
  // of the form Name1\Name2\...\Name3\Name4.ttg and if each Name? is
  // at least one character and contains only alphanumeric characters,
  // dashes, and/or underscores.

  bool CheckTTGFileName (const char *TTGFileName)
  {
    bool beginningOfWord;
    const char *p;

    // This function works as a very simple finite state automaton, with
    // two states: beginningOfWord and not-beginningOfWord. The state
    // beginningOfWord means that we're at the beginning of a file name
    // part, e.g., right after a backslash or at the beginning of the file
    // name.

    for (p = TTGFileName, beginningOfWord = true; 1; p++) {
      if (beginningOfWord) {
        // At the beginning of a word, the next character must be
        // alphanumeric. If it is, past that character we enter the
        // not-beginningOfWord state.

        if (!isalnum(*p))
          return false;
        else
          beginningOfWord = false;
      }
      else {
        // When it's not the beginning of a word, the next character
        // can be various things:
        //
        // Alphanumeric, dash, or slash: fine, state still not-beginningOfWord
        // Backslash: fine, switch to state beginningOfWord
        // Period: fine, but the rest of the name must be "ttg"
        // Anything else: bad file name

        if (!isalnum(*p)) {
          if (*p == '\\')
            beginningOfWord = true;
          else if (*p == '.')
            return !stricmp(p+1, "ttg");
          else if (*p != '-' && *p != '_')
            return FALSE;
        }
      }
    }
  }

}

/////////////////////////////////////////////////////////////////////////////
// FILE OPERATIONS
/////////////////////////////////////////////////////////////////////////////

// This function creates or opens a task type group file. The file
// name is given as 'TTGFileName'. The creation disposition
// 'CreationDisposition' can be one of CREATE_ALWAYS, CREATE_NEW,
// OPEN_ALWAYS, or OPEN_EXISTING, with the same meanings as for the
// Windows CreateFile() call. The remaining two parameters,
// 'NumTaskTypes' and 'TaskTypeDescriptions', are only used if this
// function actually creates (instead of just opening) a file. If so,
// 'NumTaskTypes' gives the number of task types the file should have,
// and 'TaskTypeDescriptions' is an array of descriptions, one per
// task type. 'TaskTypeDescriptions' can be NULL, meaning to use the
// default task type parameters for all task types; even if
// 'TaskTypeDescriptions' is not null, some of its entries may be NULL
// strings meaning to use the default task type parameters for those
// task types. This function returns a handle to the open file, which
// can be closed with CloseHandle(); if it fails, it returns
// INVALID_HANDLE_VALUE after having set the last error with
// SetLastError().

DLLEXPORT HANDLE RightSpeedCreateFile (
  LPCSTR TTGFileName,
  DWORD CreationDisposition,
  DWORD NumTaskTypes,
  LPCSTR TaskTypeDescriptions[]
  )
{
  // Make sure the file name is a valid one.

  if (!CheckTTGFileName(TTGFileName)) {
    SetLastError(RIGHTSPEED_ERROR_BAD_TTG_FILE_NAME);
    return INVALID_HANDLE_VALUE;
  }

  // Create/open the file.

  return CreateTTGFile(TTGFileName,
                       CreationDisposition,
                       NumTaskTypes,
                       TaskTypeDescriptions);
}

// This function makes an I/O control request asking that RSTask allow
// the task type group file opened by the given handle be allowed to
// be modified on disk. It returns a boolean indicating if it was
// successful.

DLLEXPORT BOOL RightSpeedAllowFileModification (
  HANDLE TTGHandle
  )
{
  RSTaskHandle h(TTGHandle);
  return h.AllowFileModification();
}

// This function makes an I/O control request asking that RSTask flush
// the in-memory, possibly modified, image of the task type group file
// opened by the given handle to disk soon. It returns a boolean
// indicating if it was successful, but since the flush is done
// asynchronously, the flush may not actually have been completed by
// the time this function returns. Indeed, this function could return
// TRUE indicating that it was successful at *requesting* a future
// flush happen soon, but then that flush could actually be scheduled
// and fail.

DLLEXPORT BOOL RightSpeedFlush (
  HANDLE TTGHandle
  )
{
  RSTaskHandle h(TTGHandle);
  return h.Flush();
}

/////////////////////////////////////////////////////////////////////////////
// TASK OPERATIONS
/////////////////////////////////////////////////////////////////////////////

// This function begins a task of the given type, belonging to the task
// type group opened by the given handle. It returns a task ID for the
// newly begun task, or RIGHTSPEED_INVALID_TASK_ID if it fails.

DLLEXPORT DWORD RightSpeedBeginTask (
  HANDLE TTGHandle,
  DWORD TaskType
  )
{
  RSTaskHandle h(TTGHandle);
  return h.BeginTask(TaskType);
}

// This function ends a task of the given ID. It does not necessarily
// have to belong to the task type group opened by the given handle, but
// the caller of this function must pass a valid open RSTask handle. This
// function returns a boolean indicating if it was successful.

DLLEXPORT BOOL RightSpeedEndTask (
  HANDLE TTGHandle,
  DWORD TaskId
  )
{
  RSTaskHandle h(TTGHandle);
  return h.EndTask(TaskId);
}

// This function ends one task and begins another. The task to end is
// given as a task ID, and the task to begin is given as a task type.
// The task to begin should belong to the task type opened by the
// given task type group handle, but the task to end does not have to
// belong to that task type group. This function returns a task ID for
// the newly begun task, or RIGHTSPEED_INVALID_TASK_ID if it fails to
// begin a new task.

DLLEXPORT DWORD RightSpeedEndTaskAndBeginAnother (
  HANDLE TTGHandle,
  DWORD TaskId,
  DWORD TaskType
  )
{
  RSTaskHandle h(TTGHandle);
  return h.EndTaskAndBeginAnother(TaskId, TaskType);
}

// This function disables (or enables) the use of automatic tasks for
// this process. It does this by disabling (or enabling) hook
// procedures in all threads of this process. Whether it disables
// or enables depends on the value of the parameter 'Disable': if it's
// TRUE, the function disables; if it's FALSE, the function enables.

DLLEXPORT void RightSpeedDisableAutomaticTasks (
  BOOL Disable
  )
{
  if (Disable)
    g.hooks.Disable();
  else
    g.hooks.Enable();
}

// This function temporarily disables (or enables) the use of
// automatic tasks for this process. It does this by setting a
// process-global variable indicating that the functionality of hooks
// should be disabled (or enabled). Whether it disables or enables
// depends on the value of the parameter 'Disable': if it's TRUE, the
// function disables; if it's FALSE, the function enables.

DLLEXPORT void RightSpeedTemporarilyDisableAutomaticTasks (
  BOOL Disable
  )
{
  if (Disable)
    g.hooks.TemporarilyDisable();
  else
    g.hooks.TemporarilyEnable();
}

// This function determines the deadline for a given task type in the
// task type group whose handle is passed. It puts this deadline, in
// milliseconds, in 'DeadlinePtr'. The function returns a boolean
// indicating if it was successful.

DLLEXPORT BOOL RightSpeedQueryDeadline (
  HANDLE TTGHandle,
  DWORD TaskType,
  double *DeadlinePtr
  )
{
  RSTaskHandle h(TTGHandle);
  return h.QueryDeadline(TaskType, DeadlinePtr);
}

// This function changes the deadline for a given task type in the
// task type group whose handle is passed. It uses 'Deadline' as the
// new deadline, in milliseconds. The function returns a boolean
// indicating if it was successful.

DLLEXPORT BOOL RightSpeedChangeDeadline (
  HANDLE TTGHandle,
  DWORD TaskType,
  double Deadline
  )
{
  RSTaskHandle h(TTGHandle);
  return h.ChangeDeadline(TaskType, Deadline);
}

// This function resets the task work sample for a given task type in
// the task type group whose handle is passed. It returns a boolean
// indicating if it was successful.

DLLEXPORT BOOL RightSpeedResetSample (
  HANDLE TTGHandle,
  DWORD TaskType
  )
{
  RSTaskHandle h(TTGHandle);
  return h.ResetSample(TaskType);
}

// This function resets the task work samples for all task types in
// the task type group whose handle is passed. It returns a boolean
// indicating if it was successful.

DLLEXPORT BOOL RightSpeedResetAllSamples (
  HANDLE TTGHandle
  )
{
  RSTaskHandle h(TTGHandle);
  return h.ResetAllSamples();
}

/////////////////////////////////////////////////////////////////////////////
// ERRORS
/////////////////////////////////////////////////////////////////////////////

// This function formats a RightSpeed error code as an English string.
// The error code to format is given in 'err', and the buffer of
// length 'bufSize' in which to put the English string is passed as
// 'buf'.  This function will format system error codes as well as
// RightSpeed-specific ones, since RightSpeed sometimes uses system
// error codes as its error codes.

DLLEXPORT DWORD RightSpeedFormatMessage (
  DWORD err,
  LPTSTR buf,
  DWORD bufSize
  )
{
  // If this error code is not in the range of valid RightSpeed error
  // codes, format it as if it is a system error code.

  if (err < RIGHTSPEED_ERROR_MIN || err > RIGHTSPEED_ERROR_MAX) {
    return FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, // flags
                         NULL,   // source: use default
                         err,
                         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), 
                         buf,
                         bufSize,
                         NULL);  // arguments: none
  }

  // Otherwise, format it assuming it's a RightSpeed error code.

  switch (err) {
  case RIGHTSPEED_ERROR_TTG_OPENING_OR_CLOSING :
    return CopyIntoBuffer(buf, bufSize,
                          "The task type group was being opened or closed "
                          "by someone else, even though I tried "
                          "repeatedly.\n");
    break;

  case RIGHTSPEED_ERROR_BAD_TTG_DIRECTORY :
    return CopyIntoBuffer(buf, bufSize,
                          "The specification of the task type group "
                          "subdirectory was invalid.\n");
    break;

  case RIGHTSPEED_ERROR_BAD_TTG_FILE_NAME :
    return CopyIntoBuffer(buf, bufSize,
                          "The task type group file name was invalid.\n");
    break;

  case RIGHTSPEED_ERROR_BAD_TTG_FILE_FORMAT :
    return CopyIntoBuffer(buf, bufSize,
                          "The task type group file had an invalid "
                          "format.\n");
    break;

  case RIGHTSPEED_ERROR_INVALID_TASK_TYPE :
    return CopyIntoBuffer(buf, bufSize, "The task type was invalid.\n");
    break;

  case RIGHTSPEED_ERROR_INTERVENING_TASK :
    return CopyIntoBuffer(buf, bufSize, "There was an intervening task.\n");
    break;

  case RIGHTSPEED_ERROR_INVALID_TASK_ID :
    return CopyIntoBuffer(buf, bufSize, "The task ID was invalid.\n");
    break;

  case RIGHTSPEED_ERROR_INVALID_REGISTRY_DATA :
    return CopyIntoBuffer(buf, bufSize, "The registry data was invalid.\n");
    break;

  case RIGHTSPEED_ERROR_INVALID_SCHEDULE :
    return CopyIntoBuffer(buf, bufSize, "The speed schedule was invalid.\n");
    break;

  case RIGHTSPEED_ERROR_TOO_MANY_ONGOING_TASKS :
    return CopyIntoBuffer(buf, bufSize,
                          "Too many tasks are currently ongoing.\n");
    break;

  case RIGHTSPEED_ERROR_WRITEBACK_DISABLED :
    return CopyIntoBuffer(buf, bufSize,
                          "Write-back on the file was disabled.\n");
    break;

  case RIGHTSPEED_ERROR_FLUSH_IN_PROGRESS :
    return CopyIntoBuffer(buf, bufSize,
                          "Flushing of the file was already in progress.\n");
    break;

  case RIGHTSPEED_ERROR_NOT_INITIALIZED :
    return CopyIntoBuffer(buf, bufSize, "RightSpeed was not initialized.\n");
    break;

  case RIGHTSPEED_ERROR_BAD_NUMBER_OF_TASK_TYPES :
    return CopyIntoBuffer(buf, bufSize,
                          "The number of task types was invalid.\n");
    break;

  case RIGHTSPEED_ERROR_BAD_TASK_TYPE_DESCRIPTION :
    return CopyIntoBuffer(buf, bufSize,
                          "There was an invalid task type description.\n");
    break;

  case RIGHTSPEED_ERROR_FS_ALREADY_FILTERED :
    return CopyIntoBuffer(buf, bufSize,
                          "The file system is already filtered.\n");
    break;

  case RIGHTSPEED_ERROR_INVALID_POLICY_TYPE :
    return CopyIntoBuffer(buf, bufSize,
                          "An invalid policy type was specified.\n");
    break;

  case RIGHTSPEED_ERROR_LOG_FILE_STARTING_OR_STOPPING :
    return CopyIntoBuffer(buf, bufSize,
                          "A log file is in the process of begin started "
                          "or stopped.\n");
    break;

  case RIGHTSPEED_ERROR_LOG_FILE_ACTIVE :
    return CopyIntoBuffer(buf, bufSize,
                          "Another log file is in use, and only one log "
                          "file can be used at a time.\n");
    break;

  case RIGHTSPEED_ERROR_UNINSTALLING :
    return CopyIntoBuffer(buf, bufSize, "RightSpeed is being uninstalled.\n");
    break;

  default :
    return CopyIntoBuffer(buf, bufSize,
                          "An unspecified RightSpeed error occurred.\n");
    break;
  }
}

// This function sets the global policy of RightSpeed according to the
// passed parameters that describe which type of policy to change and
// what value to assign to that policy. It returns a boolean indicating
// if it was successful.

BOOL RightSpeedSetGlobalPolicy (
  DWORD PolicyType,
  DWORD PolicyValue
  )
{
  switch (PolicyType) {
  case RIGHTSPEED_POLICY_DISABLE_TASKS :
    // Setting the "disable tasks" policy is simply a matter of setting
    // the 'disableTasks' boolean in the systemwide globals accordingly.

    systemwideGlobals.disableTasks = (PolicyValue != 0);
    return TRUE;
    break;

  default:
    // For any other policy, open a handle to RSTask and send the
    // policy-change request on to it.
    {
      RSTaskHandle h("\\\\.\\RSTask");
      if (h == INVALID_HANDLE_VALUE)
        return FALSE;
      else
        return h.SetGlobalPolicy(PolicyType, PolicyValue);
    }
    break;
  }
}
