// This file contains the implementation for the hook-related classes.

#include <windows.h>
#include "../Common/ErrCodes.h"
#include "Globals.h"
#include "AutoTask.h"
#include "Hooks.h"
#include "Log.h"
#include "TTGFile.h"

namespace {

////////////////////////////////////////////////////////////////////////////
// APPLICATION NAMES
////////////////////////////////////////////////////////////////////////////

  // This constant indicates the longest we expect an application name
  // to be; we'll truncate any name longer than this:

  const int MAX_APPLICATION_NAME_LENGTH = 64;

  // This function determines the name of this application. It puts
  // this name in 'appName'. It never uses a non-alphanumeric character,
  // always appends a terminating null character, and never uses more
  // than 'maxLength' bytes including the terminating null. If it
  // cannot determine the application name, it uses "Unknown", assuming
  // that 'maxLength' is sufficient to hold this.
  //
  // The application name is formed by getting the full path,
  // searching for the last part of it (e.g., netcape.exe out of
  // c:\program files\netscape\programs\netscape.exe), removing the
  // extension (e.g., removing .exe from netscape.exe), and removing
  // any non-alphanumeric characters that remain.

  void GetApplicationName (char *appName, DWORD maxLength)
  {
    // Read the full path of this application into a local buffer.

    char fullPath[MAX_PATH];
    DWORD fullPathLength = GetModuleFileName(NULL, fullPath, MAX_PATH);

    // If the call failed, use "Unknown".

    if (fullPathLength == 0) {
      strcpy(appName, "Unknown");
      return;
    }

    // Search backwards from the end until we can't go any further
    // without reaching the beginning of the path, a backslash, or a
    // colon. Use that position as the beginning of the application
    // name.

    DWORD fullPathPos = fullPathLength;
    while (fullPathPos > 0 && fullPath[fullPathPos-1] != '\\' &&
           fullPath[fullPathPos-1] != ':')
      --fullPathPos;

    // Copy all alphanumeric characters from the application name part
    // of the full path to the target application name buffer until we
    // reach a '.' indicating the beginning of the extension. Convert
    // characters to lowercase before copying them.

    DWORD appNamePos = 0;
    char c;
    while (fullPathPos < fullPathLength && fullPath[fullPathPos] != '.') {
      c = fullPath[fullPathPos];
      if (isalnum(c)) {
        appName[appNamePos++] = tolower(c);
        // If the next application name position would be the last
        // available one, terminate the loop so we still have room
        // to insert a terminating null character.
        if (appNamePos == maxLength - 1)
          break;
      }
      ++fullPathPos;
    }

    // If we didn't copy any characters, we have failed, so just use
    // "Unknown".
  
    if (appNamePos == 0) {
      strcpy(appName, "Unknown");
      return;
    }

    // Add a terminating null character to the application name and we're
    // done.

    appName[appNamePos] = '\0';
  }

////////////////////////////////////////////////////////////////////////////
// MODIFIER STATE
////////////////////////////////////////////////////////////////////////////

  // This function updates the systemwide global keyboard state
  // information to reflect a message we just intercepted. If we saw
  // a shift-released message, for instance, we update the keyboard
  // state to reflect the fact that the shift key is no longer down.

  void UpdateModifierState (MSG *msg)
  {
    if (msg->message == WM_KEYDOWN) {
      // If it's a key-down message, see if the key being pressed is a
      // shift or control key; if so, update the systemwide global
      // state to reflect the fact that that modifier key is now down.

      if (msg->wParam == VK_SHIFT)
        systemwideGlobals.shiftKeyDown = true;
      else if (msg->wParam == VK_CONTROL)
        systemwideGlobals.ctrlKeyDown = true;
    }
    else if (msg->message == WM_KEYUP) {
      // If it's a key-up message, see if the key being released is a
      // shift or control key; if so, update the systemwide global
      // state to reflect the fact that that modifier key is now up.

      if (msg->wParam == VK_SHIFT)
        systemwideGlobals.shiftKeyDown = false;
      else if (msg->wParam == VK_CONTROL)
        systemwideGlobals.ctrlKeyDown = false;
    }
  }

////////////////////////////////////////////////////////////////////////////
// MESSAGE HOOK ROUTINE
////////////////////////////////////////////////////////////////////////////

  // This function is the message hook routine used by all message
  // hooks.
  
  LRESULT CALLBACK MessageHookRoutine (int code, WPARAM wParam, LPARAM lParam)
  {
    // The 'lParam' parameter is a pointer to a message.

    MSG *msg = (MSG *) lParam;

    // Update the systemwide modifier state to reflect the receipt
    // of this message.

    UpdateModifierState(msg);

    // Have the hooks object handle this message.
    
    g.hooks.HandleMessage(wParam == PM_REMOVE, msg);

    // Call the next hook. (This is the last thing a hook routine is
    // supposed to do.)
    
    return CallNextHookEx(NULL, code, wParam, lParam);
  }

}

/////////////////////////////////////////////////////////////////////////////
// THREAD HOOKING
/////////////////////////////////////////////////////////////////////////////

// This function initializes a thread hooks object for a given thread
// ID. It begins with hooking disabled.

ThreadHooks::ThreadHooks (DWORD threadIdVal) :
  enabled(false),
  threadId(threadIdVal),
  hMessageHook(NULL)
{
}

// This function enables hooking for the thread this object is
// associated with.

void ThreadHooks::Enable ()
{
  if (!enabled) {
    hMessageHook = SetWindowsHookEx(WH_GETMESSAGE,
                                    MessageHookRoutine,
                                    NULL,  // module containing routine
                                    threadId);
    if (LOGGING_TASKS)
      g.initDeviceHandle.LogThreadHooked(threadId);
    enabled = true;
  }
}

// This function disables hooking for the thread this object is
// associated with.

void ThreadHooks::Disable ()
{
  if (enabled) {
    if (hMessageHook) {
      UnhookWindowsHookEx(hMessageHook);
      if (LOGGING_TASKS)
        g.initDeviceHandle.LogThreadUnhooked(threadId);
    }

    enabled = false;
  }
}

/////////////////////////////////////////////////////////////////////////////
// PROCESSWIDE HOOK INFORMATION
/////////////////////////////////////////////////////////////////////////////

// This function initializes all the fields of a hooks object to a
// reasonable initial state.

Hooks::Hooks () :
  enabled(false),
  functionalityEnabled(true),
  lastUnremovedMessagePtr(0),
  lastAutomaticTaskId(RIGHTSPEED_INVALID_TASK_ID)
{
  InitializeCriticalSection(&criticalSection);
}

// This function initializes hooking for the current process. It returns
// a boolean indicating if it was successful.

bool Hooks::Initialize ()
{
  // First, record that the current thread is running so that we can
  // add it to our list of running threads.

  NotifyThreadStarting(GetCurrentThreadId());

  /////////////////////////////////////////////////////////////////////
  // Open the task type group file for the systemwide automatic tasks.
  /////////////////////////////////////////////////////////////////////

  // Use Auto\System\All.ttg as the name of the TTG file to
  // create/open.

  LPCSTR systemTaskDescriptions[NUM_AUTO_SYSTEM_TASK_TYPES] =
      { AUTO_SYSTEM_TASK_TYPE_DESCRIPTIONS };
  hSystemTasks = CreateTTGFile("Auto\\System\\All.ttg",
                               OPEN_ALWAYS,
                               NUM_AUTO_SYSTEM_TASK_TYPES,
                               systemTaskDescriptions);
  if (hSystemTasks == INVALID_HANDLE_VALUE) {
    if (LOGGING_ERRORS)
      LOG_ERROR_ONE_PARAMETER(RS_LIB_ERROR_CANT_CREATE_SYSTEM_TTG,
                              GetLastError());
    return false;
  }

  // Make sure that the TTG file has the proper number of task types
  // (one).
  
  TTG_INFO ttgInfo;
  if (!hSystemTasks.QueryTTGInfoPreamble(&ttgInfo)) {
    if (LOGGING_ERRORS)
      LOG_ERROR_ONE_PARAMETER(RS_LIB_ERROR_CANT_GET_SYSTEM_TTG_INFO_PREAMBLE,
                              GetLastError());
    return false;
  }
  if (ttgInfo.numTaskTypes != 1) {
    if (LOGGING_ERRORS)
      LOG_ERROR_ONE_PARAMETER(
        RS_LIB_ERROR_SYSTEM_TTG_HAS_WRONG_NUM_TASK_TYPES,
        ttgInfo.numTaskTypes);
    return false;
  }

  //////////////////////////////////////////////////////////////////
  // Open the task type group file for this application's automatic
  // tasks.
  //////////////////////////////////////////////////////////////////

  // Determine the application name and form the TTG file name from
  // it.

  char appName[MAX_APPLICATION_NAME_LENGTH];
  GetApplicationName(appName, MAX_APPLICATION_NAME_LENGTH);
  string appTTGFileName = string("Auto\\Apps\\") + appName + ".ttg";

  // Create/open the TTG file.

  LPCSTR appTaskDescriptions[NUM_AUTO_APP_TASK_TYPES] =
      { AUTO_APP_TASK_TYPE_DESCRIPTIONS };
  hAppTasks = CreateTTGFile(appTTGFileName.c_str(),
                            OPEN_ALWAYS,
                            NUM_AUTO_APP_TASK_TYPES,
                            appTaskDescriptions);
  if (hAppTasks == INVALID_HANDLE_VALUE) {
    if (LOGGING_ERRORS)
      LOG_ERROR_ONE_PARAMETER(RS_LIB_ERROR_CANT_CREATE_APP_TTG,
                              GetLastError());
    return false;
  }

  // Make sure the TTG file has the right number of task types.
  
  if (!hAppTasks.QueryTTGInfoPreamble(&ttgInfo)) {
    if (LOGGING_ERRORS)
      LOG_ERROR_ONE_PARAMETER(RS_LIB_ERROR_CANT_GET_APP_TTG_INFO_PREAMBLE,
                              GetLastError());
    return false;
  }
  if (ttgInfo.numTaskTypes != NUM_AUTO_APP_TASK_TYPES) {
    if (LOGGING_ERRORS)
      LOG_ERROR_ONE_PARAMETER(RS_LIB_ERROR_APP_TTG_HAS_WRONG_NUM_TASK_TYPES,
                              ttgInfo.numTaskTypes);
    return false;
  }

  /////////////////////////////////////////////////////////////////////
  
  // Log the fact that we've successfully initialized hooking for this
  // process.

  if (LOGGING_TASKS)
    g.initDeviceHandle.LogProcessHooked(appName);

  return true;
}

// This function terminates hooking for the current process.

void Hooks::Close ()
{
  // Disable all hooks.

  Disable();

  // Close any handles to TTG files we opened.

  if (hSystemTasks != INVALID_HANDLE_VALUE) {
    CloseHandle(hSystemTasks);
    hSystemTasks = INVALID_HANDLE_VALUE;
  }
  if (hAppTasks != INVALID_HANDLE_VALUE) {
    CloseHandle(hAppTasks);
    hAppTasks = INVALID_HANDLE_VALUE;
  }
}

// This function adds the given thread ID to the list of threads in this
// process. If hooking is enabled for this process, it starts hooking that
// thread.

void Hooks::NotifyThreadStarting (DWORD threadId)
{
  EnterCriticalSection(&criticalSection);
  threadHooks.push_back(ThreadHooks(threadId));
  if (enabled)
    threadHooks.back().Enable();
  LeaveCriticalSection(&criticalSection);
}

// This function removes the given thread ID from the list of threads
// in this process. Also, it stops hooking that thread.

void Hooks::NotifyThreadEnding (DWORD threadId)
{
  EnterCriticalSection(&criticalSection);
  list<ThreadHooks>::iterator it;
  for (it = threadHooks.begin(); it != threadHooks.end(); ++it) {
    if (it->GetThreadId() == threadId) {
      it->Disable();
      threadHooks.erase(it);
      break;
    }
  }
  LeaveCriticalSection(&criticalSection);
}

// This function enables hooking for the current process. It starts up
// hooks for each of its threads immediately.

void Hooks::Enable ()
{
  // Do a quick check to see if we're already enabled and there's no
  // reason to enter the critical section.

  if (enabled)
    return;

  // Within the critical section, enable hooking for all threads in
  // this process and set 'enabled' to true so future threads will
  // have hooking enabled.

  EnterCriticalSection(&criticalSection);
  if (!enabled) {
    list<ThreadHooks>::iterator it;
    for (it = threadHooks.begin(); it != threadHooks.end(); ++it)
      it->Enable();

    enabled = true;
  }
  LeaveCriticalSection(&criticalSection);
}

// This function disables hooking for the current process. It stops all
// hooks for each of its threads immediately.

void Hooks::Disable ()
{
  // Do a quick check to see if we're already disabled and there's no
  // reason to enter the critical section.

  if (!enabled)
    return;

  // Within the critical section, disable hooking for all threads in
  // this process add set 'enabled' to false to future threads will not
  // have hooking enabled.

  EnterCriticalSection(&criticalSection);
  if (enabled) {
    list<ThreadHooks>::iterator it;
    for (it = threadHooks.begin(); it != threadHooks.end(); ++it)
      it->Disable();

    enabled = false;
  }
  LeaveCriticalSection(&criticalSection);
}

// This function handles the receipt of a message by one of the thread's
// message hooks. The 'removed' parameter indicates if this message was
// retrieved by a call that will remove it from the queue. The 'msg'
// parameter points to the message.

void Hooks::HandleMessage (bool removed, MSG *msg)
{
  DWORD taskType;

  // If hooking functionality is temporarily disabled, don't do anything.

  if (!functionalityEnabled)
    return;

  // If the last message received before this one wasn't removed from
  // the queue, and this message matches that last message received,
  // then we've already processed it, so there's nothing to
  // do. Except, if it is *now* being removed, we should now set
  // 'lastUnremovedMessagePtr' to 0.

  if (msg == lastUnremovedMessagePtr) {
    if (removed)
      lastUnremovedMessagePtr = 0;
    return;
  }

  // Set 'lastUnremovedMessagePtr' to either 0 or 'msg', depending
  // on if this message is being removed from the queue.

  lastUnremovedMessagePtr = removed ? 0 : msg;

  // If this is a mouse-move message, begin an automatic mouse-move
  // system task and end the last automatic task. Record the task ID
  // from beginning the task in 'lastAutomaticTaskId' so that we can
  // end it when the next automatic task begins.

  if (msg->message == WM_MOUSEMOVE || msg->message == WM_NCMOUSEMOVE) {
    if (LOGGING_TASKS)
      g.initDeviceHandle.LogAutoSystemTask();

    // This is a mouse-move message, which we treat as task type 0 of
    // the systemwide automatic task type group.

    lastAutomaticTaskId =
      hSystemTasks.EndTaskAndBeginAnother(lastAutomaticTaskId,
                                          0); // task type is 0
    if (lastAutomaticTaskId == RIGHTSPEED_INVALID_TASK_ID &&
        (LOGGING_WARNINGS))
      LOG_ERROR_ONE_PARAMETER(RS_LIB_WARNING_CANT_BEGIN_AUTO_SYSTEM_TASK,
                              GetLastError());
    return;
  }

  // Now that we know the message is not a mouse-move message, and
  // therefore doesn't correspond to an automatic system task, see if
  // it corresponds to the beginning of an automatic application
  // task.

  taskType = GetAutoTaskType(msg->message, msg->wParam);
  if (taskType == RIGHTSPEED_INVALID_TASK_TYPE)
    return;

  // If this message does correspond to an automatic application task,
  // begin that task and end the last automatic task. Record the task
  // ID from beginning the task in 'lastAutomaticTaskId' so that we
  // can end it when the next automatic task begins.

  if (LOGGING_TASKS)
    g.initDeviceHandle.LogAutoAppTask(taskType);

  lastAutomaticTaskId =
    hAppTasks.EndTaskAndBeginAnother(lastAutomaticTaskId, taskType);

  if (lastAutomaticTaskId == RIGHTSPEED_INVALID_TASK_ID &&
      (LOGGING_WARNINGS))
    LOG_ERROR_ONE_PARAMETER(RS_LIB_WARNING_CANT_BEGIN_AUTO_APP_TASK,
                            GetLastError());
}
