// This file contains the implementation of the class representing all
// parameters for a task type.

#include <windows.h>
#include <string>
#include <ctype.h>
#include "../Common/Parameters.h"
#include "../Common/TTG.h"
#include "Log.h"
#include "TaskType.h"

using namespace std;

namespace {

  //////////////////////////////////////////////////////////////////////////
  // CONSTANTS
  //////////////////////////////////////////////////////////////////////////

  // This is the number of hundreds of nanoseconds per millisecond:
  const long HNS_PER_MS = 10000;

  // This is the number of cycles per milliseconds per megahertz:
  const long CYC_PER_MS_PER_MHZ = 1000;

  //////////////////////////////////////////////////////////////////////////
  // UTILITY FUNCTIONS
  //////////////////////////////////////////////////////////////////////////

  // This function rounds the given real number, assumed to be positive,
  // to the nearest integer.

  unsigned long RoundToNearest (double x)
  {
    return (unsigned long)(x + 0.5);
  }

  //////////////////////////////////////////////////////////////////////////
  // ALGORITHM TYPES
  //////////////////////////////////////////////////////////////////////////

  // This function converts a string describing an algorithm type to
  // an algorithm type, which it returns. It returns INVALID_ALGORITHM_TYPE
  // if it cannot determine the algorithm type.
  
  long StringToAlgorithmType (string& s)
  {
    if (s == "flat")
      return ALGORITHM_TYPE_FLAT;
    if (s == "flat-ag-basic")
      return ALGORITHM_TYPE_FLAT_AG_BASIC;
    if (s == "longshort")
      return ALGORITHM_TYPE_LONGSHORT;
    if (s == "longshort-ag-basic")
      return ALGORITHM_TYPE_LONGSHORT_AG_BASIC;
    if (s == "pastpeg")
      return ALGORITHM_TYPE_PASTPEG;
    if (s == "pastpeg-ag-basic")
      return ALGORITHM_TYPE_PASTPEG_AG_BASIC;
    if (s == "longrun")
      return ALGORITHM_TYPE_LONGRUN;
    if (s == "longrun-ag-basic")
      return ALGORITHM_TYPE_LONGRUN_AG_BASIC;

    // If we couldn't match the algorithm type, return the error value
    // INVALID_ALGORITHM_TYPE.

    return INVALID_ALGORITHM_TYPE;
  }

  ///////////////////////////////////////////////////////////////////////////
  // COMMON SUBSTRUCTURES
  ///////////////////////////////////////////////////////////////////////////

  // This function initializes an aged sample given a pointer to that
  // sample and an aging factor to use.

  void InitializeAgedSample (PAGED_SAMPLE sample, double agingFactor)
  {
    sample->agingFactor = agingFactor;
    sample->n = 0;
    sample->weightedCount = 0;
    sample->weightedSum = 0;
    sample->weightedSumSq = 0;
  }

}

//////////////////////////////////////////////////////////////////////////////
// INITIALIZATION
//////////////////////////////////////////////////////////////////////////////

// This function initializes a task type to use all default parameters.

TaskType::TaskType() :
  algorithmType(DEFAULT_ALGORITHM_TYPE),
  deadline(DEFAULT_DEADLINE_MS),
  sampleIfIo(TRUE),
  speed(DEFAULT_PREDEADLINE_SPEED_MHZ),
  sampleValuesBeforeFirstReschedule(DEFAULT_SAMPLE_VALUES_BEFORE_RESCHEDULE),
  sampleValuesBetweenReschedules(DEFAULT_SAMPLE_VALUES_BETWEEN_RESCHEDULES),
  agingFactor(DEFAULT_SAMPLER_AGING_FACTOR),
  updatePeriod(DEFAULT_UPDATE_PERIOD_MS),
  percentOfDeadlinesToMake(DEFAULT_PERCENT_OF_DEADLINES_TO_MAKE)
{
}

// This function initializes a task type to use the parameters as
// described in the given description.  When it must refer to the
// maximum speed, it uses the given parameter, assumed to be the
// maximum speed available on this machine in MHz.  It returns a
// boolean indicating if it was successful, i.e., if the description
// was valid.

bool TaskType::ApplyDescription (const char *description, DWORD maxSpeed)
{
  const char *next = description, *equal;
  string parameterType, parameterValue;

  // Loop through all the space-delimited parts of the description,
  // processing each one as we come to it. We start 'next' as the
  // pointer to the description, and each time through the loop it
  // will point to NULL if we're done or a point past the part we
  // just finished processing.

  while (next) {
    // Skip any whitespace.

    while (isspace(*next))
      ++next;

    // If we've reached the end of the string, we're done.

    if (!*next)
      return true;

    /////////////////////////////////////////////////////////////
    // At this point, 'next' should point to a part in the form
    // "ParameterType=ParameterValue".
    /////////////////////////////////////////////////////////////

    // Search for the equals sign. If we can't find it, this is
    // an invalid description.

    equal = strchr(next, '=');
    if (!equal)
      return false;

    // The parameter type is the string between 'next' and the equals
    // sign.
    
    parameterType = string(next, equal - next);

    // Search for the first space after the equals sign; this will
    // be where the next part begins, or NULL if there is no next
    // part.

    next = strchr(equal + 1, ' ');

    // The parameter value is the string between the equals sign
    // and the first space after it; if there is no space after it,
    // it is the entire string after the equals sign.
    
    if (next)
      parameterValue = string(equal + 1, next - equal - 1);
    else
      parameterValue = string(equal + 1);

    // Set the field of this object corresponding to the parameter
    // type to a value corresponding to the parameter value.

    if (parameterType == "algorithm") {
      algorithmType = StringToAlgorithmType(parameterValue);
      if (algorithmType == INVALID_ALGORITHM_TYPE)
        return false;
    }
    else if (parameterType == "deadline") {
      deadline = atof(parameterValue.c_str());
    }
    else if (parameterType == "sampleIfIo") {
      if (parameterValue.size() && tolower(parameterValue[0]) == 't')
        sampleIfIo = true;
      else
        sampleIfIo = atoi(parameterValue.c_str());
    }
    else if (parameterType == "speed") {
      speed = atof(parameterValue.c_str());
    }
    else if (parameterType == "speedRelativeToMaximum") {
      double speedRelativeToMaximum = atof(parameterValue.c_str());
      if (speedRelativeToMaximum < 0 || speedRelativeToMaximum > 1)
        return false;
      speed = speedRelativeToMaximum * maxSpeed;
    }
    else if (parameterType == "agingFactor") {
      agingFactor = atof(parameterValue.c_str());
      if (agingFactor < 0 || agingFactor > 1)
        return false;
    }
    else if (parameterType == "sampleValuesBeforeFirstReschedule") {
      sampleValuesBeforeFirstReschedule = atoi(parameterValue.c_str());
      if (sampleValuesBeforeFirstReschedule <
          MIN_SAMPLE_VALUES_BEFORE_RESCHEDULE)
        return false;
    }
    else if (parameterType == "sampleValuesBetweenReschedules") {
      sampleValuesBetweenReschedules = atoi(parameterValue.c_str());
      if (sampleValuesBetweenReschedules <
          MIN_SAMPLE_VALUES_BETWEEN_RESCHEDULES)
        return false;
    }
    else if (parameterType == "updatePeriod") {
      updatePeriod = atof(parameterValue.c_str());
      if (updatePeriod <= 0 || updatePeriod > MAX_UPDATE_PERIOD_MS)
        return false;
    }
    else if (parameterType == "percentOfDeadlinesToMake") {
      percentOfDeadlinesToMake = atof(parameterValue.c_str());
      if (percentOfDeadlinesToMake < 0 || percentOfDeadlinesToMake > 100)
        return false;
    }
    else {
      // If we don't recognize the parameter type, return false
      // indicating an invalid description.

      return false;
    }
  }

  // We made it through the loop without error, so return success.

  return true;
}

//////////////////////////////////////////////////////////////////////////////
// TASK TYPE INFORMATION
//////////////////////////////////////////////////////////////////////////////

// This function returns the deadline parameter in hundreds of
// nanoseconds.

unsigned long TaskType::GetDeadlineHns () const
{
  return RoundToNearest(deadline * HNS_PER_MS);
}

// This function returns a boolean indicating if this task type uses
// a sample.

BOOLEAN TaskType::GetUseSample () const
{
  // The algorithm uses sampling if and only if its automatic schedule
  // type is not "none".

  return (AUTO_SCHEDULE_OF_ALGORITHM(algorithmType) !=
          AUTO_SCHEDULE_TYPE_NONE);
}

// This function returns the size of the algorithm data needed by this
// task type.

int TaskType::GetAlgorithmDataSize () const
{
  switch (algorithmType) {
  case ALGORITHM_TYPE_FLAT :
    return sizeof(ALGORITHM_DATA_FLAT);
    break;

  case ALGORITHM_TYPE_FLAT_AG_BASIC :
    return sizeof(ALGORITHM_DATA_FLAT_AG_BASIC);
    break;

  case ALGORITHM_TYPE_LONGSHORT :
    return sizeof(ALGORITHM_DATA_LONGSHORT);
    break;

  case ALGORITHM_TYPE_LONGSHORT_AG_BASIC :
    return sizeof(ALGORITHM_DATA_LONGSHORT_AG_BASIC);
    break;

  case ALGORITHM_TYPE_PASTPEG :
    return sizeof(ALGORITHM_DATA_PASTPEG);
    break;

  case ALGORITHM_TYPE_PASTPEG_AG_BASIC :
    return sizeof(ALGORITHM_DATA_PASTPEG_AG_BASIC);
    break;

  case ALGORITHM_TYPE_LONGRUN :
    return 0; // There is no algorithm data for this algorithm type.
    break;

  case ALGORITHM_TYPE_LONGRUN_AG_BASIC :
    return sizeof(ALGORITHM_DATA_LONGRUN_AG_BASIC);
    break;

  default:
    return 0; // We should never reach this point.
  }
}

// This function fills in a buffer containing algorithm data according
// to the parameters for this task type. The pointer 'algorithmDataAddr'
// is the buffer to fill in.

void TaskType::InitializeAlgorithmData (PVOID algorithmDataAddr) const
{
  switch (algorithmType) {
  case ALGORITHM_TYPE_FLAT :
    {
      PALGORITHM_DATA_FLAT algorithmData =
        (PALGORITHM_DATA_FLAT) algorithmDataAddr;

      algorithmData->cycPerMs = RoundToNearest(speed * CYC_PER_MS_PER_MHZ);
    }
    break;

  case ALGORITHM_TYPE_FLAT_AG_BASIC :
    {
      PALGORITHM_DATA_FLAT_AG_BASIC algorithmData =
        (PALGORITHM_DATA_FLAT_AG_BASIC) algorithmDataAddr;

      InitializeAgBasicData(&algorithmData->agBasicData);
      algorithmData->cycPerMs = RoundToNearest(speed * CYC_PER_MS_PER_MHZ);
    }
    break;

  case ALGORITHM_TYPE_LONGSHORT :
    {
      PALGORITHM_DATA_LONGSHORT algorithmData =
        (PALGORITHM_DATA_LONGSHORT) algorithmDataAddr;

      algorithmData->updatePeriodHns =
        RoundToNearest(updatePeriod * HNS_PER_MS);
    }
    break;

  case ALGORITHM_TYPE_LONGSHORT_AG_BASIC :
    {
      PALGORITHM_DATA_LONGSHORT_AG_BASIC algorithmData =
        (PALGORITHM_DATA_LONGSHORT_AG_BASIC) algorithmDataAddr;

      InitializeAgBasicData(&algorithmData->agBasicData);
      algorithmData->updatePeriodHns =
        RoundToNearest(updatePeriod * HNS_PER_MS);
    }
    break;

  case ALGORITHM_TYPE_PASTPEG :
    {
      PALGORITHM_DATA_PASTPEG algorithmData =
        (PALGORITHM_DATA_PASTPEG) algorithmDataAddr;

      algorithmData->updatePeriodHns =
        RoundToNearest(updatePeriod * HNS_PER_MS);
    }
    break;

  case ALGORITHM_TYPE_PASTPEG_AG_BASIC :
    {
      PALGORITHM_DATA_PASTPEG_AG_BASIC algorithmData =
        (PALGORITHM_DATA_PASTPEG_AG_BASIC) algorithmDataAddr;

      InitializeAgBasicData(&algorithmData->agBasicData);
      algorithmData->updatePeriodHns =
        RoundToNearest(updatePeriod * HNS_PER_MS);
    }
    break;

  case ALGORITHM_TYPE_LONGRUN :
    // There's no data for this algorithm type.
    break;

  case ALGORITHM_TYPE_LONGRUN_AG_BASIC :
    {
      PALGORITHM_DATA_LONGRUN_AG_BASIC algorithmData =
        (PALGORITHM_DATA_LONGRUN_AG_BASIC) algorithmDataAddr;

      InitializeAgBasicData(&algorithmData->agBasicData);
    }
    break;
  }
}

// This function fills in a buffer containing aged-gamma-basic data
// according to the parameters for this task type. The pointer 'data'
// is the buffer to fill in.

void TaskType::InitializeAgBasicData (PAG_BASIC_DATA data) const
{
  InitializeAgedSample(&data->sample, agingFactor);
  data->sampleValuesBeforeFirstReschedule = sampleValuesBeforeFirstReschedule;
  data->sampleValuesBetweenReschedules = sampleValuesBetweenReschedules;
}
