// UninstallRoutines.cpp contains routines for uninstalling RightSpeed.

#include "stdafx.h"
#include "../Common/Version.h"
#include "../Common/KernelUser.h"
#include <shlobj.h>
#include <winioctl.h>
#include <winsvc.h>

////////////////////////////////////////////////////////////////////////////
// MACROS
////////////////////////////////////////////////////////////////////////////

#define INITIAL_GUESS_SERVICE_CONFIG_SIZE       256

////////////////////////////////////////////////////////////////////////////
// FORWARD DEFINITIONS
////////////////////////////////////////////////////////////////////////////

BOOLEAN MyDeleteDirectory (CString directory);

////////////////////////////////////////////////////////////////////////////
// ERROR DISPLAY
////////////////////////////////////////////////////////////////////////////

// DisplayError displays, in a message box, the error message given in
// the string problem as well as the error message implied by the
// system error state.

void DisplayError (const char *problem)
{
  char *lpMsgBuf;
  CString str;

  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                NULL, GetLastError(),
                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf,
                0, NULL);
  str.Format("%s because %s", problem, lpMsgBuf);
  MessageBox(NULL, str, "Error", MB_OK | MB_ICONEXCLAMATION);

  LocalFree(lpMsgBuf);
}

////////////////////////////////////////////////////////////////////////////
// REGISTRY ACCESS
////////////////////////////////////////////////////////////////////////////

// GetMainDirectory gets the main directory name from the registry and puts
// it in the buffer given (mainDirectory, with size bufferSize).  It returns
// a boolean indicating whether it was successful.

BOOLEAN GetMainDirectory (char *mainDirectory, int bufferSize)
{
  HKEY hKey;
  LONG status;
  DWORD type, length;

  // Open the RightSpeed registry key

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        "Software\\Jay Lorch\\RightSpeed\\CurrentVersion",
                        0, KEY_QUERY_VALUE, &hKey);
  if (status != ERROR_SUCCESS) {
    SetLastError(status);
    DisplayError("Cannot open RightSpeed registry key to obtain main "
                 "directory");
    return FALSE;
  }

  // Read the MainDirectory value from the key

  length = bufferSize;
  status = RegQueryValueEx(hKey, "MainDirectory", NULL, &type,
                           (BYTE *) mainDirectory, &length);
  RegCloseKey(hKey);
  if (status != ERROR_SUCCESS) {
    SetLastError(status);
    DisplayError("Cannot read RightSpeed registry key to obtain main "
                 "directory");
    return FALSE;
  }
  if (type != REG_SZ) {
    MessageBox(NULL, "Main directory in RightSpeed registry key not a string",
               "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }

  return TRUE;
}

// GetRightSpeedVersion reads RightSpeed version number from the
// registry and puts it in the given pointer location
// (RightSpeedVersionPtr). It returns a boolean indicating whether it
// was successful.

BOOLEAN GetRightSpeedVersion (DWORD *versionPtr)
{
  DWORD status, length, type;
  HKEY hKey;

  // Open the RightSpeed registry key

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        "Software\\Jay Lorch\\RightSpeed\\CurrentVersion",
                        0, KEY_QUERY_VALUE, &hKey);
  if (status != ERROR_SUCCESS) {
    SetLastError(status);
    DisplayError("Cannot open RightSpeed registry key to obtain version "
                 "number");
    return FALSE;
  }

  // Read the Version value from the key

  length = sizeof(DWORD);
  status = RegQueryValueEx(hKey, "Version", NULL, &type,
                           (BYTE *) versionPtr, &length);
  RegCloseKey(hKey);
  if (status != ERROR_SUCCESS) {
    SetLastError(status);
    DisplayError("Cannot read RightSpeed registry key to obtain RightSpeed "
                 "version number");
    return FALSE;
  }
  if (type != REG_DWORD || length != sizeof(DWORD)) {
    MessageBox(NULL,
               "Version number in RightSpeed registry key is not the right "
               "type.", "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }

  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
// STRING ROUTINES
////////////////////////////////////////////////////////////////////////////

// RemoveRightSpeedDLL removes RIGHTSPD.DLL from the given string
// (str) listing a space-separated list of dynamic link library names.
// It returns a boolean indicating whether it found such a substring
// to remove.

BOOLEAN RemoveRightSpeedDLL (char *str)
{
  char *p, *q;

  // Loop, each time finding the next instance of the substring
  // RIGHTSPD.DLL and checking to see if it is surrounded on the left
  // and right by spaces and/or beginning-or-end-of-string. If such a
  // bounded substring is found, it is extracted.

  for (p = str; 1; p += 12) {
    // Find the next RIGHTSPD.DLL at 'p'.  If not found, return FALSE
    p = strstr(p, "RIGHTSPD.DLL");
    if (p == NULL) return FALSE;

    if (p == str) {          // Case 1: <start-of-str>RIGHTSPD.DLL
      if (*(p+12) == '\0') { // Case 1a: <start-of-str>RIGHTSPD.DLL<end-of-str>
        *p = '\0';           //  ... turn str into null string
        return TRUE;
      }
      if (*(p+12) == ' ') {    // Case 1b: <start-of-str>RIGHTSPD.DLL<space>
        for (q = p; 1; q++) {  //   ... copy stuff after space to start-of-str
          *q = *(q+13);
          if (*q == '\0') break;
        }
        return TRUE;
      }
    }
    else if (*(p-1) == ' ') {  // Case 2: <space>RIGHTSPD.DLL
      if (*(p+12) == '\0') {   // Case 2a: <space>RIGHTSPD.DLL<end-of-str>
        *(p-1) = '\0';         //  ... terminate string where space is
        return TRUE;
      }
      if (*(p+12) == ' ') {    // Case 2b: <space>RIGHTSPD.DLL<space>
        for (q = p; 1; q++) {  //  ... copy stuff after end space to where R is
          *q = *(q+13);
          if (*q == '\0') break;
        }
        return TRUE;
      }
    }
  }

  return FALSE;
}

////////////////////////////////////////////////////////////////////////////
// DIRECTORY DELETION ROUTINES
////////////////////////////////////////////////////////////////////////////

// This function returns a boolean indicating if the given file name
// is "." or "..".

BOOLEAN IsSpecialFile (char *fileName)
{
  if (fileName[0] != '.') return FALSE;
  if (fileName[1] == '\0') return TRUE;
  if (fileName[1] != '.') return FALSE;
  return (fileName[2] == '\0');
}

// RemoveContentsOfDirectory removes the contents of the directory
// named in the given string (directory). It returns a boolean
// indicating whether it was successful.

BOOLEAN RemoveContentsOfDirectory (CString directory)
{
  CString searchPattern, fullPath;
  WIN32_FIND_DATA findData;
  HANDLE findHandle;

  searchPattern = directory + "\\*.*";

  findHandle = FindFirstFile(searchPattern, &findData);
  if (findHandle == INVALID_HANDLE_VALUE)
    return TRUE;

  do {
    if (IsSpecialFile(findData.cFileName))
      continue;

    // Delete this "file", which may be a directory.

    fullPath = directory + "\\" + findData.cFileName;
    if (findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
      if (!MyDeleteDirectory(fullPath))
        return FALSE;
    }
    else {
      // Try to delete the file. If we can't, delay the delete until
      // the next reboot.
      if (!DeleteFile(fullPath) &&
          !MoveFileEx(fullPath,
                      NULL,
                      MOVEFILE_DELAY_UNTIL_REBOOT)) {
        DisplayError("Can't delete " + fullPath);
        return FALSE;
      }
    }

  } while (FindNextFile(findHandle, &findData));

  // Close the search handle and return success.

  FindClose(findHandle);
  return TRUE;
}

// This function deletes the given directory and all its contents. It
// returns a boolean indicating if it was successful.

BOOLEAN MyDeleteDirectory (CString directory)
{
  // First, try to remove the contents of the directory.

  if (!RemoveContentsOfDirectory(directory))
    return FALSE;

  // Then try to remove the directory. If we fail, attempt to remove
  // the directory at the next reboot.

  if (!RemoveDirectory(directory) &&
      !MoveFileEx(directory,
                  NULL,
                  MOVEFILE_DELAY_UNTIL_REBOOT)) {
    DisplayError("Can't remove directory " + directory);
    return FALSE;
  }

  // Return success.

  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
//                        PROGRAM FOLDER ROUTINES
////////////////////////////////////////////////////////////////////////////

// GetRightSpeedProgramFolder gets the name of the RightSpeed
// subfolder of the main program folder, whichever version it is.  It
// returns that subdirectory name in programFolder, and returns a
// boolean indicating if it was successful.

BOOLEAN GetRightSpeedProgramFolder (char *programFolder)
{
  LPITEMIDLIST idList;
  DWORD version;

  // Get the version number from the registry

  if (!GetRightSpeedVersion(&version)) return FALSE;

  // Find the main program folder

  if (SHGetSpecialFolderLocation(NULL, CSIDL_COMMON_PROGRAMS, &idList) !=
      NO_ERROR) {
    DisplayError("Could not find Start menu location");
    return FALSE;
  }
  if (!SHGetPathFromIDList(idList, programFolder)) {
    DisplayError("Could not get Start menu directory name");
    return FALSE;
  }

  // Append "RightSpeed" and the version number to the program folder name

  if (version % 256 == 0)
    sprintf(&programFolder[strlen(programFolder)], "\\RightSpeed %d.%d",
            version >> 16, (version >> 8) % 256);
  else
    sprintf(&programFolder[strlen(programFolder)], "\\RightSpeed %d.%d.%d",
            version >> 16, (version >> 8) % 256, version % 256);
  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
//                            DRIVER COMMUNICATION
////////////////////////////////////////////////////////////////////////////

// This function tells RSTask to prepare to be uninstalled. It returns a
// boolean indicating if it was successful.

BOOLEAN PrepareRSTaskForUninstall (void)
{
  HANDLE hDevice;
  DWORD returnSize;

  // Connect to RSTask to tell it to prepare for uninstallation.

  hDevice = CreateFile("\\\\.\\RSTask",
                       GENERIC_READ | GENERIC_WRITE,
                       0, // share mode
                       NULL, // security attributes
                       OPEN_EXISTING,
                       FILE_ATTRIBUTE_NORMAL,
                       NULL); // template
  if (hDevice == INVALID_HANDLE_VALUE) {
    // If we couldn't connect to RSTask, there's no need to tell it
    // to prepare for uninstallation.
    return TRUE;
  }

  // Send the I/O control request telling it to prepare for uninstallation.

  if (!DeviceIoControl(hDevice,
                       RS_TASK_IOCTL_PREPARE_FOR_UNINSTALL,
                       NULL, // input buffer
                       0, // input buffer length
                       NULL, // output buffer
                       0, // output buffer length
                       &returnSize,
                       NULL)) { // overlapped
    DisplayError("RSTask was unable to prepare for uninstallation");
    CloseHandle(hDevice);
    return FALSE;
  }

  // Close the device handle and return success.

  CloseHandle(hDevice);
  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
// LOAD ORDER GROUP TAGS
////////////////////////////////////////////////////////////////////////////

// GetTagAndGroup determines the tag and group of a given service.
// If such a tag and group exists, it puts it in tagPtr and groupName and
// returns TRUE; otherwise, it returns FALSE.

BOOLEAN GetTagAndGroup (SC_HANDLE schService, DWORD *tagPtr, char *groupName,
                        DWORD maxGroupSize)
{
  QUERY_SERVICE_CONFIG *serviceConfigInfo;
  DWORD sizeNeeded;

  // Read the configuration information for the service. The first guess
  // as to the needed size for this information could be wrong, so try
  // again if the return value indicates a bigger size is needed.

  serviceConfigInfo = (QUERY_SERVICE_CONFIG *)
    malloc(INITIAL_GUESS_SERVICE_CONFIG_SIZE);
  if (serviceConfigInfo == NULL) return FALSE;
  if (!QueryServiceConfig(schService, serviceConfigInfo,
                          INITIAL_GUESS_SERVICE_CONFIG_SIZE, &sizeNeeded)) {
    if (GetLastError() != ERROR_INSUFFICIENT_BUFFER) {
      free(serviceConfigInfo);
      return FALSE;
    }
    else {
      free(serviceConfigInfo);
      serviceConfigInfo = (QUERY_SERVICE_CONFIG *) malloc(sizeNeeded);
      if (serviceConfigInfo == NULL) return FALSE;
      if (!QueryServiceConfig(schService, serviceConfigInfo, sizeNeeded,
                              &sizeNeeded)) {
        free(serviceConfigInfo);
        return FALSE;
      }
    }
  }

  // Extract the tag and group information if it exists.

  if (serviceConfigInfo->lpLoadOrderGroup &&
      serviceConfigInfo->lpLoadOrderGroup[0] &&
      strlen(serviceConfigInfo->lpLoadOrderGroup) < maxGroupSize) {
    strcpy(groupName, serviceConfigInfo->lpLoadOrderGroup);
    *tagPtr = serviceConfigInfo->dwTagId;
    free(serviceConfigInfo);
    return TRUE;
  }
  else {
    free(serviceConfigInfo);
    return FALSE;
  }
}

// RemoveTagFromGroup removes the given tag number from the given
// group. Actually, it removes the tag from the GroupOrderList
// registry entry that indicates the order in which that group
// services should load.

BOOLEAN RemoveTagFromGroup (DWORD tag, char *groupName)
{
  DWORD err, type, length, lengthNeeded, i, j, *groupOrderList;
  HKEY hKey;

  // Open the group order list registry key.

  err = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                     "System\\CurrentControlSet\\Control\\GroupOrderList",
                     NULL, KEY_READ | KEY_WRITE, &hKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not open registry key for service group order list");
    return FALSE;
  }

  // Read the value named after the group to get the actual group order
  // list. First we read the value with no buffer to put it into, to find
  // out how long a buffer we need, if any. If there is no such key, we're
  // done. Otherwise, we read the value from the registry. Note that the
  // structure for the value is a sequence of DWORD's, the first of which
  // is the number of tags n, and the next n of which are those tags in the
  // order in which they should be started.

  lengthNeeded = 0;
  err = RegQueryValueEx(hKey, groupName, NULL, &type, NULL, &lengthNeeded);
  if (err == ERROR_FILE_NOT_FOUND) {
    RegCloseKey(hKey);
    return TRUE;
  }
  else if (err == ERROR_SUCCESS) {
    groupOrderList = (DWORD *) malloc(lengthNeeded);
    length = lengthNeeded;
    err = RegQueryValueEx(hKey, groupName, NULL, &type,
                          (LPBYTE) groupOrderList, &length);
    if (err != ERROR_SUCCESS) {
      SetLastError(err);
      DisplayError("Could not read group order list value");
      RegCloseKey(hKey);
      return FALSE;
    }
    if (type != REG_BINARY || lengthNeeded != length ||
        (groupOrderList[0]+1) * sizeof(DWORD) != length) {
      MessageBox(NULL, "Read invalid group order list value", "Error",
                 MB_OK | MB_ICONEXCLAMATION);
      RegCloseKey(hKey);
      return FALSE;
    }
  }
  else {
    SetLastError(err);
    DisplayError("Could not get size of group order list value");
    RegCloseKey(hKey);
    return FALSE;
  }

  // Search for the tag in the group order list. If it's there, remove it.

  for (i = 1; i <= groupOrderList[0]; i++) {
    if (groupOrderList[i] == tag) {
      for (j = i; j < groupOrderList[0]; j++)
        groupOrderList[j] = groupOrderList[j+1];
      groupOrderList[0]--;
      break;
    }
  }

  // Set the new group order list in the registry.

  err = RegSetValueEx(hKey, groupName, 0, REG_BINARY, (LPBYTE) groupOrderList,
                      (groupOrderList[0] + 1) * sizeof(DWORD));
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not set the group order list in the registry");
    RegCloseKey(hKey);
    return FALSE;
  }

  // Release resources used and return a status value.

  free(groupOrderList);
  RegCloseKey(hKey);

  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
//                       UNINSTALLATION SUBROUTINES
////////////////////////////////////////////////////////////////////////////

// DeleteDLL deletes the RightSpeed DLL RIGHTSPD.DLL. For this, it
// needs to know the name of the system root directory (given in
// systemRoot). It returns a boolean indicating whether it succeeded.

BOOLEAN DeleteDLL (char *systemRoot)
{
  char fileName[MAX_PATH], appInitDLLs[MAX_PATH];
  DWORD status, length, type;
  HKEY hKey;

  // Remove RightSpeed.DLL from the list of DLL's automatically loaded
  // into each application. We do this by updating the registry,
  // specifically by removing RIGHTSPD.DLL from the list of DLL's in
  // the key AppInit_DLLs in the Windows registry key.

  status = RegOpenKeyEx
             (HKEY_LOCAL_MACHINE,
              "Software\\Microsoft\\Windows NT\\CurrentVersion\\Windows", 0,
              KEY_QUERY_VALUE | KEY_SET_VALUE, &hKey);
  if (status != ERROR_SUCCESS) {
    SetLastError(status);
    DisplayError("Could not open Windows key");
    return FALSE;
  }

  length = sizeof(appInitDLLs);
  status = RegQueryValueEx(hKey, "AppInit_DLLs", 0, &type,
                           (BYTE *) appInitDLLs, &length);
  if (status != ERROR_SUCCESS) {
    RegCloseKey(hKey);
    SetLastError(status);
    DisplayError("Could not get AppInit_DLLs registry value");
    return FALSE;
  }

  if (type != REG_SZ) {
    RegCloseKey(hKey);
    MessageBox(NULL, "AppInit_DLLs registry value is not a string", "Error",
               MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }

  if (RemoveRightSpeedDLL(appInitDLLs)) {
    status = RegSetValueEx(hKey, "AppInit_DLLs", 0, REG_SZ,
                           (BYTE *) appInitDLLs, strlen(appInitDLLs) + 1);
    if (status != ERROR_SUCCESS) {
      RegCloseKey(hKey);
      SetLastError(status);
      DisplayError("Could not set AppInit_DLLs registry value");
      return FALSE;
    }
  }

  RegCloseKey(hKey);

  // Delete RIGHTSPD.DLL from the SYSTEM32 subdirectory of the system root
  // directory at reboot (not now, because it's in use).

  sprintf(fileName, "%s\\SYSTEM32\\RIGHTSPD.DLL", systemRoot);
  if (!MoveFileEx(fileName, NULL, MOVEFILE_DELAY_UNTIL_REBOOT) &&
      GetLastError() != ERROR_FILE_NOT_FOUND) {
    DisplayError("Could not delete RightSpeed.DLL from system directory");
    return FALSE;
  }

  return TRUE;
}

// DeleteDriver deletes the driver of the given name (driverName). It
// must be passed an open handle to the service control manager
// (schSCManager) and a string containing the system root directory
// name (systemRoot). It should also be given a boolean indicating
// whether it should stop the driver before deleting it
// (stopDriver). It returns a boolean indicating whether it succeeded.

BOOLEAN DeleteDriver (
  char *driverName,
  SC_HANDLE schSCManager,
  char *systemRoot,
  BOOLEAN stopDriver
  )
{
  char fileName[MAX_PATH];
  char errString[256];
  SC_HANDLE schService;
  SERVICE_STATUS serviceStatus;
  BOOLEAN hasTagAndGroup;
  DWORD tagInGroup;
  char groupName[128];
  DWORD maxGroupSize = 128;

  // Open the service.  If successful, stop the service (if stopDriver
  // is set) then delete it.  If the service belongs to a group,
  // determine that group and its tag within that group.

  schService = OpenService(schSCManager, driverName, SERVICE_ALL_ACCESS);
  if (schService == NULL) {
    if (GetLastError() != ERROR_SERVICE_DOES_NOT_EXIST) {
      sprintf(errString, "Could not open the %s service to delete it",
              driverName);
      DisplayError(errString);
      return FALSE;
    }
  }
  else {
    hasTagAndGroup = GetTagAndGroup(schService, &tagInGroup, groupName,
                                    maxGroupSize);
    if (stopDriver)
      ControlService(schService, SERVICE_CONTROL_STOP, &serviceStatus);
    DeleteService(schService);
    CloseServiceHandle(schService);
  }

  // Delete the no-longer-needed SYS file from the drivers directory

  sprintf(fileName, "%s\\SYSTEM32\\DRIVERS\\%s.SYS", systemRoot, driverName);
  if (!DeleteFile(fileName) && GetLastError() != ERROR_FILE_NOT_FOUND) {
    sprintf(errString, "Could not delete %s.sys from drivers directory",
            driverName);
    DisplayError(errString);
    return FALSE;
  }

  // Remove the no-longer-needed tag from the group in the GroupOrderList.

  if (hasTagAndGroup)
    RemoveTagFromGroup(tagInGroup, groupName);

  return TRUE;
}

// DeleteMainDirectory removes the main RightSpeed directory and all
// its contents (including the TTG and Logs subdirectories and their
// contents).  It returns a boolean indicating whether it succeeded.
// It takes as parameter the name of the main directory.

BOOLEAN DeleteMainDirectory (char *mainDirectory)
{
  return MyDeleteDirectory(mainDirectory);
}

// DeleteStartMenuShortcuts removes all the RightSpeed utility
// shortcuts from the Start menu, as well as the RightSpeed
// subdirectory that they were in.

BOOLEAN DeleteStartMenuShortcuts (void)
{
  char programFolder[MAX_PATH];

  if (!GetRightSpeedProgramFolder(programFolder)) return FALSE;

  // Remove the contents of the RightSpeed program folder, and finally
  // that folder itself

  if (!MyDeleteDirectory(programFolder)) return FALSE;

  return TRUE;
}

// DeleteRegistryInfo deletes all the information about RightSpeed in
// the registry.

BOOLEAN DeleteRegistryInfo (void)
{
  DWORD err;

  // Delete the RightSpeed\\CurrentVersion registry key.

  err = RegDeleteKey(HKEY_LOCAL_MACHINE,
                     "Software\\Jay Lorch\\RightSpeed\\CurrentVersion");
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Cannot delete RightSpeed\\CurrentVersion registry key");
    return FALSE;
  }

  // Delete the RightSpeed registry key.

  err = RegDeleteKey(HKEY_LOCAL_MACHINE,
                     "Software\\Jay Lorch\\RightSpeed");
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Cannot delete RightSpeed registry key");
    return FALSE;
  }

  // Return success.

  return TRUE;
}
